/* Copyright (c) 2004, Noah Sorscher
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 	 	Redistributions of source code must retain the above copyright
 * 	notice, this list of conditions and the following disclaimer.
 * 	 	Redistributions in binary form must reproduce the above copyright
 * 	notice, this list of conditions and the following disclaimer in the
 * 	documentation and/or other materials provided with the distribution.
 * 	 	Neither the name of the Noah Sorscher nor the names of its
 * 	contributors may be used to endorse or promote products derived from
 * 	this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
//
//  MHMovie.m
//  Hootenanny
//
//  Created by Noah Sorscher on Tue Mar 25 2003.
//  Copyright (c) 2003 __MyCompanyName__. All rights reserved.
//

#import <QTKit/QTKit.h>
#import "MHMovie.h"
#import "MHChapters.h"
#import "APIC.h"
#import <QuickTime/QuickTimeComponents.h>
#import "NSStringWebAdditions.h"
#import <sys/event.h>

#define DUMMYURL [NSURL fileURLWithPath:[[NSBundle mainBundle] pathForResource:@"dummy" ofType:@"txt"]]
#define DEBUG NO


long long Wide2LongLong(wide w);

char *rstring(char *s, int length);

@interface QTMovie (adaptations)
- (void *)QTMovie;
@end

@implementation QTMovie (adaptations)
- (void *)QTMovie
{	return [self quickTimeMovie];	}
@end


@implementation MHTrack
- (id) initWithTrack:(Track)aTrack
{
    if (!aTrack) {
        [self release];
        return nil;
    }
    myTrack = aTrack;
    return [super init];
}

- (void)makeAlterations:(NSString *)a
{
	NSLog(a);
	NSArray *list = [a componentsSeparatedByString:@"\n"];
	int i, n = [list count];
	for (i=0;i<n;i++) {
		NSRange r = [[list objectAtIndex:i] rangeOfString:@"|"];
		if (r.length == 1) {
			NSString *verb = [[list objectAtIndex:i] substringToIndex:r.location];
			NSString *object = [[list objectAtIndex:i] substringFromIndex:r.location+1];
			if ([verb isEqualToString:@"Name"]) [self setName:object];
			else if ([verb isEqualToString:@"Enabled"]) [self setEnabled:[object boolValue]];
			else if ([verb isEqualToString:@"Offset"]) [self setOffset:[object intValue]];
			else if ([verb isEqualToString:@"Duration"]) [self setDuration:[object intValue]];
			else if ([verb isEqualToString:@"FourCC"]) [self setFourCC:object];
			else if ([verb isEqualToString:@"Vendor"]) [self setVendor:object];
			else if ([verb isEqualToString:@"TimeScale"]) [self setTimeScale:[object intValue]];
			else if ([verb isEqualToString:@"Format"]) [self setFormat:object];
			else if ([verb isEqualToString:@"Preload"]) [self setPreloads:[object boolValue]];
			else if ([verb isEqualToString:@"High Quality"]) [self setHighQuality:[object boolValue]];
			else if ([verb isEqualToString:@"Layer"]) [self setLayer:(short)[object intValue]];
			else if ([verb isEqualToString:@"Frame"]) [self setDisplayFrame:NSRectFromString(object)];
			else if ([verb isEqualToString:@"Rotate"]) [self rotate:object];
			else if ([verb isEqualToString:@"Language"]) [self setLanguage:object];
			else if ([verb isEqualToString:@"Transparency"]) [self setTransparency:[object floatValue]];
			else if ([verb isEqualToString:@"Graphics Mode"]) [self setGraphicsMode:[object intValue]];
			else return;
		}
	}
	NSString *alt = [self alterations];
	while (CountUserDataType(GetTrackUserData(myTrack), kUserDataTextEditDate1))
		RemoveUserData(GetTrackUserData(myTrack), kUserDataTextEditDate1, 1);
	
	if ([alt length] && ![alt hasSuffix:@"\n"]) alt = [alt stringByAppendingString:@"\n"];
	alt = [alt stringByAppendingString:a];
	Handle h = NewHandleClear([alt length]);
	const char *buffer = [[alt dataUsingEncoding:NSMacOSRomanStringEncoding allowLossyConversion:YES] bytes];
	BlockMoveData(buffer, *h, [alt length]);
	OSErr err = AddUserDataText(GetTrackUserData(myTrack), h, kUserDataTextEditDate1, 1, langUnspecified);
	if (err) NSLog(@"error adding track user data: %ld", err);
}

- (NSString *)alterations
{
	Handle h = NewHandleClear(0);
	GetUserDataText(GetTrackUserData(myTrack), h, kUserDataTextEditDate1, 1, langUnspecified);
	NSString *s = [NSString stringWithCString:*h length:GetHandleSize(h)];
	DisposeHandle(h);
	return s;
}

- (MHTrack *)alternate
{
    Track t = GetTrackAlternate(myTrack);
    if (t == myTrack) return nil;
    return [[[MHTrack alloc] initWithTrack:t] autorelease];
}

- (BOOL) containsString:(NSString *)searchString
{
    if (![[self type] isEqualToString:@"Text"]) return NO;
    TimeValue tv = 0;
    Handle h = NewHandleClear(0);
    GetTrackNextInterestingTime(myTrack, nextTimeMediaSample + nextTimeEdgeOK, 0, fixed1, &tv, NULL);
    while (tv != -1) {
        GetMediaSample(GetTrackMedia(myTrack), h, 0, NULL, TrackTimeToMediaTime(tv, myTrack), NULL, NULL, NULL, NULL, 0, NULL, NULL);
        short size = *(short *)(*h);
        NSString *text = [NSString stringWithCString:(char *)(*h + sizeof(short)) length:size];
        if (text && [text rangeOfString:searchString options:NSCaseInsensitiveSearch].location != NSNotFound) return YES;
        GetTrackNextInterestingTime(myTrack, nextTimeMediaSample, tv, fixed1, &tv, NULL);
    }
    return NO;
}

- (NSString *)dataRef
{
	Handle h = NewHandleClear(0);
	GetUserDataText(GetTrackUserData(myTrack), h, kUserDataTextOriginalSource, 1, langUnspecified);
	char *buffer = (char *)*h;
	NSString *source = [[[NSString alloc] initWithBytes:buffer length:GetHandleSize(h) encoding:NSUTF8StringEncoding] autorelease];
	//if ([source length]) source = [NSString stringWithFormat:@" %@", source];
	
	Media m = GetTrackMedia(myTrack);
	short count = 0;
	OSErr err = GetMediaDataRefCount(m, &count);
	if (err) NSLog(@"error in GetMediaDataRefCount: %ld", err);
	//Handle h = NewHandleClear(0);
	OSType dataType = 0;
	long attr = 0;
	err = GetMediaDataRef(m, 1, &h, &dataType, &attr);
#ifdef __LITTLE_ENDIAN__
	dataType = NSSwapLong(dataType);
#endif
	NSString *type = [NSString stringWithCString:(char *)&dataType length:4];
	//NSLog(type);
	//if (dataType == AliasDataHandlerSubType) type = QTDataReferenceTypePointer;
	//else if (dataType == HandleDataHandlerSubType) type = QTDataReferenceTypeHandle;
	//else if (dataType == 0) type = nil;
	QTDataReference *qtdr = [QTDataReference dataReferenceWithDataRef:h type:type];
	NSString *s = [qtdr referenceFile];
	double size = GetHandleSize(h);
	NSString *ssize = @"B";
	if (size > 1024 * 1024 * 1024) ssize = [NSString stringWithFormat:@"%.2f GB", size / 1024 / 1024 / 1024];
	else if (size > 1024 * 1024) ssize = [NSString stringWithFormat:@"%.2f MB", size / 1024 / 1024];
	else if (size > 1024) ssize = [NSString stringWithFormat:@"%.2f KB", size / 1024];
	else ssize = [NSString stringWithFormat:@"%d B", (int)size];
	if (!s) s = [NSString stringWithFormat:@"in memory: %@ %@", ssize, source];
	return s;
}

- (NSString *)description
{
    NSString *type = [self type];
    if (![type isEqualToString:@"Text"]) return type;
    NSMutableString *string = [NSMutableString stringWithString:@""];
    TimeValue tv = 0;
    Handle h = NewHandleClear(0);
    short flags = nextTimeMediaSample + nextTimeEdgeOK;
    while (YES) {
        GetTrackNextInterestingTime(myTrack, flags, tv, fixed1, &tv, NULL);
        if (tv == -1) break;
        flags = nextTimeMediaSample;
        GetMediaSample(GetTrackMedia(myTrack), h, 0, NULL, TrackTimeToMediaTime(tv, myTrack), NULL, NULL, NULL, NULL, 0, NULL, NULL);
        short size = *(short *)(*h);
        NSString *line = [NSString stringWithCString:(char *)(*h + sizeof(short)) length:size];
        if (line && [[line whitespaceTrimmed] length]) {
            if ([string length]) [string appendString:@"|"];
            [string appendString:line];
        }
    }
    return string;
}

- (NSRect)displayFrame
{
    Fixed w, h;
    GetTrackDimensions(myTrack, &w, &h);
    Rect box;
    box.top = box.left = 0;
    box.right = Fix2Long(w);
    box.bottom = Fix2Long(h);
    MatrixRecord m;
    GetTrackMatrix(myTrack, &m);
    TransformRect(&m, &box, NULL);
    NSRect rect;
    rect.origin.x = box.left;
    rect.origin.y = box.bottom;
    rect.size.width = box.right - box.left;
    rect.size.height = box.bottom - box.top;
	
	//GetMovieMatrix(GetTrackMovie(myTrack), &m);
	//TransformRect(&m, &box, NULL);
	
    return rect;
}

- (TimeValue)duration
{
    return GetTrackDuration(myTrack);
}

- (BOOL)enabled
{
    return GetTrackEnabled(myTrack);
}

- (NSString *)format
{
    ImageDescriptionHandle h = (ImageDescriptionHandle)NewHandle(sizeof(SoundDescription));
    GetMediaSampleDescription(GetTrackMedia(myTrack), 1, (SampleDescriptionHandle)h);
    OSType t = (*h)->cType;
    GetMediaHandlerDescription(GetTrackMedia(myTrack), &t, 0, 0);
    switch (t) {
        case VideoMediaType: return CFStringCreateWithPascalString(nil, (*h)->name, kCFStringEncodingMacRoman); // [NSString stringWithCString: (char *)&(*h)->name + 1 length: *(char *)&(*h)->name];
        default: break;
    }
	return nil;
}

- (NSString *)fourCC
{
    SampleDescriptionHandle h = (SampleDescriptionHandle)NewHandle(sizeof(SoundDescription));
    GetMediaSampleDescription(GetTrackMedia(myTrack), 1, h);
    long type = (*(SoundDescriptionHandle)h)->dataFormat;
    if (!type) type = (*(ImageDescriptionHandle)h)->cType;
    if (type) {
        char *buffer = (char *)&type;
#ifdef __LITTLE_ENDIAN__
		buffer = rstring(buffer, 4);
#endif
        NSString *output = [NSString stringWithCString:buffer];
		//int l = [output length];
        if ([output hasPrefix:@"ms"]) {
            output = [NSString stringWithFormat:@"ms\\%d\\%d", (int)buffer[2], (int)buffer[3]];
            if (buffer[2] == 0 && buffer[3] == 85) output = [output stringByAppendingString:@" (mp3)"];
        }
        return output;
    }
    return nil;
}

- (long)graphicsMode
{
	MediaHandler mh = GetMediaHandler(GetTrackMedia(myTrack));
	long mode = -1;
	RGBColor color;
	MediaGetGraphicsMode(mh, &mode, &color);
	//NSLog(@"%@: %ld", [self name], mode);
	return mode;
	/*
			// Boolean modes
			// src modes are used with bitmaps and text;
			// pat modes are used with lines and shapes
			srcCopy               =0
			srcOr                 =1
			srcXor                =2
			srcBic                =3
			notSrcCopy            =4
			notSrcOr              =5
			notSrcXor             =6
			notSrcBic             =7
			patCopy               =8
			patOr                 =9
			patXor                =10
			patBic                =11
			notPatCopy            =12
			notPatOr              =13
			notPatXor             =14
			notPatBic             =15
			// Text dimming
			grayishTextOr         =49
			// Highlighting
			hilite                =50
			hilitetransfermode    =50
			// Arithmetic modes
			blend                 =32
			addPin                =33
			addOver               =34
			subPin                =35
			addMax                =37
			adMax                 =37
			subOver               =38
			adMin                 =39
			ditherCopy            =64
			// Transparent mode
			transparent           =36
	 
	 graphicsModeStraightAlpha	= 256
	 graphicsModePreWhiteAlpha	= 257
	 graphicsModePreBlackAlpha	= 258
	 graphicsModeComposition	= 259
	 graphicsModeStraightAlphaBlend	= 260
	 graphicsModePreMulColorAlpha	= 261
	 graphicsModePerComponentAlpha	= 272
	 
	*/
}

- (BOOL)highQuality
{
    long hints = hintsHighQuality;
    GetMediaPlayHints(GetTrackMedia(myTrack), &hints);
    long hints2 = hints & hintsHighQuality;             // if I just return this, it always evaluates to false for some unknown reason.  does this mean my compiler is smoking crack?
    return hints2 > 0;
}

- (NSString *)language
{
    short lang = GetMediaLanguage(GetTrackMedia(myTrack));
    return [tLanguages numberToString:lang];
}

- (short)layer
{
    return GetTrackLayer(myTrack);
}

- (double) mbSize
{
    wide w;
    OSErr err = GetTrackDataSize64(myTrack, 0, GetTrackDuration(myTrack), &w);
    if (err) NSLog(@"error getting track data size: %d", err);
    double mb = Wide2LongLong(w) / 1024.0 / 1024.0;
	return mb;
}

- (NSString *)name
{
    NSString *name = nil;
    Handle h = NewHandleClear(0);
    GetUserData(GetTrackUserData(myTrack), h, kUserDataName, 1);
    int size = GetHandleSize(h);
    if (size) {
        char *buffer = (char *)*h;
        name = [NSString stringWithCString:buffer length:size];
    }
    return name;
}

- (TimeValue)offset
{
    return GetTrackOffset(myTrack);
}

- (BOOL)preloads
{
	long preloadFlags, defaultHints;
	TimeValue start, duration;
	GetTrackLoadSettings(myTrack, &start, &duration, &preloadFlags, &defaultHints);
	return (preloadFlags & preloadAlways || preloadFlags & preloadOnlyIfEnabled);
}

- (Track)QTTrack
{
    return myTrack;
}

- (void)rotate:(NSString *)degrees
{
	NSRect f = [self displayFrame];
	Fixed x = Long2Fix((long)(f.origin.x + f.size.width / 2));
	Fixed y = Long2Fix((long)(f.origin.y - f.size.height / 2));
	MatrixRecord m;
	GetTrackMatrix(myTrack, &m);
	if ([degrees hasPrefix:@"Flip"]) {
		if ([degrees hasSuffix:@"Vertical"]) ScaleMatrix(&m, fixed1, Long2Fix(-1), Long2Fix(0), y);
		else ScaleMatrix(&m, Long2Fix(-1), fixed1, x, Long2Fix(0));
		SetTrackMatrix(myTrack, &m);
		return;
	}
	long n = [degrees intValue];
	if (n) {
		RotateMatrix(&m, Long2Fix(n), x, y);
		SetTrackMatrix(myTrack, &m);
	}
}

- (void)setAlterations:(NSString *)a
{
	
}

- (void)setAlternate:(MHTrack *)track
{
    if (track) SetTrackAlternate(myTrack, [track QTTrack]);
}

- (void)setDisplayFrame:(NSRect)frame
{
	NSRect oldFrame = [self displayFrame];
	MatrixRecord m;
	GetTrackMatrix(myTrack, &m);
	float dw = frame.size.width - oldFrame.size.width;
	float dh = frame.size.height - oldFrame.size.height;
	if (dw || dh) {
		Fixed scaleX = FixRatio(frame.size.width, oldFrame.size.width);
		Fixed scaleY = FixRatio(frame.size.height, oldFrame.size.height);
		ScaleMatrix(&m, scaleX, scaleY, Long2Fix(0), Long2Fix(0));
		SetTrackMatrix(myTrack, &m);
	}
	oldFrame = [self displayFrame];
	int dx = frame.origin.x - oldFrame.origin.x;
	int dy = frame.origin.y - oldFrame.origin.y;
	if (dx || dy) {
		Fixed deltaH = Long2Fix(dx);
		Fixed deltaV = Long2Fix(dy);
		TranslateMatrix(&m, deltaH, deltaV);
	}
	SetTrackMatrix(myTrack, &m);
}

- (void)setDuration:(TimeValue)duration
{
    OSErr err = ScaleTrackSegment(myTrack, GetTrackOffset(myTrack), GetTrackDuration(myTrack), duration);
    if (err) NSLog(@"error scaling track segment: %d", err);
}

- (void)setEnabled:(BOOL)flag
{
    SetTrackEnabled(myTrack, flag);
}

- (void)setFormat:(NSString *)format
{
    if ([format length] > 31) format = [format substringToIndex:31];
    ImageDescriptionHandle h = (ImageDescriptionHandle)NewHandleClear(sizeof(ImageDescription));
    GetMediaSampleDescription(GetTrackMedia(myTrack), 1, (SampleDescriptionHandle)h);
    CopyCStringToPascal([format cString], (*h)->name);
    SetMediaSampleDescription(GetTrackMedia(myTrack), 1, (SampleDescriptionHandle)h);
}

- (void)setFourCC:(NSString *)fourcc
{
    if (!fourcc) return;
    char buffer[] = {0, 0, 0, 0};
    NSScanner *scan = [NSScanner scannerWithString:fourcc];
    int spot = 0;
    while (![scan isAtEnd] && spot < 4) {
        if ([scan scanString:@"\\" intoString:nil]) {
            int n = -1;
            [scan scanInt:&n];
            if (n != -1)
                buffer[spot++] = n;
        } else {
            buffer[spot++] = [fourcc characterAtIndex:[scan scanLocation]];
            [scan setScanLocation:[scan scanLocation]+1];
        }
    }
    OSType type = 0;
    type += buffer[0] * 256 * 256 * 256;
    type += buffer[1] * 256 * 256;
    type += buffer[2] * 256;
    type += buffer[3];
    Media media = GetTrackMedia(myTrack);
    SampleDescriptionHandle desc = (SampleDescriptionHandle)NewHandle(sizeof(SampleDescription));
    GetMediaSampleDescription(media, 1, desc);
    MoveHHi((Handle)desc);
    OSType trackType;
    GetMediaHandlerDescription(media, &trackType, 0, 0);
    switch (trackType) {
        case VideoMediaType: case MPEGMediaType: (*(ImageDescriptionHandle)desc)->cType = type; break;
        case SoundMediaType: default: (*desc)->dataFormat = type; break;
    }
    SetMediaSampleDescription(media, 1, desc);
}

- (void)setGraphicsMode:(long)mode
{
	//unsigned short i = trans * 65535;
	//long mode = blend;
	//if (!i) mode = srcCopy;
	RGBColor c;
	long oldMode;
	//c.red = c.green = c.blue = 65535 - i;
	MediaHandler mh = GetMediaHandler(GetTrackMedia(myTrack));
	MediaGetGraphicsMode(mh, &oldMode, &c);
	ComponentResult err = MediaSetGraphicsMode(mh, mode, &c);
	if (err) NSLog(@"graphics mode error: %d", err);
}

- (void)setHighQuality:(BOOL)flag
{
    long hint = 0;
    if (flag) hint = hintsHighQuality;
    SetMediaPlayHints(GetTrackMedia(myTrack), hint, hintsHighQuality);
}

- (BOOL)setLanguage:(NSString *)lang
{
    short newLang = [tLanguages stringToNumber:lang];
    if (newLang == -1) return NO;
    SetMediaLanguage(GetTrackMedia(myTrack), newLang);
    return YES;
}

- (void)setLayer:(short)layer
{
    SetTrackLayer(myTrack, layer);
}

- (void)setName:(NSString *)name
{
    if (!name || ![name isKindOfClass:[NSString class]]) return;
    char *buffer = malloc(sizeof(char) * ([name length] + 1));
    [name getCString:buffer];
    SetUserDataItem(GetTrackUserData(myTrack), buffer, [name length], kUserDataName, 1);
    free(buffer);
}

- (void)setOffset:(TimeValue)offset
{
    SetTrackOffset(myTrack, offset);
}

- (void)setPreloads:(BOOL)p
{
	if (p)
		SetTrackLoadSettings(myTrack, -1, 0, preloadAlways, 0);
	else
		SetTrackLoadSettings(myTrack, -1, 0, 0, 0);
}

- (void)setSource:(NSString *)s
{
	int i = 0;
	while (i = CountUserDataType(GetTrackUserData(myTrack), kUserDataTextOriginalSource))
		RemoveUserData(GetTrackUserData(myTrack), kUserDataTextOriginalSource, 1);
	Handle h = NewHandleClear([s length]);
	BlockMoveData([s UTF8String], *h, [s length]);
	AddUserDataText(GetTrackUserData(myTrack), h, kUserDataTextOriginalSource, 1, langUnspecified);
	DisposeHandle(h);
}

- (void)setTimeScale:(TimeScale)ts
{
    SetMediaTimeScale(GetTrackMedia(myTrack), ts);
}

- (void)setTransparency:(float)trans
{
	unsigned short i = trans * 65535;
	long mode = blend;
	if (!i) mode = srcCopy;
	RGBColor c;
	MediaHandler mh = GetMediaHandler(GetTrackMedia(myTrack));
	ComponentResult err = MediaGetGraphicsMode(mh, &mode, &c);
	if (i && mode != blend && mode != transparent && mode != 260)
		mode = blend;
	c.red = c.green = c.blue = 65535 - i;
	err = MediaSetGraphicsMode(mh, mode, &c);
	if (err) NSLog(@"transparency error: %d", err);
}

- (void)setVendor:(NSString *)ven
{
    if (!ven) return;
    char buffer[] = {0, 0, 0, 0};
    NSScanner *scan = [NSScanner scannerWithString:ven];
    int spot = 0;
    while (![scan isAtEnd] && spot < 4) {
        if ([scan scanString:@"\\" intoString:nil]) {
            int n = -1;
            [scan scanInt:&n];
            if (n != -1)
                buffer[spot++] = n;
        } else {
            buffer[spot++] = [ven characterAtIndex:[scan scanLocation]];
            [scan setScanLocation:[scan scanLocation]+1];
        }
    }
    OSType type = 0;
    type += buffer[0] * 256 * 256 * 256;
    type += buffer[1] * 256 * 256;
    type += buffer[2] * 256;
    type += buffer[3];
    Media media = GetTrackMedia(myTrack);
    SampleDescriptionHandle desc = (SampleDescriptionHandle)NewHandle(sizeof(SampleDescription));
    GetMediaSampleDescription(media, 1, desc);
    MoveHHi((Handle)desc);
    OSType trackType;
    GetMediaHandlerDescription(media, &trackType, 0, 0);
    switch (trackType) {
        case VideoMediaType: case MPEGMediaType: (*(ImageDescriptionHandle)desc)->vendor = type; break;
        //case SoundMediaType: default: (*desc)->dataFormat = type; break;
    }
    SetMediaSampleDescription(media, 1, desc);
}

- (BOOL)subtitleType
{
	return [[NSArray arrayWithObjects:@"SRT ", @"SSA ", @"ASS ", @"SPU ", nil] containsObject:[self fourCC]];
}

- (float)transparency
{
	MediaHandler mh = GetMediaHandler(GetTrackMedia(myTrack));
	long mode = 0;
	RGBColor color;
	color.red = 65535;
	MediaGetGraphicsMode(mh, &mode, &color);
	//NSLog(@"%.1g, %.1g, %.1g", (float)color.red/65535, (float)color.green/65535, (float)color.blue/65535);
	return 1 - (float)color.red / 65535;
}

- (NSString *)type
{
    OSType t;
    GetMediaHandlerDescription(GetTrackMedia(myTrack), &t, 0, 0);
    char *buffer;
    switch (t) {
        case SoundMediaType: return @"Sound";
        case VideoMediaType: return @"Video";
        case SpriteMediaType: return @"Sprite";
        case MPEGMediaType: return @"MPEG";
        case TextMediaType: return @"Text";
        case TimeCodeMediaType: return @"TimeCode";
        default:
            buffer = (char *)&t;
            return [NSString stringWithCString:buffer length:4];
    }
}

- (BOOL) isSoundType
{
    OSType t;
    GetMediaHandlerDescription(GetTrackMedia(myTrack), &t, 0, 0);
    return (SoundMediaType == t);
}

- (MHTrack *)chapterTrack{
    Track cTrack = GetTrackReference(myTrack, kTrackReferenceChapterList, 1);
    if (cTrack)
        return [[[MHTrack alloc] initWithTrack:cTrack] autorelease];
    return nil;
}

- (void)remove
{
    // first delete any references to the track
    Movie mov = GetTrackMovie(myTrack);
    long i, n = GetMovieTrackCount(mov);
    for (i=1;i<=n;i++) {
        Track track = GetMovieIndTrack(mov, i);
        if (track && track != myTrack) {
            OSType type = 0L;
            type = GetNextTrackReferenceType(track, type);
            while (type != 0L) {
                long j, refCount = GetTrackReferenceCount(track, type);
                for (j = refCount; j >= 1; j--) {
                    Track aRefTrack = GetTrackReference(track, type, j);
                    if (aRefTrack == myTrack)
                        DeleteTrackReference(track, type, j);
                }
                type = GetNextTrackReferenceType(track, type);
            }
        }
    }
    DisposeMovieTrack(myTrack);
    myTrack = nil;
}

- (TimeScale)timeScale
{
    return GetMediaTimeScale(GetTrackMedia(myTrack));
}

- (NSString *)vendor
{
    SampleDescriptionHandle h = (SampleDescriptionHandle)NewHandleClear(sizeof(SoundDescription));
    GetMediaSampleDescription(GetTrackMedia(myTrack), 1, h);
    long type = (*(ImageDescriptionHandle)h)->vendor;
    if (!type) type = (*(SoundDescriptionHandle)h)->vendor;
    if (type) {
        char *buffer = (char *)&type;
#ifdef __LITTLE_ENDIAN__
		buffer = rstring(buffer, 4);
#endif
        NSString *output = [NSString stringWithCString:buffer length:4];
        //if ([output length] > 4) output = [output substringToIndex:4];
        if ([output hasPrefix:@"ms"]) {
            output = [NSString stringWithFormat:@"ms\\%d\\%d", (int)buffer[2], (int)buffer[3]];
            //if (buffer[2] == 0 && buffer[3] == 85) output = [output stringByAppendingString:@" (mp3)"];
        }
        return output;
    }
    return nil;
}

- (float)volume
{
	float f = 0;
	GetTrackAudioGain(myTrack, &f, 0);
	return f;
}

@end

@implementation MHMovie


//void CopyCStringToPascal (const char *theSrc, Str31 theDst);

BOOL asks = NO;
BOOL ignores = YES;
NSArray *languages = nil;

static OSType theKUserDataTypes[] = {kUserDataTextAlbum , kUserDataTextArtist , kUserDataTextAuthor , kUserDataTextChapter , kUserDataTextComment , kUserDataTextComposer , kUserDataTextCopyright , kUserDataTextCreationDate , kUserDataTextDescription , kUserDataTextDirector , kUserDataTextDisclaimer , kUserDataTextEncodedBy , kUserDataTextFullName , kUserDataTextGenre , kUserDataTextHostComputer , kUserDataTextInformation , kUserDataTextKeywords , kUserDataTextMake , kUserDataTextModel , kUserDataTextOriginalArtist , kUserDataTextOriginalFormat , kUserDataTextOriginalSource , kUserDataTextPerformers , kUserDataTextProducer , kUserDataTextProduct , kUserDataTextSoftware , kUserDataTextSpecialPlaybackRequirements , kUserDataTextTrack , kUserDataTextWarning , kUserDataTextWriter , kUserDataTextURLLink , kUserDataTextEditDate1, '\251web', '\251rat' };

static BOOL WriteMovieToPath(Movie movie, NSString *path, BOOL flatten);
Track GetBiggestTrack(Movie qtMovie, OSType type);

+ (NSArray *)supportedMetadata
{
    return [NSArray arrayWithObjects: @"Album", @"Artist", @"Author", @"Chapter", @"Comment", @"Composer", @"Copyright", @"Creation Date",
			@"Description", @"Director", @"Disclaimer", @"Encoded By", @"Title", @"Genre", @"Host Computer", @"Information", @"Keywords",
			@"Make", @"Model", @"Original Artist", @"Original Format", @"Original Source", @"Performers", @"Producer", @"Product", @"Software",
			@"Special Playback Requirements", @"Track", @"Warning", @"Writer", @"URL Link", @"Edit Date", @"Website", @"Rating", nil];

}

+ (NSArray *)itmsKeys
{
	NSString *g = [NSString stringWithFormat:@"%c", '\251'];
	return [NSArray arrayWithObjects: [g stringByAppendingString:@"ART"],
	@"aART",	[g stringByAppendingString:@"gen"], [g stringByAppendingString:@"day"],	@"desc", [g stringByAppendingString:@"grp"],	@"tvsh",	@"tven",	
	@"tvsn",	@"tves",	@"sonm",	@"soal",	@"soar",	@"soaa",	@"sosn",	@"cpil",	@"disk",	@"trkn",	@"tmpo", 
		[g stringByAppendingString:@"lyr"], @"soco", [g stringByAppendingString:@"wrt"], [g stringByAppendingString:@"cmt"],
		[g stringByAppendingString:@"alb"], [g stringByAppendingString:@"nam"], @"rtng", [g stringByAppendingString:@"too"], @"cprt",
			@"keyw", @"catg", @"ldes", @"purl", @"egid", @"stik", @"pgap", @"purd", @"tvnn", nil];
}

+ (NSArray *)itmsStrings
{
	return [NSArray arrayWithObjects:@"Artist", @"Album Artist",	@"Genre",	@"Year",	@"Description",	@"Grouping",
	@"Show",	@"Episode",	@"Season Number",	@"Episode Number",	@"Sort Name",	@"Sort Album",	@"Sort Artist",	@"Sort Album Artist",
			@"Sort Show",	@"Compliation",	@"Disk Number",	@"Track Number", @"BPM", @"Lyrics", @"Sort Composer", @"Composer", @"Comment",
			@"Album", @"Title", @"Advisory", @"Software", @"Copyright", @"Keywords", @"Category", @"Information", @"Feed URL", @"Ep URL",
			@"Kind", @"Gapless", @"Purchase Date", @"TV Network Name", nil];
}

+ (NSArray *)qtKeys
{
	return [NSArray arrayWithObjects:@"com.apple.quicktime.year", nil];
}

+ (NSArray *)qtStrings
{
	return [NSArray arrayWithObjects:@"Year", nil];
}

+ (NSString *)itmsTypeToString:(NSString *)t
{
	NSArray *ignore = [NSArray arrayWithObjects:@"AllF", @"SelO", @"play", @"aqui", @"ptv ", @"aclo", nil];
	if ([ignore containsObject:t]) return nil;
	
	if ([t hasPrefix:@"com.apple.quicktime."]) {
		t = [[t substringFromIndex:20] capitalizedString];
		if ([t isEqualToString:@"Displayname"]) t = @"Title";
		else if ([t isEqualToString:@"Creationdate"]) t = @"Creation Date";
		if ([t hasPrefix:@"Player."] || [t isEqualToString:@"Version"]) return nil;
		return t;
	}
	//if ([t hasPrefix:@"com.apple.iTunes"]) return nil;

	NSString *g = [NSString stringWithFormat:@"%c", '\251'];
	NSArray *types = [MHMovie itmsKeys];
	int i = [types indexOfObject:t], n = [types count];
	if (i < n && i >= 0) return [[MHMovie itmsStrings] objectAtIndex:i];
	
	
	types = [NSArray arrayWithObjects:
	[NSArray arrayWithObjects:@"Artist", [g stringByAppendingString:@"ART"], nil],
					  [NSArray arrayWithObjects:@"Album Artist", @"aART", nil],
					  [NSArray arrayWithObjects:@"Genre", [g stringByAppendingString: @"gen"], nil],
					  [NSArray arrayWithObjects:@"Year", [g stringByAppendingString:@"day"], nil],
					  [NSArray arrayWithObjects:@"Description", @"desc", nil],
					  [NSArray arrayWithObjects:@"Grouping", [g stringByAppendingString:@"grp"], nil],
					  [NSArray arrayWithObjects:@"Show", @"tvsh", nil],
					  [NSArray arrayWithObjects:@"Episode", @"tven", nil],
					  [NSArray arrayWithObjects:@"Season Number", @"tvsn", nil],
					  [NSArray arrayWithObjects:@"Episode Number", @"tves", nil],
					  [NSArray arrayWithObjects:@"Sort Name", @"sonm", nil],
					  [NSArray arrayWithObjects:@"Sort Album", @"soal", nil],
					  [NSArray arrayWithObjects:@"Sort Artist", @"soar", nil],
					  [NSArray arrayWithObjects:@"Sort Album Artist", @"soaa", nil],
					  [NSArray arrayWithObjects:@"Sort Show", @"sosn", nil],
					  [NSArray arrayWithObjects:@"Compliation", @"cpil", nil],
					  [NSArray arrayWithObjects:@"Disk Number", @"disk", nil],
					  [NSArray arrayWithObjects:@"Track Number", @"trkn", nil],
					  [NSArray arrayWithObjects:@"BPM", @"tmpo", nil],
					  [NSArray arrayWithObjects:@"Lyrics", [g stringByAppendingString:@"lyr"], nil],
					  nil];
	n = [types count];
	
	
	for (i=0;i<n;i++)
		if ([[types objectAtIndex:i] containsObject:t])
			return [[types objectAtIndex:i] objectAtIndex:0];
	NSArray *sup = [MHMovie supportedMetadata];
	n = [sup count];
	for (i=0;i<n;i++) {
		OSType o = theKUserDataTypes[i];
#ifdef __LITTLE_ENDIAN__
		o = NSSwapLong(o);
#endif
		NSString *s = [NSString stringWithCString:(char *)&o length:4];
		if ([s isEqual:t])
			return [sup objectAtIndex:i];
	}
	return t;
}

+ (void)setAsksForMissingFiles:(BOOL)flag
{
    asks = flag;
}

+ (void)setIgnoresMissingFiles:(BOOL)flag
{
    ignores = flag;
}

+ (void)setLanguages:(NSArray *)langs
{
    languages = [[NSArray alloc] initWithArray:langs];
}

+ (NSArray *)blindspotlight
{
	return [NSArray arrayWithObjects:@"Chapter", @"Creation Date", @"Disclaimer", @"Edit Date", @"Encoded By", @"Host Computer", @"Make", @"Model",
			@"Original Artist", @"Product", @"Rating", @"Special Playback Requirements", @"Track", @"URL Link", @"Warning", @"Website", @"Writer", @"Year", nil];
}

- (BOOL)movieShouldTask:(id)movie
{
	if ([[NSApp delegate] respondsToSelector:@selector(movieShouldTask:)])
		return [[NSApp delegate] movieShouldTask:movie];
	return NO;
}

- (void)goToNextChapter
{
	double seconds = [self currentTimeIndexSeconds];
	NSString *lang = nil;
	if ([[self chapters] multilingual]) lang = [self currentLanguage];
	seconds = [[self chapters] nextTimeForTime:seconds language:lang];
	if (seconds == -1) seconds = [self durationSeconds] - 0.5;
	[self setCurrentTimeIndexSeconds:seconds];
}

- (void)goToPrevChapter
{
	double seconds = [self currentTimeIndexSeconds];
	NSString *lang = nil;
	if ([[self chapters] multilingual]) lang = [self currentLanguage];
	double newTime = [[self chapters] prevTimeForTime:seconds language:lang];
	if (seconds - newTime < 1) newTime = [[self chapters] prevTimeForTime:newTime language:lang];
	if (newTime == -1) newTime = [self durationSeconds] - 0.5;
	[self setCurrentTimeIndexSeconds:newTime];
}

- (id)qtKitMovie
{
	return mainMovie;
}

- (Movie)quickTimeMovie
{
	return [self QTMovie];
}

- (BOOL)hasSubs
{
	NSArray *subs = [NSArray arrayWithObjects:@"SRT ", @"SSA ", @"ASS ", @"SPU ", nil];
	int i, n = [self trackCount];
	for (i=0;i<n;i++) {
		NSString *f = [[self trackAtIndex:i] fourCC];
		if ([subs containsObject:[f uppercaseString]]) return YES;
	}
	return NO;
}

- (MHMovie *)initWithURL:(NSURL *)aURL byReference:(BOOL)ref
{
    //NSLog(@"MHMovie initWithURL:byReference");
    //self = [super initWithURL:DUMMYURL byReference:YES];
	alterations = nil;
	subUpdateDate = nil;
	kqfd = 0;
    isOffline = NO;
	wasOffline = NO;
    saveInPlaceUnSafe = NO;
    listNumbers = nil;
    _sortedKeys = nil;
	editedFields = nil;
	//myQTKitMovie = nil;
	statusString = nil;
    previewPic = posterPic = nil;
    if ([aURL isFileURL])
        stringPath = [aURL path];
    else
        stringPath = [aURL absoluteString];
    [stringPath retain];
	NSError *err = nil;
	mainMovie = [[QTMovie movieWithFile:stringPath error:&err] retain];
	if (err) NSLog(@"Error opening movie at %@: %@", stringPath, err);
	myRefMovie = nil; //[MHUtil uniquePath:[[MHUtil supportDir] stringByAppendingPathComponent:[stringPath lastPathComponent]]];
	//[mainMovie writeToFile:myRefMovie withAttributes:nil];
	err = nil;
	//mainMovie = [[QTMovie movieWithFile:myRefMovie error:&err] retain];
	//if (err) NSLog(@"error opening reference movie at %@: %@", myRefMovie, err);
	[(QTMovie *)mainMovie setDelegate:self];
    //mainMovie = [[NSMovie alloc] initWithMovie:[myQTKitMovie quickTimeMovie]];
    if (nil == mainMovie || nil == self) {
        [self release];
        return nil;
    }
    data = [[NSMutableDictionary alloc] init];
    metadata = [[NSMutableDictionary alloc] init];
    info = [[NSMutableDictionary alloc] init];
    tempFiles = nil;
    isChanged = chaptersNeedUpdating = NO;
    //NSAutoreleasePool *pool2 = [[NSAutoreleasePool alloc] init];
    Chapters = [[MHChapters alloc] initFromMovie:self];
    //[pool2 release];
    Menus = nil;
    [data setObject:stringPath forKey:@"File Path"];
    NSString *filename = [stringPath lastPathComponent];
    //if (![filename hasSuffix:@".mov"]) filename = [filename stringByAppendingString:@".mov"];
    [info setObject:filename forKey:@"File Name"];
    [self readMetadata];
    [self gatherData];
    if (![self objectForKey:@"Title"])
        [self setObject:[self stringValue] forKey:@"Title"];
    
	if (![self hasSubs]) [self addSeparateSubs];
	
    return self;
}

- (MHMovie *)initWithURL:(NSURL *)aURL
{
    return (MHMovie *)[self initWithURL:aURL byReference:NO];
}

- (MHMovie *)initWithPath:(NSString *)path
{
    NSURL *aURL;
    if ([path hasPrefix:@"/"])
        aURL = [NSURL fileURLWithPath:path];
    else
        aURL = [NSURL URLWithString:path];
    return [self initWithURL:aURL];
    // open movie at path
}

- (MHMovie *)initWithDefaultsEntry:(NSDictionary *)entry
{
    //NSLog(@"MHMovie initWithDefaultsEntry");
    //NSURL *aURL;
	//[[NSOpenPanel openPanel] runModalForTypes:nil];
	alterations = nil;
	subUpdateDate = nil;
	kqfd = 0;
    isOffline = NO;
	wasOffline = YES;
    saveInPlaceUnSafe = NO;
    listNumbers = nil;
    _sortedKeys = nil;
	editedFields = nil;
    mainMovie = nil;
	statusString = nil;
    //mainQTMovie = nil;
    data = [[NSMutableDictionary alloc] init];
    metadata = [[NSMutableDictionary alloc] init];
    info = [[NSMutableDictionary alloc] init];
    previewPic = posterPic = nil;
    stringPath = [entry objectForKey:@"File Path"];
    if (![[NSFileManager defaultManager] fileExistsAtPath:stringPath]) {
        NSData *fsrefD = [entry objectForKey:@"FSRef"];
        if (fsrefD) {
            UInt8 *newPath = malloc(2048 * sizeof(char));
            FSRef *ref = (FSRef *)[fsrefD bytes];
            OSErr err = FSRefMakePath(ref, newPath, 2048);
            if (!err && newPath) {
                stringPath = [NSString stringWithUTF8String:newPath];
                [data setObject:stringPath forKey:@"File Path"];
            }
            free (newPath);
        }
    }
    [stringPath retain];
	myRefMovie = nil; //[entry objectForKey:@"myRefMovie"];
	//if (!myRefMovie) myRefMovie = [[MHUtil uniquePath:[[MHUtil supportDir] stringByAppendingPathComponent:[entry objectForKey:@"File Name"]]] retain];
    if (![[NSFileManager defaultManager] fileExistsAtPath:stringPath]) {
        if (ignores) {
            [self release];
            return nil;
        }
        if (asks && [self askForMovie]) {
        } else {
            stringPath = [[[NSBundle mainBundle] pathForResource:@"dummy" ofType:@"txt"] retain];
            isOffline = YES;
        }
    }
    
    tempFiles = [entry objectForKey:@"tempFiles"];
    if (tempFiles)
        tempFiles = [NSMutableArray arrayWithArray:tempFiles];
    isChanged = chaptersNeedUpdating = NO;
        
    // open metadata and file data
    
    NSArray *bad = [NSArray arrayWithObjects:@"File Path", @"chapters", @"Seconds Elapsed", @"Seconds", @"Movie Length", @"File Size", @"Original Height", @"Saved Height", @"Original Width", @"Saved Width", @"FPS", @"Video Format", @"Video FourCC", @"Video Vendor", @"Video Width", @"Video Height", @"Video Bitrate", @"Audio Channels", @"Audio Format", @"Audio Sample Rate", @"Audio Bitrate", @"otherLangs", @"menus", @"Cover Image", @"File Name", @"File Creation Date", @"File Modification Date", /*@"Source Name",*/ @"Timecode", @"isChanged", @"Poster", nil];
    //[metadata addEntriesFromDictionary:entry];
    int i, n = [[entry allKeys] count];
    for (i=0;i<n;i++) {
        NSString *key = [[entry allKeys] objectAtIndex:i];
        id value = [entry objectForKey:key];
		if ([key isEqualToString:@"Full Name"]) key = @"Title";
        if (![bad containsObject:key]) {
            if ([value isKindOfClass:[NSString class]])
                [metadata setObject:[NSMutableDictionary dictionaryWithObject:value forKey:key] forKey:@"English"];
            else if ([value isKindOfClass:[NSDictionary class]])
                [metadata setObject:[NSMutableDictionary dictionaryWithDictionary:value] forKey:key];
        }
    }
    NSArray *infos = [NSArray arrayWithObjects:@"File Name", @"Poster", nil];
    NSArray *nots = [NSArray arrayWithObjects:@"chapters", @"otherLangs", @"isChanged", nil];
    n = [bad count];
    for (i=0;i<n;i++) {
        NSString *key = [bad objectAtIndex:i];
        id value = [entry objectForKey:key];
        if (value) {
            if ([infos containsObject:key] && ![info objectForKey:key])
                [info setObject:value forKey:key];
            else if (![nots containsObject:key] && ![data objectForKey:key])
                [data setObject:value forKey:key];
        }
    }
    if ([[entry objectForKey:@"isChanged"] boolValue]) isChanged = YES;
    [data removeObjectForKey:@"Cover Image"];
    NSString *cover = [entry objectForKey:@"Cover Image"];
    if (cover) [self setImageFile:cover];
    
    Chapters = [[MHChapters alloc] init];
    NSArray *chapArray = [entry objectForKey:@"chapters"];
    if (chapArray) {
        [Chapters addObjectsFromArray:chapArray];
        if ([[entry objectForKey:@"multiChaps"] boolValue])
            //[Chapters setMultilingual:YES languages:[self languages]];
            [Chapters setMultilingual:YES languages:nil];
    }
    NSDictionary *menus = [entry objectForKey:@"menus"];
    if (menus)
        Menus = [MHUtil spritesFromPlist:menus];
	
	if (!isOffline && ![self hasSubs]) [self addSeparateSubs];
	if ([entry objectForKey:@"editedFields"])
		editedFields = [[NSMutableArray arrayWithArray:[entry objectForKey:@"editedFields"]] retain];
	
	NSArray *alters = [entry objectForKey:@"alterations"];
	if (alters) {
		NSArray *adds = [[alters objectAtIndex:0] componentsSeparatedByString:@"\n"];
		n = [adds count];
		for (i=0;i<n;i++) {
			NSString *s = [adds objectAtIndex:i];
			if ([s hasPrefix:@"Add File: "])
				[self addMovie:[s substringFromIndex:10]];
			else if ([s hasPrefix:@"Remove Track "])
				[self removeTrackAtIndex:[[s substringFromIndex:13] intValue]];
		}
		n = [alters count];
		if ([self trackCount] < n-1) n = [self trackCount];
		for (i=1;i<n;i++)
			[[self trackAtIndex:i-1] makeAlterations:[alters objectAtIndex:i]];
	}
	
    return self;
}

- (void *)QTMovie
{
	NSString *filePath = [data objectForKey:@"File Path"];
	BOOL hasFile = NO;
	if (filePath && [[NSFileManager defaultManager] fileExistsAtPath:filePath]) hasFile = YES;
    if (!mainMovie || (isOffline && hasFile)) {
		if (hasFile) {
			NSError *err = nil;
			mainMovie = [[QTMovie movieWithFile:filePath error:&err] retain];
			if (err) NSLog(@"error opening movie at %@: %@", filePath, err);
			[(QTMovie *)mainMovie setDelegate:self];
		}
        if (mainMovie) {
            isOffline = NO;
			NSNumber *n = [info objectForKey:@"Poster"];
			//NSLog(@"poster: %@", n);
			if (n) [mainMovie setAttribute:[NSValue valueWithQTTime:QTMakeTime([n longValue], GetMovieTimeScale([mainMovie quickTimeMovie]))] forKey:QTMoviePosterTimeAttribute];
            [self gatherData];
            if ([(MHChapters *)[self chapters] count])
                [[self chapters] addToMovie:self];
			if (![self hasSubs]) [self addSeparateSubs];
			double seconds = [[self objectForKey:@"Seconds Elapsed"] doubleValue];
			if (seconds) {
				QTTime t = [(QTMovie *)mainMovie currentTime];
				t.timeValue = seconds *t.timeScale;
				[(QTMovie *)mainMovie setCurrentTime:t];
			}
        } else {
            isOffline = YES;
            mainMovie = [[QTMovie movieWithURL:DUMMYURL error:nil] retain];
			return [(QTMovie *)mainMovie quickTimeMovie];
        }
    }
    if (chaptersNeedUpdating) {
        chaptersNeedUpdating = NO;
        [[self chapters] addToMovie:self];
    }
    return [(QTMovie *)mainMovie quickTimeMovie];
}

- (void)setMovie:(NSMovie *)movie
{
    [mainMovie release];
	if ([movie isKindOfClass:[QTMovie class]])
		mainMovie = [movie retain];
	else
		mainMovie = [[QTMovie movieWithQuickTimeMovie:[movie QTMovie] disposeWhenDone:NO error:nil] retain];
    [self gatherData];
}
/*
- (id)retain
{
    int before = [self retainCount];
    id me = [super retain];
    int after = [self retainCount];
    NSLog(@"%d retain %d", before, after);
    return me;
}

- (oneway void)release
{
    int before = [self retainCount];
    [super release];
    int after = 0;
    if (before > 1) after = [self retainCount];
    NSLog(@"%d release %d", before, after);
}
*/
- (void) dealloc
{
	[[NSNotificationCenter defaultCenter] removeObserver:self];
	NSArray *a = [NSArray arrayWithObjects:data, metadata, info, listNumbers, Chapters, stringPath, mainMovie, nil];
	int i, n = [a count];
    data = nil;
    metadata = nil;
    info = nil;
    listNumbers = nil;
	Chapters = nil;
    stringPath = nil;
    if (tempFiles) {
        n = [tempFiles count];
        NSFileManager *fm = [NSFileManager defaultManager];
        for (i=0;i<n;i++) {
            [fm removeFileAtPath:[tempFiles objectAtIndex:i] handler:nil];
        }
        [tempFiles release];
    }
	mainMovie = nil;
	n = [a count];
	for (i=0;i<n;i++)
		[[a objectAtIndex:i] performSelector:@selector(release) withObject:nil afterDelay:1];
	
	
	return;  // ever since QTKit, QTMovie idleAllMovies crashes shortly after a MHMovie is released, even if the QTMovie is not released.
    [super dealloc];
}

- (BOOL)addMovie:(NSString *)path
{
	MHMovie *newMovie = [[MHMovie alloc] initWithPath:path];
	if (!newMovie) {
		NSArray *subs = [NSArray arrayWithObjects:@"ssa", @"ass", @"srt", @"sub", @"idx", @"smi", @"sami", nil];
		if (![subs containsObject:[[path pathExtension] lowercaseString]]) return NO;
		NSString *avipath = [self aviPathForSize:NSMakeSize(0, 0)];
		NSString *subPath = [[avipath stringByDeletingPathExtension] stringByAppendingPathExtension:[path pathExtension]];
		[[NSFileManager defaultManager] copyPath:path toPath:subPath handler:nil];
		newMovie = [[MHMovie alloc] initWithPath:avipath];
		[[newMovie trackAtIndex:0] remove];
		if ([newMovie trackCount] < 1) return NO;
	}
	int i = [self trackCount], n = i + [newMovie trackCount];
	SetMovieSelection([newMovie QTMovie], 0, GetMovieDuration([newMovie QTMovie]));
	AddMovieSelection([self QTMovie], [newMovie QTMovie]);
	if (n-i == 1) [[self trackAtIndex:i] makeAlterations:[NSString stringWithFormat:@"Name|%@", [[path lastPathComponent] stringByDeletingPathExtension]]];
	for (;i<n;i++) {
		[[self trackAtIndex:i] setOffset:0];
		[[self trackAtIndex:i] setSource:path];
	}
	[self gatherData];
	if (!alterations) alterations = @"";
	NSString *a = [[alterations stringByAppendingFormat:@"Add File: %@\n", path] retain];
	[alterations release];
	alterations = a;
	[newMovie release];
	
	[[NSNotificationCenter defaultCenter] removeObserver:self]; // we want to be notified, but not multiple/dozens of times
	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(fileWasModified:) name:@"MHModified" object:nil];
	[[NSApp delegate] watchFile:path];
	
	return YES;
}

- (NSURL *)URL
{
    if ([stringPath hasPrefix:@"/"])
        return [NSURL fileURLWithPath:stringPath];
    else
        return [NSURL URLWithString:stringPath];
}

- (BOOL) copyFileMetadata:(NSString *)path
{
	if ([path isKindOfClass:[NSMenuItem class]] && [path representedObject])
		path = [path representedObject];
	MHMovie *o = [[MHMovie alloc] initWithPath:path];
	NSArray *keys = [o allMetadataKeys];
	int i, n = [keys count];
	for (i=0;i<n;i++)
		[self setObject:[o objectForKey:[keys objectAtIndex:i]] forKey:[keys objectAtIndex:i]];
	[o release];
}

- (BOOL)isFileWritable
{
	NSString *aFilePath = [data objectForKey:@"File Path"];
	return [[NSFileManager defaultManager] isWritableFileAtPath:aFilePath];
}

- (NSArray *)subFilesForPath:(NSString *)path
{
	NSArray *sufs = [NSArray arrayWithObjects:@"srt", @"ssa", @"ass", @"idx", @"sub", nil];
	NSString *d = [path stringByDeletingLastPathComponent];
	NSString *pref = [[path lastPathComponent] stringByDeletingPathExtension];
	NSFileManager *fm = [NSFileManager defaultManager];
	NSArray *files = [fm directoryContentsAtPath:d];
	int i, n = [files count];
	NSMutableArray *array = [NSMutableArray array];
	for (i=0;i<n;i++) {
		NSString *f = [files objectAtIndex:i], *e = [[f pathExtension] lowercaseString];
		if ([f hasPrefix:pref] &&
			[sufs containsObject:e])
			[array addObject:[d stringByAppendingPathComponent:f]];
	}
	n = [array count];
	if (!n) return nil;
	BOOL vobsub = NO;
	for (i=0;i<n;i++)
		if ([[[array objectAtIndex:i] lowercaseString] hasSuffix:@".idx"]) vobsub = YES;
	if (!vobsub)
		for (i=n-1;i>=0;i--)
			if ([[[array objectAtIndex:i] lowercaseString] hasSuffix:@".sub"])
				[array removeObjectAtIndex:i];
	return array;
}

-(void)noteSubFrames
{
	NSRect r = NSMakeRect(0,0,0,0);
	int i, n = [self trackCount];
	for (i=0;i<n;i++) {
		if ([[self trackAtIndex:i] subtitleType]) {
			r = [[self trackAtIndex:i] displayFrame];
			break;
		}
	}
	if (r.size.width > 1) {
		[[NSUserDefaults standardUserDefaults] setObject:NSStringFromRect(r) forKey:@"subFrame"];
		[[NSUserDefaults standardUserDefaults] setObject:[NSDate dateWithTimeIntervalSinceNow:0] forKey:@"subFrameDate"];
	}
}

-(BOOL)removeSubs:(NSString *)path
{
	NSArray *types = [NSArray arrayWithObjects:@"SRT ", @"SSA ", @"ASS ", @"SPU ", @"SPU ", nil];
	NSArray *exts = [NSArray arrayWithObjects:@".srt", @".ssa", @".ass", @".idx", @".sub", nil];
	int i, j;
	for (i=[self trackCount]-1; i>0;i--) {
		MHTrack *t = [self trackAtIndex:i];
		if (path) {
			for (j=0;j<[exts count];j++)
				if ([[path lowercaseString] hasSuffix:[exts objectAtIndex:j]])
					if ([[t fourCC] isEqualToString:[types objectAtIndex:j]]) {
						//NSLog(@"removing track %d", i);
						//[t remove];
						[self removeTrackAtIndex:i];
					} else {
						//NSLog(@"not removing track %d", i);
					}

		} else if ([types containsString:[t fourCC]]) {
			//[t remove];
			[self removeTrackAtIndex:i];
		}
	}
}
/*
void socketCallBack2 (CFSocketRef socketref, CFSocketCallBackType type, CFDataRef address, const void *data, void *info)
{
	MHMovie *me = (MHMovie *)info;
	struct kevent event;
	if (kevent(me->kqfd, NULL, 0, &event, 1, NULL) == -1) {
		NSLog(@"could not pick up event.");
		return;
	}
	
		NSString *path = (NSString *)event.udata;
		if (!path) return;
		
	NSLog(@"(%d) detect write: %@", me->kqfd, path);
	
		NSDate *newD = [[[NSFileManager defaultManager] fileAttributesAtPath:path traverseLink:NO] objectForKey:NSFileModificationDate];
		NSTimeInterval gap = [newD timeIntervalSinceNow];
		if (gap < -2) return; // if it's older than 2 seconds
		
		if (![[NSUserDefaults standardUserDefaults] objectForKey:@"always reimport subs"]) {
			NSString *infoText = [NSString stringWithFormat:@"The subtitles file for this movie has just been edited.  Do you want to re-import it?"];
			NSAlert *alert = [NSAlert alertWithMessageText:@"Re-import Subtitles" defaultButton:@"Re-import" alternateButton:@"No" otherButton:nil informativeTextWithFormat:infoText];
			if ([alert respondsToSelector:@selector(setShowsSuppressionButton:)])
				[alert setShowsSuppressionButton:YES];
			int result = [alert runModal];
			if ([alert respondsToSelector:@selector(suppressionButton)]) {
				if ([[alert suppressionButton] state]) {
					if (result == NSAlertDefaultReturn) 
						[[NSUserDefaults standardUserDefaults] setBool:YES forKey:@"always reimport subs"];
					else if (result == NSAlertAlternateReturn)
						[[NSUserDefaults standardUserDefaults] setBool:NO forKey:@"always reimport subs"];
					[[NSUserDefaults standardUserDefaults] synchronize];
				}
			}
			if (result == NSAlertAlternateReturn) return;
		} else if (![[NSUserDefaults standardUserDefaults] boolForKey:@"always reimport subs"]) return;
		
	// delete old event
		//int filefd = open ([path UTF8String], O_RDONLY);
		//EV_SET(&event, filefd, EVFILT_WRITE, EV_DELETE, 0, 0, NULL);
	event.flags = EV_DELETE;
		if (kevent(me->kqfd, &event, 1, NULL, 0, NULL) == -1) NSLog(@"could not delete event");
		
		NSLog(@"Reloading sub after detected write: %@", path);
		[me noteSubFrames];
		[me removeSubs:nil];
		[me addSeparateSubs];
}
*/
- (void) fileWasModified:(NSNotification *)note
{
	NSArray *paths = nil;
	if ([note isKindOfClass:[NSNotification class]]) {
		NSString *path = [note object];
		if ([path isKindOfClass:[NSString class]])
			paths = [NSArray arrayWithObject:[note object]];
		else if ([path isKindOfClass:[NSArray class]])
			paths = path;
		else {
			NSLog(@"fileWasModified: %@", [path className]);
			return;
		}
	//NSString *path = [note object];
	//if (![path isKindOfClass:[NSString class]]) {
	//	NSLog(@"file modified notification object not string but %@", [path className]);
	//	return;
	//}
	} else if ([note isKindOfClass:[NSArray class]]) paths = note;
	else {
		NSLog(@"filewasmodified: %@", [note className]);
		paths = [self subFilesForPath:stringPath];
	}
	// look for added files
	int i;
	NSMutableDictionary *foundFiles = [NSMutableDictionary dictionaryWithCapacity:[paths count]]; // key = file path, value = array of tracks, represented by string lists of the alterations made
	for (i=[self trackCount]-1;i>=0;i--) {
		MHTrack *track = [self trackAtIndex:i];
		NSString *dataRef = [track dataRef];
		if (![dataRef hasPrefix:@"/"]) {
			NSRange r = [dataRef rangeOfString:@"/"];
			if (r.length == 1) dataRef = [dataRef substringFromIndex:r.location];
		}
		if ([paths containsObject:dataRef]) {
			NSMutableArray *list = [foundFiles objectForKey:dataRef];
			if (!list) {
				list = [NSMutableArray array];
				[foundFiles setObject:list forKey:dataRef];
			}
			NSString *alters = [track alterations];
			if (!alters) alters = @"";
			[list insertObject:alters atIndex:0];
			//[track remove];
			[self removeTrackAtIndex:i];
		}
	}
	NSArray *keys = [foundFiles allKeys];
	int n = [keys count];
	for (i=0;i<n;i++) {
		int oldTrackCount = [self trackCount];
		//MHMovie *newMovie = [[MHMovie alloc] initWithPath:[keys objectAtIndex:i]];
		[self addMovie:[keys objectAtIndex:i]];
		int newTrackCount = [self trackCount];
		int j, m = [[foundFiles objectForKey:[keys objectAtIndex:i]] count] + oldTrackCount;
		if (newTrackCount < m) m = newTrackCount;
		for (j=oldTrackCount;j<m;j++)
			[[self trackAtIndex:j] makeAlterations:[[foundFiles objectForKey:[keys objectAtIndex:i]] objectAtIndex:j-oldTrackCount]];
		/*SetMovieSelection([newMovie QTMovie], 0, GetMovieDuration([newMovie QTMovie]));
		AddMovieSelection([self QTMovie], [newMovie QTMovie]);
		for (j=oldTrackCount;j<[self trackCount];j++) {
			[[self trackAtIndex:j] setOffset:0];
			[[self trackAtIndex:j] setSource:[keys objectAtIndex:i]];
			NSString *alters = [[foundFiles objectForKey:[keys objectAtIndex:i]] objectAtIndex:j-oldTrackCount];
			[[self trackAtIndex:j] makeAlterations:alters];
		}*/
	}
	
	
	// look for subs
	NSString *prefix = [[stringPath lastPathComponent] stringByDeletingPathExtension];
	n = [paths count];
	for (i=0;i<n;i++) {
		if ([[[paths objectAtIndex:i] lastPathComponent] hasPrefix:prefix]) {
			//NSDate *newD = [[[NSFileManager defaultManager] fileAttributesAtPath:[paths objectAtIndex:i] traverseLink:NO] objectForKey:NSFileModificationDate];
			//if ([subUpdateDate compare:newD] != NSOrderedDescending)
				break; // our last update was not more recent than the mod date | edit: this check was moved upstream
		}
	}
	if (i == n) return; // didn't find a modified file
	
	NSLog(@"Reloading sub after detected write: %@", [paths objectAtIndex:i]);
	
	n = [self trackCount];
	NSMutableArray *alters = [NSMutableArray array];
	for (i=0;i<n;i++) {
		if ([[self trackAtIndex:i] subtitleType])
			[alters addObject: [[self trackAtIndex:i] alterations]];
	}
	
	[self removeSubs:nil];
	if ([NSApp respondsToSelector:@selector(setPresentationOptions:)]) // 10.6 and later cause subs to lose transparency on previously watched QTMovie objects. this is a >=10.6 method
		[self performSelectorOnMainThread:@selector(reloadMovie) withObject:nil waitUntilDone:YES];
	[self performSelectorOnMainThread:@selector(addSeparateSubs) withObject:nil waitUntilDone:YES];
	int j = 0;
	n = [self trackCount];
	for (i=0;i<n;i++) {
		if ([[self trackAtIndex:i] subtitleType] && j < [alters count]) {
			[[self trackAtIndex:i] makeAlterations:[alters objectAtIndex:j]];
			j++;
		}
	}
	//subUpdateDate = [[NSDate dateWithTimeIntervalSinceNow:0] retain];
	
}

- (void)reloadMovie
{
	NSError *e = nil;
	//Movie q = [mainMovie quickTimeMovie];
	//QTMovie *m = [[QTMovie alloc] initWithQuickTimeMovie:q disposeWhenDone:NO error:&e];
	QTMovie *m = [[QTMovie movieWithData:[mainMovie movieFormatRepresentation] error:&e] retain];
	if (e) NSLog(@"error in reloadMovie: %@", e);
	[m setDelegate:self];
	[mainMovie performSelector:@selector(release) withObject:nil afterDelay:10];
	mainMovie = m;
}

- (NSString *)aviPathForSize:(NSSize)size
{
	NSFileManager *fm = [NSFileManager defaultManager];
	NSString *d = [[MHUtil supportDir] stringByAppendingPathComponent:@"subserve"];
	//[fm removeFileAtPath:d handler:nil];
	if (![fm fileExistsAtPath:d])
		[fm createDirectoryAtPath:d attributes:nil];
	
	if (size.width < 1 || size.height < 1) size = [self displayFrame].size;
	/*if ([[[NSUserDefaults standardUserDefaults] objectForKey:@"subFrameDate"] timeIntervalSinceNow] > -2) {
		NSString *val = [[NSUserDefaults standardUserDefaults] objectForKey:@"subFrame"];
		if (val) {
			NSRect r = NSRectFromString(val);
			if (r.size.width > 1 && r.size.height > 1) {
				size = r.size;
				frame = r;
			}
		}
	}*/
	// create avi
	NSString *avipath = [NSString stringWithFormat:@"%@.avi", NSStringFromSize(size)];
	//path = [[[[data objectForKey:@"File Path"] lastPathComponent] stringByDeletingPathExtension] stringByAppendingString:@".avi"];
	avipath = [d stringByAppendingPathComponent:avipath];
	if (![fm fileExistsAtPath:avipath]) {
		NSString *iPath = [[NSBundle mainBundle] pathForResource:@"Add" ofType:@"png"];
		if (iPath) {
			NSTask *myTask = [[NSTask alloc] init];
			[myTask setLaunchPath:[[NSBundle mainBundle] pathForResource:@"qtFlatten" ofType:nil]];
			[myTask setArguments:[NSArray arrayWithObjects:iPath, avipath, @"-w", [NSString stringWithFormat:@"%d", (int)size.width], @"-h", [NSString stringWithFormat:@"%d", (int)size.height], nil]];
			[myTask launch];
			[myTask waitUntilExit];
		}
	}
	// remove old subs
	NSArray *found = [fm directoryContentsAtPath:d];
	//NSString *target = [[[path lastPathComponent] stringByDeletingPathExtension] stringByAppendingPathExtension:@"avi"];
	int i, n = [found count];
	for (i=0;i<n;i++) {
		//if ([[found objectAtIndex:i] isEqualToString:target]) {	NSLog(@"not removing %@", target);	} else
		if (![[found objectAtIndex:i] hasSuffix:@".avi"]) {
			//NSLog(@"removing %@ (%@)", [found objectAtIndex:i], target);
			[fm removeFileAtPath:[d stringByAppendingPathComponent:[found objectAtIndex:i]] handler:nil];
		}
	}
	return avipath;
}

- (BOOL) addSeparateSubs:(NSArray *)files  forceSize:(NSSize)size
{
	NSString *path = [data objectForKey:@"File Path"];
	if (!path) return NO;
	if (!files) files = [self subFilesForPath:path];
	//NSLog(@"add subs: %@", files);
	if (![files count]) return NO;
	
	NSRect frame = [self displayFrame];
	if (size.width && size.height) frame.size = size;
	else size = frame.size;
	
	NSFileManager *fm = [NSFileManager defaultManager];
	
	NSString *avipath = [self aviPathForSize:size];
	NSString *target = [[[path lastPathComponent] stringByDeletingPathExtension] stringByAppendingPathExtension:@"avi"];
	
	// subs file
	/*	if (!kqfd) {
		kqfd = kqueue ();
		NSLog(@"%@ - %d", stringPath, kqfd);
	}*/
	int i, n = [files count];
	target = [[target lastPathComponent] stringByDeletingPathExtension];
	if (n) {
		[[NSNotificationCenter defaultCenter] removeObserver:self]; // we want to be notified, but not multiple/dozens of times
		[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(fileWasModified:) name:@"MHModified" object:nil];
	}
	for (i=0;i<n;i++) {
		NSMutableString *name = [NSMutableString stringWithString:[[files objectAtIndex:i] lastPathComponent]];
		[name replaceOccurrencesOfString:target withString:NSStringFromSize(size) options:nil range:NSMakeRange(0, [name length])];
		path = [[avipath stringByDeletingLastPathComponent] stringByAppendingPathComponent:name];
		[fm copyPath:[files objectAtIndex:i] toPath:path handler:nil];
		
		//if (
		[[NSApp delegate] watchFile:[files objectAtIndex:i]];
		//) {
			//NSString *noteName = [NSString stringWithFormat:@"MHModified %@", [files objectAtIndex:i]];
		
		//}
		
		
		/*/ listen for changes
		int dirfd = open ([[files objectAtIndex:i] UTF8String], O_RDONLY);
		struct kevent fevent;
		EV_SET (&fevent, dirfd, EVFILT_WRITE, EV_ADD | EV_CLEAR | EV_ENABLE, 0, 0, [[files objectAtIndex:i] copy]);
		if (kevent(kqfd, &fevent, 1, NULL, 0, NULL) == -1) NSLog(@"could not kevent watching %@.", [files objectAtIndex:i]);*/
	}/*
	CFSocketContext context = { 0, self, NULL, NULL, NULL };
	runLoopSocket = CFSocketCreateWithNative(NULL, kqfd, kCFSocketReadCallBack, socketCallBack, &context);
	if (!runLoopSocket) NSLog(@"could not CFSocketCreateWithNative");
	CFRunLoopSourceRef rls = CFSocketCreateRunLoopSource(NULL, runLoopSocket, 0);
	if (!rls) NSLog(@"could not create a run loop source");
	CFRunLoopAddSource(CFRunLoopGetCurrent(), rls, kCFRunLoopDefaultMode);
	CFRelease (rls);*/
	
	
	//if (![fm fileExistsAtPath:avipath])
	//	avipath = [[NSBundle mainBundle] pathForResource:@"subserve" ofType:@"avi"];
	
	//unsigned long long fsize = [[fm fileAttributesAtPath:avipath traverseLink:NO] fileSize];
	
	//path = [d stringByAppendingPathComponent:[[[[data objectForKey:@"File Path"] lastPathComponent] stringByDeletingPathExtension] stringByAppendingString:@".avi"]];
	//[fm copyPath:avipath toPath:path handler:nil];
	//i=0;
	//while([[fm fileAttributesAtPath:path traverseLink:NO] fileSize] < fsize) {
	//	i++;
	//}
	//if (i) NSLog(@"waited %d for file to copy", i);
	
	MHMovie *m = [[MHMovie alloc] initWithPath:avipath];
	if (!m) {
		NSLog(@"could not open subserve");
		return NO;
		m = [[MHMovie alloc] initWithPath:avipath];
	}
	BOOL success = NO;
	if ([m trackCount] > 1) {
		success = YES;
		NSRect newRect = [m displayFrame];		
		//NSLog(@"newRect: %@", NSStringFromRect(newRect));
		int oldCount = [self trackCount];
		TimeValue oldDur = GetMovieDuration([self QTMovie]);
		[[m trackAtIndex:0] remove]; // remove the dummy avi video track
		
		//NSRect r1 = [self displayFrame];
		//NSLog(@"(%g, %g) %g x %g", r1.origin.x, r1.origin.y, r1.size.width, r1.size.height);
		SetMovieSelection([m QTMovie], 0, GetMovieDuration([m QTMovie]));
		SetMovieSelection([self QTMovie], 0, 0);
		AddMovieSelection([self QTMovie], [m QTMovie]);
		/*
		QTMovie *q = [m qtKitMovie];
		QTTime noTime = QTMakeTime(0, [[self qtKitMovie] timeScale]);
		QTTimeRange fullRange = QTMakeTimeRange(noTime, [q duration]);
		[[self qtKitMovie] setAttribute:[NSNumber numberWithBool:YES] forKey:QTMovieEditableAttribute];
		n = [[q tracks] count];
		for (i=0;i<n;i++)
			[[self qtKitMovie] insertSegmentOfTrack:[[q tracks] objectAtIndex:i] timeRange:fullRange atTime:noTime];
		*/
		/*
		MHTrack *newTrack = [self trackAtIndex:[self trackCount]-1];
		NSRect r2 = [newTrack displayFrame];
		float ratio = r2.size.width / r2.size.height;
		r1.origin.y -= 10;
		r1.size.height -= 10;
		// now fit r2 into r1
		if (r2.size.height > r1.size.height) {
			r2.size.height = r1.size.height;
			r2.size.width = r2.size.height * ratio;
		}
		if (r2.size.width > r1.size.width) {
			r2.size.width = r1.size.width;
			r2.size.height = r2.size.width / ratio;
		}
		r2.origin.x = (r1.size.width - r2.size.width) / 2;
		r2.origin.y = r1.size.height;
		
		[newTrack setDisplayFrame:r2];
		
		n = [self trackCount];
		for (i=oldCount;i<n;i++)
			//if ([[[self trackAtIndex:i] fourCC] isEqualToString:@"SPU "])
				[[self trackAtIndex:i] setDisplayFrame:r2];
		*/
		SetMovieSelection([self QTMovie], oldDur, GetMovieDuration([self QTMovie]) - oldDur);
		ClearMovieSelection([self QTMovie]);
		n = [self trackCount];
		for (i=oldCount;i<n;i++) {
			//[[self trackAtIndex:i] setGraphicsMode:adMax]; // in >=10.6 in movies that have already played on screen, sub backgrounds draw black :(
			//if (![avipath hasSuffix:@"subserve.avi"]) frame.size = [[self trackAtIndex:i] displayFrame].size;
			//[[self trackAtIndex:i] setDisplayFrame:frame];
		}
	}
	[m release];
	return success;
}

- (BOOL) addSeparateSubs
{
	if ([self addSeparateSubs:[self subFilesForPath:[data objectForKey:@"File Path"]]]) return YES;
	//[self performSelector:@selector(addSeparateSubs) withObject:nil afterDelay:1];
	return NO;
}

- (BOOL) addSeparateSubs:(NSArray *)files
{
	return [self addSeparateSubs:files forceSize:NSMakeSize(0, 0)];
}

- (BOOL) addSeparateAudio
{
    if ([self trackCount] > 1) return NO;
    NSString *dirPath = [[data objectForKey:@"File Path"] stringByDeletingLastPathComponent];
    if (!dirPath) return NO;
    NSArray *files = [[NSFileManager defaultManager] directoryContentsAtPath:dirPath];
    int i, n = [files count];
    BOOL success = NO;
    NSArray *imageSufs = [NSArray arrayWithObjects:@".jpg", @".jpeg", @".gif", @".tif", @".tiff", @".pct", @".pict", @".png", nil];
	NSString *scrpt = nil;
    for (i=0;i<n;i++) {
        NSString *newPath = [dirPath stringByAppendingPathComponent:[files objectAtIndex:i]];
        if (![newPath isEqualToString:[data objectForKey:@"File Path"]]) {
            if ([newPath hasExtension:[NSArray arrayWithObjects:@"mov", @"mp3", @"m4a", nil]]) {
            //if ([[newPath lowercaseString] hasSuffix:@".mov"] || [[newPath lowercaseString] hasSuffix:@".mp3"] || [[newPath lowercaseString] hasSuffix:@".m4a"]) {
                MHMovie *newMovie = [[MHMovie alloc] initWithPath:newPath];
                if ([newMovie trackCount] == 1 && [[newMovie trackAtIndex:0] isSoundType]) {
                    SetMovieSelection([newMovie QTMovie], 0, GetMovieDuration([newMovie QTMovie]));
                    AddMovieSelection([self QTMovie], [newMovie QTMovie]);
                    [[self trackAtIndex:[self trackCount]-1] setName:[[newPath lastPathComponent] extensionRemoved]];
					[[self trackAtIndex:[self trackCount]-1] setLanguage:[[newPath lastPathComponent] extensionRemoved]];
                }
                [newMovie release];
                success = YES;
            } else if (![(MHChapters *)Chapters count] && [[newPath lowercaseString] hasSuffix:@".txt"]) {
                [Chapters parseTextFile:[NSURL fileURLWithPath:newPath]];
                success = YES;
			} else if ([[newPath lowercaseString] hasSuffix:@".mhscript"]) {
				scrpt = [NSString stringWithContentsOfFile:newPath];
            } else {
                int i, n = [imageSufs count];
                for (i=0;i<n;i++)
                    if ([[newPath lowercaseString] hasSuffix:[imageSufs objectAtIndex:i]]) {
                        [self setImageFile:newPath];
                        break;
                    }
            }
        }
    }
	if (scrpt) [self mhscript:scrpt];
	
    if (success) [self gatherData];
    chaptersNeedUpdating = YES;
    return success;
}

- (void) mhscript:(NSString *)input
{
	NSArray *list = [input componentsSeparatedByCharactersInSet:[NSCharacterSet characterSetWithCharactersInString:@"\r\n"]];
	int i, n = [list count];
	for (i=0;i<n;i++) {
		NSString *line = [list objectAtIndex:i];
		NSArray *words = [line componentsSeparatedByString:@"}{"];
		int k = [words count];
		if (!k) continue;
		NSString *subject = nil, *verb = nil, *object = nil;
		if ([[words objectAtIndex:0] isEqual:@"track"]) {
			if (k > 1) {
				int index = [[words objectAtIndex:1] intValue];
				if (index >= 0 && index < [self trackCount]) subject = [self trackAtIndex:index];
				if (k > 2) verb = [words objectAtIndex:2];
				if (k > 3) object = [words objectAtIndex:3];
				if ([verb isEqual:@"name"] && object) [subject setName:object];
				if ([verb isEqual:@"language"] && object) [subject setLanguage:object];
			}
		} else if ([[words objectAtIndex:0] isEqual:@"create alternate group"]) {
			int i, n = [self trackCount];
			NSMutableArray *selected = [NSMutableArray array];
			for (i=0;i<n;i++)
				if ([[self trackAtIndex:i] isSoundType])
					[selected addObject:[self trackAtIndex:i]];
			n = [selected count];
			if (n > 1) {
				MHTrack *lastTrack = nil, *thisTrack;
				for (i=0;i<n;i++) {
					thisTrack = [selected objectAtIndex:i];
					[lastTrack setAlternate:thisTrack];
					lastTrack = thisTrack;
				}
				[lastTrack setAlternate:[selected objectAtIndex:0]];
			}
		} else if ([[words objectAtIndex:0] isEqual:@"metadata"]) {
			if (k > 1) verb = [words objectAtIndex:1];
			if (k > 2) object = [words objectAtIndex:2];
			if (verb && object) [self setObject:object forKey:verb];
		}
	}
}

- (BOOL) containsString:(NSString *)searchString
{
    //NSLog(@"MHMovie containsString");
    if (nil == searchString) return NO;
    NSArray *cats = [metadata allValues];
    int i, n = [cats count];
    for (i=0;i<n;i++) {
        NSDictionary *cat = [cats objectAtIndex:i];
        if ([cat isKindOfClass:[NSDictionary class]]) {
            int j, m = [[cat allValues] count];
            for (j=0;j<m;j++)
                if ([[[cat allValues] objectAtIndex:j] rangeOfString:searchString options:NSCaseInsensitiveSearch].length)
                    return YES;
        }
    }
    if (nil != Chapters)
        if ([Chapters containsString:searchString])
            return YES;
    return NO;
}

- (BOOL) saveToPath:(NSString *)path flatten:(BOOL)f
{
    //NSLog(@"MHMovie saveToPath");
    NSString *newPath;
    BOOL isDir;
    int suffix = 1;
    
    newPath = path;
    while ([[NSFileManager defaultManager] fileExistsAtPath:newPath isDirectory:&isDir])
        newPath = [NSString stringWithFormat:@"%@ %d", path, suffix++];
    
    NSUserDefaults *defs = [NSUserDefaults standardUserDefaults];
    NSString *textDump = [defs objectForKey:@"textDump"];
    /*
    if (!f) {
        [defs setObject:newPath forKey:@"textDump"];
        //[defs removeObjectForKey:@"textDump"];
        [Chapters addToMovie:self];
    }
    */
    [self updateQTMovie];
    
    BOOL wrote = WriteMovieToPath([self QTMovie], newPath, f);
	[self setChaptersNeedUpdating];
	if (!wrote)
        return NO;
    if (textDump)
        [defs setObject:textDump forKey:@"textDump"];
    
    // Custom Icon
    //IconFamily *icon = nil;
    NSString *iconPath = [info objectForKey:@"Cover Image"];
    if (nil == iconPath)
        return YES; 
    /*
    if ([iconPath hasPrefix:@"icns:"])
        icon = [IconFamily iconFamilyWithContentsOfFile:[iconPath substringFromIndex:5]];
    
    else if ([iconPath hasPrefix:@"icon:"])
        icon = [IconFamily iconFamilyWithIconOfFile:[iconPath substringFromIndex:5]];
        
    else if ([iconPath hasPrefix:@"frame:"]) {
        TimeValue tv = [[iconPath substringFromIndex:6] intValue];
        PicHandle picHandle = GetMoviePict([self QTMovie], tv);
        NSImage *image = nil;
        image = [MHUtil picHandleToImage:picHandle];
        if (image != nil)
            icon = [IconFamily iconFamilyWithThumbnailsOfImage:image];
            
    } else      */
        NSImage *image = [[NSImage alloc] initWithContentsOfFile:iconPath];
        
    if (image != nil)
        [[NSWorkspace sharedWorkspace] setIcon:image forFile:newPath options:NSExclude10_4ElementsIconCreationOption];
    
	[image release];
    return YES;
}

- (BOOL) saveInPlace
{
	[self updateQTMovie];
	
	if ([[self qtKitMovie] canUpdateMovieFile]) {
		[Chapters setIgnoresTempFile:YES];
		if (![Chapters addToQTMovie:[[self qtKitMovie] quickTimeMovie]])
			NSLog(@"failed to add chapters");
		[Chapters setIgnoresTempFile:NO];
		
		return [[self qtKitMovie] updateMovieFile];
	}
	NSRunAlertPanel(@"Save-In-Place not supported", @"This file is not compatible with saving in-place.  Please choose a different save style", @"OK", nil, nil);
    return NO;
    FSRef fsRef;
    FSSpec fsSpec;
    OSErr err = noErr;
    short resRefNum;
    
    //[self updateQTMovie];
    err = FSPathMakeRef((const UInt8 *)[stringPath fileSystemRepresentation], &fsRef, NULL /* isDirectory */);
    if (err != noErr)
        NSLog(@"failed to make FSRef  error: %ld", err);
    err = FSGetCatalogInfo(&fsRef, kFSCatInfoNone, NULL /* catalogInfo */, NULL /* outName */, &fsSpec, NULL /* parentRef */);
    if (err != noErr)
        NSLog(@"failed to make FSSpec  error: %ld", err);
    FSSpecPtr movieFile = &fsSpec;
    err = OpenMovieFile(movieFile, &resRefNum, fsWrPerm);
    if (err != noErr)
        NSLog(@"failed to open movie file  error: %ld", err);
    Movie fMovie = nil;
    short resID = movieInDataForkResID;
    Str255 resName;
    err = NewMovieFromFile(&fMovie, resRefNum, &resID, resName, newMovieActive, NULL);
    if (err != noErr)
        NSLog(@"failed to create movie  error: %ld", err);
    
    // metadata
    //[self updateQTMovie];
    err = CopyMovieUserData([self QTMovie], fMovie, kQTCopyUserDataReplace);
    if (err != noErr)
        NSLog(@"failed to copy user data.  error: %ld", err);
    
    
    // chapters - remove dependency on textDump file.
    [Chapters setIgnoresTempFile:YES];
    if (![Chapters addToQTMovie:fMovie])
        NSLog(@"failed to add chapters");
    [Chapters setIgnoresTempFile:NO];
	
    // tracks, 4CCs, etc -- doesn't seem to work
    //int trackCount = GetMovieTrackCount(fMovie);
    //int i;
    
    //for (i=0;i<trackCount;i++)
    //    DisposeMovieTrack(GetMovieIndTrack(fMovie, 1));
    
    //SetMovieSelection([self QTMovie], 0, GetMovieDuration([self QTMovie]));
    //AddMovieSelection(fMovie, [self QTMovie]);
    
    err = UpdateMovieResource(fMovie, resRefNum, movieInDataForkResID, NULL);
    if (err != noErr)
        NSLog(@"failed to update movie resource.  eror: %ld", err);
    DisposeMovie(fMovie);
    CloseMovieFile(resRefNum);
    return YES;
}

- (void)cancelSave
{
    //NSLog(@"cancel save");
    [NSThread exit];
}

- (MHTrack *)trackAtIndex:(int)index
{
    Track t = GetMovieIndTrack([self QTMovie], index + 1);
    if (!t) return nil;
    return [[[MHTrack alloc] initWithTrack:t] autorelease];
}

- (int)trackCount
{
    return GetMovieTrackCount([self QTMovie]);
}

- (void)removeTrackAtIndex:(int)i
{
	[[self trackAtIndex:i] remove];
	if (!alterations) alterations = @"";
	NSString *a = [[alterations stringByAppendingFormat:@"Remove Track %d\n", i] retain];
	[alterations release];
	alterations = a;
	[self gatherData];
}

- (NSArray *)tracksOfType:(NSString *)type
{
    if (!type) return nil;
    NSMutableArray *array = [NSMutableArray array];
    int i, n = [self trackCount];
    for (i=0;i<n;i++) {
        MHTrack *track = [self trackAtIndex:i];
        if ([[track type] isEqualToString:type])
            [array addObject:track];
    }
    return array;
}

- (NSArray *)languages
{
    NSMutableArray *array = [NSMutableArray array];
    int i, n = [self trackCount];
    for (i=0;i<n;i++) {
        MHTrack *track = [self trackAtIndex:i];
        NSString *l = [track language];
        if (![array containsObject:l])
            [array addObject:l];
    }
    return array;
}

- (NSString *)currentLanguage
{
    int i, n = [self trackCount];
    for (i=0;i<n;i++) {
        MHTrack *track = [self trackAtIndex:i];
        if ([track alternate] && [track enabled])
            return [track language];
    }
    for (i=0;i<n;i++) {
        MHTrack *track = [self trackAtIndex:i];
        if ([track enabled])
            return [track language];
    }
    return @"Unspecified";
}

- (NSRect)displayFrame
{
    NSRect rect = NSMakeRect(0,0,0,0);
    int i, n = [self trackCount];
	float x, y;
    BOOL first = YES;
    for (i=0;i<n;i++) {
        MHTrack *track = [self trackAtIndex:i];
        if (![track isSoundType] && [track enabled]) {
            NSRect r = [track displayFrame];
            if (first || r.origin.x < rect.origin.x) {
				x = NSMaxX(rect);
                rect.origin.x = r.origin.x;
				rect.size.width = x - r.origin.x;
			}
            if (first || r.origin.y > rect.origin.y) {
				y = rect.origin.y - rect.size.height;
                rect.origin.y = r.origin.y;
				rect.size.height = r.origin.y - y;
			}
            if (first || NSMaxX(r) > NSMaxX(rect)) {
                rect.size.width = NSUnionRect(r, rect).size.width;
			}
            if (first || r.origin.y - r.size.height < rect.origin.y - rect.size.height)
                rect.size.height = rect.origin.y - (r.origin.y - r.size.height);
            first = NO;
        }
    }
    return rect;
}

- (NSString *)summary
{
    //NSLog(@"MHMovie summary");
    NSArray *keys = [metadata allKeys];
    int i, n = [keys count];
    NSString *key, *value;
    NSMutableString *output = [NSMutableString stringWithCapacity:1000];
    for (i=0;i<n;i++) {
        key = [keys objectAtIndex:i];
        value = [[metadata objectForKey:key] objectForKey:@"English"];
        [output appendString:[NSString stringWithFormat:@"%@: %@\n\n", key, value]];
    }
    return output;
}

- (NSString *)data
{
    //NSLog(@"MHMovie data");
    NSMutableString *output = [NSMutableString stringWithCapacity:1000];
    [output appendString:[NSString stringWithFormat:@"%@\n%@\n%.2f MB\n%@\n\n", [info objectForKey:@"File Name"], [data objectForKey:@"File Path"], [[data objectForKey:@"File Size"] doubleValue], [data objectForKey:@"Movie Length"]]];
    NSString *vCodec = [data objectForKey:@"Video Format"],
            *vFCC = [data objectForKey:@"Video FourCC"],
            *vVendor = [data objectForKey:@"Video Vendor"],
            *fps = [[data objectForKey:@"FPS"] stringValue],
            *vBitrate = [[data objectForKey:@"Video Bitrate"] stringValue],
            *w = [[data objectForKey:@"Video Width"] stringValue],
            *h = [[data objectForKey:@"Video Height"] stringValue],
            *aCodec = [data objectForKey:@"Audio Format"],
            *aBitrate = [[data objectForKey:@"Audio Bitrate"] stringValue],
            *aSampleRate = [[data objectForKey:@"Audio Sample Rate"] stringValue];
	[output appendString:@"Video:\n"];
	if ([vCodec length]) [output appendFormat:@"\"%@,\" ", vCodec];
	if ([vFCC isEqualToString:@"avc1"]) vFCC = @"avc1 (h264)";
	if ([vFCC length]) [output appendFormat:@"%@, ", vFCC];
	if ([vVendor length]) [output appendFormat:@"%@, ", vVendor];
	if ([fps intValue]) [output appendFormat:@"%@ FPS, ", fps];
	if ([vBitrate intValue]) [output appendFormat:@"%@ KB/s, ", vBitrate];
	[output appendFormat:@"%@ x %@\n\nAudio:\n", w, h];
	if ([aCodec isEqualToString:@"none"]) {
		[output appendString:@"No Audio."];
	} else {
		if ([aCodec length]) [output appendFormat:@"%@, ", aCodec];
		if ([aBitrate intValue]) [output appendFormat:@"%@ Kb/s, ", aBitrate];
		if ([aSampleRate intValue]) [output appendFormat:@"%@ Hz, ", aSampleRate];
		
		//[output appendString:[NSString stringWithFormat:@"Video:\n%@, %@, %@, %@ FPS, %@ KB/s, %@ x %@\n\nAudio:\n%@, %@ Kb/s, %@ Hz", vCodec, vFCC, vVendor, fps, vBitrate, w, h, aCodec, aBitrate, aSampleRate]];
		NSNumber *channels = [data objectForKey:@"Audio Channels"];
		if ([channels intValue] == 1)
			[output appendString:@"mono"];
		else if ([channels intValue] == 2)
			[output appendString:@"stereo"];
		else
			[output appendString:[NSString stringWithFormat:@"%.1f channels", [channels floatValue]]];
	}
    return output;
}

- (NSString *)stringValue
{
    //NSLog(@"MHMovie stringValue");
    NSString *title = [self objectForKey:@"Title" language:@"English"];
	if (!title) title = [self objectForKey:@"Title"];
    if (title) return title;
    if (nil == title)
        title = [info objectForKey:@"File Name"];
    if (nil == title)
        title = [[data objectForKey:@"File Path"] lastPathComponent];
    if (nil == title)
        title = [stringPath lastPathComponent];
    return [title extensionRemoved];
}

- (NSString *)description
{
    return [self stringValue];
}

- (TimeValue)duration
{
    return GetMovieDuration([self QTMovie]);
}

- (double)durationSeconds
{
    return (double)GetMovieDuration([self QTMovie]) / (double)GetMovieTimeScale([self QTMovie]);
}

- (NSString *)languageOfPictureResource:(pictureResource)r
{
    NSString *description = [NSString stringWithCString:r.description];
    NSRange range = [description rangeOfString:@":"];
    if (range.length) return [description substringToIndex:range.location];
    if (-1 != [tLanguages stringToNumber:description]) return description;
    return @"English";
}

- (BOOL) hasImage
{
	long count = CountUserDataType(GetMovieUserData([self QTMovie]), 'APIC');
	if (count) return YES;
	QTMetaDataRef qref = nil;
	OSErr err = QTCopyMovieMetaData([self QTMovie], &qref);
	QTMetaDataItem item = kQTMetaDataItemUninitialized;
	OSType key = kQTMetaDataCommonKeyArtwork;
	err = QTMetaDataGetNextItem(qref, kQTMetaDataStorageFormatiTunes, item, kQTMetaDataKeyFormatCommon, &key, sizeof(key), &item);
	if (item != kQTMetaDataItemUninitialized) return YES;
	return NO;
}

- (NSImage *)image
{
	NSImage *image = [self imageForLanguage:[self currentLanguage]];
	if (image) return image;
    NSArray *langs = [tLanguages stringLangs];
	int i, n = [langs count];
	for (i=0;i<n;i++) {
		image = [self imageForLanguage:[langs objectAtIndex:i]];
		if (image) return image;
	}
	return nil;
}

/*  
    V2.3-2.4 encoding
    Text encoding      $xx
    MIME type          <text string> $00
    Picture type       $xx
    Description        <text string according to encoding> $00 (00)
    Picture data       <binary data> */

- (NSImage *)imageForLanguage:(NSString *)language
{
	if ([stringPath isEqualToString:[[NSBundle mainBundle] pathForResource:@"dummy" ofType:@"txt"]]) return nil;
	
	/////////////////////////////// APIC
    Handle h = NewHandleClear(0);
    NSImage *image = nil;
	OSErr err = 0;
    long i, count = CountUserDataType(GetMovieUserData([self QTMovie]), 'APIC');
    for (i=1;i<=count;i++) {
        err = GetUserData(GetMovieUserData([self QTMovie]), h, 'APIC', i);
        if (!err) {
            NSString *lang = nil;
            pictureResource r = parseTag(h);
            NSString *description = [NSString stringWithCString:r.description];
            NSRange range = [description rangeOfString:@":"];
            if (range.length) {
                lang = [description substringToIndex:range.location];
                description = [description substringFromIndex:range.location + 1];
            } else if (-1 != [tLanguages stringToNumber:description]) lang = description;
            else lang = @"English";
            
            if (!image || [language isEqualToString:lang]) {
                image = [[[NSImage alloc] initWithData:[NSData dataWithBytes:*(r.pictureData) length:GetHandleSize(r.pictureData)]] autorelease];
                [image setName:description];
            }
        }
    }
    DisposeHandle(h);	
	/////////////////////////////////  QT7
	QTMetaDataRef qref = nil;
	err = QTCopyMovieMetaData([self QTMovie], &qref);
	QTMetaDataItem item = kQTMetaDataItemUninitialized;
	OSType key = kQTMetaDataCommonKeyArtwork;
	err = QTMetaDataGetNextItem(qref, kQTMetaDataStorageFormatiTunes, item, kQTMetaDataKeyFormatCommon, &key, sizeof(key), &item);
	ByteCount size;
	char *artData = nil;
	err = QTMetaDataGetItemValue(qref, item, nil, 0, &size);
	artData = malloc(size);
	err = QTMetaDataGetItemValue(qref, item, artData, size, &size);
	QTPropertyValueType dataType;
	err = QTMetaDataGetItemProperty(qref, item, kPropertyClass_MetaDataItem, kQTMetaDataItemPropertyID_DataType, sizeof(dataType), &dataType, nil);
	NSData *d = [NSData dataWithBytes:artData length:size];
	NSImage *image2 = [[[NSImage alloc] initWithData:d] autorelease];
	switch (dataType) {
		case kQTMetaDataTypeJPEGImage:
			[image2 setName:@"iTunes.jpg"]; break;
		case kQTMetaDataTypePNGImage:
			[image2 setName:@"iTunes.png"]; break;
		case kQTMetaDataTypeBMPImage:
			[image2 setName:@"iTunes.bmp"]; break;
		case 12: // GIF(?)
			break;
		default: break;
	}
    free(artData);
	if (image2)
		return image2;
	if (image)
		return image;
	//return nil;
	if ([[[NSUserDefaults standardUserDefaults] objectForKey:@"showPreviewPic"] boolValue]) {
		if ([[NSUserDefaults standardUserDefaults] boolForKey:@"showPosterPic"]) {
			if (!posterPic) {
				posterPic = [[mainMovie posterImage] retain];
				//NSLog(@"poster - %@", [self stringValue]);
			}
			return posterPic;
		}
		if (previewPic) return previewPic;
		TimeValue tv = [self timeIndex];
		TimeScale ts = GetMovieTimeScale([self QTMovie]);
		if (tv < 1 || tv > GetMovieDuration([self QTMovie])) tv = GetMovieDuration([self QTMovie])/5;
		previewPic = [[mainMovie frameImageAtTime:QTMakeTime(tv, GetMovieTimeScale([self QTMovie]))] retain];
		//NSLog(@"preview %g - %@", (float)tv / (float)ts, [self stringValue]);
		if (previewPic) return previewPic;
	}
	return nil;
}

- (void)removeImageForLanguage:(NSString *)language
{
	if (![language isKindOfClass:[NSString class]]) language = nil;
	//////////////////////////////////// APIC
    Handle h = NewHandleClear(0);
    UserData ud = GetMovieUserData([self QTMovie]);
    long i, count = CountUserDataType(ud, 'APIC');
    for (i=count; i>0; i--) {
        OSErr err = GetUserData(ud, h, 'APIC', i);
        if (!err) {
            if (!language || [language isEqualToString:[self languageOfPictureResource:parseTag(h)]]) {
                OSErr err = RemoveUserData(ud, 'APIC', i);
                if (err) NSLog(@"error removing APIC data: %d", err);
            }
        }
    }
    DisposeHandle(h);
    if (!language) [metadata removeObjectForKey:@"Cover Image"];
    else [[metadata objectForKey:@"Cover Image"] removeObjectForKey:language];
	[info removeObjectForKey:@"Cover Image"];
	
	////////////////////////////////////// QT7
	
	QTMetaDataRef qref = nil;
	OSErr err = QTCopyMovieMetaData([self QTMovie], &qref);
	QTMetaDataItem item = kQTMetaDataItemUninitialized, item2 = item;
	OSType key = kQTMetaDataCommonKeyArtwork;
	while (!err) {
		err = QTMetaDataGetNextItem(qref, kQTMetaDataStorageFormatiTunes, item, kQTMetaDataKeyFormatWildcard, &key, sizeof(key), &item);
		if (item2)
			QTMetaDataRemoveItem(qref, item2);
		item2 = kQTMetaDataItemUninitialized;
		if (!err) {
			item2 = item;
			/*if (language) { // find out if it's the right language
				OSType propType = 0;
				ByteCount size = 0;
				UInt32 flags = 0;
				err = QTMetaDataGetItemPropertyInfo(qref, item, kPropertyClass_MetaDataItem, kQTMetaDataItemPropertyID_Locale, &propType, &size, &flags);
				char *value = malloc(size);
				err = QTMetaDataGetItemProperty(qref, item, kPropertyClass_MetaDataItem, kQTMetaDataItemPropertyID_Locale, size, value, &size);
				if (!err) {
					if ([language isEqualToString:[NSString stringWithCString:value length:size]] || [[language lowercaseString] hasPrefix:[NSString stringWithCString:value length:size]])
						item2 = item;
				}
			} else { // erase it
				item2 = item;
			}*/
		}
	}
}

- (void)setImageFile:(NSString *)path
{
    [self setImageFile:path language:@"English"];
}

- (void)setImageFile:(NSString *)path language:(NSString *)language
{
    OSErr err = noErr;
    if (!path || ![path isKindOfClass:[NSString class]] || ![path length]) {
        [self removeImageForLanguage:language];
        goto BAIL;
    }
	if ([language isEqualToString:@"Unspecified"]) language = @"English";  // it doesn't like unspecified?
	
    NSData *imageData = [NSData dataWithContentsOfFile:path];
    if (!imageData) return;
	
	UInt32 dataType = kQTMetaDataTypeBMPImage; // apparently it doesn't matter anyway...
	
    NSMutableArray *mimeArray = [NSMutableArray array];
    [mimeArray addObject:[NSArray arrayWithObjects:@"image/jpeg", @"jpg", @"jpeg", nil]];
    [mimeArray addObject:[NSArray arrayWithObjects:@"image/bmp", @"bmp", nil]];
    [mimeArray addObject:[NSArray arrayWithObjects:@"image/gif", @"gif", nil]];
    //[mimeArray addObject:[NSArray arrayWithObjects:@"image/tiff", @"tif", @"tiff", nil]];
    [mimeArray addObject:[NSArray arrayWithObjects:@"image/png", @"png", nil]];
	UInt32 types[] = {kQTMetaDataTypeJPEGImage, 12 /* might be GIF */, kQTMetaDataTypeBMPImage, kQTMetaDataTypePNGImage};
	int i, n = [mimeArray count];
    for (i=0;i<n;i++) {
        NSArray *mimes = [mimeArray objectAtIndex:i];
        if ([mimes containsObject:[[path extension] lowercaseString]]) {
            dataType = types[i];
            break;
        }
    }
	
	QTMetaDataRef qref = nil;
	QTMetaDataItem item = kQTMetaDataItemUninitialized;
	UInt8 *artwork = [imageData bytes];
	ByteCount size = [imageData length];
	OSType key = kQTMetaDataCommonKeyArtwork;
	err = QTCopyMovieMetaData([self QTMovie], &qref);
	if (!err) {
		err = QTMetaDataAddItem(qref, kQTMetaDataStorageFormatiTunes, kQTMetaDataKeyFormatCommon, (const UInt8 *)&key, sizeof(key), artwork, size, kQTMetaDataTypeJPEGImage, &item);
		//const char *langCodeStr = [language cStringUsingEncoding:NSMacOSRomanStringEncoding]; // use the short code, eg "en", "de", etc
		//status = QTMetaDataSetItemProperty(qref, item, kPropertyClass_MetaDataItem, kQTMetaDataItemPropertyID_Locale, strlen(langCodeStr) + 1, langCodeStr);
		QTMetaDataRelease(qref);
	}
	
	if (err)
        NSLog(@"error adding image to user data: %ld", err);
    else {
        NSMutableDictionary *cat = [metadata objectForKey:@"Cover Image"];
        if (!cat) {
            cat = [NSMutableDictionary dictionary];
            [metadata setObject:cat forKey:@"Cover Image"];
        }
        [cat setObject:[path lastPathComponent] forKey:language];
		[info setObject:path forKey:@"Cover Image"];
    }
	
	goto BAIL;
    
	/////////////////  APIC stuff don't use
	
    NSString *description = nil;
    if (!language) language = @"English";
    description = [NSString stringWithFormat:@"%@:%@", language, [path lastPathComponent]];
    
    pictureResource r;
    
    r.textEncoding = 0;
    r.mimeType = "image/jpeg";
    r.picType = CoverFront;
    r.description = (char *)[description cString];
    r.pictureData = nil;
    
    
    n = [mimeArray count];
    for (i=0;i<n;i++) {
        NSArray *mimes = [mimeArray objectAtIndex:i];
        if ([mimes containsObject:[[path extension] lowercaseString]]) {
            r.mimeType = (char *)[[mimes objectAtIndex:0] cString];
            break;
        }
    }
    Handle h = NewHandleClear([imageData length]);
    BlockMoveData([imageData bytes], *h, [imageData length]);
    r.pictureData = h;
    h = createTag(r);
    
    // add it to the movie object
    err = AddUserData(GetMovieUserData([self QTMovie]), h, 'APIC');
    if (err)
        NSLog(@"error adding image to user data: %ld", err);
    else {
        NSMutableDictionary *cat = [metadata objectForKey:@"Cover Image"];
        if (!cat) {
            cat = [NSMutableDictionary dictionary];
            [metadata setObject:cat forKey:@"Cover Image"];
        }
        [cat setObject:[path lastPathComponent] forKey:language];
		[info setObject:path forKey:@"Cover Image"];
    }
    
    //if (previewPic) [previewPic release];
    //previewPic = [[NSImage alloc] initWithContentsOfFile:path];
BAIL:
    if (_sortedKeys) [_sortedKeys release];
    _sortedKeys = nil;
    [[NSNotificationCenter defaultCenter] postNotificationName:@"movieLanguageChangedNotification" object:self];
}

- (id)objectForKey:(id)anObject language:(NSString *)lang
{
	if (statusString) return statusString;
    return [[metadata objectForKey:anObject] objectForKey:lang];
}

- (id)objectForKey:(id)anObject
{
    //NSLog(@"MHMovie objectForKey");
	if (statusString) return statusString;
    NSDictionary *cat = [metadata objectForKey:anObject];
    NSArray *sysLangs = [tLanguages systemLanguages];
	NSString *sysLang = nil, *value = nil;
    int i;
    for (i=0;i<[sysLangs count];i++) {
        sysLang = [sysLangs objectAtIndex:i];
        value = [cat objectForKey:sysLang];
        if (value) break;
    }
    //NSString *value = [cat objectForKey:@"English"];
    if (!value)
        value = [data objectForKey:anObject];
    if (!value)
        value = [info objectForKey:anObject];
    return value;
}

- (BOOL)setObject:(id)anObject forKey:(id)key language:(NSString *)language
{
    id oldValue = nil;
    if (!key) return NO;
    if (!language || ![language length]) return [self setObject:anObject forKey:key];
    if ([key isEqual:@"Cover Image"]) {
        [self setImageFile:anObject language:language];
        goto BAIL;
    }
	if (!editedFields) editedFields = [[NSMutableArray array] retain];
	if (![editedFields containsObject:key]) [editedFields addObject:key];
    NSMutableDictionary *cat = [metadata objectForKey:key];
    if (!cat) {
        cat = [NSMutableDictionary dictionary];
        [metadata setObject:cat forKey:key];
    }
    oldValue = [cat objectForKey:language];
    if (!oldValue) oldValue = @"";
    if (!anObject) anObject = @"";
    if ([oldValue isKindOfClass:[NSString class]] && [anObject isKindOfClass:[NSString class]] && [oldValue isEqualToString:anObject])
        return NO;
    if ([oldValue isKindOfClass:[NSNumber class]] && [anObject isKindOfClass:[NSNumber class]] && [[oldValue stringValue] isEqualToString:[anObject stringValue]])
        return NO;
    if ([anObject isKindOfClass:[NSString class]] && [anObject isEqualToString:@""])
        [cat removeObjectForKey:language];
    else
        [cat setObject:anObject forKey:language];
BAIL:
    [self setEdited:YES];
    if (_sortedKeys) [_sortedKeys release];
    _sortedKeys = nil;
    [[NSNotificationCenter defaultCenter] postNotificationName:@"liveUpdateNotification" object:nil userInfo:[NSDictionary dictionaryWithObject:self forKey:key]];
    return YES;
}

- (BOOL)setObject:(NSString *)o forKey:(NSString *)k tableView:(NSTableView *)t
{
	NSUndoManager *u = [[t window] undoManager];
	[[u prepareWithInvocationTarget:self] setObject:[self objectForKey:k] forKey:k tableView:t];
	[u setActionName:[NSString stringWithFormat:@"Set %@", k]];
	[self setObject:o forKey:k];
	[t reloadData];
}

- (BOOL)setObject:(NSString *)anObject forKey:(NSString *)key
{
    //NSLog(@"MHMovie setObject:forKey:");
    id oldValue = nil;
    if (!key) return NO;
    if ([[NSArray arrayWithObjects:@"File Name", @"Poster", nil] containsObject:key]) {
        if (!anObject || ([anObject isKindOfClass:[NSString class]] && ![anObject length])) [info removeObjectForKey:key];
        else [info setObject:anObject forKey:key];
		if ([key isEqualToString:@"Poster"]) {
			[posterPic release];
			posterPic = nil;
		}
        return YES;
    } else if ([key isEqualToString:@"Cover Image"]) {
        [self setImageFile:anObject language:nil];
        goto BAIL;
    } else if ([key isEqualToString:@"Source Name"]) {
		[self setSourceName:anObject];
		goto BAIL;
	}
	if (!editedFields) editedFields = [[NSMutableArray array] retain];
	if (![editedFields containsObject:key]) [editedFields addObject:key];
    NSMutableDictionary *cat = [metadata objectForKey:key];
    if (!cat) {
        cat = [NSMutableDictionary dictionary];
        [metadata setObject:cat forKey:key];
    }
    NSArray *sysLangs = [tLanguages systemLanguages];
    NSString *sysLang = nil;
    if ([sysLangs count])
        sysLang = [sysLangs objectAtIndex:0];
    if (!sysLang) sysLang = @"English";
    oldValue = [cat objectForKey:sysLang];
    if (oldValue == nil)
        oldValue = @"";
    if (anObject == nil)
        anObject = @"";
    if ([oldValue isKindOfClass:[NSString class]] && [anObject isKindOfClass:[NSString class]] && [oldValue isEqualToString:anObject])
        return NO;
    if ([oldValue isKindOfClass:[NSNumber class]] && [anObject isKindOfClass:[NSNumber class]] && [[oldValue stringValue] isEqualToString:[(NSNumber *)anObject stringValue]])
        return NO;
    if ([anObject isKindOfClass:[NSString class]] && [anObject isEqualToString:@""])
        [cat removeObjectForKey:sysLang];
    else
        [cat setObject:anObject forKey:sysLang];
BAIL:
    [self setEdited:YES];
    if (_sortedKeys) [_sortedKeys release];
    _sortedKeys = nil;
    [[NSNotificationCenter defaultCenter] postNotificationName:@"liveUpdateNotification" object:nil userInfo:[NSDictionary dictionaryWithObject:self forKey:key]];
    return YES;
}

- (void)addEntriesFromDictionary:(NSDictionary *)otherDictionary
{
    NSArray *keys = [otherDictionary allKeys];
    int i, n = [keys count];
    for (i=0;i<n;i++) {
        NSString *anObject = [otherDictionary objectForKey:[keys objectAtIndex:i]];
        if (anObject)
            [self setObject:anObject forKey:[keys objectAtIndex:i]];
    }
    if (_sortedKeys) [_sortedKeys release];
    _sortedKeys = nil;
}

- (BOOL)setDataCode:(NSString *)newC forKey:(NSString *)key
{
    if (!newC) return NO;
    if (![newC isKindOfClass:[NSString class]]) return NO;
    const char *buffer = [newC cString];
    //Str255 pString;
    //CopyCStringToPascal(buffer, pString);
    if ([newC length] < 4) newC = [newC stringByAppendingString:@"    "];
    NSString *oldC = [data objectForKey:key];
    if (!oldC) return NO;
    if ([oldC isEqualToString:newC])
        return NO;
    OSType type = [newC characterAtIndex:0] * 256 * 256 * 256;
    type += [newC characterAtIndex:1] * 256 * 256;
    type += [newC characterAtIndex:2] * 256;
    type += [newC characterAtIndex:3];
    Track track = GetBiggestTrack([self QTMovie], VideoMediaType);
    if (!track)
        track = GetBiggestTrack([self QTMovie], MPEGMediaType);
    if (!track)
        return NO;
    Media media = GetTrackMedia(track);
    SampleDescriptionHandle desc = (SampleDescriptionHandle)NewHandle(sizeof(SampleDescription));
    GetMediaSampleDescription(media, 1, desc);
    ImageDescription *desc2 = *(ImageDescriptionHandle)desc;
    MoveHHi((Handle)desc);
    if ([key isEqualToString:@"Video FourCC"])
        desc2->cType = type;
    else if ([key isEqualToString:@"Video Vendor"])
        desc2->vendor = type;
    else if ([key isEqualToString:@"Video Format"])
        CopyCStringToPascal(buffer, desc2->name);
    else
        return NO;
    SetMediaSampleDescription(media, 1, desc);
    [data setObject:[NSString stringWithString:newC] forKey:key];
    [self setEdited:YES];
    saveInPlaceUnSafe = YES;
    return YES;
}

- (NSArray *)allKeys
{
    //NSLog(@"MHMovie allKeys");
    NSMutableArray *keys = [NSMutableArray array];
    [keys addObjectsFromArray:[metadata allKeys]];
    [keys addObjectsFromArray:[data allKeys]];
    [keys addObjectsFromArray:[info allKeys]];
    return keys;
}

- (NSArray *)allMetadataKeys
{
    return [metadata allKeys];
}

- (void)setListNumber:(int)number
{
    if (!listNumbers) listNumbers = [[NSMutableArray alloc] init];
    [listNumbers addObject:[NSNumber numberWithInt:number]];
}

- (NSDictionary *)defsEntry
{
    //NSLog(@"MHMovie defsEntry");
    NSMutableDictionary *entry = [NSMutableDictionary dictionary];
    [entry addEntriesFromDictionary:metadata];
    [entry addEntriesFromDictionary:data];
    [entry addEntriesFromDictionary:info];
	if (myRefMovie) [entry setObject:myRefMovie forKey:@"myRefMovie"];
    if (tempFiles) [entry setObject:tempFiles forKey:@"tempFiles"];
    if (nil != Chapters) {
        [entry setObject:[(MHChapters *)Chapters defsEntry] forKey:@"chapters"];
        [entry setObject:[NSNumber numberWithBool:[Chapters multilingual]] forKey:@"multiChaps"];
    }
    if (Menus)
        [entry setObject:[MHUtil plist:Menus] forKey:@"menus"];
    if (listNumbers)
        [entry setObject:[NSArray arrayWithArray:listNumbers] forKey:@"listNumbers"];
    [listNumbers release];
    listNumbers = nil;
    NSMutableArray *alters = [NSMutableArray array];
	if (!alterations) alterations = @"";
	[alters addObject:alterations];
	int i, n = [self trackCount];
	for (i=0;i<n;i++)
		[alters addObject:[[self trackAtIndex:i] alterations]];
	[entry setObject:alters forKey:@"alterations"];
	
    FSRef fsRef;
    NSString *path = [data objectForKey:@"File Path"];
    if (path) {
        const char *fsrep = [path fileSystemRepresentation];
        OSErr err = FSPathMakeRef((const UInt8 *)fsrep, &fsRef, NULL /* isDirectory */);
        if (!err) [entry setObject:[NSData dataWithBytes:&fsRef length:sizeof(fsRef)] forKey:@"FSRef"];
    }
    
    if (isChanged) [entry setObject:[NSNumber numberWithBool:YES] forKey:@"isChanged"];
	if (editedFields) [entry setObject:editedFields forKey:@"editedFields"];
    
    return entry;
}

- (void)setChapters:(id)chapters
{
    //[chapters setTimeScaleFromMovie:self];
    Chapters = chapters;
}

- (id)chapters
{
    return Chapters;
}

- (void)setChaptersNeedUpdating
{
    chaptersNeedUpdating = YES;
    [[NSNotificationCenter defaultCenter] postNotificationName:@"liveUpdateNotification" object:nil userInfo:[NSDictionary dictionaryWithObject:self forKey:@"Chapters"]];
}

- (void)setMenu:(id)menu
{	Menus = menu;	}

- (id)menu
{	return Menus;	}

- (void)saveCurrentHeight
{
    //NSLog(@"MHMovie saveCurrentHeight");
    Rect boxRect;
    GetMovieBox([self QTMovie], &boxRect);
    [data setObject:[NSNumber numberWithInt:boxRect.right - boxRect.left] forKey:@"Saved Width"];
    [data setObject:[NSNumber numberWithInt:boxRect.bottom - boxRect.top] forKey:@"Saved Height"];
}

- (void)addTempFilePath:(NSString *)aPath
{
    if (!aPath) return;
    if (!tempFiles) tempFiles = [[NSMutableArray alloc] init];
    [tempFiles addObject:aPath];
}

- (void)setStatus:(NSString *)status
{
	statusString = status;
}

- (BOOL)saveInPlaceUnSafe
{
    return saveInPlaceUnSafe;
}

- (void)setSaveInPlaceUnSafe
{
    saveInPlaceUnSafe = YES;
}

- (BOOL)isEdited
{
    return isChanged;
}
- (void)setEdited:(BOOL)e
{
    isChanged = e;
	/*
	if (e) {
		if (![mainMovie updateMovieFile])
			NSLog(@"updateMovieFile = NO.  %@", myRefMovie);
	}
	*/
}

- (BOOL)isOffline
{
    return isOffline;
}

- (double)currentTimeIndexSeconds
{
    return (double)[self currentTimeIndex] / (double)GetMovieTimeScale([self QTMovie]);
}

- (TimeValue)currentTimeIndex
{
    TimeValue tv;
    TimeRecord tr;
    tv = GetMovieTime([self QTMovie], &tr);
    return tv;
}

- (TimeValue)timeIndex
{
	double secs = [[data objectForKey:@"Seconds Elapsed"] doubleValue];
    TimeScale scale = GetMovieTimeScale([self QTMovie]);
	return secs * (double)scale;
}

- (void)setCurrentTimeIndexSeconds:(double)seconds
{
    [self setCurrentTimeIndex:seconds * (double)GetMovieTimeScale([self QTMovie])];
}

- (void)setCurrentTimeIndex:(TimeValue)samples
{
    TimeRecord tr;
    GetMovieTime([self QTMovie], &tr);
    tr.value.lo = samples;
    SetMovieTime([self QTMovie], &tr);
}

- (void)setTimeIndex:(TimeValue)tv
{
	if (isOffline) return;
    double secs = (double)tv / (double)GetMovieTimeScale([self QTMovie]);
    //NSLog(@"set time: %g", secs);
    [data setObject:[NSNumber numberWithDouble:secs] forKey:@"Seconds Elapsed"];
	//if (tv != [[info objectForKey:@"Poster"] longValue]) {
		//[previewPic release];
		//previewPic = nil;
		//NSLog(@"no preview - %@", [self stringValue]);
	//}
}

- (void)thumb
{
	NSImage *i = previewPic;
	previewPic = nil;
	[i release];
}

- (void)setSize:(int)size
{
    Rect boxRect;
    boxRect.top = boxRect.left = 0;
    boxRect.bottom = [[data objectForKey:@"Correct Height"] intValue];
    boxRect.right = [[data objectForKey:@"Correct Width"] intValue];
    switch (size) {
        case 1: boxRect.bottom *= 2;
                boxRect.right *= 2;
                break;
        case 2: boxRect.bottom = [[data objectForKey:@"Saved Height"] intValue];
                boxRect.right = [[data objectForKey:@"Saved Width"] intValue];
                break;
        default: break;
    }
    if (DEBUG)
	NSLog(@"setSize: %dx%d", boxRect.right, boxRect.bottom);
    if (boxRect.bottom && boxRect.right)
	SetMovieBox([self QTMovie], &boxRect);
}

- (void)updateQTMovie
{
    int i, j, m, n;
    NSString *value, *defaultValue, *theKey, *aLang;
    NSMutableArray *keys = [NSMutableArray arrayWithArray:[metadata allKeys]];
	[keys sortUsingSelector:@selector(compare:)];
	if ([keys containsObject:@"Rating"]) {
		[keys removeObject:@"Rating"];
		[keys insertObject:@"Rating" atIndex:0];
	}
    
	[self nukeQTKeys];
	while (CountUserDataType(GetMovieUserData([self QTMovie]), '\251MTH'))
		RemoveUserData(GetMovieUserData([self QTMovie]), '\251MTH', 1);
	
	
    n = [keys count];
    for (i=0; i<n; i++) {
        theKey = [keys objectAtIndex:i];
        if ([theKey isEqualToString:@"Cover Image"]) continue;
        NSDictionary *cat = [metadata objectForKey:theKey];
        m = [[cat allKeys] count];
        defaultValue = nil;
        [self removeQTDataForKey:theKey];
        for (j=0;j<m;j++) {
            aLang = [[cat allKeys] objectAtIndex:j];
            value = [cat objectForKey:aLang];
            [self setQTData:value forKey:theKey forLanguage:aLang];
            if (!defaultValue) defaultValue = value;
        }
        NSArray *langs = [tLanguages selectedLanguages];
        if (![tLanguages toCopy])
            langs = [NSArray arrayWithObject:[NSNumber numberWithInt:0]]; // if we're not copying, then at least we're copying to english
        m = [langs count];
        if (defaultValue) {
            for (j=0;j<m;j++) {
                aLang = [tLanguages numberToString:[[langs objectAtIndex:j] intValue]];
                if (![[cat allKeys] containsObject:aLang])
                    [self setQTData:defaultValue forKey:theKey forLanguage:aLang];
            }
        }
    }
    // remove extras, just in case
    n = [[[self class] supportedMetadata] count];
    for (i=0;i<n;i++) {
        theKey = [[[self class] supportedMetadata] objectAtIndex:i];
        if (![metadata objectForKey:theKey])
            [self removeQTDataForKey:theKey];
    }
}

- (void)setMTHData:(NSString *)value forKey:(NSString *)key forLanguage:(NSString *)language
{
	short regionTag = [tLanguages stringToNumber:language];
	if (regionTag == -1) return;
	UserData ud = GetMovieUserData([self QTMovie]);
	OSErr err = noErr;
	OSType type = '\251MTH';
	Handle h = NewHandleClear(0);
	long i, count = CountUserDataType(ud, type);
	for (i=1;i<=count;i++)
		if (noErr != GetUserDataText(ud, h, type, i, regionTag)) break;
	DisposeHandle(h);
	if (i>count) {
		h = NewHandleClear(0);
		err = AddUserData(ud, h, type);
		DisposeHandle(h);
	}
	char *buffer = malloc(sizeof(char) * ([key length] + [value length] + 2));
	[key getCString:buffer];
	[value getCString:buffer+[key length]+1];
	h = NewHandleClear([key length]+[value length]+1);
	BlockMoveData(buffer, *h, [key length]+[value length]+1);
	err = AddUserDataText(ud, h, type, i, regionTag);
	if (err) NSLog(@"error adding MHT data: %d, %@, %@, %@", err, value, key, language);
	DisposeHandle(h);
	free(buffer);
}

- (NSString *)keyFormatOfQItem:(QTMetaDataItem)qitem qref:(QTMetaDataRef)qref
{
	ByteCount propertySize = 0;
	ByteCount sizeUsed = 0;
	//QTPropertyValueType
	OSType outPropType = 0;
	UInt32 outFlags = 0;
	// get the size of the property: keyformat
	OSErr err = QTMetaDataGetItemPropertyInfo(qref, qitem, kPropertyClass_MetaDataItem, kQTMetaDataItemPropertyID_KeyFormat, &outPropType, &propertySize, &outFlags);
	// allocate memory to hold the property value
	//QTPropertyValuePtr 
	char *qvalue = malloc(propertySize);
	// get the property of the metadata item: key format
	err = QTMetaDataGetItemProperty(qref, qitem, kPropertyClass_MetaDataItem, kQTMetaDataItemPropertyID_KeyFormat, propertySize, qvalue, &sizeUsed);
	NSString *keyFormat = [[[NSString alloc] initWithBytes:qvalue length:sizeUsed encoding:NSUTF8StringEncoding] autorelease];
#ifdef __LITTLE_ENDIAN__
	keyFormat = [keyFormat reverse];
#endif
	NSLog(keyFormat);
	return keyFormat;
}


- (void)nukeQTKeys
{
	NSMutableArray *keys = [NSMutableArray arrayWithArray:[MHMovie itmsKeys]];
	if (editedFields) [keys addObjectsFromArray:editedFields];
	[keys addObject:@"meta"];
	QTMetaDataItem item = kQTMetaDataItemUninitialized, itemToRemove = kQTMetaDataItemUninitialized;
	QTMetaDataRef qref;
	OSErr err = QTCopyMovieMetaData([self QTMovie], &qref);
	ByteCount propertySize = 0;
	ByteCount sizeUsed = 0;
	UInt32 outFlags = 0;
	while (!err) {
		err = QTMetaDataGetNextItem(qref, kQTMetaDataStorageFormatWildcard, item, kQTMetaDataKeyFormatWildcard, nil, 0, &item);
		if (err || item == kQTMetaDataItemUninitialized) break;
		if (itemToRemove != kQTMetaDataItemUninitialized)
			QTMetaDataRemoveItem(qref, itemToRemove);
		itemToRemove = kQTMetaDataItemUninitialized;
		if (item == kQTMetaDataItemUninitialized) break;
		// get the size of the property: key
		err = QTMetaDataGetItemPropertyInfo(qref, item, kPropertyClass_MetaDataItem, kQTMetaDataItemPropertyID_Key, nil, &propertySize, &outFlags);
		// allocate memory to hold the property value
		QTPropertyValuePtr qvalue = malloc(propertySize);
		// get the property of the metadata item: key
		err = QTMetaDataGetItemProperty(qref, item, kPropertyClass_MetaDataItem, kQTMetaDataItemPropertyID_Key, propertySize, qvalue, &sizeUsed);
		NSString *keyString = [NSString stringWithCString:qvalue length:sizeUsed];
#ifdef __LITTLE_ENDIAN__
		if (sizeUsed == 4)
			keyString = [keyString reverse];
#endif
		//NSLog(@"found: %@", keyString);
		free(qvalue);
		if ([keys containsObject:keyString] || [keys containsObject:[MHMovie itmsTypeToString:keyString]]) {
		//if ([keyString isEqualToString:@"meta"]) {
			//NSLog(@"remove");
			itemToRemove = item;  // save it until we have the next one, then kill it
		}
	}
}

- (BOOL)matchRating:(NSString **)v
{
	// handy: http://shadowofged.blogspot.com/2008/06/itunes-content-ratings.html
	NSArray *ratings = [NSArray arrayWithObjects:@"NR", @"Not Rated", @"G", @"PG", @"M", @"R", @"Unrated", @"UR", @"TV", @"NC-17", nil];
	if ([*v containsString:@"|"]) return YES;
	int i, n = [ratings count];
	for (i=0;i<n;i++) {
		if ([*v hasPrefix:[ratings objectAtIndex:i]]) {
			*v = [NSString stringWithFormat:@"|%@", *v];
			return YES;
		}
	}
	return NO;
}

- (void)setQTData:(NSString *)value forKey:(NSString *)key forLanguage:(NSString *)language
{
    if (!key || !language || !value) return;
	if ([key isEqualToString:@"meta"]) return;
    OSErr err = noErr;
    //NSLog(@"adding value: %@\nfor key: %@\nfor language: %@", value, key, language);
    UserData ud = GetMovieUserData([self QTMovie]);
    int length = [value length]; // + 1;
    Handle h = nil;
	
	NSArray *ikeys = [MHMovie itmsKeys],
	*istrings = [MHMovie itmsStrings];
	
	BOOL success = NO;
	UInt32 dataFormat = kQTMetaDataTypeUTF8;
	if ([value hasPrefix:@"{int}"]) {
		dataFormat = kQTMetaDataTypeSignedIntegerBE;
		value = [value substringFromIndex:5];
	} else if ([value hasPrefix:@"{float}"]) {
		dataFormat = kQTMetaDataTypeFloat32BE;
		value = [value substringFromIndex:7];
	} else if ([value hasPrefix:@"{binary}"]) {
		dataFormat = kQTMetaDataTypeBinary;
		if ([value characterAtIndex:8] == ' ') value = [value substringFromIndex:9];
		else value = [value substringFromIndex:8];
	}
	
	NSData *dk, *dv = nil;
	SInt16 *bytes = nil;
	SInt8 my1;
	SInt16 my2;
	SInt32 my4;
	if (dataFormat == kQTMetaDataTypeFloat32BE) {
		Float32 myF = [value floatValue];
		dv = [NSData dataWithBytes:&myF length:sizeof(myF)];
	} else if (dataFormat == kQTMetaDataTypeBinary) {
		NSArray *myBinary = [value componentsSeparatedByString:@" "];
		int i=0, n = [myBinary count];
		char *myBytes = malloc(n*sizeof(char));
		for (;i<n;i++)
			myBytes[i] = [[myBinary objectAtIndex:i] intValue];
		dv = [NSData dataWithBytes:myBytes length:n*sizeof(char)];
	}
	
    if ([key isEqualToString:@"WLOC"]) {
        h = NewHandleClear(2*sizeof(short));
        short coords[2];
        sscanf([value cString], "%hd, %hd", &coords[1], &coords[0]);
        BlockMoveData(coords, *h, 2*sizeof(short));
        RemoveUserData(ud, 'WLOC', 1);
        if (noErr != AddUserData(ud, h, 'WLOC'))
            NSLog(@"error adding UserData: %@, WLOC", value);
		success = YES;
/*	} else if ([key isEqualToString:@"Auto Fullscreen"]) { // moved to tTracks
		h = NewHandleClear(4*sizeof(unichar));
		unichar letters[4];
		letters[0] = 768;
		letters[1] = letters[2] = 0;
		letters[3] = 256;
		BlockMoveData(letters, *h, 4*sizeof(unichar));
		err = AddUserData(ud, h, 'ptv '); */
    }
	
	NSString *ikey;
	if ([istrings containsObject:key]) {
		
		if (![key isEqualToString:@"Composer"] || [self objectForKey:@"Author"]) {// QT reading bug: itsk-composer masks udta-author.  udta-composer without itsk-composer works fine in iTunes and QT.  QT 7.6.9
			
			ikey = [ikeys objectAtIndex:[istrings indexOfObject:key]];
			//dv = nil;
			
			if ([ikey isEqualToString:@"trkn"] || [ikey isEqualToString:@"disk"]) {
				bytes = malloc(4*sizeof(UInt16));
				bytes[0] = bytes[1] = bytes[2] = bytes[3] = 0;
				sscanf([value cString], "%hd of %hd", &bytes[1], &bytes[2]);
#ifdef __LITTLE_ENDIAN__
				bytes[1] = NSSwapShort(bytes[1]);
				bytes[2] = NSSwapShort(bytes[2]);
#endif
				dv = [NSData dataWithBytes:bytes length:4*sizeof(UInt16)];
				dataFormat = kQTMetaDataTypeBinary;
			} else if ([ikey isEqualToString:@"cpil"] || [ikey isEqualToString:@"stik"] || [ikey isEqualToString:@"pgap"] || [ikey isEqualToString:@"rtng"]) {
				if ([ikey isEqualToString:@"stik"]) {
					if ([[value lowercaseString] hasPrefix:@"tv"]) my1 = 10;
					else if ([[value lowercaseString] hasPrefix:@"mo"]) my1 = 0;
					else if ([[value lowercaseString] hasPrefix:@"mu"]) my1 = 6;
					else if ([[value lowercaseString] hasPrefix:@"ri"]) my1 = 14;
					else if ([[value lowercaseString] hasPrefix:@"po"]) {
						ikey = @"pcst";
						my1 = 1;
					} else if ([[value lowercaseString] isEqualToString:@"it"]) {
						ikey = @"itnu";
						my1 = 1;
					} else my1 = [value intValue];
				} else if ([ikey isEqualToString:@"rtng"]) {
					if ([[value lowercaseString] hasPrefix:@"e"]) my1 = 4;
					else if ([[value lowercaseString] hasPrefix:@"c"]) my1 = 2;
					else my1 = [value intValue];
				} else
					my1 = [value intValue];
				dv = [NSData dataWithBytes:&my1 length:sizeof(UInt8)];
				dataFormat = kQTMetaDataTypeUnsignedIntegerBE;
				if ([ikey isEqualToString:@"cpil"]) dataFormat = kQTMetaDataTypeSignedIntegerBE;
			} else if ([ikey isEqualToString:@"tmpo"]) {
				my2 = EndianS16_NtoB( [value intValue] );
				dv = [NSData dataWithBytes:&my2 length:sizeof(my2)];
				dataFormat = kQTMetaDataTypeSignedIntegerBE;
			} else if ([ikey isEqualToString:@"tvsn"] || [ikey isEqualToString:@"tves"] || dataFormat == kQTMetaDataTypeSignedIntegerBE) {
				my4 = EndianS32_NtoB( [value intValue] );
				dv = [NSData dataWithBytes:&my4 length:sizeof(my4)];
				dataFormat = kQTMetaDataTypeSignedIntegerBE;
			} else if ([ikey isEqualToString:[NSString stringWithFormat:@"%cgen", '\251']] && [[MHMovie gnre] containsObject:value]) {
				// note: might need to add "gnre"
				my2 = [[MHMovie gnre] indexOfObject:value];
				my2 = EndianS16_NtoB( my2 );
				dv = [NSData dataWithBytes:&my2 length:sizeof(my2)];
				dataFormat = kQTMetaDataTypeBinary;
			} else if (!dv) {
				dv = [value dataUsingEncoding:NSUTF8StringEncoding];
			}
#ifdef __LITTLE_ENDIAN__
			ikey = [ikey reverse];
#endif
			dk = [ikey dataUsingEncoding:NSMacOSRomanStringEncoding];
			QTMetaDataRef qref;
			err = QTCopyMovieMetaData([self QTMovie], &qref);
			QTMetaDataItem newItem = kQTMetaDataItemUninitialized;
			err = QTMetaDataAddItem(qref, kQTMetaDataStorageFormatiTunes, kQTMetaDataKeyFormatiTunesShortForm, [dk bytes], [dk length], [dv bytes], [dv length], dataFormat, &newItem);
			if (err) NSLog(@"error %ld adding %@ %@", err, key, value);
			if ([language length] > 2) language = [tLanguages abbrev:language];
			const char *lcode = [language cStringUsingEncoding:NSMacOSRomanStringEncoding];
			err = QTMetaDataSetItemProperty(qref, newItem, kPropertyClass_MetaDataItem, kQTMetaDataItemPropertyID_Locale, strlen(lcode)+1, lcode);
			if (err) NSLog(@"error %ld adding %@ %@", err, key, value);
			else success = YES;
		}
	}
		
        h = NewHandleClear(length);
        OSType theType = [self stringToOSType:key];
        if (theType) {
			short regionTag = [tLanguages stringToNumber:language];
			if (regionTag != -1) {
				//char *buffer = malloc(sizeof(char) * (length + 1));
				
				const char *buffer = [[value dataUsingEncoding:NSMacOSRomanStringEncoding allowLossyConversion:YES] bytes];
				//[value getSafeCString:buffer];
				if (buffer) {
					BlockMoveData(buffer, *h, length);
					//free(buffer);
					//buffer = nil;
					long index = [key hasIndex];
					long count = CountUserDataType(ud, theType);
					if (index) {
						while ((count = CountUserDataType(ud, theType)) <= index) {
							Handle h2 = NewHandleClear(0);
							//**h2 = 'X';
							err = AddUserData(ud, h2, theType);
							if (err) NSLog(@"error adding UserData instance: %d", err);
							else success = YES;
							DisposeHandle(h2);
						}
					}
					index++;
					//count = CountUserDataType(ud, theType);
					if (err = AddUserDataText(ud, h, theType, index, regionTag))
						NSLog(@"error adding UserData:%d, %@, %@, %@", err, value, key, language);
					else success = YES;
					
					if ([key isEqualToString:@"Rating"] && [self matchRating:&value]) { // Oh jeez what a mess.  This helped: http://atomicparsley.sourceforge.net/mpeg-4files.html
						// this is for iTunes but it's stored in udta
						while (CountUserDataType(ud, 'meta')) {
							err = RemoveUserData(ud, 'meta', 1);
						}
						char *buffer = malloc(118 + [value length]);
						
						char stem[] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x22, 0x68, 0x64, 0x6C, 0x72, 0x00, 0x00, 0x00, 0x00,
							0x00, 0x00, 0x00, 0x00, 0x6D, 0x64, 0x69, 0x72, 0x61, 0x70, 0x70, 0x6C, 0x00, 0x00, 0x00, 0x00,
							0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x5F, 0x69, 0x6C, 0x73, 0x74, 0x00, 0x00,
							0x00, 0x57, 0x2D, 0x2D, 0x2D, 0x2D, 0x00, 0x00, 0x00, 0x1C, 0x6D, 0x65, 0x61, 0x6E, 0x00, 0x00,
							0x00, 0x00, 0x63, 0x6F, 0x6D, 0x2E, 0x61, 0x70, 0x70, 0x6C, 0x65, 0x2E, 0x69, 0x54, 0x75, 0x6E,
							0x65, 0x73, 0x00, 0x00, 0x00, 0x14, 0x6E, 0x61, 0x6D, 0x65, 0x00, 0x00, 0x00, 0x00, 0x69, 0x54,
							0x75, 0x6E, 0x45, 0x58, 0x54, 0x43, 0x00, 0x00, 0x00, 0x1F, 0x64, 0x61, 0x74, 0x61, 0x00, 0x00,
						0x00, 0x01, 0x00, 0x00, 0x00, 0x00 };
						int i;
						for (i=0;i<118;i++)
							buffer[i] = stem[i];
						buffer[41] = 80 + [value length];
						buffer[49] = 72 + [value length];
						buffer[105] = 16 + [value length];
						for (i=0;i<[value length];i++)
							buffer[118+i] = [value characterAtIndex:i];
						
						//NSError *e = nil;
						//NSData *d = [NSData dataWithContentsOfFile:@"/rat.data" options:nil error:&e];
						Handle h3 = NewHandleClear(118 + [value length]);
						BlockMoveData(buffer, *h3, 118 + [value length]);
						err = AddUserData(ud, h3, 'meta');
					}
				}
			}
		}
	
	ikey = nil;
	if (!success) ikey = key; // if it's an unknown format, we're going to wing it
	istrings = [NSArray arrayWithObjects:@"Year", @"Author", 
				//@"Rating",
				nil];
	ikeys = [NSArray arrayWithObjects:@"com.apple.quicktime.year", @"com.apple.quicktime.author",
			 //@"com.apple.iTunes",
			 nil];
	if ([istrings containsObject:key])
		ikey = [ikeys objectAtIndex:[istrings indexOfObject:key]];
	
	if (ikey) {
		dk = [ikey dataUsingEncoding:NSMacOSRomanStringEncoding];
		QTMetaDataRef qref;
		err = QTCopyMovieMetaData([self QTMovie], &qref);
		QTMetaDataItem newItem = kQTMetaDataItemUninitialized;
		if (![ikey hasPrefix:@"com.apple.iTunes"]) { // if we know it's itunes we won't bother with qt
			if (dataFormat == kQTMetaDataTypeUTF8) dataFormat = kQTMetaDataTypeMacEncodedText; // qt likes mac encoded text
			if (!dv || dataFormat == kQTMetaDataTypeMacEncodedText)
				dv = [value dataUsingEncoding:NSMacOSRomanStringEncoding];
			err = QTMetaDataAddItem(qref, kQTMetaDataStorageFormatQuickTime, kQTMetaDataKeyFormatQuickTime, [dk bytes], [dk length], [dv bytes], [dv length], dataFormat, &newItem);
			if (err) NSLog(@"error %ld adding %@ %@", err, key, value);
			if ([language length] > 2) language = [tLanguages abbrev:language];
			const char *lcode = [language cStringUsingEncoding:NSMacOSRomanStringEncoding];
			err = QTMetaDataSetItemProperty(qref, newItem, kPropertyClass_MetaDataItem, kQTMetaDataItemPropertyID_Locale, strlen(lcode)+1, lcode);
			if (err) NSLog(@"error %ld adding %@ %@", err, key, value);
			else success = YES;
		}
		if (![ikey hasPrefix:@"com.apple.quicktime"]) { // if we know it's qt we won't bother with iTunes
			if (dataFormat == kQTMetaDataTypeMacEncodedText) dataFormat = kQTMetaDataTypeUTF8; // iTunes likes UTF text
			QTMetaDataKeyFormat format = kQTMetaDataKeyFormatiTunesLongForm;
			if ([ikey length] == 4) {
				format = kQTMetaDataKeyFormatiTunesShortForm;
#ifdef __LITTLE_ENDIAN__
				ikey = [ikey reverse];
#endif
			}
			dk = [ikey dataUsingEncoding:NSUTF8StringEncoding];
			if (!dv || dataFormat == kQTMetaDataTypeUTF8)
				dv = [value dataUsingEncoding:NSUTF8StringEncoding];
			err = QTMetaDataAddItem(qref, kQTMetaDataStorageFormatiTunes, format, [dk bytes], [dk length], [dv bytes], [dv length], dataFormat, &newItem);
			if (err) NSLog(@"error %ld adding %@ %@", err, key, value);
			if ([language length] > 2) language = [tLanguages abbrev:language];
			const char *lcode = [language cStringUsingEncoding:NSMacOSRomanStringEncoding];
			err = QTMetaDataSetItemProperty(qref, newItem, kPropertyClass_MetaDataItem, kQTMetaDataItemPropertyID_Locale, strlen(lcode)+1, lcode);
			if (err) NSLog(@"error %ld adding %@ %@", err, key, value);
			else success = YES;
		}
	}
	
	
	
	if (!success) // only roll our own if none else fit
		[self setMTHData:value forKey:key forLanguage:language];
    DisposeHandle(h);
}

- (void) removeQTDataForKey:(NSString *)key
{
    if (!key) return;
	if ([key isEqualToString:@"Year"]) {
		QTMetaDataItem item = kQTMetaDataItemUninitialized, itemToRemove = kQTMetaDataItemUninitialized;
		QTMetaDataRef qref;
		OSErr err = QTCopyMovieMetaData([self QTMovie], &qref);
		ByteCount propertySize = 0;
		ByteCount sizeUsed = 0;
		UInt32 outFlags = 0;
		while (!err) {
			err = QTMetaDataGetNextItem(qref, kQTMetaDataStorageFormatQuickTime, item, kQTMetaDataKeyFormatQuickTime, nil, 0, &item);
			if (err || item == kQTMetaDataItemUninitialized) break;
			if (itemToRemove != kQTMetaDataItemUninitialized)
				QTMetaDataRemoveItem(qref, itemToRemove);
			itemToRemove = kQTMetaDataItemUninitialized;
			if (item == kQTMetaDataItemUninitialized) break;
			// get the size of the property: key
			err = QTMetaDataGetItemPropertyInfo(qref, item, kPropertyClass_MetaDataItem, kQTMetaDataItemPropertyID_Key, nil, &propertySize, &outFlags);
			// allocate memory to hold the property value
			QTPropertyValuePtr qvalue = malloc(propertySize);
			// get the property of the metadata item: key
			err = QTMetaDataGetItemProperty(qref, item, kPropertyClass_MetaDataItem, kQTMetaDataItemPropertyID_Key, propertySize, qvalue, &sizeUsed);
			NSString *keyString = [NSString stringWithCString:qvalue length:sizeUsed];
			free(qvalue);
			if ([keyString isEqualToString:@"com.apple.quicktime.year"])
				itemToRemove = item;  // save it until we have the next one, then kill it
		}
		return;
	}
    OSType theType = [self stringToOSType:key];
    int index = [key hasIndex];
    UserData ud = GetMovieUserData([self QTMovie]);
    if (!index)
        while (CountUserDataType(ud, theType))
            RemoveUserData(ud, theType, 1);
    else
        RemoveUserData(ud, theType, index+1);
}

- (BOOL) learn:(NSString *)newString key:(NSString *)key
{
    if (![newString length] || ![key length]) return NO;
    NSUserDefaults *defs = [NSUserDefaults standardUserDefaults];
    NSMutableDictionary *learned = [NSMutableDictionary dictionary];
    [learned addEntriesFromDictionary:[defs objectForKey:@"learned"]];
    NSMutableArray *cat = [NSMutableArray array];
    [cat addObjectsFromArray:[learned objectForKey:key]];
    NSArray *values = [newString componentsSeparatedByString:@", "];
    int i, n = [values count];
    for (i=0;i<n;i++) {
        NSString *value = [values objectAtIndex:i];
        if ([value length]) {
            if (NSNotFound == [cat indexOfObject:value])
                [cat addObject:value];
            else { // readd the value to the top because it was used again
                [cat removeObject:value];
                [cat addObject:value];
            }
        }
    }
    [learned setObject:cat forKey:key];
    [defs setObject:learned forKey:@"learned"];
    [defs synchronize];
    return YES;
}

- (void)makeSortedKeys
{
    if (_sortedKeys) [_sortedKeys release];
    _sortedKeys = [[NSMutableArray alloc] init];
    [_sortedKeys addObjectsFromArray:[metadata allKeys]];
    int i, n = [_sortedKeys count];
    for (i=n-1;i>=0;i--) {
        NSString *aKey = [_sortedKeys objectAtIndex:i];
        if ([aKey isEqualToString:@"WLOC"] || ![(NSDictionary *)[metadata objectForKey:aKey] count])
            [_sortedKeys removeObjectAtIndex:i];
    }
    [_sortedKeys sortUsingSelector:@selector(caseInsensitiveCompare:)];
	
    NSMutableArray *supported = [NSMutableArray arrayWithArray:[[self class] supportedMetadata]];
	if ([[NSUserDefaults standardUserDefaults] boolForKey:@"itSwitch"])
		[supported addObjectsFromArray:[MHMovie itmsStrings]];
	[supported sortUsingSelector:@selector(caseInsensitiveCompare:)];
    n = [supported count];
    for (i=0;i<n;i++) {
        if (![_sortedKeys containsObject:[supported objectAtIndex:i]])
            [_sortedKeys addObject:[supported objectAtIndex:i]];
    }
    [_sortedKeys removeObject:@"Title"];
    [_sortedKeys insertObject:@"Title" atIndex:0];
	if (![_sortedKeys containsObject:@"Year"])
		[_sortedKeys addObject:@"Year"];
	
}

// table view
- (int)numberOfRowsInTableView:(NSTableView *)aTableView
{
	if (!_sortedKeys) [self makeSortedKeys];
    return [_sortedKeys count];
    //return [[MHMovie supportedMetadata] count];
}

- (id)tableView:(NSTableView *)aTableView objectValueForTableColumn:(NSTableColumn *)aTableColumn row:(int)rowIndex
{
    if (!_sortedKeys) [self makeSortedKeys];
    NSString *value, *idString = [aTableColumn identifier];
    //NSArray *supported = [MHMovie supportedMetadata];
    if ([idString isEqualToString:@"Category"])
        return [_sortedKeys objectAtIndex:rowIndex];
    NSString *key = [_sortedKeys objectAtIndex:rowIndex];
    NSDictionary *dict = [metadata objectForKey:key];
    value = [dict objectForKey:idString];
    int i, n = [[tLanguages selectedLanguages] count];
    if ([tLanguages toCopy]) {
        for (i=0;i<n && !value;i++) {
            NSString *lang = [tLanguages numberToString:[[[tLanguages selectedLanguages] objectAtIndex:i] intValue]];
            value = [dict objectForKey:lang];
        }
    }
    //NSLog(@"%@: %@, %@ - %@", key, dict, value, idString);
    return value;
}

- (void)tableView:(NSTableView *)aTableView setObjectValue:(id)anObject forTableColumn:(NSTableColumn *)aTableColumn row:(int)rowIndex
{
    if (!_sortedKeys) [self makeSortedKeys];
    //if (!anObject || [anObject isEqualToString:@""]) return;
    NSString *lang = [aTableColumn identifier];
    NSString *key = [_sortedKeys objectAtIndex:rowIndex];
	if (!editedFields) editedFields = [[NSMutableArray array] retain];
	if (![editedFields containsObject:key]) [editedFields addObject:key];
    [self setObject:anObject forKey:key language:lang];
	[self learn:anObject key:key];
	[[NSNotificationCenter defaultCenter] postNotificationName:@"movieEditNotification" object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:anObject, @"object", key, @"key", lang, @"lang", nil]];
}

- (BOOL) tableView: (NSTableView *) tableView shouldEditTableColumn: (NSTableColumn *) tableColumn row: (int) rowIndex
{
    return ![[tableColumn identifier] isEqualToString:@"Category"];
}

- (void)tableView:(NSTableView *)aTableView willDisplayCell:(id)aCell forTableColumn:(NSTableColumn *)aTableColumn row:(int)rowIndex
{
    if ([[aTableColumn identifier] isEqualToString:@"Category"]) {
        NSArray *infoValues = [NSArray arrayWithObjects:@"Album", @"Author", @"Comment", @"Copyright", @"Description", @"Director", @"Title", @"Information", @"Performers", @"Writer", nil];
        //NSFontTraitMask mask = NSUnboldFontMask | NSUnitalicFontMask;
		float size = [[NSUserDefaults standardUserDefaults] floatForKey:@"fontSize"];
		if (!size) size = 12;
		NSFont *font = [NSFont userFontOfSize:size];
        if ([infoValues containsObject:[aCell stringValue]])
			font = [[NSFontManager sharedFontManager] convertFont:font toHaveTrait: NSBoldFontMask];
            //mask = NSBoldFontMask | NSUnitalicFontMask;
		else if (![[MHMovie supportedMetadata] containsObject:[aCell stringValue]])
			font = [[NSFontManager sharedFontManager] convertFont:font toHaveTrait: NSItalicFontMask];
			//mask = NSUnboldFontMask | NSItalicFontMask;
        [aCell setFont:font];
    } else {
		
	}
}

- (BOOL)tableView:(NSTableView *)tableView shouldShowCellExpansionForTableColumn:(NSTableColumn *)tableColumn row:(NSInteger)row
{
	return NO;
}

- (NSString *)tableView:(NSTableView *)tview toolTipForCell:(NSCell *)cell rect:(NSRectPointer)rect tableColumn:(NSTableColumn *)aTableColumn row:(NSInteger)row mouseLocation:(NSPoint)p
{
	if (![[cell stringValue] hasPrefix:@"{binary}"]) return [cell stringValue];
	NSArray *c = [[cell stringValue] componentsSeparatedByString:@" "];
	int i, d, n = [c count];
	NSMutableString *s = [NSMutableString stringWithCapacity:n];
	for (i=0;i<n;i++) {
		d = [[c objectAtIndex:i] intValue];
		if (!d) [s appendString:@"."];
		else if (d < 32 || d > 126) [s appendFormat:@"%d,", d];
		else [s appendFormat:@"%c", d];
	}
	return s;
}

BOOL recursion = NO;

- (void)textDidChange:(NSNotification *)aNotification
{
    //NSLog(@"textDidChange");
	if (recursion) return;
    NSTextView *box = [aNotification object];
    if (![[NSUserDefaults standardUserDefaults] boolForKey:@"autoComplete"]) return;
    NSString *item, *item2, *string = [NSString stringWithString:[box string]];
    NSRange selection = [box selectedRange];
    int initialLength = [string length];
    if ([string isEqualToString:@""] || [string hasSuffix:@", "] || selection.location != initialLength) return;
    NSRange aRange = [string rangeOfString:@", " options:NSBackwardsSearch];
    if (aRange.length != 0)
        string = [string substringFromIndex:aRange.location+2];
	id sview = box;
	while (YES) {
		sview = [sview superview];
		if ([sview isKindOfClass:[NSTableView class]]) break;
		if (!sview) return;
	}
	int row = [sview editedRow];
	if (row == -1) return;
	NSString *category = [_sortedKeys objectAtIndex:row];
	
    NSArray *list = [[[NSUserDefaults standardUserDefaults] objectForKey:@"learned"] objectForKey:category];
    if (!list) return;
    int i;
    recursion = YES;
    for (i=[list count]-1;i>=0;i--) {
        item = [list objectAtIndex:i];
        item2 = [item stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]];
        if ([item2 hasPrefix:string])
            item = item2;
        else if (![item hasPrefix:string])
            item = nil;
        if (item) {
            item = [item substringFromIndex:[string length]];
            [box insertText:item];
            selection.length = [item length];
            [box setSelectedRange:selection];
            break;
        }
    }
    recursion = NO;
}

- (void)parseiTunesValue:(NSString *)v key:(NSString *)k language:(NSString *)l
{
	if ([v hasPrefix:@"mpaa|"] || [v containsString:@"-tv|"] || [v hasPrefix:@"|"]) {
		k = @"Rating";
	} else if ([v hasPrefix:@"<?xml"]) {
		return;
	}
	NSMutableDictionary *d = [metadata objectForKey:k];
	if (!d) {
		d = [NSMutableDictionary dictionary];
		[metadata setObject:d forKey:k];
	}
	[d setObject:v forKey:l];
}

- (BOOL)isAsciiFriendly:(char *)buffer length:(long)length
{
	int i;
	for (i=0;i<length;i++) {
		char ch = buffer[i];
		if (ch != '\t' && ch < ' ') return NO;
		if (ch > '~') return NO;
	}
	return YES;
}

+ (NSArray *)gnre
{
	return [NSArray arrayWithObjects: @"", @"Blues", @"", @"Country", @"Dance", @"", @"", @"", @"", @"Jazz", @"", @"New Age", @"", @"", @"Pop", @"", @"", @"", @"Rock", @"Techno", @"Industrial", @"Alternative", @"", @"", @"", @"Soundtrack", @"", @"", @"", @"", @"", @"", @"Trance", @"Classical", @"", @"", @"House", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"Electronic", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"Folk", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"", @"Easy Listening", @"Country", @"Dance", @"" @"Children's Music", @"", @"", @"", @"Easy Listening", nil];
}

- (void)readQTKitMetadata
{
	//NSLog(@"entering...");
	QTMetaDataRef qref;
	QTMovie *q = [QTMovie movieWithFile:stringPath error:nil];
	if (!q)
		q = [self qtKitMovie];
	OSErr err = QTCopyMovieMetaData([q quickTimeMovie], &qref);
	//ItemCount outCount = 0;
	//err = QTMetaDataGetItemCountWithKey(qref, kQTMetaDataStorageFormatWildcard, kQTMetaDataKeyFormatWildcard, nil, 0, &outCount);
	//NSLog(@"count: %d", outCount);
	QTMetaDataItem qitem = kQTMetaDataItemUninitialized;
	//outCount = 0;
	while (!err) {
		//int inKey = 0;
		err = QTMetaDataGetNextItem(qref, kQTMetaDataStorageFormatWildcard, qitem, kQTMetaDataKeyFormatWildcard, nil, 0, &qitem);
		//if (qitem) outCount++;
		ByteCount propertySize = 0;
		ByteCount sizeUsed = 0;
		//QTPropertyValueType
		OSType outPropType = 0;
		UInt32 outFlags = 0;
		//OSType testType = 0;
		
		// get the size of the property: keyformat
		err = QTMetaDataGetItemPropertyInfo(qref, qitem, kPropertyClass_MetaDataItem, kQTMetaDataItemPropertyID_KeyFormat, &outPropType, &propertySize, &outFlags);
		// allocate memory to hold the property value
		//QTPropertyValuePtr 
		char *qvalue = malloc(propertySize);
		// get the property of the metadata item: key format
		err = QTMetaDataGetItemProperty(qref, qitem, kPropertyClass_MetaDataItem, kQTMetaDataItemPropertyID_KeyFormat, propertySize, qvalue, &sizeUsed);
		NSString *keyFormat = [[[NSString alloc] initWithBytes:qvalue length:sizeUsed encoding:NSUTF8StringEncoding] autorelease];
#ifdef __LITTLE_ENDIAN__
		keyFormat = [keyFormat reverse];
#endif
		if ([keyFormat isEqualToString:@"udta"]) continue;
		// get the size of the property: data type
		err = QTMetaDataGetItemPropertyInfo(qref, qitem, kPropertyClass_MetaDataItem, kQTMetaDataItemPropertyID_DataType, &outPropType, &propertySize, &outFlags);
		// allocate memory to hold the property value
		 qvalue = malloc(propertySize);
		// get the property of the metadata item: data type
		err = QTMetaDataGetItemProperty(qref, qitem, kPropertyClass_MetaDataItem, kQTMetaDataItemPropertyID_DataType, propertySize, qvalue, &sizeUsed);
		UInt32 dataType = *(UInt32 *)qvalue;
		//NSLog(@"size: %d, value: %d", (int)sizeUsed, (int)dataType);
		
		// get the size of the property: key
		err = QTMetaDataGetItemPropertyInfo(qref, qitem, kPropertyClass_MetaDataItem, kQTMetaDataItemPropertyID_Key, &outPropType, &propertySize, &outFlags);
		// allocate memory to hold the property value
		qvalue = malloc(propertySize);
		// get the property of the metadata item: key
		err = QTMetaDataGetItemProperty(qref, qitem, kPropertyClass_MetaDataItem, kQTMetaDataItemPropertyID_Key, propertySize, qvalue, &sizeUsed);
		//testType = outPropType;
		//char *buffer = qvalue;
#ifdef __LITTLE_ENDIAN__
		switch (outPropType) {
			case 'code': case 'itsk': case 'itlk': *(int *)qvalue = NSSwapInt(*(int *)qvalue);
		}
#endif
		NSString *keyString = [NSString stringWithCString:qvalue length:sizeUsed];
		free(qvalue);
		if (!keyString || ![keyString length] || [keyString isEqualToString:@"meta"])
			continue;
		
		//if ([keyString isEqualToString:@"nsav"])
			//NSRunAlertPanel(@"Warning", @"Warning: this file won't be able to be saved because it is protected by the \"nsav\" flag", @"OK", nil, nil);
		// get the size of the property: locale
		err = QTMetaDataGetItemPropertyInfo(qref, qitem, kPropertyClass_MetaDataItem, kQTMetaDataItemPropertyID_Locale, &outPropType, &propertySize, &outFlags);
		// allocate memory to hold the property value
		qvalue = malloc(propertySize);
		// get the property of the metadata item: locale
		err = QTMetaDataGetItemProperty(qref, qitem, kPropertyClass_MetaDataItem, kQTMetaDataItemPropertyID_Locale, propertySize, qvalue, &sizeUsed);
		if (sizeUsed) sizeUsed--; // it gives a trailing \000 we don't need
		NSString *langString = [NSString stringWithCString:qvalue length:sizeUsed];
		free(qvalue);
		if ([langString length] < 3) langString = [tLanguages abbrev:langString];
		if (![langString length]) langString = @"English";
		// get the size of the property: value
		err = QTMetaDataGetItemPropertyInfo(qref, qitem, kPropertyClass_MetaDataItem, kQTMetaDataItemPropertyID_Value, &outPropType, &propertySize, &outFlags);
		// allocate memory
		qvalue = malloc(propertySize);
		// get the property of the metadata item: value
		err = QTMetaDataGetItemProperty(qref, qitem, kPropertyClass_MetaDataItem, kQTMetaDataItemPropertyID_Value, propertySize, qvalue, &sizeUsed);
		//buffer = qvalue;
		NSString *valueString = nil;
		int myNumber = 0;
		Float32 myFloat = 0;
		int i;
		SInt8 my1;
		SInt16 my2;
		SInt32 my4;
		switch (dataType) {
			//case 21: case 22:
			case 0: // binary
				if (([keyString isEqualToString:@"disk"] || [keyString isEqualToString:@"trkn"]) && sizeUsed >= 6) {
					valueString = [NSString stringWithFormat:@"%hu of %hu", EndianU16_NtoB( ((UInt16 *)qvalue)[1]), EndianU16_NtoB( ((UInt16 *)qvalue)[2])];
					break;
				} else if ([keyString isEqualToString:[NSString stringWithFormat:@"%cgen", '\251']] && sizeUsed >= 2) {
					// note might need to add "gnre"
					my2 = *(SInt16 *)qvalue;
					my2 = EndianU16_NtoB( my2 );
					if (my2 > 0 && my2 < [[MHMovie gnre] count]) {
						valueString = [[MHMovie gnre] objectAtIndex:my2];
						if ([valueString length]) break;
					}
					valueString = [NSString stringWithFormat:@"%hu", my2];
					break;
				} else if ([keyString isEqualToString:@"purl"] || [keyString isEqualToString:@"egid"] || [self isAsciiFriendly:qvalue length:sizeUsed]) {
					valueString = [NSString stringWithCString:qvalue length:sizeUsed];
					break;
				}
				valueString = [NSMutableString stringWithCapacity:sizeUsed];
				[valueString appendString:@"{binary}"];
				//NSLog(@"sizeUsed: %ld", sizeUsed);
				for (i=0;i<sizeUsed;i++)
					[valueString appendFormat:@" %d", (int)((UInt8 *)qvalue)[i]];
				break;
			case 2: // UTF16 untested
				valueString = [[[NSString alloc] initWithBytes:qvalue length:sizeUsed encoding:0x90000100] autorelease];
				valueString = [NSString stringWithFormat:@"{UTF16}%@", valueString]; break;
			case 3: // MacEncodingText untested
			case 1: // UTF8
				valueString = [[[NSString alloc] initWithBytes:qvalue length:sizeUsed encoding:NSUTF8StringEncoding] autorelease]; break;
			case 12: // GIF
			case 13: // JPEG
			case 14: // PNG
			case 27: // BMP
				break;
			case 21: // SignedIntegerBE
			case 22: // UnsignedIntegerBE
				if (sizeUsed == 1) {
					my1 = *(SInt8 *)qvalue;
					myNumber = my1;
					if ([keyString isEqualToString:@"stik"]) {
						if (my1 == 10) { valueString = @"TV"; break; 
						} else if (my1 == 6) { valueString = @"Music Video"; break;
						} else if (my1 == 14) { valueString = @"Ring Tone"; break;
						} else if (my1 == 0) { valueString = @"Movie"; break; }
					} else if ([keyString isEqualToString:@"pcst"] && my1) {
						valueString = @"Podcast";
						keyString = @"stik";
						break;
					} else if ([keyString isEqualToString:@"itnu"] && my1) {
						valueString = @"iTunes U";
						keyString = @"stik";
						break;
					} else if ([keyString isEqualToString:@"rtng"]) {
						if (my1 == 2) valueString = @"Clean";
						else if (my1 == 4) valueString = @"Explicit";
						else valueString = [NSString stringWithFormat:@"{int}%d", my1];
						break;
					}
				} else if (sizeUsed == 2) {
					//my1 = *(SInt8 *)qvalue;
					my2 = EndianS16_NtoB(*(SInt16 *)qvalue);
					myNumber = my2;
				} else if (sizeUsed == 4) {
					//my2 = *(SInt16 *)qvalue;
					my4 = EndianS32_NtoB(*(SInt32 *)qvalue);
					myNumber = my4;
				}
				//myNumber = (int)*(SInt32 *)qvalue;
				valueString = [NSString stringWithFormat:@"{int}%d", myNumber]; break;
			case 23: // Float32BE untested
			case 24: // Float64BE untested
				myFloat = *(Float32 *)qvalue;
				valueString = [NSString stringWithFormat:@"{float}%f", myFloat]; break;
			case 28: // QT Medadata
				break;
		}
		free(qvalue);
		//NSLog(@"%@ --- %@ --- %d", keyString, valueString, dataType);
		//NSLog(@"value: %@", valueString);
		keyString = [MHMovie itmsTypeToString:keyString];
		if ([keyString hasPrefix:@"com.apple.iTunes"]) [self parseiTunesValue:valueString key:keyString language:langString];
		else if ([keyString length] && [valueString length]) {
		//if ([keyString isEqualToString:@"com.apple.quicktime.year"] && [valueString length]) {
			NSMutableDictionary *cat = [metadata objectForKey:keyString];
			if (!cat) {
				cat = [NSMutableDictionary dictionary];
				[metadata setObject:cat forKey:keyString];
			}
			[cat setObject:valueString forKey:langString];
		} //else NSLog(keyString);
		//NSLog(@"%@\n%@", keyString, valueString);
	}
	//NSLog(@"returned: %d", outCount);
}

- (void)readMTHMetadata
{
	UserData ud = GetMovieUserData([self QTMovie]);
	OSType type = '\251MTH';
	short itlRegionTag = 0;
	Handle h = NewHandleClear(0);
	long i, count = CountUserDataType(ud, type);
	for (i=1;i<=count;i++) {
		for (itlRegionTag=0; itlRegionTag <= 150; itlRegionTag++) {
			if (noErr == GetUserDataText(ud, h, type, i, itlRegionTag)) {
				NSString *lang = [tLanguages numberToString:itlRegionTag];
				int j, length = GetHandleSize(h);
				for (j=0;j<length;j++)
					if (((char *)*h)[j] == '\0') break;
				NSString *key = [NSString stringWithCString:(char *)*h length:j];
				if (j < length) {
					NSString *value = [NSString stringWithCString:(char *)*h+j+1 length:length-j-1];
					if ([key length] && [value length]) {
						NSMutableDictionary *cat = [metadata objectForKey:key];
						if (!cat) {
							cat = [NSMutableDictionary dictionary];
							[metadata setObject:cat forKey:key];
						}
						[cat setObject:value forKey:lang];
					}
				}
			}
		}
	}
}

- (void)readMetadata
{
	[self readQTKitMetadata];
    //NSLog(@"MHMovie readMetadata");
    UserData theUserData = GetMovieUserData([self QTMovie]);
    OSErr err = noErr;
    Handle dataHandle = NewHandle(0);
    short m, itlRegionTag = langEnglish;
    NSString *keyString = nil, *lang = nil;
    NSMutableDictionary *cat = nil;
    long index = 0;
	NSString *optg = [NSString stringWithFormat:@"%c", '\251'];
    
    OSType nextType = GetNextUserDataType(theUserData, 0);
    while (nextType != 0) {
		NSString *type = [NSString stringWithFormat:@"%c%c%c%c",
					  ((char *)&nextType)[0],
					  ((char *)&nextType)[1],
					  ((char *)&nextType)[2],
					  ((char *)&nextType)[3]];
#ifdef __LITTLE_ENDIAN__
		type = [NSString stringWithFormat:@"%c%c%c%c",
				[type characterAtIndex:3], [type characterAtIndex:2],
				[type characterAtIndex:1], [type characterAtIndex:0]];
#endif
        if ([type isEqual:@"WLOC"]) {
            err = GetUserData(theUserData, dataHandle, nextType, 1);
            if (!err) {
                long size = GetHandleSize(dataHandle);
                if (size >= 2 * sizeof(short)) {
                    short *buffer = (short *)*dataHandle;
                    cat = [metadata objectForKey:@"WLOC"];
                    if (!cat) {
                        cat = [NSMutableDictionary dictionary];
                        [metadata setObject:cat forKey:@"WLOC"];
                    }
                    [cat setObject:[NSString stringWithFormat:@"%d, %d", buffer[1], buffer[0]] forKey:@"English"]; // WLOC coords seem to be flipped for some reason.
                }
            }
        } else if ([type isEqual:@"APIC"]) {
            m = CountUserDataType(theUserData, nextType);
            for (index=1;index<=m;index++) {
                err = GetUserData(theUserData, dataHandle, nextType, index);
                if (!err) {
                    pictureResource r = parseTag(dataHandle);
                    NSString *description = [NSString stringWithCString:r.description];
                    NSRange range = [description rangeOfString:@":"];
                    if (range.length) {
                        lang = [description substringToIndex:range.location];
                        description = [description substringFromIndex:range.location+1];
                    } else if (-1 != [tLanguages stringToNumber:description]) lang = description;
                    else lang = @"English";
                    if (![description length]) description = @"a picture";
                    cat = [metadata objectForKey:@"Cover Image"];
                    if (!cat) {
                        cat = [NSMutableDictionary dictionary];
                        [metadata setObject:cat forKey:@"Cover Image"];
                    }
                    [cat setObject:description forKey:lang];
                }
            }
		} else if ([type isEqual:[optg stringByAppendingString:@"MTH"]]) {
			[self readMTHMetadata];
        } else if ([type hasPrefix:optg]) {
            m = CountUserDataType(theUserData, nextType);
            for (index=1; index<=m; index++) {
                itlRegionTag = 0; // langEnglish
                while (itlRegionTag <= 150) {
                    if (noErr == GetUserDataText(theUserData, dataHandle, nextType, index, itlRegionTag)) {
                        //if (noErr == GetUserData(theUserData, dataHandle, nextType, index)) {
                        NSString *lang = [tLanguages numberToString:itlRegionTag];
                        keyString = [self osTypeToString:nextType];
                        if (keyString) {
                            if (index > 1)
                                keyString = [keyString stringByAppendingFormat:@" %ld", index - 1];
                            NSString *value = [NSString stringWithCString:(char *)*dataHandle length:GetHandleSize(dataHandle)];
                            if ([value length]) {
                                cat = [metadata objectForKey:keyString];
                                if (!cat) {
                                    cat = [NSMutableDictionary dictionary];
                                    [metadata setObject:cat forKey:keyString];
                                }
                                [cat setObject:value forKey:lang];
                            }
                        }
                    }
                    itlRegionTag++;
                }
            }
/*		} else if ([type isEqualToString:@"ptv "]) { // auto fullscreen // this has been moved to tTracks
			err = GetUserData(theUserData, dataHandle, nextType, 1);
            if (!err) {
				unichar *samp = (unichar *)*dataHandle;
                long size = GetHandleSize(dataHandle);
				if (size >= 4*sizeof(unichar)) {
					if (256 == samp[3]) { // 768, 0, 0, 256
						cat = [metadata objectForKey:@"Auto Fullscreen"];
						if (!cat) {
							cat = [NSMutableDictionary dictionary];
							[metadata setObject:cat forKey:@"Auto Fullscreen"];
						}
						[cat setObject:@"YES" forKey:@"English"];
					}
				}
            } */
        } else { // unknown type
			NSArray *ignore = [NSArray arrayWithObjects:@"AllF", @"SelO", @"play", @"aqui", @"ptv ", @"aclo", nil];
			if (![ignore containsObject:type]) {
				err = GetUserData(theUserData, dataHandle, nextType, 1);
				long size = GetHandleSize(dataHandle);
				int i = 0;
				if (size > 3) i = (int)((UInt8 *)*dataHandle)[3];
				if ([type isEqualToString:@"nsav"] && i == 1)
					NSRunAlertPanel(@"Warning", @"Warning: this file won't be able to be saved because it is protected by the \"nsav\" flag", @"OK", nil, nil);
				NSString *v = nil;
				if (size && [self isAsciiFriendly:*dataHandle length:size])
					v = [NSString stringWithCString:(char *)*dataHandle length:size];
				else {
					//NSLog(@"size: %ld", size);
					v = [NSMutableString stringWithCapacity:size];
					[v appendString:@"{binary}"];
					for (i=0;i<size;i++)
						[v appendFormat:@" %d", (int)((UInt8 *)*dataHandle)[i]];
				}
				if ([v length]) {
					cat = [metadata objectForKey:type];
					if (!cat) {
						cat = [NSMutableDictionary dictionary];
						[metadata setObject:cat forKey:type];
					}
					[cat setObject:v forKey:@"English"];
				}
			}
			
			//NSLog(@"Ingoring unknown metadata type: %@, %d, %d", type, size, ch);
			
			//NSLog(@"Ignoring unknown metadata type: %c%c%c%c", ((char *)(&nextType))[0], ((char *)(&nextType))[1], ((char *)(&nextType))[2], ((char *)(&nextType))[3]);
		}
        nextType = GetNextUserDataType(theUserData, nextType);
    }
	//[self readQTKitMetadata];
    [_sortedKeys release];
    _sortedKeys = nil;
}

- (NSString *)osTypeToString:(OSType)type
{
    int i, n = [[MHMovie supportedMetadata] count];
    for (i=0;i<n;i++) {
        if (type == theKUserDataTypes[i])
            return [[MHMovie supportedMetadata] objectAtIndex:i];
    }
    char *buffer = (char *)&type;
    //if (buffer[0] != '\251')
    //    return nil;
    NSString *string = [NSString stringWithCString:buffer length:4];
    // check if it's just the caps version
    NSString *lowString = [string lowercaseString];
    OSType lowType = [self stringToOSType:lowString];
    for (i=0;i<n;i++)
        if (lowType == theKUserDataTypes[i])
            return [[MHMovie supportedMetadata] objectAtIndex:i];
    return string;
}

- (OSType)stringToOSType:(NSString *)s
{
    if (nil == s || [s isEqualToString:@""]) return 0;
    int i;
    char target = '\251';
    
    char first = [s characterAtIndex:0];
    if (first == target) {
        while ([s length] < 4)
            s = [s stringByAppendingString:@" "];
        long output = '\251' * 256 * 256 * 256;
        output += [s characterAtIndex:1] * 256 * 256;
        output += [s characterAtIndex:2] * 256;
        output += [s characterAtIndex:3];
        OSType type = output;
        return type;
    }
    i = [[MHMovie supportedMetadata] indexOfObject:s];
    if (NSNotFound == i) {
        NSString *last = [[s componentsSeparatedByString:@" "] lastObject];
        if ([last intValue]) s = [s substringToIndex:[s length] - [last length] - 1];
        i = [[MHMovie supportedMetadata] indexOfObject:s];
    }
    if (NSNotFound != i) return theKUserDataTypes[i];
    
    return 0;
}

- (NSRect) combineRect:(NSRect)aRect withRect:(NSRect)bRect
{
    // for top-left referenced rects
    NSRect r;
    // we want the lesser x
    if (aRect.origin.x < bRect.origin.x) r.origin.x = aRect.origin.x;
    else r.origin.x = bRect.origin.x;
    // we want the greater y;
    if (aRect.origin.y > bRect.origin.y) r.origin.y = aRect.origin.y;
    else r.origin.y = bRect.origin.y;
    // we want the greater x+width
    if (aRect.origin.x + aRect.size.width > bRect.origin.x + bRect.size.width)
        r.size.width = aRect.origin.x + aRect.size.width - r.origin.x;
    else r.size.width = bRect.origin.x + bRect.size.width - r.origin.x;
    // we want the lesser y-height;
    if (aRect.origin.y - aRect.size.height < bRect.origin.y - bRect.size.height)
        r.size.height = aRect.origin.y - aRect.size.height + r.origin.y;
    else r.size.height = bRect.origin.y - bRect.size.height + r.origin.y;
    
    return r;
}

- (void)gatherData
{
	// file system stuff
	NSDictionary *attributes = [[NSFileManager defaultManager] fileAttributesAtPath:stringPath traverseLink:YES];
	if (attributes) {
		NSDate *creationDate = [attributes objectForKey:NSFileCreationDate];
		[data setObject:creationDate forKey:@"File Creation Date"];
                NSDate *modificationDate = [attributes objectForKey:NSFileModificationDate];
                [data setObject:modificationDate forKey:@"File Modification Date"];
	}
	
    //NSLog(@"MHMovie gatherData");
    Movie mov = [self QTMovie];
    if ([data objectForKey:@"Seconds Elapsed"])
        [self setCurrentTimeIndexSeconds:[[data objectForKey:@"Seconds Elapsed"] doubleValue]];
    else
        [data setObject:[NSNumber numberWithDouble:0] forKey:@"Seconds Elapsed"];
    TimeValue duration = GetMovieDuration(mov);
    TimeScale scale = GetMovieTimeScale(mov);
    [data setObject:[NSNumber numberWithInt:duration/scale] forKey:@"Seconds"];
    NSString *movieLength = [MHUtil nSecondsToString:(double)duration/(double)scale];
    int i = -1;
    char ch;
    while (YES) {
        ch = [movieLength characterAtIndex:++i];
        if (ch != '0' && ch != ':') break;
    }
    OSErr theErr = noErr;
    wide dataSize;
    double mbSize;
    [data setObject:[movieLength substringFromIndex:i] forKey:@"Movie Length"];
    theErr = GetMovieDataSize64(mov, 0, duration, &dataSize);
    if (theErr != noErr) mbSize = -1;
    else mbSize = (double)Wide2LongLong(dataSize) / 1024.0 / 1024.0;
    [data setObject:[NSNumber numberWithDouble:mbSize] forKey:@"File Size"];
    Rect boxRect;
    GetMovieBox(mov, &boxRect);
    [data setObject:[NSNumber numberWithInt:boxRect.bottom-boxRect.top] forKey:@"Original Height"];
    [data setObject:[NSNumber numberWithInt:boxRect.bottom-boxRect.top] forKey:@"Saved Height"];
    [data setObject:[NSNumber numberWithInt:boxRect.right-boxRect.left] forKey:@"Original Width"];
    [data setObject:[NSNumber numberWithInt:boxRect.right-boxRect.left] forKey:@"Saved Width"];
    //NSLog(@"original: %dx%d", boxRect.right-boxRect.left, boxRect.bottom-boxRect.top);
    
    [data setObject:[self fpsForMovie:mov] forKey:@"FPS"];
    
    Track theTrack = nil;
    Media aMedia;
    char *buffer;
    //*width = *height = *vBitrate = *aBitrate = 0;
    
    // Video
    NSString *vCodec, *vFourCC, *vVendor;
    NSNumber *width, *height, *vBitrate;
    ImageDescriptionHandle vDesc = (ImageDescriptionHandle)NewHandle(sizeof(ImageDescription));
    
    //if (DEBUG)
//	NSLog(@"getting video...");
        theTrack = GetBiggestTrack(mov, VideoMediaType);    
        if (theTrack == nil) {
	   // NSLog(@"no video, getting mpeg...");
            theTrack = GetBiggestTrack(mov, MPEGMediaType);
        }
	if (theTrack == nil) {
	   // NSLog(@"no video track found");
            vCodec = vFourCC = vVendor = @"none";
            width = height = vBitrate = [NSNumber numberWithInt:0];
        } else {
	    MHTrack *track = [[[MHTrack alloc] initWithTrack:theTrack] autorelease];
            aMedia = GetTrackMedia(theTrack);
            GetMediaSampleDescription(aMedia, 1, (SampleDescriptionHandle)vDesc);
            MoveHHi((Handle)vDesc); //HLock((Handle)vDesc);
            height = [NSNumber numberWithInt:(*vDesc)->height];
            width = [NSNumber numberWithInt:(*vDesc)->width];
            buffer = (char *)&(*vDesc)->name + 1;
			int bufSize = *(*vDesc)->name;
//#ifdef __LITTLE_ENDIAN__
//buffer = rstring(buffer, bufSize);
//#endif
            vCodec = [NSString stringWithCString:buffer length:bufSize];
            buffer = (char *)&(*vDesc)->cType;
#ifdef __LITTLE_ENDIAN__
buffer = rstring(buffer, 4);
#endif
            vFourCC = [NSString stringWithCString:buffer];
            buffer = (char *)&(*vDesc)->vendor;
#ifdef __LITTLE_ENDIAN__
buffer = rstring(buffer, 4);
#endif
            vVendor = [NSString stringWithCString:buffer];
			double mb = [track mbSize];
			mb /= (double)GetTrackDuration(theTrack) / (double)scale;
			mb *= 1024;
            vBitrate = [NSNumber numberWithInt:mb];
        }
        NSRect frame = NSMakeRect(0,0,0,0);
        int n = [self trackCount];
        for (i=0;i<n;i++) {
            NSRect aFrame = [[self trackAtIndex:i] displayFrame];
            frame = [self combineRect:frame withRect:aFrame];
        }
        [data setObject:[NSNumber numberWithInt:frame.size.height] forKey:@"Correct Height"];
        [data setObject:[NSNumber numberWithInt:frame.size.width] forKey:@"Correct Width"];
        //NSLog(@"correct: %gx%g", frame.size.width, frame.size.height);
    // Audio
    NSString *aCodec;
    NSNumber *aChannels, *sampleRate, *aBitrate;
    
    SoundDescriptionHandle sDesc = (SoundDescriptionHandle)NewHandle(sizeof(SoundDescription));
        theTrack = nil;
        theTrack = GetBiggestTrack(mov, SoundMediaType);
        if (nil == theTrack)
            theTrack = GetBiggestTrack(mov, MPEGMediaType);
        if (nil == theTrack) {
            aCodec = @"none";
            aChannels = sampleRate = aBitrate = [NSNumber numberWithInt:0];
        } else {
            aMedia = GetTrackMedia(theTrack);
            GetMediaSampleDescription(aMedia, 1, (SampleDescriptionHandle)sDesc);
            MoveHHi((Handle)sDesc); HLock((Handle)sDesc);
            aChannels = [NSNumber numberWithInt:(*sDesc)->numChannels];
			NSString *sum = [[QTTrack trackWithQuickTimeTrack:theTrack error:nil] attributeForKey:QTTrackFormatSummaryAttribute];
			NSScanner *scan = [NSScanner scannerWithString:sum];
			[scan scanUpToString:@", " intoString:nil];
			[scan scanString:@", " intoString:nil];
			[scan scanUpToString:@" " intoString:&sum];
			if (![sum isEqualToString:@"Mono, "] && ![sum isEqualToString:@"Stereo, "] && [sum floatValue]) aChannels = [NSNumber numberWithFloat:[sum floatValue]];
            buffer = (char *)&(*sDesc)->dataFormat;
#ifdef __LITTLE_ENDIAN__
buffer = rstring(buffer, 4);
#endif
            aCodec = [NSString stringWithCString:buffer];
			TimeScale ts = GetMediaTimeScale(aMedia);
			if (ts % 10) { // don't know why we have to hunt around on some of these
				UnsignedFixed r = (*sDesc)->sampleRate;
				ts = r >> 16;
				if (ts % 10) {
					r = EndianU32_NtoB((*sDesc)->sampleRate);
					ts = r >> 16;
				}
			}
			//NSLog(@"%@: %ld", [self stringValue], ts);
			/*long long c = GetMediaSampleCount(aMedia);
			double rate = (double)c / GetTrackDuration(theTrack) * scale * 1024;
			rate += 0.49; // rounding*/
			sampleRate = [NSNumber numberWithLong:ts];
			TimeValue dura = GetTrackDuration(theTrack);
			long size = GetTrackDataSize(theTrack, 0, dura);
			if (size && dura && scale)
				aBitrate = [NSNumber numberWithInt:size * 8 * scale / dura / 1024]; // integer math, 0 when duration < timescale (anything under 1 second) when ordered to make human sense
            buffer = (char *)&(*sDesc)->vendor;
        }
    
    NSArray *objects = [NSArray arrayWithObjects:vCodec, vFourCC, vVendor, width, height, vBitrate, aChannels, aCodec, sampleRate, aBitrate, nil];
    NSArray *keys = [NSArray arrayWithObjects:@"Video Format", @"Video FourCC", @"Video Vendor", @"Video Width", @"Video Height", @"Video Bitrate", @"Audio Channels", @"Audio Format", @"Audio Sample Rate", @"Audio Bitrate", nil];
    n = [objects count];
    for (i=0;i<n;i++) {
        [data setObject:[objects objectAtIndex:i] forKey:[keys objectAtIndex:i]];
    }
	int aCount = 0;
	while (YES) {
		theTrack = GetMovieIndTrackType(mov, aCount+1, SoundMediaType, movieTrackMediaType);
		if (nil == theTrack) break;
		else aCount++;
	}
	[data setObject:[NSNumber numberWithInt:aCount] forKey:@"Audio Count"];
    
    // TimeCode stuff
    UserData			myUserData;
    MediaHandler myHandler = GetMediaHandler(GetTrackMedia(GetMovieIndTrackType(mov, 1, TimeCodeMediaType, movieTrackMediaType)));
    if (myHandler) {
        theErr = TCGetCurrentTimeCode(myHandler, NULL, NULL, NULL, &myUserData);
        if (!theErr) {
            Handle h = NewHandleClear(0);
            GetUserDataText(myUserData, h, TCSourceRefNameType, 1, langEnglish);
            NSString *name = [NSString stringWithCString:*h length:GetHandleSize(h)];
            [data setObject:name forKey:@"Source Name"];
            if (h) DisposeHandle(h);
            DisposeUserData(myUserData);
        }
        TimeCodeDef			myTCDef;
        TimeCodeRecord                  myTCRec;
        theErr = TCGetCurrentTimeCode(myHandler, NULL, &myTCDef, &myTCRec, NULL);
        if (!theErr) {
            Str255		myString;
            theErr = TCTimeCodeToString(myHandler, &myTCDef, &myTCRec, myString);
            if (!theErr) {
                NSString *time = [NSString stringWithCString:(char *)myString+1 length:*myString];
                [data setObject:time forKey:@"Timecode"];
            }	
        }
    }
    [[NSNotificationCenter defaultCenter] postNotificationName:@"liveUpdateNotification" object:nil userInfo:[NSDictionary dictionaryWithObject:self forKey:@"ALL"]];
}

- (void)setSourceName:(NSString *)s
{
	UserData			myUserData;
	TimeCodeDescriptionHandle tcdH;
	Media				myMedia = GetTrackMedia(GetMovieIndTrackType([self QTMovie], 1, TimeCodeMediaType, movieTrackMediaType));
	OSErr theErr = BeginMediaEdits(myMedia);
    MediaHandler myHandler = GetMediaHandler(myMedia);
	Handle h = NewHandleClear([s length]);
	Handle h2 = NewHandleClear([s length]);
	Handle h3 = NewHandleClear([s length]);
	int n;
    if (myHandler) {
        theErr = TCGetCurrentTimeCode(myHandler, NULL, NULL, NULL, &myUserData);
		//myUserData = GetMovieUserData([self QTMovie]);
		n = CountUserDataType(myUserData, TCSourceRefNameType);
		//theErr = NewUserData(&myUserData);
        if (!theErr) {
			const char *buffer = [[s dataUsingEncoding:NSMacOSRomanStringEncoding allowLossyConversion:YES] bytes];
			if (buffer) {
				BlockMoveData(buffer, *h, [s length]);
				theErr = AddUserDataText(myUserData, h, TCSourceRefNameType, 1, langEnglish);
				//theErr = GetUserData(myUserData, h2, TCSourceRefNameType, 1);
				n = CountUserDataType(myUserData, TCSourceRefNameType);
				tcdH = (TimeCodeDescriptionHandle)NewHandleClear(sizeof(TimeCodeDescription));
				GetMediaSampleDescription( myMedia, 1, (SampleDescriptionHandle)tcdH);
				theErr = GetUserDataText(myUserData, h2, TCSourceRefNameType, 1, langEnglish);
				n = CountUserDataType(myUserData, TCSourceRefNameType);
				theErr = TCSetSourceRef(myHandler, tcdH, myUserData);
				theErr = GetUserDataText(myUserData, h3, TCSourceRefNameType, 1, langEnglish);
			}
		}
	}
	theErr = EndMediaEdits(myMedia);
	
	UserData u;
	theErr = TCGetCurrentTimeCode(myHandler, nil, nil, nil, &u);
	theErr = GetUserDataText(u, h3, TCSourceRefNameType, 1, langEnglish);
	theErr = GetUserDataText(myUserData, h3, TCSourceRefNameType, 1, langEnglish);
	n = CountUserDataType(u, TCSourceRefNameType);
	
	[self gatherData];
}

- (NSNumber *)fpsForMovie:(Movie)theMovie
{
    TimeValue 			start = GetMovieDuration(theMovie) / 2;
    TimeValue			end = start + (GetMovieTimeScale(theMovie) * 10);
    TimeValue			duration = GetMovieDuration(theMovie);
	
	if (end > duration) {
		start = 0;
		end = duration;
	}
	int numFrames = [MHUtil framesBetweenStart:start end:end movie:[self QTMovie]];
	
	double fps = (double)numFrames;
	double seconds = end - start;
	seconds /= GetMovieTimeScale(theMovie);
	fps /= seconds;
	
	//fps += 0.49; // for proper rounding
	
	return [NSNumber numberWithFloat:fps];
}

/*
- (IconFamily *)iconFromFile:(NSString *)path
{
    if (path == nil) return nil;
    IconFamily *icon = nil;
    NSImage *image = nil;
    if ([path hasSuffix:@".icns"])
        icon = [IconFamily iconFamilyWithContentsOfFile:path];
    else {
        if ([path hasPrefix:@"http:"] || [path hasPrefix:@"ftp:"])
            image = [[NSImage alloc] initWithContentsOfURL:[NSURL URLWithString:path]];
        else
            image = [[NSImage alloc] initWithContentsOfFile:path];
        if (image != nil) {
            [image retain];
            icon = [IconFamily iconFamilyWithThumbnailsOfImage:image];
        }
    }
    return icon;
}
*/
- (BOOL)askForMovie
{
    NSOpenPanel *theOpenPanel = [NSOpenPanel openPanel];
    int result;
    
    [theOpenPanel setTitle:[NSString stringWithFormat:@"Please Locate %@",[stringPath lastPathComponent]]];
    [theOpenPanel setCanChooseFiles:YES];
    [theOpenPanel setCanChooseDirectories:NO];
    [theOpenPanel setResolvesAliases:YES];
    [theOpenPanel setAllowsMultipleSelection:NO];
    result = [theOpenPanel runModalForDirectory:nil file:nil types:nil];
    if (result == NSOKButton) {
        stringPath = [[theOpenPanel filenames] objectAtIndex:0];
        return YES;
    }
    return NO;
}

int IsChapterTrack(Track track) {
    Movie mov = GetTrackMovie(track);
    int i, n = GetMovieTrackCount(mov);
    for (i=1;i<=n;i++) {
        if (GetTrackReference(GetMovieIndTrack(mov, i), kTrackReferenceChapterList, 1) == track) return i;
    }
    return NO;
}


// from MyMovie.m
// http://developer.apple.com/samplecode/Sample_Code/Cocoa/bMoviePalette/MyMovie.m.htm
static BOOL WriteMovieToPath(Movie movie, NSString *path, BOOL flatten)
{
    // get QT version
    long version;
    OSErr err;
    FSRef    fsRef;
    FSSpec   fsSpec;
    short    resRefNum;
    err = Gestalt(gestaltQuickTime, &version);
    if ((noErr == err) && version >= 0x06408000)
    {
        // QT version is 6.4 or higher
        Handle dataRef = NULL;
        OSType dataRefType;
        DataHandler dataHandler;
        
        err = QTNewDataReferenceFromFullPathCFString((CFStringRef)path, kQTNativeDefaultPathStyle, 0, &dataRef, &dataRefType);
        if (flatten) {
            Movie mov = FlattenMovieDataToDataRef(movie, flattenAddMovieToDataFork | flattenForceMovieResourceBeforeMovieData | flattenCompressMovieResource, dataRef, dataRefType, FOUR_CHAR_CODE('TVOD'), smSystemScript, createMovieFileDeleteCurFile);
            if (!mov) NSLog(@"couldn't flatten");
        } else { // not flattening
            Track aTrack;
            // find any chapter tracks
            int i, j, n = GetMovieTrackCount(movie);
            NSMutableArray *chapTrackNumbers = [NSMutableArray array];
            NSMutableArray *targTrackNumbers = [NSMutableArray array];
            for (i=1;i<=n;i++) {
                aTrack = GetMovieIndTrack(movie, i);
                int target = IsChapterTrack(aTrack);
                if (target) {
                    [chapTrackNumbers addObject:[NSNumber numberWithInt:i]];
                    [targTrackNumbers addObject:[NSNumber numberWithInt:target]];
                }
            }
            // adjust the target numbers for moving chapters to the start
            NSMutableArray *newTargTrackNumbers = [NSMutableArray array];
            for (i=0;i<[targTrackNumbers count];i++) {
                //int subtractor = 0;
                int targNum = [[targTrackNumbers objectAtIndex:i] intValue];
                // subtract one for every chapter track below it
                for (j=0;j<[chapTrackNumbers count];j++)
                    if ([[chapTrackNumbers objectAtIndex:j] intValue] < targNum)
                        targNum--;
                // add the number of chapter tracks, since they will all be at the start
                targNum += [chapTrackNumbers count];
                [newTargTrackNumbers addObject:[NSNumber numberWithInt:targNum]];
            }
            targTrackNumbers = newTargTrackNumbers;
            
            // save the chapter tracks to a new movie
            if ([chapTrackNumbers count]) {
                MovieEditState srcState = NewMovieEditState(movie);
                
                // remove chapter references
                for (i=1;i<GetMovieTrackCount(movie);i++) {
                    aTrack = GetTrackReference(GetMovieIndTrack(movie, i), kTrackReferenceChapterList, 1);
                    if (aTrack)
                        err = DeleteTrackReference(GetMovieIndTrack(movie, i), kTrackReferenceChapterList, 1);
                }
                
                // remove non-chapter tracks
                for (i=GetMovieTrackCount(movie);i>0;i--) {
                    if (![chapTrackNumbers containsObject:[NSNumber numberWithInt:i]])
                        DisposeMovieTrack(GetMovieIndTrack(movie, i));
                }
                
                // save and revert
                WriteMovieToPath(movie, path, YES);
                err = UseMovieEditState(movie, srcState);
                err = DisposeMovieEditState(srcState);
                srcState = NewMovieEditState(movie);
                
                // remove chapter tracks
                for (i=[chapTrackNumbers count]-1;i>=0;i--) {
                    int index = [[chapTrackNumbers objectAtIndex:i] intValue];
                    DisposeMovieTrack(GetMovieIndTrack(movie, index));
                }
                
                // update new saved movie
                err = FSPathMakeRef((const UInt8 *)[path fileSystemRepresentation], &fsRef, NULL);
                err = FSGetCatalogInfo(&fsRef, kFSCatInfoNone, NULL, NULL, &fsSpec, NULL);
                err = OpenMovieFile(&fsSpec, &resRefNum, fsWrPerm);
                Movie newMov = nil;
                short resID = movieInDataForkResID;
                Str255 resName;
                err = NewMovieFromFile(&newMov, resRefNum, &resID, resName, newMovieActive, NULL);
                
                SetMovieSelection(movie, 0, 0);
                SetMovieSelection(newMov, 0, 0);
                AddMovieSelection(newMov, movie);
                err = CopyMovieUserData(movie, newMov, kQTCopyUserDataReplace);
                
                // restore chapter references
                for (i=0;i<[chapTrackNumbers count];i++) {
                    int targT = [[targTrackNumbers objectAtIndex:i] intValue];
                    AddTrackReference(GetMovieIndTrack(newMov, targT), GetMovieIndTrack(newMov, i+1), kTrackReferenceChapterList, NULL);
                }
                
                err = UpdateMovieResource(newMov, resRefNum, movieInDataForkResID, NULL);
				if (err) NSLog(@"error %ld writing file %@", err, path);
                DisposeMovie(newMov);
                CloseMovieFile(resRefNum);
                
                // revert again
                err = UseMovieEditState(movie, srcState);
                err = DisposeMovieEditState(srcState);
                srcState = nil;
                
                if (err) return NO;
                return YES;
            } else { // no chapters
                err = CreateMovieStorage(dataRef, dataRefType, FOUR_CHAR_CODE('TVOD'), smSystemScript, 0, // createMovieFileDeleteCurFile,
                    &dataHandler, NULL);
                //NSMovie *refMov = [[NSMovie alloc] initWithURL:[NSURL fileURLWithPath:path] byReference:NO];
                err = UpdateMovieInStorage(movie, dataHandler);
                //AddMovieToStorage(movie, dataHandler);
                err = CloseMovieStorage(dataHandler);
                //CloseComponent(dataHandler);
                /* rsrc fork style
                if (![[NSFileManager defaultManager] fileExistsAtPath:path]) [@"" writeToFile:path atomically:NO];
                err = FSPathMakeRef([path fileSystemRepresentation], &fsRef, NULL);
                err = FSGetCatalogInfo(&fsRef, kFSCatInfoNone, NULL, NULL, &fsSpec, NULL);
                short resId = 0, newResId = 0;
                Movie newmovie = NULL;
                err = CreateMovieFile(&fsSpec, 'TVOD', smSystemScript, nil, &resId, &newmovie);
                err = AddMovieResource(movie, resId, &newResId, nil);
                CloseMovieFile(resId);
                */
            }
        }
        DisposeHandle(dataRef);
        if (err) return NO;
        return YES;
    }
    // QT is less than 6.4
    if (!flatten) {
        Handle h = NewHandle(0);
        PutMovieIntoHandle(movie, h);
        HLock(h);
        NSData *d = [NSData dataWithBytes:*h length:GetHandleSize(h)];
        DisposeHandle(h);
        [d writeToFile:path atomically:YES];
        return YES;
    }
    // flattening, old QT
    NSString *newPath = path;
    NSFileManager *fm = [NSFileManager defaultManager];
    
    // old QT don't like long file names
    /*
    newPath = [NSString stringWithFormat:@"%@~",path];
    newName = [newPath lastPathComponent];
    if ([newName length] >= 25) {
        newName = [newName substringToIndex:25];
        newFolder = [newPath stringByDeletingLastPathComponent];
        newPath = [newFolder stringByAppendingPathComponent:newName];
        tempPath = newPath;
        int count = 1;
        while ([fm fileExistsAtPath:tempPath])
            tempPath = [NSString stringWithFormat:@"%@%d", newPath, count++];
        newPath = tempPath;
    }
    */
    if (![fm fileExistsAtPath:path])
        [[NSString stringWithString:@"this file is empty"] writeToFile:path atomically:NO];
    err = FSPathMakeRef((const UInt8 *)[newPath fileSystemRepresentation], &fsRef, NULL);
    if (err) return NO;
    err = FSGetCatalogInfo(&fsRef,
                kFSCatInfoNone,
                NULL, // catalogInfo
                NULL, // outName
                &fsSpec, NULL // parentRef
                );

    if (err) return NO;
    short resId;
    FlattenMovie(movie,
            flattenAddMovieToDataFork | flattenForceMovieResourceBeforeMovieData | flattenCompressMovieResource,
            &fsSpec,'TVOD',
            smSystemScript,
            createMovieFileDeleteCurFile | createMovieFileDontCreateResFile,
                &resId,nil);
    CloseMovieFile(resId);
     /*
        old ref mov method
                Movie newmovie;
                short resId;
                OSType creatorType = FOUR_CHAR_CODE('TVOD');
                err = CreateMovieFile(&fsSpec, creatorType, smSystemScript, nil, &resId, &newmovie); 
                err = AddMovieResource (movie, resId, &newResId, nil);
                CloseMovieFile(resId);
    */
    rename([newPath fileSystemRepresentation],[path fileSystemRepresentation]);
    //[[NSFileManager defaultManager] movePath:newPath toPath:path handler:nil];
    return YES;
        /* Clean up here */
    //unlink([newPath fileSystemRepresentation]);
}

Track GetBiggestTrack(Movie qtMovie, OSType type) {
    Track aTrack, theTrack;
    aTrack = theTrack = nil;
    int i = 1;
    wide w;
    if (nil == qtMovie) return nil;
    w.hi = w.lo = 0;
    long long newsize, oldsize;
    newsize = oldsize = 0;
    OSErr err = noErr;
    aTrack = GetMovieIndTrackType(qtMovie, i++, type, movieTrackMediaType);
    while (aTrack != nil) {
        err = GetTrackDataSize64(aTrack, 0, GetTrackDuration(aTrack), &w);
	newsize = Wide2LongLong(w);
	if (DEBUG)
	    NSLog(@"size %d: %lld", i-1, newsize);
        if (!theTrack || newsize > oldsize) {
            oldsize = newsize;
            theTrack = aTrack;
        }
        aTrack = GetMovieIndTrackType(qtMovie, i++, type, movieTrackMediaType);
    }
    if (DEBUG)
	NSLog(@"GetBiggestTrack: %d tracks tested", i-2);
    return theTrack;
}

long long Wide2LongLong(wide w) {
    long long result = (((long long)w.hi)<<32)|w.lo;
    return result;
}

char *rstring(char *s, int length) {
	char *buffer = malloc(sizeof(char)*(length+1));
	buffer[length] = '\0';
	int i, j;
	for (i=0,j=length-1;i<length;i++,j--) {
		buffer[j] = s[i];
	}
	return buffer;
}

@end
