//
//  MHRectView.m
//  Hootenanny
//
//  Created by Noah Sorscher on 7/24/10.
//  Copyright 2010 __MyCompanyName__. All rights reserved.
//

#import "MHRectView.h"
#import "tTracks.h"

@implementation MHRectView

- (id)initWithFrame:(NSRect)frame {
    self = [super initWithFrame:frame];
    if (self) {
        // Initialization code here.
		myRect = NSMakeRect(0, 0, 0, 0);
		dragging = 0;
		coord = NSMakeRect(0, 0, 1.0, 1.0);
    }
    return self;
}

- (BOOL)shouldDelayWindowOrderingForEvent:(NSEvent *)theEvent; 
{
	return YES;
}

- (BOOL)acceptsFirstMouse:(NSEvent *)theEvent; 
{
	return YES; 
}

- (void)setRect:(NSRect)r
{
	myRect = r;
	
	
	// we are going to scale the rect now, instead of at every drawing frame, since the scale can't change anyway
	myRect.origin.x *= coord.size.width;
	myRect.origin.y *= coord.size.height;
	myRect.size.width *= coord.size.width;
	myRect.size.height *= coord.size.height;
	
	
	float h = [self frame].size.height + coord.origin.y; // h is the height relative to the QT origin
	myRect.origin.y = h - myRect.origin.y; // Cocoa origin is lower left and QT origin is upper left; this puts the rect into Cocoa format
	//myRect.origin.y += coord.origin.y; // coord.origin is a correction factor for when r has an origin that's not in the corner
	[self setNeedsDisplay:YES];
	//NSLog(@"setRect: %@", NSStringFromRect(myRect));
}

- (void)setCoord:(NSRect)c
{
	coord = c; 
	coord.origin.x *= coord.size.width;
	coord.origin.y *= coord.size.height;
	[self setNeedsDisplay:YES];
	//NSLog(@"setScale: %gx%g", scale.width, scale.height);
}

- (void)drawRect:(NSRect)dirtyRect {
    // Drawing code here.
	//NSLog(@"drawRect: %@", NSStringFromRect(myRect));
	NSRect r = myRect;
	
	r.origin.x -= coord.origin.x;
	//r.origin.y -= coord.origin.y;
	
	[[NSColor redColor] set];
	[NSBezierPath setDefaultLineWidth:2];
	[NSBezierPath strokeRect:r];
	//NSLog(@"drawRect: %@", NSStringFromRect(r));
}

- (void)mouseDown:(NSEvent *)e
{
	[NSApp preventWindowOrdering]; 
	NSPoint p = [e locationInWindow];
	p.x += coord.origin.x;
	
	if (NSPointInRect(p, NSMakeRect(myRect.origin.x - 5, myRect.origin.y - 5, 10, myRect.size.height + 10)))
		dragging = 1; // left side
	else if (NSPointInRect(p, NSMakeRect(myRect.origin.x - 5, myRect.origin.y - 5, myRect.size.width + 10, 10)))
		dragging = 2; // bottom side
	else {
		int x = myRect.origin.x + myRect.size.width,
		y = myRect.origin.y + myRect.size.height;
		if (NSPointInRect(p, NSMakeRect(x-5, myRect.origin.y - 5, 10, myRect.size.height + 10)))
			dragging = 3; // right side
		else if (NSPointInRect(p, NSMakeRect(myRect.origin.x - 5, y-5, myRect.size.width + 10, 10)))
			dragging = 4; // top side
		else {
			dragging = 0;
		}

	}
}

- (void)mouseDragged:(NSEvent *)e
{
	NSPoint p = [e locationInWindow];
	p.x += coord.origin.x;
	switch (dragging) {
		case 1: // left side
			myRect.size.width += myRect.origin.x - p.x;
			myRect.origin.x = p.x;
			break;
		case 2: // bottom side
			myRect.size.height += myRect.origin.y - p.y;
			myRect.origin.y = p.y;
			break;
		case 3: // right side
			myRect.size.width = p.x - myRect.origin.x;
			break;
		case 4: // top side
			myRect.size.height = p.y - myRect.origin.y;
			break;
		default:
			myRect.origin.x = p.x - myRect.size.width / 2;
			myRect.origin.y = p.y - myRect.size.height / 2;
	}
	if (myRect.size.height < 1) myRect.size.height = 1;
	if (myRect.size.width < 1) myRect.size.width = 1;
	NSRect r = myRect;
	float h = [self frame].size.height + coord.origin.y;
	r.origin.y = h - r.origin.y;
	
	r.origin.x /= coord.size.width;
	r.size.width /= coord.size.width;
	r.origin.y /= coord.size.height;
	r.size.height /= coord.size.height;
	
	[tTracks setTrackRect:r];
	[self setNeedsDisplay:YES];
}

@end
