/* Copyright (c) 2004, Noah Sorscher
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 	 	Redistributions of source code must retain the above copyright
 * 	notice, this list of conditions and the following disclaimer.
 * 	 	Redistributions in binary form must reproduce the above copyright
 * 	notice, this list of conditions and the following disclaimer in the
 * 	documentation and/or other materials provided with the distribution.
 * 	 	Neither the name of the Noah Sorscher nor the names of its
 * 	contributors may be used to endorse or promote products derived from
 * 	this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
#import "MHSprite.h"
#import "QTImage.h"
#import "SpriteDict.h"
#define TRACK_DURATION 1000
#define DEBUG NO

@interface NSOutlineView (spriteStuff)
- (id) selectedItem;
- (void)selectItem:(id)item;
@end

@implementation NSOutlineView (spriteStuff)

- (id) selectedItem
{
    int index = [self selectedRow];
    if (index == -1) return nil;
    return [self itemAtRow:index];
}

- (void)selectItem:(id)item
{
	int row = [self rowForItem:item];
	NSIndexSet *set = [NSIndexSet indexSetWithIndex:row];
	[self selectRowIndexes:set byExtendingSelection:NO];
}

@end

@interface NSMutableArray (duh)
- (void) addObjectsFromArrayIf:(NSArray *)a prefix:(NSString *)pref suffix:(NSString *)suf;
- (void) addObjectsFromArrayIf:(NSArray *)a prefix:(NSString *)pref;
@end

@implementation NSMutableArray (duh)

- (void) addObjectsFromArrayIf:(NSArray *)a
{
    int i, n = [a count];
    for (i=0;i<n;i++) {
        if (![self containsObject:[a objectAtIndex:i]])
            [self addObject:[a objectAtIndex:i]];
    }
}

- (void) addObjectsFromArrayIf:(NSArray *)a prefix:(NSString *)pref suffix:(NSString *)suf
{
    int i, n = [a count];
    for (i=0;i<n;i++) {
        NSString *s = [a objectAtIndex:i];
        if (suf) s = [s stringByAppendingString:@")"];
        if ([s hasCaseInsensitivePrefix:pref] && ![self containsObject:s]) {
            [self addObject:s];
        }
    }
}

- (void) addObjectsFromArrayIf:(NSArray *)a prefix:(NSString *)pref
{
    [self addObjectsFromArrayIf:a prefix:pref suffix:nil];
}

@end


@implementation MHSprite

BOOL isAwake = NO;
id mhSpriteNibInstance = nil;

- (void)windowWillClose:(NSNotification *)aNotification
{
	cancelled = YES;
    [spriteMovieView stop:self];
    [theOutline autosaveNow];
}

- (NSRect)windowWillUseStandardFrame:(NSWindow *)sender defaultFrame:(NSRect)defaultFrame
{
    return [self rightSize];
}

- (void)flagsChanged:(NSNotification *)note
{
    NSEvent *event = [NSApp currentEvent];
    unsigned int flags = [event modifierFlags];
    NSToolbarItem *item = [[[spriteWindow toolbar] items] objectAtIndex:2];
    if (flags & NSAlternateKeyMask) {
        [item setAction:@selector(duplicateIncrementing:)];
        [item setLabel:@"Duplicate (Incrementing)"];
    } else {
        [item setAction:@selector(duplicate:)];
        [item setLabel:@"Duplicate"];
    }
}

BOOL backspace = NO;

- (void) backSpace
{
    backspace = YES;
}

- (NSArray *)textView:(NSTextView *)tview completions:(NSArray *)words forPartialWordRange:(NSRange)charRange indexOfSelectedItem:(int *)index
{
	*index = -1;
	NSDictionary *actions = [NSDictionary dictionaryWithContentsOfFile:[[NSBundle mainBundle] pathForResource:@"SpriteActions" ofType:@"plist"]];
	NSArray *keys = [actions allKeys];
	NSMutableArray *list = [NSMutableArray array];
	// find the line and word
	NSRange r = [[tview string] lineRangeForRange:charRange];
	if (r.location == NSNotFound) return nil;
	NSString *line = [[tview string] substringWithRange:r];
        r.location = charRange.location - r.location;
        r.length = charRange.length;
	NSString *word = [line substringWithRange:r];
        
        int i, n;
        // complete the possible correct commands
        if (r.location > 1) { // there's a previous clause
            line = [line substringToIndex:r.location];
            if ([line hasSuffix:@"."]) line = [line substringToIndex:[line length]-1];
            r = [line rangeOfString:@"." options:NSBackwardsSearch];
            if (r.length) line = [line substringFromIndex:r.location+1];
            // first check if it's one we can complete args to
            if ([line hasSuffix:@"("]) {
                NSArray *cmds = [NSArray arrayWithObjects:@"GoToChapter(", @"TrackNamed(", @"SetLanguage(", nil];
                n = [cmds count];
                for (i=0;i<n;i++) {
                    NSString *cmd = [cmds objectAtIndex:i];
                    if ([line hasCaseInsensitivePrefix:cmd]) {
                        //word = [word substringFromIndex:[cmd length]];
                        NSArray *args = nil;
                        switch (i) {
                            case 0: args = [[(MHMovie *)[spriteMovieView movie] chapters] chapterNames]; break;
                            case 1: args = [self getTracksOfType:0]; break;
                            case 2: args = [tLanguages stringLangs]; break;
                            default: break;
                        }
                        [list addObjectsFromArrayIf:args prefix:word suffix:@")"];
                        return list;
                    }
                }
            }
            n = [keys count];
            for (i=0;i<n;i++) {
                NSString *key = [keys objectAtIndex:i];
                if ([line hasCaseInsensitivePrefix:key]) {
                    NSArray *values = [actions objectForKey:key];
                    if ([values containsObject:@"*"])
                        [list addObjectsFromArrayIf:keys prefix:word];
                    else
                        [list addObjectsFromArrayIf:values prefix:word];
                }
            }
        } else { // it's the first clause of the line
            [list addObjectsFromArrayIf:keys prefix:word];
        }
        
	return list;
}
/*
- (void) autoComplete
{
    NSLog(@"autocomplete");
    if (backspace) {
        backspace = NO;
        return;
    }
    NSString *oldString = [qScriptBox string];
    if (![oldString length]) return;
    NSRange r = [oldString rangeOfString:@"." options:NSBackwardsSearch];
    int dotLocation = 0;
	if (r.length) dotLocation = r.location;
	r = [oldString rangeOfString:@"\n" options:NSBackwardsSearch];
	if (r.length && r.location > dotLocation) dotLocation = r.location;
    if (dotLocation > [oldString length]-2) return;
    if (dotLocation) {
        dotLocation++;
        oldString = [oldString substringFromIndex:dotLocation];
    }
    NSArray *possibilities = [NSArray arrayWithObjects:@"TrackNamed(", @"GoToChapter(", @"SetEnabled(", @"SetMovieLanguage(", @"TrackAtIndex(", @"GoToPage(", @"GoToTime(", @"SetRate(", @"SetVolume(", @"OnMouseEnter", @"OnMouseExit", @"OnFrameLoad", @"OnIdle", @"StartPlaying", @"StepForward", @"StepBackward", @"GoToBeginning", @"GoToEnd", @"SaveTime", @"GoToSavedTime", @"EnterFullScreen", @"ExitFullScreen", @"Close", nil];
    
    NSString *ext = [oldString extensionLessSharedPrefix:possibilities];
    if ([ext length]) { // complete a command
        oldString = [qScriptBox string];
        r.location = [oldString length];
        r.length = [ext length];
        [qScriptBox setString:[oldString stringByAppendingString:ext]];
        [qScriptBox setSelectedRange:r];
    } else { // a complete command
        int i, n = 4;
        for (i=0;i<n;i++) {
            NSString *pos = [possibilities objectAtIndex:i];
            if ([oldString hasCaseInsensitivePrefix:pos]) {
                NSArray *args = nil;
                switch (i) {
                    case 0: args = [self getTracksOfType:0]; break;
                    case 1: args = [[(MHMovie *)[spriteMovieView movie] chapters] chapterNames]; break;
                    case 2: args = [NSArray arrayWithObjects:@"yes", @"no", @"true", @"false", nil]; break;
                    case 3: args = [tLanguages stringLangs]; break;
                    default: break;
                }
                if (args) {
                    oldString = [oldString substringFromIndex:[pos length]];
                    ext = [oldString extensionLessSharedPrefix:args];
                    if ([ext length]) {
                        oldString = [qScriptBox string];
                        r.location = [oldString length];
                        r.length = [ext length] + 1;
                        [qScriptBox setString:[NSString stringWithFormat:@"%@%@)",oldString, ext]];
                        [qScriptBox setSelectedRange:r];
                    }
                }
            }
        }
    }
}
*/
- (void)textDidChange:(NSNotification *)aNotification
{
    //[self autoComplete];
    [NSObject cancelPreviousPerformRequestsWithTarget:qScriptBox];
    [qScriptBox performSelector:@selector(complete:) withObject:self afterDelay:1.5];
    //NSMutableDictionary *item = [theOutline selectedItem];
    
	int i, n = [theOutline numberOfRows];
	for (i=0;i<n;i++) {
		if ([theOutline isRowSelected:i]) {
			NSMutableDictionary *item = [theOutline itemAtRow:i];
			if ([[item objectForKey:@"isLeaf"] boolValue]) {
				NSString *string = [[qScriptBox string] copy];
				if (string) [item setObject:string forKey:@"Action"];
			}
		}
	}
    [qScriptBox setTextColor:[NSColor blackColor]];
}

+ (id)nibInstance
{
    return mhSpriteNibInstance;
}

- (void)postUpdate:(NSString *)string done:(double)done
{
    if (string) {
        [progText setStringValue:string];
		if (done >= 0) [progBar setDoubleValue:done];
    }
}

+ (id)goWithMovie:(NSMovie *)aMovie
{
    if (!isAwake) {
        NSWindowController *winCon = [[NSWindowController alloc] initWithWindowNibName:@"SpriteEditor"];
        [winCon showWindow:self];
    }
    //aMovie = [[NSMovie alloc] initWithURL:[NSURL fileURLWithPath:@"/mov.mov"] byReference:NO];
    [mhSpriteNibInstance setMovie:aMovie];
    [mhSpriteNibInstance open];
    return mhSpriteNibInstance;
}

- (void)populateMenu:(NSMenu *)aMenu
{
    NSArray *titles = [NSArray arrayWithObjects:@"New Sprite", @"New Page", @"Delete Selection", @"Duplicate Selection", @"Duplicate (Incrementing)", @"Wizard", @"Update Movie", @"Advanced Options", nil];
    void *actions[] = { @selector(newSprite:), @selector(newPage:), @selector(deletePage:), @selector(duplicate:), @selector(duplicateIncrementing:), @selector(wizard:), @selector(updateMovie:), @selector(more:) };
    int i, n = [titles count];
    for (i=0;i<n;i++) {
	NSMenuItem *item = [aMenu addItemWithTitle:[titles objectAtIndex:i] action:actions[i] keyEquivalent:@""];
	[item setTarget:self];
	if (i == 8) // quickMode
	    quickModeMenu = item;
    }
}

+ (void)setMovie:(NSMovie *)aMovie
{
    [mhSpriteNibInstance setMovie:aMovie];
}

+ (void)setX:(int)x Y:(int)y
{
    [mhSpriteNibInstance iSetX:x Y:y];
}

+ (NSColor *)bgColor
{
    return [mhSpriteNibInstance color];
}

+ (NSSize)naturalSize
{
    return [mhSpriteNibInstance size];
}

+ (id)selectedItem		{	return [mhSpriteNibInstance selectedTableItem];	}
+ (int)selectedImage            {       return [mhSpriteNibInstance selectedImage];     }
+ (void)selectItem:(id)item	{	[mhSpriteNibInstance selectItem:item];		}
+ (void)selectItem:(id)item byExtendingSelection:(BOOL)flag {	[mhSpriteNibInstance selectItem:item byExtendingSelection:flag];	}
+ (void)reloadOutline		{	[mhSpriteNibInstance reload];			}
+ (BOOL)qsMode			{       return [mhSpriteNibInstance qsMode];		}
+ (void)newSpriteWithRect:(NSRect)rect  { [mhSpriteNibInstance newSpriteWithRect:rect]; }
- (BOOL)qsMode			{       return [quickModeMenu state];			}
- (void)reload			{	[theOutline reloadData];			}
+ (NSMenu *)rightClickMenu {	return [mhSpriteNibInstance rightClickMenu];	}

- (NSMenu *)rightClickMenu
{
	NSMutableDictionary *theHit = [self selectedTableItem];
	if (!theHit) return nil;
	NSMenu *menu = [[[NSMenu alloc] init] autorelease];
    // action
    NSArray *actions = [NSArray arrayWithObjects:@"None", @"Go To Chapter", @"Set Audio Track", @"Set Text Track", @"Step Forward", @"Step Backward", @"Go To Page", @"Set Volume", @"Enter Fullscreen", @"Exit Fullscreen", @"Close", @"Start Playing", @"Set Rate", @"Play Selection", @"Set Selection", @"Set Selection By Chapter", @"Save Time", @"Go To Saved Time", @"Set Sprite Visible", @"Rotate Sprite", @"Scale Sprite", @"Move Sprite", @"Set Sprite Opacity", nil];
    NSArray *keys = [NSArray arrayWithObjects: @"none", @"GoToChapter(", @"// set audio track", @"// set text track", @"stepforward", @"stepbackward", @"gotopage(", @"SetVolume(", @"EnterFullScreen", @"ExitFullScreen", @"Close", @"StartPlaying", @"SetRate(", @"PlaySelection", @"SetSelection(", @"SetSelectionByChapter(", @"SaveTime", @"GoToSavedTime",@"SetVisible", @"Rotate", @"Scale", @"Translate", @"SetOpacity", nil];
    NSString *action = [theHit objectForKey:@"Action"];
    if (!action) action = @"";
    NSScanner *scan = [NSScanner scannerWithString:action];
    [scan setCaseSensitive:NO];
    int i, n = [actions count];
    for (i=0;i<n;i++) {
	NSMenuItem *anItem = [[[NSMenuItem alloc] initWithTitle:[actions objectAtIndex:i] action:@selector(setAction:) keyEquivalent:@""] autorelease];
	[anItem setTarget:self];
        if (i==0 && ![action length]) [anItem setState:YES];
        else if ([scan scanString:[keys objectAtIndex:i] intoString:nil])
            [anItem setState:YES];
	[menu addItem:anItem];
        // values
        NSArray *values = nil;
        switch (i) {
            case 1: case 15: values = [self chapterNames]; break;
            case 2: values = [self getTracksOfType:SoundMediaType]; break;
            case 3: values = [self getTracksOfType:TextMediaType]; break;
            case 6: values = [self pageNumbers]; break;
			case 7: values = [NSArray arrayWithObjects:@"0", @"60", @"120", @"180", @"256", nil]; break;
			case 18: case 19: case 20: case 21: case 22: values = [self spriteNames:theHit]; break;
            default: break;
        }
        if (values) {
            NSString *value = nil;
            int x = -1, j, m = [values count];
            if ([anItem state]) { // if this is our action type
                switch (i) {
                    case 1: [scan scanUpToString:@")" intoString:&value]; break;
                    case 2: case 3:
                        while (![scan isAtEnd]) {
                            [scan scanPastString:@"TrackNamed("];
                            [scan scanUpToString:@")" intoString:&value];
                            if ([scan scanString:@".SetEnabled(" intoString:nil] && ([scan scanString:@"YES" intoString:nil] || [scan scanString:@"TRUE" intoString:nil]))
                                break;
                            else value = nil;
                        } break;
                    case 6:
                        [scan scanInt:&x];
                        if (x >=0 && x < m) value = [NSString stringWithFormat:@"%d", x];
                        break;
                    default: break;
                }
            }
            NSMenu *vMenu = [[[NSMenu alloc] init] autorelease];
            [anItem setSubmenu:vMenu];
            for (j=0;j<m;j++) {
                NSString *aValue = [values objectAtIndex:j];
                NSMenuItem *vItem = [[[NSMenuItem alloc] initWithTitle:aValue action:@selector(setValue:) keyEquivalent:@""] autorelease];
                if (i == 6) // go to page
                    aValue = [NSString stringWithFormat:@"%d", [aValue intValue]];
                if (value && [value isEqualToString:aValue])
                    [vItem setState:YES];
                [vItem setTarget:self];
                [vMenu addItem:vItem];
            }
        }
    }
	return menu;
}

- (void)newSpriteWithRect:(NSRect)rect
{
    if (rect.size.width < 1 || rect.size.height < 1) return;
	
	[self newSprite:nil];
	NSMutableDictionary *sprite = [theOutline itemAtRow:[theOutline selectedRow]];
	/*
    NSMutableDictionary *menu = [[spriteMovieView movie] menu];
    if (!menu)
        menu = [self newMenu];
    NSMutableDictionary *track = nil;
    int index = [theOutline selectedRow];
    if (index != -1) {
        NSMutableDictionary *item = [theOutline itemAtRow:index];
        if (![[item objectForKey:@"isLeaf"] boolValue])
            track = item;
        else
            track = [item objectForKey:@"parent"];
    } else {
        NSArray *tracks = [menu objectForKey:@"tracks"];
        if ([tracks count])
            track = [tracks objectAtIndex:0];
    }
    if (!track) {
        [self newPage:nil];
        track = [[menu objectForKey:@"tracks"] objectAtIndex:0];
    }
    NSMutableDictionary *sprite = [NSMutableDictionary dictionary];
    [sprite setObject:@"Sprite" forKey:@"Name"];
    [sprite setObject:[NSNumber numberWithBool:YES] forKey:@"isLeaf"];
    [sprite setObject:track forKey:@"parent"];
    [[track objectForKey:@"sprites"] addObject:sprite];
	 */
    [sprite setObject:[NSString stringWithFormat:@"%d", (int)rect.origin.x] forKey:@"X"];
    [sprite setObject:[NSString stringWithFormat:@"%d", (int)rect.origin.y] forKey:@"Y"];
    [sprite setObject:[NSString stringWithFormat:@"%d", (int)rect.size.width] forKey:@"Width"];
    [sprite setObject:[NSString stringWithFormat:@"%d", (int)rect.size.height] forKey:@"Height"];
    
    [theOutline reloadData];
    //[theOutline expandItem:track];
    //[theOutline selectRowIndexes:[NSIndexSet indexSetWithIndex:[theOutline rowForItem:sprite]] byExtendingSelection:NO];
    [theImageView setNeedsDisplay:YES];
}

- (void)selectItem:(id)item { [self selectItem:item byExtendingSelection:NO];	}

- (void)selectItem:(id)item byExtendingSelection:(BOOL)flag
{
    int row = [theOutline rowForItem:item];
    if (row == -1) {
        id parent = [item objectForKey:@"parent"];
        if (parent) {
            [theOutline expandItem:parent];
            row = [theOutline rowForItem:item];
        }
    }
    if (row != -1)
        [theOutline selectRowIndexes:[NSIndexSet indexSetWithIndex:row] byExtendingSelection:flag];
}

- (id)selectedTableItem
{
    if ([theOutline selectedRow] == -1) return nil;
    return [theOutline itemAtRow:[theOutline selectedRow]];
}

- (int)selectedImage
{   return [imageSelect indexOfSelectedItem];   }

- (NSSize)size
{
    return naturalSize;
}

- (NSColor *)color
{
    return [bgColor color];
}

- (IBAction)setColor:(id)sender
{
    NSColor *color = [bgColor color];
	NSData *dColor = [NSArchiver archivedDataWithRootObject:color];
    [[NSUserDefaults standardUserDefaults] setObject:dColor forKey:@"spriteBGColor"];
    //NSMutableDictionary *item = [theOutline itemAtRow:[theOutline selectedRow]];
	int i, n = [theOutline numberOfRows];
	for (i=0;i<n;i++) {
		if ([theOutline isRowSelected:i]) {
			NSMutableDictionary *item = [theOutline itemAtRow:i];
			if ([[item objectForKey:@"isLeaf"] boolValue])
				item = [item objectForKey:@"parent"];
			[item setObject:dColor forKey:@"color"];
		}
	}
    [theImageView setNeedsDisplay:YES];
    [[NSColor blackColor] set];
}

- (NSImage *)imageForSprite:(NSMutableDictionary *)sprite key:(NSString *)key
{
	NSMutableDictionary *menu = [[spriteMovieView movie] menu];
	if (!menu)
		menu = [self newMenu];
	QTImage *qi = [menu objectForKey:@"qi"];
	if (!qi) {
		qi = [[[QTImage alloc] init] autorelease];
		[menu setObject:qi forKey:@"qi"];
	}
	NSSize size = NSMakeSize([[sprite objectForKey:@"Width"] intValue], [[sprite objectForKey:@"Height"] intValue]);
	if (![[sprite objectForKey:@"isLeaf"] boolValue]) size = naturalSize;
    NSColor *aColor = nil;
    NSData *dColor = [sprite objectForKey:@"keyColor"];
    if (dColor) aColor = [NSUnarchiver unarchiveObjectWithData:dColor];
    TimeValue time = 0;
    NSString *path = [sprite objectForKey:key];
	if (!path) return nil;
	if ([path hasPrefix:@"\""]) {
		path = [path substringFromIndex:1];
		if ([path hasSuffix:@"\""])
			path = [path substringToIndex:[path length]-1];
		NSImage *i = [sprite textify:path];
		return i;
    } else if ([path hasPrefix:@"/"]) {
		if ([path hasSuffix:@".mov"]) {
			time = -2;
		}
	} else {
        time = [path intValue];
        path = [[[spriteMovieView movie] URL] path];
  /*  } else {
		size.width /= 2;
		size.height /= 2;*/
	}
	
	[imageLock lock];
	//[[MHUtil imageHelper] openMovie:[NSURL fileURLWithPath:path]];
	//NSData *data = [[MHUtil imageHelper] getImage:time sender:nil transparency:aColor];
	NSImage *image = /*[NSUnarchiver unarchiveObjectWithData:data]; //*/[qi imageForPath:path transparency:aColor time:time size:size];
	[imageLock unlock];
	
   // NSImage *image = [[QTImage alloc] initWithFile:path transparency:aColor time:time];
    return image;
}

- (IBAction)setImage:(id)sender
{
    NSArray *anArray = [NSArray arrayWithObjects:@"Image", @"Rollover", @"Click", nil];
    NSArray *keys = [NSArray arrayWithObjects:@"image", @"rimage", @"cimage", nil];
    //NSMutableDictionary *item = [theOutline itemAtRow:[theOutline selectedRow]];
	
	NSImage *image = nil;
	NSString *key = nil;
	int dest = [imageSelect indexOfSelectedItem];
	int type = [imageType selectedRow];
    if (type == 0) { // Movie Frame
        Movie mov = [[spriteMovieView movie] QTMovie];
        TimeValue tv;
        TimeRecord tr;
        tv = GetMovieTime(mov, &tr);
        key = [NSString stringWithFormat:@"%ld", tv];
        //[item setObject:key forKey:[anArray objectAtIndex:dest]];
        //image = [self imageForSprite:item key:[anArray objectAtIndex:dest]];
        //if (image) [item setObject:image forKey:[keys objectAtIndex:dest]];
    } else { // File
		NSOpenPanel *theOpenPanel = [NSOpenPanel openPanel];
		int result;
    
		[theOpenPanel setCanChooseFiles:YES];
		[theOpenPanel setCanChooseDirectories:NO];
		[theOpenPanel setResolvesAliases:YES];
		[theOpenPanel setAllowsMultipleSelection:NO];
		result = [theOpenPanel runModalForDirectory:nil file:nil types:nil];
		if (result == NSOKButton) {
			key = [[theOpenPanel filenames] objectAtIndex:0];
			/*
			[item setObject:choice forKey:[anArray objectAtIndex:dest]];
			image = [self imageForSprite:item key:[anArray objectAtIndex:dest]];
				if (image) {
					[item setObject:image forKey:[keys objectAtIndex:dest]];
					[item setObject:[NSNumber numberWithInt:[image size].width] forKey:@"Width"];
					[item setObject:[NSNumber numberWithInt:[image size].height] forKey:@"Height"];
				}
			}*/
		}
    }
	NSMutableArray *selectedItems = [NSMutableArray array];
	
	if (!key) return;
	int i, n = [theOutline numberOfRows];
	for (i=0;i<n;i++) {
		if ([theOutline isRowSelected:i]) {
			NSMutableDictionary *item = [theOutline itemAtRow:i];
			[selectedItems addObject:item];
			if ([[item objectForKey:@"isLeaf"] boolValue]) {
				[item setObject:key forKey:[anArray objectAtIndex:dest]];
				if (!image)
					image = [self imageForSprite:item key:[anArray objectAtIndex:dest]];
				if (image) {
					[item setObject:image forKey:[keys objectAtIndex:dest]];
					if (type) {
						[item setObject:[NSNumber numberWithInt:[image size].width] forKey:@"Width"];
						[item setObject:[NSNumber numberWithInt:[image size].height] forKey:@"Height"];
					}
				}
			}
		}
	}
	
    [theOutline reloadData];
    //[theOutline expandItem:[item objectForKey:@"parent"]];
    //[theOutline selectRowIndexes:[NSIndexSet indexSetWithIndex:[theOutline rowForItem:item]] byExtendingSelection:NO];
    [theImageView setNeedsDisplay:YES];
}

- (NSMutableDictionary *)newMenu
{
    MHMovie *movie = (MHMovie *)[spriteMovieView movie];
    if (!movie) return nil;
    NSMutableDictionary *menu = [movie menu];
    if (menu) [menu release];
    menu = [[NSMutableDictionary alloc] init];
    [menu setObject:[NSMutableArray array] forKey:@"tracks"];
	[menu setObject:[[[QTImage alloc] init] autorelease] forKey:@"qi"];
    [movie setMenu:menu];
    return menu;
}

- (NSString *)newNameForName:(NSString *)name list:(NSArray *)list
{
	int i, n = [list count], index = 1;
	NSString *try = name;
	while (YES) {
		for (i=0;i<n;i++)
			if ([[[list objectAtIndex:i] objectForKey:@"Name"] isEqualToString:try]) {
				try = nil;
				break;
			}
		if (try) return try;
		try = [NSString stringWithFormat:@"%@ %d", name, index++];
	}
}

- (IBAction)newPage:(id)sender
{
    MHMovie *aMovie = (MHMovie *)[spriteMovieView movie];
    if (nil == aMovie) return;
    NSMutableDictionary *menu = [[spriteMovieView movie] menu];
    if (!menu)
        menu = [self newMenu];
    NSMutableDictionary *track = [[NSMutableDictionary alloc] init];
	/*
	int i, index = 0;
	NSString *name = nil;
	while (YES) {
		name = [NSString stringWithFormat:@"Page %d", index++];
		for (i=0;i<[[menu objectForKey:@"tracks"] count];i++) {
			if ([[[[menu objectForKey:@"tracks"] objectAtIndex:i] objectForKey:@"Name"] isEqualToString:name]) {
				name = nil;
				break;
			}
		}
		if (name)
			break;
	}*/
	NSString *name = [self newNameForName:@"Page" list:[menu objectForKey:@"tracks"]];
	
    //MHSpriteTrack *track = [(MHSpriteTrack *)[MHSpriteTrack alloc] initWithMovie:aMovie];
    //[track setObject:[NSArchiver archivedDataWithRootObject:[bgColor color]] forKey:@"bgColor"];
    [track setObject:name forKey:@"Name"];
    [track setObject:[NSMutableArray array] forKey:@"sprites"];
    [[menu objectForKey:@"tracks"] addObject:track];
    [theOutline reloadData];
    [theOutline selectRowIndexes:[NSIndexSet indexSetWithIndex:[theOutline rowForItem:track]] byExtendingSelection:NO];
    [self setColor:nil];
}

- (IBAction)newSprite:(id)sender
{
    NSMutableDictionary *menu = [[spriteMovieView movie] menu];
    if (!menu)
        menu = [self newMenu];
    if (!menu) return;
    NSMutableDictionary *track = nil;
    int index = [theOutline selectedRow];
    if (index != -1) {
        NSMutableDictionary *item = [theOutline itemAtRow:index];
        if (![[item objectForKey:@"isLeaf"] boolValue])
            track = item;
        else
            track = [item objectForKey:@"parent"];
    } else {
        NSArray *tracks = [menu objectForKey:@"tracks"];
        if ([tracks count])
            track = [tracks objectAtIndex:0];
    }
    if (!track) {
        [self newPage:nil];
        track = [[menu objectForKey:@"tracks"] objectAtIndex:0];
    }
    NSMutableDictionary *sprite = [NSMutableDictionary dictionary];
	NSString *name = [self newNameForName:@"Sprite" list:[track objectForKey:@"sprites"]];
    [sprite setObject:name forKey:@"Name"];
    [sprite setObject:[NSNumber numberWithBool:YES] forKey:@"isLeaf"];
    [sprite setObject:track forKey:@"parent"];
    [[track objectForKey:@"sprites"] addObject:sprite];
    [theOutline reloadData];
    [theOutline expandItem:track];
    [theOutline selectRowIndexes:[NSIndexSet indexSetWithIndex:[theOutline rowForItem:sprite]] byExtendingSelection:NO];
    [theImageView setNeedsDisplay:YES];
}

- (IBAction)deletePage:(id)sender
{
    [loadLock lock];
    NSMutableDictionary *menu = [[spriteMovieView movie] menu];
    if (!menu)
        menu = [self newMenu];
	id previousItem = [theOutline itemAtRow:[theOutline selectedRow]];
	if ([[previousItem objectForKey:@"isLeaf"] boolValue])
		previousItem = [previousItem objectForKey:@"parent"];
	
	int j, i = [theOutline numberOfRows] - 1;
	for (;i>=0;i--) {
		if ([theOutline isRowSelected:i]) {
			NSMutableDictionary *item = [theOutline itemAtRow:i];
			NSMutableArray *list = [[item objectForKey:@"parent"] objectForKey:@"sprites"];
			if (!list) list = [menu objectForKey:@"tracks"];
			j = [list count] - 1;
			for (;j>=0;j--)
				if ([list objectAtIndex:j] == item)
					[list removeObjectAtIndex:j];
		}
	}
    [loadLock unlock];
    [theOutline reloadData];
	
	NSMutableIndexSet *set = [NSMutableIndexSet indexSet];
	i = [theOutline numberOfRows] - 1;
	for (;i>=0;i--)
		if ([theOutline itemAtRow:i] == previousItem)
			[set addIndex:i];
	
	[theOutline selectRowIndexes:set byExtendingSelection:NO];
    [theImageView setNeedsDisplay:YES];
}

- (NSMutableDictionary *)duplicate:(id)sender
{
    NSDictionary *item = [theOutline itemAtRow:[theOutline selectedRow]];
    if (!item) return nil;
    NSMutableDictionary *newItem = [[NSMutableDictionary alloc] init];
    NSArray *keys = [item allKeys];
    NSArray *values = [item allValues];
    int i, n = [keys count];
    if ([[item objectForKey:@"isLeaf"] boolValue]) {
        for (i=0;i<n;i++)
            [newItem setObject:[values objectAtIndex:i] forKey:[keys objectAtIndex:i]];
        [[[item objectForKey:@"parent"] objectForKey:@"sprites"] addObject:newItem];
    } else {
        for (i=0;i<n;i++) {
            NSString *key = [keys objectAtIndex:i];
            if (![key isEqualToString:@"sprites"])
                [newItem setObject:[values objectAtIndex:i] forKey:[keys objectAtIndex:i]];
        }
        NSMutableArray *newSprites = [[NSMutableArray alloc] init];
        [newItem setObject:newSprites forKey:@"sprites"];
        NSArray *oldSprites = [item objectForKey:@"sprites"];
        if (oldSprites) {
            NSArray *skeys, *svalues;
            int j, k;
            n = [oldSprites count];
            for (i=0;i<n;i++) {
                NSDictionary *aSprite = [oldSprites objectAtIndex:i];
                NSMutableDictionary *newSprite = [[NSMutableDictionary alloc] init];
                [newSprites addObject:newSprite];
                skeys = [aSprite allKeys];
                svalues = [aSprite allValues];
                k = [skeys count];
                for (j=0;j<k;j++)
                    [newSprite setObject:[svalues objectAtIndex:j] forKey:[skeys objectAtIndex:j]];
				[newSprite setObject:newItem forKey:@"parent"];
            }
        }
        NSMutableDictionary *menu = [[spriteMovieView movie] menu];
        if (!menu)
            menu = [self newMenu];
        [[menu objectForKey:@"tracks"] addObject:newItem];
    }
    [theOutline reloadData];
    [self selectItem:newItem];
    [theImageView setNeedsDisplay:YES];
    [qScriptBox setTextColor:[NSColor blackColor]];
    return newItem;
}

- (NSString *)nextNameInArray:(NSArray *)names afterName:(NSString *)name
{
    int i, n = [names count];
    NSString *pref = nil;
    for (i=0;i<n;i++) {
        NSString *s = [names objectAtIndex:i];
        if ([s isEqualToString:name]) {
            if (i<n-1) return [names objectAtIndex:i+1];
            else return s;
        }
        if ([s hasPrefix:name]) pref = s;
    }
    return pref;
}

- (IBAction)duplicateIncrementing:(id)sender
{
    NSMutableDictionary *newItem = [self duplicate:nil];
    if ([[newItem objectForKey:@"isLeaf"] boolValue]) {
        NSString *action = [newItem objectForKey:@"Action"];
        NSString *value = nil;
        if (action) {
            NSScanner *scan = [NSScanner scannerWithString:action];
            [scan setCaseSensitive:NO];
            if ([scan scanString:@"gotochapter(" intoString:nil]) {
                [scan scanUpToString:@")" intoString:&value];
                NSArray *chNames = [[(MHMovie *)[spriteMovieView movie] chapters] chapterNames];
                if (chNames && value) {
                    value = [self nextNameInArray:chNames afterName:value];
                    action = [NSString stringWithFormat:@"GoToChapter(%@);", value];
                    [newItem setObject:action forKey:@"Action"];
                    return;
                }
            } else if ([scan scanString:@"// set " intoString:nil]) {
                NSArray *tracks = nil;
                int actionKind = 0;
                if ([scan scanString:@"audio track" intoString:nil]) {
                    tracks = [self getTracksOfType:SoundMediaType];
                    actionKind = 2;
                } else if ([scan scanString:@"text track" intoString:nil]) {
                    tracks = [self getTracksOfType:TextMediaType];
                    actionKind = 3;
                }
                if (tracks) {
                    while (![scan isAtEnd]) {
                        [scan scanPastString:@"tracknamed("];
                        [scan scanUpToString:@")" intoString:&value];
                        [scan scanPastString:@"setenabled("];
                        if ([scan scanString:@"yes" intoString:nil] || [scan scanString:@"true" intoString:nil]) {
                            value = [self nextNameInArray:tracks afterName:value];
                            action = [self stringForAction:actionKind value:value];
                            [newItem setObject:action forKey:@"Action"];
                            return;
                        }
                    }
                }
            } else if ([scan scanString:@"gotopage(" intoString:nil]) {
                int page = -1;
                if ([scan scanInt:&page] && page != -1) {
                    page++;
                    action = [NSString stringWithFormat:@"GoToPage(%d);", page];
                    [newItem setObject:action forKey:@"Action"];
                    return;
                }
            }
        }
    }
}

- (IBAction)updateMovie:(id)sender
{
    MHMovie *movie = (MHMovie *)[spriteMovieView movie];
    if ([movie isOffline]) return;
    [NSThread detachNewThreadSelector:@selector(updateMovieThreaded:) toTarget:self withObject:movie];
}

- (void)updateMovieThreaded:(MHMovie *)movie
{
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
    NSString *plist, *input, *resource;
    input = [[movie URL] path];
    if (![input hasExtension:[NSArray arrayWithObject:@"mov"]]) {
        input = [MHUtil uniquePath:[[MHUtil supportDir] stringByAppendingPathComponent:[input lastPathComponent]]];
        [movie saveToPath:input flatten:NO];
    }
    //input = [MHUtil uniquePath:[[MHUtil supportDir] stringByAppendingPathComponent:[[movie stringValue] lastPathComponent]]];
    plist = [MHUtil uniquePath:[[MHUtil supportDir] stringByAppendingPathComponent:@"plist.xml"]];
    resource = [[NSBundle mainBundle] pathForResource:@"four1" ofType:@"png"];
    resource = [resource stringByDeletingLastPathComponent];
    resource = [resource stringByAppendingPathComponent:@"spriteHelper2.app/Contents/MacOS/spriteHelper2"];
    NSArray *array = [NSArray arrayWithObjects:input, plist, nil];
    NSDictionary *menu = [movie menu];
    if (!menu) {
        NSBeep();
        return;
    }
    menu = [self plist:menu];
    [menu writeToFile:plist atomically:NO];
    //[menu writeToFile:@"/plist.xml" atomically:NO];
    //[movie saveToPath:input flatten:NO];
    [NSApp beginSheet: progWindow modalForWindow: spriteWindow modalDelegate: nil didEndSelector: nil contextInfo: nil];
    updateTask = [NSTask launchedTaskWithLaunchPath:resource arguments:array];
    [updateTask waitUntilExit];
    updateTask = nil;
    [progWindow orderOut: NULL];
    [progText setStringValue:@""];
    [NSApp endSheet: progWindow];
    if (!DEBUG)
        [[NSFileManager defaultManager] removeFileAtPath:plist handler:nil];
    [spriteMovieView performSelectorOnMainThread:@selector(stop:) withObject:nil waitUntilDone:NO];
    [spriteMovieView performSelectorOnMainThread:@selector(gotoBeginning:) withObject:self waitUntilDone:NO];
    [pool release];
}

-(NSMutableDictionary *)plist:(NSDictionary *)input
{
    NSMutableDictionary *newDict = [NSMutableDictionary dictionary];
    NSArray *tracks = [input objectForKey:@"tracks"];
    NSMutableArray *newTracks = [NSMutableArray array];
    [newDict setObject:newTracks forKey:@"tracks"];
    NSArray *keys, *values;
    NSString *key, *value;
    int i, n = [tracks count], si = 0;
    long *startTimes = malloc(sizeof(long)*n), totalTime = 0;
    //NSLog(@"%d start times", n);
	TimeValue dur = GetMovieTimeScale([[spriteMovieView movie] QTMovie]);
    for (i=0;i<n;i++) {
        NSString *src = [[tracks objectAtIndex:i] objectForKey:@"Image"];
        startTimes[i] = totalTime + dur / 10;
		//NSLog(@"%ld", startTimes[i]);
	//NSLog(@"start time %d: %d", i, totalTime);
        if ([src hasSuffix:@".mov"]) {
            NSMovie *aMovie = [[NSMovie alloc] initWithURL:[NSURL fileURLWithPath:src] byReference:NO];
			TimeScale aTs = GetMovieTimeScale([aMovie QTMovie]);
			TimeScale bTs = GetMovieTimeScale([[spriteMovieView movie] QTMovie]);
			float ratio = (float)bTs / (float)aTs;
			//NSLog(@"%ld %ld", aTs, bTs);
            totalTime += (float)GetMovieDuration([aMovie QTMovie]) * ratio;
	    //else
            //  totalTime -= TRACK_DURATION;
        } else {
            totalTime += dur;
        }
    }
    for (i=0;i<n;i++) {
        NSDictionary *track = [tracks objectAtIndex:i];
        NSMutableDictionary *newTrack = [NSMutableDictionary dictionary];
        [newTracks addObject:newTrack];
        keys = [track allKeys];
        values = [track allValues];
        int j;
        for (j=[keys count]-1;j>=0;j--) {
            key = [keys objectAtIndex:j];
            value = [values objectAtIndex:j];
            if (![[self stringClass:value] isEqualToString:@"Unknown!"] && ![key isEqualToString:@"sprites"])
                [newTrack setObject:value forKey:key];
        }
        NSArray *sprites = [track objectForKey:@"sprites"];
        NSMutableArray *newSprites = [NSMutableArray array];
        [(NSMutableDictionary *)[newTracks objectAtIndex:i] setObject:newSprites forKey:@"sprites"];
        for (j=[sprites count]-1;j>=0;j--) {
            NSDictionary *sprite = [sprites objectAtIndex:j];
            NSMutableDictionary *newSprite = [NSMutableDictionary dictionary];
            [newSprites insertObject:newSprite atIndex:0];
            keys = [sprite allKeys];
            values = [sprite allValues];
            int k;
            for (k=[keys count]-1;k>=0;k--) {
                key = [keys objectAtIndex:k];
                value = [values objectAtIndex:k];
				if ([key isEqualToString:@"Name"] && [value isEqualToString:@"Sprite"])
					value = [value stringByAppendingFormat:@" %d", ++si];
                if (![[self stringClass:value] isEqualToString:@"Unknown!"] && ![key isEqualToString:@"parent"])
                    [newSprite setObject:value forKey:key];
            }
            NSString *action = [sprite objectForKey:@"Action"];
            if (action) {
                NSScanner *scan = [NSScanner scannerWithString:action];
                [scan setCaseSensitive:NO];
				NSString *buffer;
				NSMutableString *newAction = [NSMutableString stringWithString:@""];
				int pageno = -1;
				while (![scan isAtEnd]) {
					//pageno = -1;
					buffer = nil;
					[scan scanUpToString:@"GoToPage(" intoString:&buffer];
					if (buffer)
						[newAction appendString:buffer];
					[scan scanString:@"GoToPage(" intoString:nil];
					[scan scanInt:&pageno];
					if (pageno >= 0) {
						[newAction appendFormat:@"GoToTime(%ld", startTimes[pageno]];
					}
				}
				if (pageno >= 0 && [[[tracks objectAtIndex:pageno] objectForKey:@"Image"] hasSuffix:@".mov"])
					[newAction appendString:@"\nStartPlaying;"];
				action = [NSString stringWithString:newAction];
				scan = [NSScanner scannerWithString:action];
				[scan setCaseSensitive:NO];
				newAction = [NSMutableString stringWithString:@""];
				while (![scan isAtEnd]) {
					buffer = nil;
					[scan scanUpToString:@"OnFrameLoad" intoString:&buffer];
					if (buffer)
						[newAction appendString:buffer];
					if ([scan isAtEnd])
						break;
					[newAction appendString:@"OnFrameLoad."];
					[scan scanPastString:@"."];
					if (![scan scanString:@"Track" intoString:&buffer] && ![scan scanString:@"Sprite" intoString:&buffer])
						[newAction appendFormat:@"SpriteNamed(%@).", [newSprite objectForKey:@"Name"]]; // OnFrameLoad surprisingly goes into the keyframe not the sprite, so if the sprite has actions for itself, then that must be specified
					else
						[newAction appendString:buffer];
				}
                [newSprite setObject:newAction forKey:@"Action"];
            }
        }
    }
    /*
    NSMutableArray *trackNames = [NSMutableArray array];
    [newDict setObject:trackNames forKey:@"trackNames"];
    long index;
    Handle h = NewHandleClear(0);
    for (index = 1; index <= GetMovieTrackCount([[spriteMovieView movie] QTMovie]); index++) {
        Track aTrack = GetMovieIndTrack([[spriteMovieView movie] QTMovie], index);
        GetUserData(GetTrackUserData(aTrack), h, kUserDataName, 1);
        int size = GetHandleSize(h);
        char *buffer = (char *)*h;
        NSString *aName = [NSString stringWithCString:buffer length:size];
        [trackNames addObject:aName];
    }
    DisposeHandle(h);
    */
    free(startTimes);
    return newDict;
}

- (NSString *)stringClass:(id)input
{
    if ([input isKindOfClass:[NSString class]])
        return @"String";
    if ([input isKindOfClass:[NSNumber class]])
        return @"Number";
    if ([input isKindOfClass:[NSData class]])
        return @"Data";
    if ([input isKindOfClass:[NSDate class]])
        return @"Date";
    if ([input isKindOfClass:[NSDictionary class]])
        return @"Dictionary";
    if ([input isKindOfClass:[NSArray class]])
        return @"Array";
    return @"Unknown!";
}

- (int)getTrackNumber:(NSString *)trackName
{
    if (![trackName isKindOfClass:[NSString class]]) return 0;
    Movie mov = [[spriteMovieView movie] QTMovie];
    if (!mov) return 0;
    long i;
    Handle h = NewHandleClear(0);
    for (i=1;i<=GetMovieTrackCount(mov);i++) {
        GetUserData(GetTrackUserData(GetMovieIndTrack(mov, i)), h, kUserDataName, 1);
        int size = GetHandleSize(h);
        if (size) {
            char *buffer = (char *)*h;
            NSString *string = [NSString stringWithCString:buffer length:size];
            if (string && [string isEqualToString:trackName]) return i;
        }
    }
    return 0;
}

- (void)submitMenus:(NSData *)data
{
    if (!data) return;
    NSMovie *menuMovie = [NSUnarchiver unarchiveObjectWithData:data];
    if (!menuMovie) return;
    MHMovie *movie = (MHMovie *)[spriteMovieView movie];
    if ([alwaysDelete state])
        [self deleteSpriteTracks:self];
    float thresh = [preloadSize floatValue];
    wide dataSize;
    OSErr theErr = GetMovieDataSize64([menuMovie QTMovie], 0, GetMovieDuration([menuMovie QTMovie]), &dataSize);
    float mbSize = (double)dataSize.lo / 1024.0 / 1024.0;
    Track aTrack;
    long count = GetMovieTrackCount([menuMovie QTMovie]), index = 1;
    if (!theErr && mbSize < thresh)
	for (aTrack = GetMovieIndTrack([menuMovie QTMovie], index); index <= count; aTrack = GetMovieIndTrack([menuMovie QTMovie], index++))
	    SetTrackLoadSettings(aTrack, -1, 0, preloadAlways, 0);
    InsertMovieSegment([menuMovie QTMovie], [movie QTMovie], 0, GetMovieDuration([menuMovie QTMovie]), 0);
    [spriteMovieView performSelectorOnMainThread:@selector(setMovie:) withObject:nil waitUntilDone:NO];
    [spriteMovieView performSelectorOnMainThread:@selector(setMovie:) withObject:movie waitUntilDone:NO];
    [[movie chapters] getChaptersFromMovie:movie];
    //[[movie chapters] offsetAll:(double)GetMovieDuration([menuMovie QTMovie]) / (double)GetMovieTimeScale([movie QTMovie])];
    [movie setSaveInPlaceUnSafe];
}

- (void)setMovieToSize:(NSSize)size
{
    //NSLog(@"Set Movie Width: %d Height: %d", (int)size.width, (int)size.height);
    Rect box;
    if (!size.height)
        size.height = 40;
    box.left = box.top = 0;
    box.right = size.width;
    box.bottom = size.height;
    Movie mov = [[spriteMovieView movie] QTMovie];
    SetMovieBox(mov, &box);
}

- (IBAction)done:(id)sender
{
}

- (IBAction)dump:(id)sender
{
    //[[spriteMovieView movie] saveToPath:@"/out.mov" flatten:NO];
}

- (IBAction)more:(id)sender
{
    if ([moreDrawer state] != NSDrawerClosedState) // don't ask me why but this is right
        [moreButton setTitle:@"More Options"];
    else
        [moreButton setTitle:@"Fewer Options"];
    [moreDrawer toggle:self];
}

- (IBAction)quickMode:(id)sender
{
    BOOL qMode = ![quickModeMenu state];
    [quickModeMenu setState:qMode];
    [theImageView resetCursorRects];
    if ([spriteWindow isVisible])
	[[NSCursor crosshairCursor] push];
    /*
    if (floor(NSAppKitVersionNumber) > NSAppKitVersionNumber10_2) {
	if (qMode)
	    [[NSCursor crosshairCursor] push];
	else
	    [NSCursor pop];
    }
    */
}

- (IBAction)savePrefs:(id)sender
{
    [self prefs:1];
}

- (IBAction)visibleSwitch:(id)sender
{
    //NSMutableDictionary *item = [theOutline selectedItem];
	//NSIndexSet *set = [theOutline selectedRowIndexes];
	int i, n = [theOutline numberOfRows];
	for (i=0;i<n;i++) {
		if ([theOutline isRowSelected:i]) {
			NSMutableDictionary *item = [theOutline itemAtRow:i];
			if ([[item objectForKey:@"isLeaf"] boolValue]) {
				[item setObject:[NSNumber numberWithBool:[visibleSwitch state]] forKey:@"invisible"];
			}
		}
    }
}

- (IBAction)transparencySwitch:(id)sender
{
	//NSLog(@"trans switch");
    BOOL state = [transparencySwitch state];
    if ([transparencyColor respondsToSelector:@selector(setHidden:)])
	[transparencyColor setHidden:!state];
	NSArray *images = [NSArray arrayWithObjects:@"image", @"rimage", @"cimage", nil];
	NSMutableArray *tracks = [NSMutableArray array];
	
	int i, n = [theOutline numberOfRows];
	for (i=0;i<n;i++) {
		if ([theOutline isRowSelected:i]) {
			NSMutableDictionary *item = [theOutline itemAtRow:i];
    //NSMutableDictionary *item = [theOutline selectedItem];
			if ([[item objectForKey:@"isLeaf"] boolValue]) {
				NSColor *aColor = nil, *bColor = nil;
				NSData *dColor = [item objectForKey:@"keyColor"];
				if (dColor) bColor = [NSUnarchiver unarchiveObjectWithData:dColor];
				if (!state) {
					[item removeObjectForKey:@"keyColor"];
					if (bColor) { // there was a change
						[item removeObjectsForKeys:images];
						[tracks addObject:[item objectForKey:@"parent"]];
					}
				} else {
					aColor = [transparencyColor color];
					dColor = [NSArchiver archivedDataWithRootObject:aColor];
					if (dColor)
						[item setObject:dColor forKey:@"keyColor"];
					if (![aColor isEqual:bColor]) { // there was a change
						[item removeObjectsForKeys:images];
						[tracks addObject:[item objectForKey:@"parent"]];
					}
				}
			}
		}
    }
	[NSThread detachNewThreadSelector:@selector(reloadImages:) toTarget:self withObject:tracks];
    [[NSColor blackColor] set];
}

- (void) setTransparencyForSprite:(NSMutableDictionary *)sprite
{
    NSArray *keys = [NSArray arrayWithObjects:@"Image", @"Rollover", @"Click", nil];
    NSArray *keys2 = [NSArray arrayWithObjects:@"image", @"rimage", @"cimage", nil];
    int i, n = [keys count];
    for (i=0;i<n;i++) {
        NSImage *image = [self imageForSprite:sprite key:[keys objectAtIndex:i]];
        if (image) {
            [sprite setObject:image forKey:[keys2 objectAtIndex:i]];
            [theImageView setNeedsDisplay:YES];
        }
    }
}

- (IBAction)loopSwitch:(id)sender
{
    BOOL state = [loopSwitch state];
    //NSMutableDictionary *item = [theOutline selectedItem];
	
	int i, n = [theOutline numberOfRows];
	for (i=0;i<n;i++) {
		if ([theOutline isRowSelected:i]) {
			NSMutableDictionary *item = [theOutline itemAtRow:i];
	
			if ([[item objectForKey:@"isLeaf"] boolValue])
				item = [item objectForKey:@"parent"];
			if (state)
				[item setObject:[NSNumber numberWithBool:YES] forKey:@"loop"];
			else
				[item removeObjectForKey:@"loop"];
		}
	}
}

- (IBAction)opaqSlider:(id)sender
{
	//NSMutableDictionary *item = [theOutline selectedItem];
	
	int i, n = [theOutline numberOfRows];
	for (i=0;i<n;i++) {
		if ([theOutline isRowSelected:i]) {
			NSMutableDictionary *item = [theOutline itemAtRow:i];
			[item setObject:[NSNumber numberWithFloat:[sender floatValue]] forKey:@"opacity"];
		}
	}
	[theImageView setNeedsDisplay:YES];
}

// flag: 0 = get from defs, 1 = save to defs
- (void)prefs:(int)flag
{
    //NSLog(@"sprite");
    NSArray *boxes = [NSArray arrayWithObjects: alwaysDelete, byNumber, nil];
    NSArray *strings = [NSArray arrayWithObjects: @"alwaysDeleteSprites", @"spritesByNumber", nil];
    NSUserDefaults *defs = [NSUserDefaults standardUserDefaults];
    int i, n = [boxes count];
    bool state;
    for (i=0;i<n;i++) {
        if (flag == 0) {
            state = [defs boolForKey:[strings objectAtIndex:i]];
            [[boxes objectAtIndex:i] setState:state];
        } else if (flag == 1) {
            [defs setBool:[[boxes objectAtIndex:i] state] forKey:[strings objectAtIndex:i]];
        }
    }
    NSArray *floats = [NSArray arrayWithObjects:preloadSize, wizSize, nil];
    strings = [NSArray arrayWithObjects:@"preloadSize", @"wizSize", nil];
    n = [floats count];
    for (i=0;i<n;i++) {
        if (flag) [defs setFloat:[[floats objectAtIndex:i] floatValue] forKey:[strings objectAtIndex:i]];
        else if ([defs objectForKey:[strings objectAtIndex:i]]) [[floats objectAtIndex:i] setFloatValue:[defs floatForKey:[strings objectAtIndex:i]]];
    }
    [defs synchronize];
    //NSLog(@"sprite");
}

- (void)restorePrefs
{
    [self prefs:0];
}

- (IBAction)deleteSpriteTracks:(id)sender
{
    NSMovie *movie = [spriteMovieView movie];
    Movie mov = [movie QTMovie];
    if (!mov) return;
    OSType type;
    Track track;
    long i, n = GetMovieTrackCount(mov);
    for (i=1;i<=n;i++) {
        track = GetMovieIndTrack(mov, i);
        GetMediaHandlerDescription(GetTrackMedia(track), &type, nil, nil);
        if (type == SpriteMediaType)
            DisposeMovieTrack(track);
    }
    n = GetMovieTrackCount(mov);
    TimeValue offset1, offset2 = GetMovieDuration(mov);
    for (i=1;i<=n;i++) {
        track = GetMovieIndTrack(mov, i);
        offset1 = GetTrackOffset(track);
        if (offset1 < offset2)
            offset2 = offset1;
    }
    DeleteMovieSegment(mov, 0, offset2);
    [spriteMovieView setMovie:nil];
    [spriteMovieView setMovie:movie];
    [(MHMovie *)movie setSaveInPlaceUnSafe];
}

- (void)syncWidgets:(NSMutableDictionary *)sprite
{
	//NSLog(@"sync: %@", [sprite objectForKey:@"Name"]);
	NSNumber *opq = [sprite objectForKey:@"opacity"];
	float o = 1;
	if (opq) o = [opq floatValue];
	[[NSColor blackColor] set];
	[opaqSlider setFloatValue:o];
    [qScriptBox setString:@""];
    [actionValue setEnabled:NO];
    [actionValue removeAllItems];
    if (![[sprite objectForKey:@"isLeaf"] boolValue]) {
        // disable everything
        [setImageButton setEnabled:NO];
        [actionType setEnabled:NO];
        [actionType selectItemAtIndex:0];
        [transparencySwitch setState:NO];
        [transparencySwitch setEnabled:NO];
	if ([transparencyColor respondsToSelector:@selector(setHidden:)])
	    [transparencyColor setHidden:YES];
        return;
    }
	[actionType setEnabled:YES];
    MHChapters *ch;
    NSMovie *movie = [spriteMovieView movie];
    NSMutableDictionary *menu = [movie menu];
    if (!menu)
        menu = [self newMenu];
    NSArray *tracks = [menu objectForKey:@"tracks"];
    NSString *action = [sprite objectForKey:@"Action"];
    if (![action length]) [actionType selectItemAtIndex:0];
    else {
        [qScriptBox setString:action];
        NSString *buffer = nil;
        NSScanner *scan = [NSScanner scannerWithString:action];
        [scan setCaseSensitive:NO];
        if ([scan scanString:@"gotochapter(" intoString:nil] || [scan scanString:@"SetSelectionByChapter(" intoString:&buffer]) {
            [actionType selectItemAtIndex:1];
			if (buffer)
				[actionType selectItemAtIndex:15];
            ch = [[[MHChapters alloc] initFromMovie:movie] autorelease];
            [actionValue addItemsWithTitles:[ch chapterNames]];
            [actionValue setEnabled:YES];
			buffer = nil;
            [scan scanUpToString:@")" intoString:&buffer];
            if (buffer && [actionValue itemWithTitle:buffer])
                [actionValue selectItemWithTitle:buffer];
        } else if ([scan scanString:@"// set" intoString:nil]) {
            OSType type = 0;
            [actionValue setEnabled:YES];
            if ([scan scanString:@"audio track" intoString:nil]) {
                type = SoundMediaType;
                [actionType selectItemAtIndex:2];
            } else if ([scan scanString:@"text track" intoString:nil]) {
                type = TextMediaType;
                [actionType selectItemAtIndex:3];
            }
            if (type)
                [actionValue addItemsWithTitles:[self getTracksOfType:type]];
            while (![scan isAtEnd]) {
                buffer = nil;
                [scan scanPastString:@"tracknamed("];
                [scan scanUpToString:@")" intoString:&buffer];
                [scan scanPastString:@"setenabled("];
                if ([scan scanString:@"yes" intoString:nil] || [scan scanString:@"true" intoString:nil]) {
                    if (buffer && [actionValue itemWithTitle:buffer])
                        [actionValue selectItemWithTitle:buffer];
                    break;
                }
            }
        } else if ([scan scanString:@"stepforward" intoString:nil]) {
            [actionType selectItemAtIndex:4];
        } else if ([scan scanString:@"stepbackward" intoString:nil]) {
            [actionType selectItemAtIndex:5];
        } else if ([scan scanString:@"gotopage(" intoString:nil]) {
            [actionType selectItemAtIndex:6];
            int i, n = [tracks count];
            for (i=0;i<n;i++)
                [actionValue addItemWithTitle:[NSString stringWithFormat:@"%d", i]];
            [actionValue setEnabled:YES];
            int v = 0;
            [scan scanInt:&v];
            if (v >= 0 && v < [actionValue numberOfItems])
                [actionValue selectItemAtIndex:v];
        } else {
			NSArray *actions = [NSArray arrayWithObjects:@"SetVolume", @"EnterFullscreen", @"ExitFullScreen", @"Close", @"StartPlaying", @"SetRate", @"PlaySelection", @"SetSelection", @"SetSelectionByChapter", @"SaveTime", @"GoToSavedTime", @"SetVisible", @"Rotate", @"Scale", @"Translate", @"SetOpacity", nil];
			if ([scan scanString:@"SpriteNamed(" intoString:nil]) {
				[scan scanUpToString:@")" intoString:&buffer];
				[scan scanString:@")" intoString:nil];
				[scan scanString:@"." intoString:nil];
			}
			int i = [actions count]-1;
			for (;i>=0;i--) {
				if ([scan scanString:[actions objectAtIndex:i] intoString:nil]) {
					[actionType selectItemAtIndex:i+7];
					if (i>=11) {
						[actionValue setEnabled:YES];
						[actionValue addItemsWithTitles:[self spriteNames:[theOutline selectedItem]]];
						if (!buffer) buffer = [[theOutline selectedItem] objectForKey:@"Name"];
						[actionValue selectItemWithTitle:buffer];
					}
					break;
				}
			}
		}
    }
    
    NSData *dColor = [sprite objectForKey:@"keyColor"];
    if (dColor) {
        NSColor *kColor = [NSUnarchiver unarchiveObjectWithData:dColor];
        [transparencySwitch setState:YES];
	if ([transparencyColor respondsToSelector:@selector(setHidden:)])
	    [transparencyColor setHidden:NO];
        [transparencyColor setColor:kColor];
    } else {
        [transparencySwitch setState:NO];
	if ([transparencyColor respondsToSelector:@selector(setHidden:)])
	    [transparencyColor setHidden:YES];
    }
    [visibleSwitch setState:[[sprite objectForKey:@"invisible"] boolValue]];
}

- (IBAction)updateWidgets:(id)sender
{
    int row = [theOutline selectedRow];
    NSMutableDictionary *item = nil;
    MHChapters *ch;
    NSMovie *movie = [spriteMovieView movie];
    NSMutableDictionary *menu = [movie menu];
    if (!menu)
        menu = [self newMenu];
    NSArray *tracks = [menu objectForKey:@"tracks"];
    int i, n;
    [[NSColor blackColor] set];
	
    if (-1 != row)
        item = [theOutline itemAtRow:row];
    if (!item) return;
    if (![[item objectForKey:@"isLeaf"] boolValue]) {
		[loopSwitch setState:[[item objectForKey:@"loop"] boolValue]];
        // disable everything
        [setImageButton setEnabled:NO];
        [actionType setEnabled:NO];
        [actionType selectItemAtIndex:0];
        [actionValue setEnabled:NO];
        [actionValue removeAllItems];
        [viewImageButton setEnabled:NO];
		[transparencySwitch setState:NO];
        [transparencySwitch setEnabled:NO];
		if ([transparencyColor respondsToSelector:@selector(setHidden:)])
			[transparencyColor setHidden:YES];
        [loadTemplateButton setEnabled:NO];
        [saveTemplateButton setEnabled:NO];
        [visibleSwitch setState:NO];
        [visibleSwitch setEnabled:NO];
    } else {
        [setImageButton setEnabled:YES];
        [viewImageButton setEnabled:YES];
        [actionType setEnabled:YES];
        [transparencySwitch setEnabled:YES];
        [loadTemplateButton setEnabled:YES];
        [saveTemplateButton setEnabled:YES];
        [visibleSwitch setEnabled:YES];
		if ([[[item objectForKey:@"parent"] objectForKey:@"Image"] hasSuffix:@".mov"]) {
			[transparencySwitch setState:NO];
			[transparencySwitch setEnabled:NO];
			if ([transparencyColor respondsToSelector:@selector(setHidden:)])
				[transparencyColor setHidden:YES];
		}
        
        NSArray *tracksOfType;
        
        int actionSelection = [actionType indexOfSelectedItem];
        switch (actionSelection) {
            case 1: case 15: // gotoChapter
            [actionValue removeAllItems];
            ch = [[MHChapters alloc] initFromMovie:movie];
            [actionValue addItemsWithTitles:[ch chapterNames]];
            [actionValue setEnabled:YES];
            if ([actionValue numberOfItems])
                [actionValue selectItemAtIndex:0];
            break;
            case 2: // set audio track
            [actionValue removeAllItems];
            [actionValue setEnabled:YES];
            tracksOfType = [self getTracksOfType:SoundMediaType];
            if (tracksOfType && [tracksOfType count])
                [actionValue addItemsWithTitles:tracksOfType];
            else
                [actionValue setEnabled:NO];
            if ([actionValue numberOfItems])
                [actionValue selectItemAtIndex:0];
            break;
            case 3: // set text track
            [actionValue removeAllItems];
            [actionValue setEnabled:YES];
            tracksOfType = [self getTracksOfType:TextMediaType];
            if (tracksOfType && [tracksOfType count])
                [actionValue addItemsWithTitles:tracksOfType];
            else
                [actionValue setEnabled:NO];
            if ([actionValue numberOfItems])
                [actionValue selectItemAtIndex:0];
            break;
            case 6: // go to page
            [actionValue removeAllItems];
            n = [tracks count];
            for (i=0;i<n;i++) {
                NSString *name = [[tracks objectAtIndex:i] objectForKey:@"Name"];
                if (!name) name = @"";
                [actionValue addItemWithTitle:[NSString stringWithFormat:@"%d. %@",i,name]];
            }
            [actionValue setEnabled:YES];
            if ([actionValue numberOfItems])
                [actionValue selectItemAtIndex:0];
            break;/*
			case 18: //sprite actions
				[actionValue removeAllItems];
				NSDictionary *page = [theOutline selectedItem];
				if ([[page objectForKey:@"isLeaf"] boolValue]) page = [page objectForKey:@"parent"];
				n = [[page objectForKey:@"sprites"] count];
				for (i=0;i<n;i++)
					[actionValue addItemWithTitle:[[[page objectForKey:@"sprites"] objectAtIndex:i] objectForKey:@"Name"]];
			break;*/
            default: // none, step forward, step backward
            [actionValue removeAllItems];
            [actionValue setEnabled:NO];
            break;
        }
    }
}

- (NSArray *)spriteNames:(NSDictionary *)item
{
	NSDictionary *parent = [item objectForKey:@"parent"];
	if (!parent) parent = item;
	NSArray *sprites = [parent objectForKey:@"sprites"];
	int i, n = [sprites count];
	NSMutableArray *a = [NSMutableArray arrayWithCapacity:n];
	for (i=0;i<n;i++)
		[a addObject:[[sprites objectAtIndex:i] objectForKey:@"Name"]];
	return a;
}

- (NSArray *)chapterNames
{
    MHChapters *ch = [[MHChapters alloc] initFromMovie:[spriteMovieView movie]];
    return [ch chapterNames];
}

- (NSArray *)pageNumbers
{
    NSDictionary *menu = [[spriteMovieView movie] menu];
    if (!menu)
	menu = [self newMenu];
    NSArray *tracks = [menu objectForKey:@"tracks"];
    int i, n = [tracks count];
    NSMutableArray *array = [NSMutableArray array];
    for (i=0;i<n;i++) {
	NSString *name = [[tracks objectAtIndex:i] objectForKey:@"Name"];
	if (!name) name = @"";
	[array addObject:[NSString stringWithFormat:@"%d. %@", i, name]];
    }
    return array;
}

- (IBAction)updateData:(id)sender
{
    //Movie mov = [[spriteMovieView movie] QTMovie];
    //Rect box;
    //GetMovieBox(mov, &box);
    //NSMutableDictionary *item = [theOutline selectedItem];
    //if (!item || ![[item objectForKey:@"isLeaf"] boolValue]) return;
    int type = [actionType indexOfSelectedItem];
    NSString *value = [actionValue titleOfSelectedItem];
    NSString *actionString = [self stringForAction:type value:value];
	int i, n = [theOutline numberOfRows];
	NSMutableDictionary *item = nil;
	for (i=0;i<n;i++) {
		if ([theOutline isRowSelected:i]) {
			item = [theOutline itemAtRow:i];
			if ([[item objectForKey:@"isLeaf"] boolValue]) {
				if (![actionString length]) [item removeObjectForKey:@"Action"];
				else [item setObject:actionString forKey:@"Action"];
			}
		}
	}
    [self syncWidgets:item];
    [theImageView setNeedsDisplay:YES];
}
    
- (NSString *)stringForAction:(int)type value:(NSString *)value
{
    NSString *actionString = @"";
    NSMutableString *name = nil;
    BOOL hit = NO;
    int i, n;
    MHTrack *track = nil;
    MHMovie *movie = (MHMovie *)[spriteMovieView movie];
	NSArray *names;
    switch (type) {
        case 0: // none
            break;
        case 1: // go to chapter
            actionString = @"GoToChapter(";
	    if (value) actionString = [actionString stringByAppendingString:value];
            actionString = [actionString stringByAppendingString:@");\nStartPlaying;"];
	    break;
        case 2: // set audio track
            actionString = @"// set audio track\n";
            n = [movie trackCount];
            for (i=0;i<n;i++) {
                track = [movie trackAtIndex:i];
                if ([track isSoundType] && [track name]) {
                    name = [NSMutableString stringWithString:[track name]];
                    hit = [value isEqualToString:name];
                    if ([byNumber state])
                        actionString = [actionString stringByAppendingFormat:@"TrackAtIndex(%d)", i];
                    else {
                        //[name replaceOccurrencesOfString:@")" withString:@"\\)" options:NSLiteralSearch range:NSMakeRange(0, [name length])];
                        actionString = [actionString stringByAppendingFormat:@"TrackNamed(%@)", name];
                    }
                    actionString = [actionString stringByAppendingFormat:@".SetEnabled(%@);\n", (hit ? @"YES" : @"NO")];
                }
            }
			if ([actionString hasSuffix:@"\n"]) actionString = [actionString substringToIndex:[actionString length]-1];
	    break;
        case 3: // set text track
            actionString = @"// set text track\n";
			names = [self getTracksOfType:TextMediaType];
            n = [names count];
            for (i=0;i<n;i++) {
				name = [names objectAtIndex:i];
				hit = [value isEqualToString:name];
				if ([byNumber state]) actionString = [actionString stringByAppendingFormat:@"TrackAtIndex(%d)", i];
				else {
					//[name replaceOccurrencesOfString:@")" withString:@"\\)" options:NSLiteralSearch range:NSMakeRange(0, [name length])];
					actionString = [actionString stringByAppendingFormat:@"TrackNamed(%@)", name];
				}
				actionString = [actionString stringByAppendingFormat:@".SetEnabled(%@);\n", (hit ? @"YES" : @"NO")];
			}
			if ([actionString hasSuffix:@"\n"]) actionString = [actionString substringToIndex:[actionString length]-1];
			break;
        case 4: // step forward
            actionString = @"StepForward;";
            break;
        case 5: // step backward
            actionString = @"StepBackward;";
            break;
        case 6: // go to page
            actionString = [NSString stringWithFormat:@"GoToPage(%d);", [value intValue]];
            break;
		case 7: // set volume
			if (value) actionString = [NSString stringWithFormat:@"SetVolume(%d);", [value intValue]];
			else actionString = @"SetVolume(256);";
			break;
		case 8: // enter fullscreen
			actionString = @"EnterFullScreen;";
			break;
		case 9: // exit fullscreen
			actionString = @"ExitFullScreen;";
			break;
		case 10: // close
			actionString = @"Close;";
			break;
		case 11: // start
			actionString = @"StartPlaying;";
			break;
		case 12: // set rate
			if (value) actionString = [NSString stringWithFormat:@"SetRate(%g);", [value floatValue]];
			else actionString = @"SetRate(0.5);";
			break;
		case 13: // play selection
			actionString = @"PlaySlection";	break;
		case 14: // set selection
			actionString = @"SetSelection(600, 1200);";		break;
		case 15: // set selection by name
			if (value) actionString = [NSString stringWithFormat:@"SetSelectionByChapter(%@);", value];
			else actionString = @"SetSelectionByChapter();";
			break;
		case 16: // save time
			actionString = @"SaveTime;";		break;
		case 17: // go to saved time
			actionString = @"GoToSavedTime;";	break;
		case 18: // set sprite visible
			if (value) actionString = [NSString stringWithFormat:@"SpriteNamed(%@).SetVisible(YES);", value];
			else actionString = @"SetVisible(YES);";
			break;
		case 19: // rotate sprite
			if (value) actionString = [NSString stringWithFormat:@"SpriteNamed(%@).Rotate(90);", value];
			else actionString = @"Rotate(90);";
			break;
		case 20: // scale
			if (value) actionString = [NSString stringWithFormat:@"SpriteNamed(%@).Scale(1.1, 1.1);", value];
			else actionString = @"Scale(1.1, 1.1);";
			break;
		case 21: // translate
			if (value) actionString = [NSString stringWithFormat:@"SpriteNamed(%@).Translate(2, 2);", value];
			else actionString = @"Translate(2, 2);";
			break;
		case 22: // opacity
			if (value) actionString = [NSString stringWithFormat:@"SpriteNamed(%@).SetOpacity(0.5);", value];
			else actionString = @"SetOpacity(0.5);";
			break;
        default: break;
    }
    return actionString;
}

- (IBAction)selectActionType:(id)sender
{
    [self updateWidgets:nil];
    [self updateData:nil];
    [theOutline reloadData];
}

- (IBAction)setAction:(id)sender
{
    NSString *title = [sender title];
    NSArray *titles = [NSArray arrayWithObjects:@"None", @"Go To Chapter", @"Set Audio Track", @"Set Text Track", @"Step Forward", @"Step Backward", @"Go To Page", @"Set Volume", @"Enter Fullscreen", @"Exit Fullscreen", @"Close", @"Start Playing", @"Set Rate", @"Play Selection", @"Set Selection", @"Set Selection By Chapter", @"Save Time", @"Go To Saved Time", @"Set Sprite Visible", @"Rotate Sprite", @"Scale Sprite", @"Move Sprite", @"Set Sprite Opacity", nil];
    NSArray *actions = [NSArray arrayWithObjects:@"", @"GoToChapter()", @"// set audio track\nTrackNamed().SetEnabled();", @"// set text track\nTrackNamed().SetEnabled();", @"StepForward;", @"StepBackward;", @"GoToPage(0);", @"SetVolume(125);", @"EnterFullScreen;", @"ExitFullScreen;", @"Close;", @"StartPlaying;", @"SetRate(0.5)", @"PlaySelection;", @"SetSelection(600, 1200)", @"SetSelectionByChapter();", @"SaveTime;", @"GoToSavedTime;", @"SetVisible(YES);", @"Rotate(90);", @"Scale(1.1, 1.1);", @"Translate(2, 2);", @"SetOpacity(0.5);", nil];
    int i, n = [titles count];
    for (i=0;i<n;i++) {
		if ([title isEqualToString:[titles objectAtIndex:i]]) {
			int j, m = [theOutline numberOfRows];
			for (j=0;j<m;j++) {
				if ([theOutline isRowSelected:j]) {
					NSMutableDictionary *item = [theOutline itemAtRow:j];
					if ([[item objectForKey:@"isLeaf"] boolValue])
						[item setObject:[actions objectAtIndex:i] forKey:@"Action"];
				}
			}
		}
    }
    [theOutline reloadData];
    [self updateWidgets:nil];
    [self syncWidgets:(NSMutableDictionary *)[theOutline selectedItem]];
}

- (IBAction)selectActionValue:(id)sender
{
    [self updateData:nil];
}

- (IBAction)setValue:(id)sender
{
    NSString *title = [sender title];
    int i = 0, n = 0;
    if (!title) return;
    if ([sender isKindOfClass:[NSMenuItem class]]) {
        NSMenu *submenu = [sender menu];
        NSMenu *supermenu = [submenu supermenu];
        n = [supermenu numberOfItems];
        for (i=0;i<n;i++) {
            if ([[supermenu itemAtIndex:i] submenu] == submenu) {
                //[self setAction:[supermenu itemAtIndex:i]];
                break;
            }
        }
    }
    NSString *string = nil;
    if (i<n)
        string = [self stringForAction:i value:title];
	n = [theOutline numberOfRows];
	for (i=0;i<n;i++) {
		if ([theOutline isRowSelected:i]) {
			NSMutableDictionary *item = [theOutline itemAtRow:i];
			if ([[item objectForKey:@"isLeaf"] boolValue]) {
				if (string)
					[item setObject:string forKey:@"Action"];
				else
					[item removeObjectForKey:@"Action"];
			}
		}
	}
    [theOutline reloadData];
    [self syncWidgets:(NSMutableDictionary *)[theOutline selectedItem]];
}

- (IBAction)viewFrame:(id)sender
{
    id item = [theOutline itemAtRow:[theOutline selectedRow]];
    //if (![[item objectForKey:@"isLeaf"] boolValue]) return;
    int sel = [imageSelect indexOfSelectedItem];
    NSString *buffer = nil;
    TimeValue tv = 0;
    switch (sel) {
        case 0: // main
            buffer = [item objectForKey:@"Image"];
            break;
        case 1: // rollover
            buffer = [item objectForKey:@"Rollover"];
            break;
        case 2: // click
            buffer = [item objectForKey:@"Click"];
            break;
        default: break;
    }
    if (!buffer) return;
    if ([buffer isKindOfClass:[NSString class]] && [buffer hasPrefix:@"/"]) return;
    tv = [buffer intValue];
    TimeRecord tr;
    Movie mov = [[spriteMovieView movie] QTMovie];
    GetMovieTime(mov, &tr);
    tr.value.lo = tv;
    SetMovieTime(mov, &tr);
	MoviesTask(mov, 0);
}

- (IBAction)setBackground:(id)sender
{
	NSString *key = nil;
	NSImage *image = nil;
    int type = [imageType selectedRow];
    if (type == 0) { // movie frame
        NSMovie *aMovie = [spriteMovieView movie];
        Movie mov = [aMovie QTMovie];
        //id helper = [MHUtil imageHelper];
        //int height = [spriteMovieView sizeForMagnification:1.0].height - 16;
        //[helper setHeight:height];
        //[helper openMovie:[aMovie URL]];
        TimeRecord tr;
        TimeValue tv = GetMovieTime(mov, &tr);
        //NSData *data = [helper getImage:tv sender:@"mhSprite"];
        //if (data) {
		//NSImage *image = [NSUnarchiver unarchiveObjectWithData:data];
		//if (image)
		//[item setObject:image forKey:@"bgImage"];
        key = [NSString stringWithFormat:@"%ld",tv]; // forKey:@"bgImageSource"];
        //image = [self imageFromFrame:tv height:height];
        //if (image) [item setObject:image forKey:@"bgImage"];
		//[item setObject:[NSString stringWithFormat:@"%d",tv] forKey:@"Image"];
        //}
    } else { // file
        NSOpenPanel *theOpenPanel = [NSOpenPanel openPanel];
		
        [theOpenPanel setCanChooseFiles:YES];
        [theOpenPanel setCanChooseDirectories:NO];
        [theOpenPanel setResolvesAliases:YES];
        [theOpenPanel setAllowsMultipleSelection:NO];
        //NSArray *extensions = [NSArray arrayWithObjects:@"pict", @"PICT", @"pct", @"PCT", @"mov", nil];
        int result = [theOpenPanel runModalForDirectory:nil file:nil types:nil];
        if (result == NSOKButton) {
            key = [[theOpenPanel filenames] objectAtIndex:0];
            if ([key hasSuffix:@".mov"]) {
                //MHMovie *menuMovie = [[[MHMovie alloc] initWithPath:choice] autorelease];
				//Rect box;
				//GetMovieBox([menuMovie QTMovie], &box);
                //TimeValue time = GetMovieDuration([menuMovie QTMovie]) / 2;
                //image = [self imageFromFrame:time height:box.bottom - box.top movieURL:[menuMovie URL]];
            } else
                image = [[NSImage alloc] initWithContentsOfFile:key];
    	}
	}
	
	if (!key) return;
    //NSMutableDictionary *item = [theOutline itemAtRow:[theOutline selectedRow]];
	int i, n = [theOutline numberOfRows];
	for (i=0;i<n;i++) {
		if ([theOutline isRowSelected:i]) {
			NSMutableDictionary *item = [theOutline itemAtRow:i];
			if ([[item objectForKey:@"isLeaf"] boolValue])
				item = [item objectForKey:@"parent"];
			[item setObject:key forKey:@"bgImageSource"];
			[item setObject:key forKey:@"Image"];
			if (!image)
				image = [self imageForSprite:item key:@"Image"];
			if (image)
				[item setObject:image forKey:@"bgImage"];
		}
	}
    [theOutline reloadData];
    [self outlineViewSelectionDidChange:nil];
    //[theImageView setNeedsDisplay:YES];
}

- (NSImage *)imageFromFrame:(TimeValue)time height:(int)height
{
    return [self imageFromFrame:time height:height movie:(MHMovie *)[spriteMovieView movie]];
}

- (NSImage *)imageFromFrame:(TimeValue)time height:(int)height movie:(MHMovie *)movie
{
    return [self imageFromFrame:time height:height movieURL:[movie URL]];
}

- (NSImage *)imageFromFrame:(TimeValue)time height:(int)height movieURL:(NSURL *)url
{
	//return [[QTImage alloc] initWithFile:[url path] transparency:nil time:time];
	
	NSMutableDictionary *menu = [[spriteMovieView movie] menu];
	if (!menu)
		menu = [self newMenu];
	QTImage *qi = [menu objectForKey:@"qi"];
	if (!qi) {
		qi = [[[QTImage alloc] init] autorelease];
		[menu setObject:qi forKey:@"qi"];
	}
	return [qi imageForPath:[url path] transparency:nil time:time];
	
	/*
    id helper = [MHUtil imageHelper];
    if (!height)
        height = 80;
    [helper setHeight:height];
    [helper openMovie:url];
    NSData *data = [helper getImage:time sender:@"mhSprite"];
    if (data)
        return [NSUnarchiver unarchiveObjectWithData:data];
    return nil;*/
}

- (BOOL)hasSprites
{
    NSArray *pages = [[[spriteMovieView movie] menu] objectForKey:@"tracks"];
    int i, n = [pages count];
    for (i=0;i<n;i++) {
        if ([(NSArray *)[[pages objectAtIndex:i] objectForKey:@"sprites"] count]) return YES;
        if ([[pages objectAtIndex:i] objectForKey:@"bgImageSource"]) return YES;
    }
    return NO;
}

- (IBAction)wizard:(id)sender
{
	int choice;
    MHChapters *ch = [[[MHChapters alloc] initFromMovie:[spriteMovieView movie]] autorelease];
    if (![ch count] && [[self getTracksOfType:SoundMediaType] count] < 2 && [[self getTracksOfType:TextMediaType] count] < 1) {
		choice = NSRunAlertPanel(@"I Need Chapters", @"This movie has no chapters.  Wizard uses the movie's chapters to construct interactive menus.  Please add chapters before running Wizard.", @"Add Chapters", @"Cancel", nil);
		if (choice == NSAlertDefaultReturn) {
			NSMenu *m = [[[NSApp mainMenu] itemWithTitle:@"View"] submenu];
			choice = [m indexOfItemWithTitle:@"Chapters"];
			[m performActionForItemAtIndex:choice];
		}
		return;
	}
	if ([self hasSprites]) {
        choice = NSRunAlertPanel(@"Erase Existing Sprites?", @"Wizard uses the movie's chapters to construct interactive menus.  It will replace the existing sprites.  Do you want to overwrite?", @"Overwrite", @"Cancel", nil);
        if (choice == NSAlertAlternateReturn) return;
    }
    [progText setStringValue:@""];
    [progBar setDoubleValue:0];
    [NSApp beginSheet: progWindow modalForWindow: spriteWindow modalDelegate: nil didEndSelector: nil contextInfo: nil];
    [NSThread detachNewThreadSelector:@selector(wizardThreaded) toTarget:self withObject:nil];
}

- (NSMutableDictionary *)addSelectionPage:(OSType)type
{
	MHMovie *aMovie = (MHMovie *)[spriteMovieView movie];
	NSMutableDictionary *menu = [aMovie menu];
	if (!menu) {
		menu = [self newMenu];
	}
	[self newPage:nil];
	int w = naturalSize.width / 4;
	if (w > 100) w = 100;
	int h = naturalSize.height / 8;
	NSString *button = [[NSBundle mainBundle] pathForResource:@"Empty" ofType:@"mhsprite"];
	NSMutableDictionary *track = [[menu objectForKey:@"tracks"] lastObject], *newSprite = nil;
	int ran = random() % 100;
	TimeValue tv = (float)GetMovieDuration([aMovie QTMovie]) / 100.0 * ran;
	NSString *bg = [NSString stringWithFormat:@"%ld", tv];
	[track setObject:bg forKey:@"Image"];
	[track setObject:bg forKey:@"bgImageSource"];
	[track setObject:[NSNumber numberWithFloat:0.5] forKey:@"opacity"];
	int i, x, y;
	NSMutableArray *myTracks = [self getTracksOfType:type];
	if (type == TextMediaType) {
		[myTracks insertObject:@"None" atIndex:0];
		/* this check has been moved upstream
		for (i=[myTracks count]-1;i>=0;i--)
			if ([[myTracks objectAtIndex:i] hasPrefix:@"Chapter"])
				[myTracks removeObjectAtIndex:i];*/
	}
	NSString *typeString = @"";
	NSString *name = nil, *action = nil;
	if (type == SoundMediaType) typeString = @"Audio ";
	else if (type == TextMediaType) typeString = @"Subtitle ";
	for (i=0;i<[myTracks count];i++) {
		newSprite = [NSMutableDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithBool:YES], @"isLeaf", track, @"parent", [NSString stringWithFormat:@"%d", h], @"Height", [NSString stringWithFormat:@"%d", w], @"Width", nil];
		x = naturalSize.width / 2;
		[newSprite setObject:[NSString stringWithFormat:@"%d", x] forKey:@"X"];
		y = naturalSize.height / 8 * (i+1) * 1.5;
		[newSprite setObject:[NSString stringWithFormat:@"%d", y] forKey:@"Y"];
		name = [myTracks objectAtIndex:i];
		if (!name) name = [NSString stringWithFormat:@"%@Track %d", typeString, i+1];
		[newSprite setObject:name forKey:@"Name"];
		int actType = 0;
		if (type == SoundMediaType) actType = 2;
		else if (type == TextMediaType) actType = 3;
		action = [self stringForAction:actType value:name];
		[newSprite setObject:@"NoHitTest" forKey:@"Action"];
		name = [NSString stringWithFormat:@"\"{textColor:0,0,0}{backColor:65535,65535,65535}%@\"", name];
		[newSprite setObject:name forKey:@"Image"];
		[newSprite setObject:[NSArchiver archivedDataWithRootObject:[NSColor whiteColor]] forKey:@"keyColor"];
		[[track objectForKey:@"sprites"] addObject:newSprite];
		newSprite = [NSMutableDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithBool:YES], @"isLeaf", track, @"parent", [NSString stringWithFormat:@"%d", h+20], @"Height", [NSString stringWithFormat:@"%d", w+10], @"Width", [NSString stringWithFormat:@"%d", x-5], @"X", [NSString stringWithFormat:@"%d", y-15], @"Y", [NSString stringWithFormat:@"Sprite %d", i+1], @"Name", action, @"Action", nil];
		[self readSpriteTemplateFromPath:button toSprite:newSprite];
		[[track objectForKey:@"sprites"] insertObject:newSprite atIndex:i*2];
		
	}
	newSprite = [NSMutableDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithBool:YES], @"isLeaf", track, @"parent", [NSString stringWithFormat:@"%d", 17], @"Height", [NSString stringWithFormat:@"%d", w+10], @"Width", [NSString stringWithFormat:@"%d", x-5], @"X", [NSString stringWithFormat:@"%d", y + h+15], @"Y", @"Play Movie", @"Name", @"\"Play Movie\"", @"Image", @"\"{backColor:0,0,65535}Play Movie\"", @"Rollover", nil];
	if ([[aMovie chapters] count]) {
		action = [[aMovie chapters] nameForTime:0];
		if ([action length]) {
			action = [NSString stringWithFormat:@"GoToChapter(%@);\nStartPlaying;", action];
			[newSprite setObject:action forKey:@"Action"];
		}
	}
	//[self readSpriteTemplateFromPath:button toSprite:newSprite];
	[[track objectForKey:@"sprites"] addObject:newSprite];
	// label
	newSprite = [NSMutableDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithBool:YES], @"isLeaf", track, @"parent", [NSString stringWithFormat:@"%d", (int)naturalSize.width /8], @"X", [NSString stringWithFormat:@"%d", (int)naturalSize.height /3], @"Height", @"Label", @"Name", nil];
	y = naturalSize.height / 4;
	w = naturalSize.width / 4;
	if (w < 135) w = 135;
	[newSprite setObject:[NSString stringWithFormat:@"%d", w] forKey:@"Width"];
	[newSprite setObject:[NSString stringWithFormat:@"%d", y] forKey:@"Y"];
	if (type == SoundMediaType) name = @"Select Audio";
	else if (type == TextMediaType) name = @"Select Subtitles";
	else name = @"Select Tracks";
	[newSprite setObject:[NSString stringWithFormat:@"\"{size:32}%@\"", name] forKey:@"Image"];
	[[track objectForKey:@"sprites"] addObject:newSprite];
	[track setObject:name forKey:@"Name"];
	/*
	newSprite = [NSMutableDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithBool:YES], @"isLeaf", track, @"parent", @"5", @"X", @"5", @"Y", [NSString stringWithFormat:@"%d", (int)naturalSize.height], @"Height", [NSString stringWithFormat:@"%d", (int)naturalSize.width], @"Width", @"Title", @"Name", [NSString stringWithFormat: @"\"{size:24}%@\"", [aMovie stringValue]], @"Image", [NSArchiver archivedDataWithRootObject:[NSColor blackColor]], @"keyColor", nil];
	[[track objectForKey:@"sprites"] addObject:newSprite];*/
	return track;
}

- (void)wizardThreaded
{
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
    MHMovie *aMovie = (MHMovie *)[spriteMovieView movie];
    if (!aMovie) goto bail;
    cancelled = NO;
    workingPath = [[aMovie objectForKey:@"File Path"] stringByDeletingLastPathComponent];
    NSFileManager *fm = [NSFileManager defaultManager];
    float scalar = [wizSize floatValue];
    if (!scalar) scalar = 1;
    int width = naturalSize.width / 3 - 20;
    int height = naturalSize.height * width / naturalSize.width;
    width = (float) width * scalar;
    height = (float) height * scalar;
    //id helper = [MHUtil imageHelper];
    //[helper setHeight:height];
    //[fHeight setIntValue:height];
    //[helper openMovie:[aMovie URL]];
    MHChapters *ch = [[[MHChapters alloc] initFromMovie:aMovie] autorelease];
    NSMutableDictionary *track;
    NSMutableDictionary *menu = [aMovie menu];
    if (!menu)
        menu = [self newMenu];
    int i, n, count = [ch numberOfRowsInTableView:nil];
    n = [(NSArray *)[menu objectForKey:@"tracks"] count];
    NSMutableArray *bgs = [NSMutableArray array], *bgss = [NSMutableArray array];
    for (i=0;i<n;i++) {
        id obj = [[[menu objectForKey:@"tracks"] objectAtIndex:i] objectForKey:@"bgImage"];
        if (obj)
            [bgs addObject:obj];
        else
            [bgs addObject:[NSNull null]];
    }
    for (i=0;i<n;i++) {
        id obj = [[[menu objectForKey:@"tracks"] objectAtIndex:i] objectForKey:@"bgImageSource"];
        if (obj)
            [bgss addObject:obj];
        else
            [bgss addObject:[NSNull null]];
    }
	
	[loadLock lock];
	[[menu objectForKey:@"tracks"] removeAllObjects];
	[loadLock unlock];
	
    int numPages = count / 6;
    if (count % 6)
        numPages++;
	
	NSMutableDictionary *play1 = nil, *play2 = nil;
    NSArray *audioTracks = [self getTracksOfType:SoundMediaType];
    if (1 < [audioTracks count]) {
        // audio track selection
		play1 = [self addSelectionPage:SoundMediaType];
		numPages++;
		[self reloadImages:[menu objectForKey:@"tracks"]];
    }
	int ntracks;
    NSMutableArray *textTracks = [self getTracksOfType:TextMediaType];
	/* this check has been moved upstream
	for (i=[textTracks count]-1;i>=0;i--)
		if ([[textTracks objectAtIndex:i] hasPrefix:@"Chapter"])
			[textTracks removeObjectAtIndex:i];*/
	if ([textTracks count]) {
		if ([audioTracks count] > 1) // add the last page's arrow
			[self addNavButton:@"next" destination:1];
		// text track selection
		play2 = [self addSelectionPage:TextMediaType];
		numPages++;
		ntracks = [[menu objectForKey:@"tracks"] count];
		if (ntracks > 1) [self addNavButton:@"prev" destination:ntracks-2];
		[self reloadImages:[menu objectForKey:@"tracks"]];
    }
	
	if (![[aMovie chapters] count]) {
		NSString *command = [NSString stringWithFormat:@"GoToTime(%d);\nStartPlaying;", numPages * TRACK_DURATION + 1];
		if (play1) {
		play1 = [play1 objectForKey:@"sprites"];
		for (i=0;i<[play1 count];i++)
			if ([[[(NSArray *)play1 objectAtIndex:i] objectForKey:@"Name"] isEqualToString:@"Play Movie"]) {
				[[(NSArray *)play1 objectAtIndex:i] setObject:command forKey:@"Action"];
				break;
			}
		}
		if (play2) {
			play2 = [play2 objectForKey:@"sprites"];
			for (i=0;i<[play2 count];i++)
				if ([[[(NSArray *)play2 objectAtIndex:i] objectForKey:@"Name"] isEqualToString:@"Play Movie"]) {
					[[(NSArray *)play2 objectAtIndex:i] setObject:command forKey:@"Action"];
					break;
				}
		}
	}
	
	if (count) {
		ntracks = [[menu objectForKey:@"tracks"] count];
		if (ntracks) [self addNavButton:@"next" destination:ntracks];
	
	    //if (count) {
		//while ([[menu objectForKey:@"tracks"] count] > 1) [[menu objectForKey:@"tracks"] removeLastObject];
        // scene selection
        i = 0;
        while (YES) {
            if (cancelled) goto bail;
            if (i+6 <= count)
                n = 6;
            else
                n = count % 6;
            //[progText setStringValue:[NSString stringWithFormat:@"Creating Page %d of %d", tally+1, numPages]];
            //[progText performSelectorOnMainThread:@selector(setStringValue:) withObject:[NSString stringWithFormat:@"Creating Page %d of %d", tally+1, numPages] waitUntilDone:NO];
            //[progBar setDoubleValue:(double)(i+1)/(double)count];
            [progBar performSelectorOnMainThread:@selector(setDoubleValueObject:) withObject:[NSNumber numberWithDouble:(double)(i+1)/(double)count] waitUntilDone:NO];
            track = [self addPageWithChapters:ch startingAt:i count:n];
			NSString *name = [self newNameForName:@"Scene Selection" list:[menu objectForKey:@"tracks"]];
			[track setObject:name forKey:@"Name"];
            if (cancelled) goto bail;
            NSColor *color = [bgColor color];
            if (color)
                [track setObject:[NSArchiver archivedDataWithRootObject:color] forKey:@"color"];
            NSString *bgpic = [workingPath stringByAppendingPathComponent:@"bg.pct"];
            if ([fm fileExistsAtPath:bgpic]) {
                NSImage *im = [[NSImage alloc] initWithContentsOfFile:bgpic];
                if (im) {
                    [track setObject:bgpic forKey:@"bgImageSource"];
                    [track setObject:im forKey:@"bgImage"];
                }
            }
            if (cancelled) goto bail;
			ntracks = [[menu objectForKey:@"tracks"] count];
            if (ntracks > 1)
                [self addNavButton:@"prev" destination:(ntracks - 2)];
            if (ntracks < numPages) {
                [self addNavButton:@"next" destination:(ntracks)];
            }
            i += 6;
            if (i >= count) break;
        }
    }
    
    n = [(NSArray *)[menu objectForKey:@"tracks"] count];
    //m = [bgs count];
    id obj, defIm = nil, defSource = nil;
    for (i=0;i<n; i++) {
        track = [[menu objectForKey:@"tracks"] objectAtIndex:i];
        if (i >= [bgs count]) {
            if (defIm)
                [track setObject:defIm forKey:@"bgImage"];
            if (defSource) {
                [track setObject:defSource forKey:@"bgImageSource"];
				[track setObject:defSource forKey:@"Image"];
			}
			if (i) {
				obj = [[[menu objectForKey:@"tracks"] objectAtIndex:i-1] objectForKey:@"opacity"];
				if (obj) [track setObject:obj forKey:@"opacity"];
			}
        } else {
            obj = [bgs objectAtIndex:i];
            if (![obj isKindOfClass:[NSNull class]]) {
                [track setObject:obj forKey:@"bgImage"];
                if (!defIm)
                    defIm = obj;
            }
            obj = [bgss objectAtIndex:i];
            if (![obj isKindOfClass:[NSNull class]]) {
                [track setObject:obj forKey:@"bgImageSource"];
				[track setObject:obj forKey:@"Image"];
                if (!defSource)
                    defSource = obj;
            }
        }
    }
    
bail:
	//[[menu objectForKey:@"tracks"] removeObjectAtIndex:0];
    //[theOutline reloadData];
    [theOutline performSelectorOnMainThread:@selector(reloadData) withObject:nil waitUntilDone:YES];
    //[theImageView setNeedsDisplay:YES];
    [theImageView performSelectorOnMainThread:@selector(setNeedsDisplayObject:) withObject:[NSNumber numberWithBool:YES] waitUntilDone:NO];
    
    //[progWindow orderOut: NULL];
    [progWindow performSelectorOnMainThread:@selector(orderOut:) withObject:self waitUntilDone:NO];
    //[NSApp endSheet: progWindow];
    [NSApp performSelectorOnMainThread:@selector(endSheet:) withObject:progWindow waitUntilDone:NO];
    
    [pool release];
}

- (IBAction)cancel:(id)sender
{
    cancelled = YES;
    [updateTask terminate];
}

- (IBAction)loadTemplate:(id)sender
{
    NSOpenPanel *theOpenPanel = [NSOpenPanel openPanel];
    NSMutableDictionary *item = [theOutline itemAtRow:[theOutline selectedRow]];
    if (!item || ![[item objectForKey:@"isLeaf"] boolValue]) return;
    
    [theOpenPanel setCanChooseFiles:YES];
    [theOpenPanel setCanChooseDirectories:NO];
    [theOpenPanel setResolvesAliases:YES];
    [theOpenPanel setAllowsMultipleSelection:NO];
    if (NSOKButton != [theOpenPanel runModalForDirectory:nil file:nil types:[NSArray arrayWithObject:@"mhsprite"]]) return;
    NSString *path = [[theOpenPanel filenames] objectAtIndex:0];
    [self readSpriteTemplateFromPath:path toSprite:item];
    [theOutline reloadData];
    [theImageView setNeedsDisplay:YES];
    [self updateWidgets:nil];
    [self syncWidgets:item];
}

- (IBAction)saveTemplate:(id)sender
{
    NSSavePanel *sp = [NSSavePanel savePanel];
    int runResult = [sp runModalForDirectory:nil file:@"Sprite Template.mhsprite"];
    if (runResult != NSOKButton) return;
    NSString *path = [sp filename];
    if (![path hasSuffix:@".mhsprite"])
        path = [path stringByAppendingString:@".mhsprite"];
    NSDictionary *item = [theOutline itemAtRow:[theOutline selectedRow]];
    if (!item || ![[item objectForKey:@"isLeaf"] boolValue]) return;
    [self writeSpriteTemplate:item toPath:path];
}

- (BOOL)writeSpriteTemplate:(NSDictionary *)sprite toPath:(NSString *)path
{
    BOOL failure = NO;
    NSFileManager *fm = [NSFileManager defaultManager];
    if (![fm createDirectoryAtPath:path attributes:nil]) {
        NSLog(@"failed to create mhsprite dir");
        failure = YES;
    }
    NSArray *anArray = [NSArray arrayWithObjects:@"Image", @"Rollover", @"Click", nil];
    NSMutableDictionary *info = [NSMutableDictionary dictionary];
    NSString *aPath, *newPath;
    int i;
    for (i=0;i<3;i++) {
        aPath = [sprite objectForKey:[anArray objectAtIndex:i]];
        if ([aPath hasPrefix:@"/"]) {
            newPath = [path stringByAppendingPathComponent:[anArray objectAtIndex:i]];
            if (![fm copyPath:aPath toPath:newPath handler:nil]) {
                NSLog(@"faile to copy file:\n%@", aPath);
                failure = YES;
            }
        } else
            [info setObject:aPath forKey:[anArray objectAtIndex:i]];
    }
    anArray = [NSArray arrayWithObjects:@"Action", @"X", @"Y", @"Width", @"Height", @"Name", @"keyColor", nil];
    for (i=0;i<[anArray count];i++) {
        id aValue = [sprite objectForKey:[anArray objectAtIndex:i]];
        if (aValue)
            [info setObject:aValue forKey:[anArray objectAtIndex:i]];
    }
    newPath = [path stringByAppendingPathComponent:@"info.xml"];
    if (![info writeToFile:newPath atomically:YES]) {
        NSLog(@"failed to write info file");
        failure = YES;
    }
    return !failure;
}

- (BOOL)readSpriteTemplateFromPath:(NSString *)path toSprite:(NSMutableDictionary *)aSprite
{
    NSString *aPath = [path stringByAppendingPathComponent:@"info.xml"];
    NSDictionary *aDict = [NSDictionary dictionaryWithContentsOfFile:aPath];
    if (!aDict) return NO;
    
    NSArray *anArray = [NSArray arrayWithObjects:@"Image", @"Rollover", @"Click", nil];
   // NSArray *imageArray = [NSArray arrayWithObjects:@"image", @"rimage", @"cimage", nil];
    int i;
    for (i=0;i<3;i++) {
        NSString *newPath = [path stringByAppendingPathComponent:[anArray objectAtIndex:i]];
		[aSprite setObject:newPath forKey:[anArray objectAtIndex:i]];
        
    }
    anArray = [aDict allKeys];
    for (i=0;i<[anArray count];i++) {
        id value = [aDict objectForKey:[anArray objectAtIndex:i]];
        if (value)
            [aSprite setObject:value forKey:[anArray objectAtIndex:i]];
    }
	[self reloadImages:[NSArray arrayWithObject:[aSprite objectForKey:@"parent"]]];
    return YES;
}

- (NSMutableDictionary *)addPageWithChapters:(MHChapters *)ch startingAt:(int)index count:(int)count
{
    int x = 0, y = 0;
    float scalar = [wizSize floatValue];
    if (!scalar) scalar = 1;
    int width = naturalSize.width / 3 - 20;
    int height = naturalSize.height * width / naturalSize.width;
    width = (float) width * scalar;
    height = (float) height * scalar;
    MHMovie *aMovie = (MHMovie *)[spriteMovieView movie];
    Movie mov = [aMovie QTMovie];
    if (!mov) return nil;
    [self newPage:nil];
    int i, j, k = 0;
    NSImage *image;
    NSData *data;
    //id helper = [MHUtil imageHelper];
	int rand = random() % 100;
	//NSLog(@"%d", rand);
    TimeValue tv = GetMovieDuration(mov) / 100 * rand, tv2, originalTV;
    TimeRecord tr;
    originalTV = GetMovieTime(mov, &tr);
    NSMutableDictionary *s;
    NSMutableDictionary *menu = [aMovie menu];
    if (!menu)
        menu = [self newMenu];
    NSMutableDictionary *track = [[menu objectForKey:@"tracks"] lastObject];
	[track setObject:[NSString stringWithFormat:@"%ld", (long)tv] forKey:@"Image"];
	[track setObject:[NSString stringWithFormat:@"%ld", (long)tv] forKey:@"bgImageSource"];
	image = [self imageForSprite:track key:@"Image"];
	if (image) [track setObject:image forKey:@"bgImage"];
	[track setObject:[NSNumber numberWithFloat:0.3] forKey:@"opacity"];
    OSType theType = VideoMediaType;
    short flags = nextTimeSyncSample;
    for (i = index; i < index + count; i++) {
        //[progText setStringValue:[NSString stringWithFormat:@"Chapter %d of %d", i, [ch numberOfRowsInTableView:nil]]];
        if (cancelled) break;
        [progText performSelectorOnMainThread:@selector(setStringValue:) withObject:[NSString stringWithFormat:@"Chapter %d of %d", i, [ch numberOfRowsInTableView:nil]] waitUntilDone:NO];
        //[progBar setDoubleValue:(double)(i+1)*100/(double)[ch numberOfRowsInTableView:nil]];
        [progBar performSelectorOnMainThread:@selector(setDoubleValueObject:) withObject:[NSNumber numberWithDouble:(double)(i+1)/(double)[ch numberOfRowsInTableView:nil]] waitUntilDone:NO];
        tv = [ch timeForChapterAtIndex:i] * (double)GetMovieTimeScale(mov);
        for (j=0;j<20;j++) {
            GetMovieNextInterestingTime(mov, flags, 1, &theType, tv, fixed1, &tv2, nil);
            if (tv2 > 0) tv = tv2;
			else break;
        }
        if (i== index + count - 1)
            tv2 = GetMovieDuration(mov);
        else
            tv2 = [ch timeForChapterAtIndex:i+1] * (double)GetMovieTimeScale(mov);
        tv2 = tv + (tv2 - tv) / 2;
        s = [[NSMutableDictionary alloc] init];
        [[track objectForKey:@"sprites"] addObject:s];
        //tr.value.lo = tv;
        //SetMovieTime(mov, &tr);
        //MoviesTask(mov, 0);
        [s setObject:[NSString stringWithFormat:@"Sprite %d", ++k] forKey:@"Name"];
        [s setObject:[NSString stringWithFormat:@"%ld", tv] forKey:@"Image"];
        [s setObject:[NSString stringWithFormat:@"%ld", tv2] forKey:@"Rollover"];
        [s setObject:[NSNumber numberWithBool:YES] forKey:@"isLeaf"];
        [s setObject:track forKey:@"parent"];
        /*		data = [helper getImage:tv sender:@"mhSprite"];
        if (data) {
            image = [NSUnarchiver unarchiveObjectWithData:data];
            [s setObject:image forKey:@"image"];
        }*/
		
		NSMutableDictionary *menu = [[spriteMovieView movie] menu];
		if (!menu)
			menu = [self newMenu];
		QTImage *qi = [menu objectForKey:@"qi"];
		if (!qi) {
			qi = [[[QTImage alloc] init] autorelease];
			[menu setObject:qi forKey:@"qi"];
		}
        
		switch (count) {
            case 1:
                x = (naturalSize.width - width) / 2;
                break;
            case 2: case 4:
                if (i % 2)
                    x = naturalSize.width * 3 / 4 - width / 2;
                else
                    x = naturalSize.width / 4 - width / 2;
                break;
            case 3: case 5: case 6:
                if (i % 3 == 0)
                    x = naturalSize.width / 6 - width / 2 + 7;
                else if (i % 3 == 1)
                    x = naturalSize.width / 2 - width / 2;
                else if (i % 3 == 2)
                    x = naturalSize.width * 5 / 6 - width / 2 - 5;
                break;
            default: break;
        }
        if (count < 4)
            y = (naturalSize.height - height) / 3;
        else {
            if (i - index < (count + 1) / 2 || (1 - index == 2 && count == 5))
                y = height / 4 - 5;
            else
                y = naturalSize.height - height * 1.5 - 5;
        }
        [s setObject:[NSString stringWithFormat:@"%d", x] forKey:@"X"];
        [s setObject:[NSString stringWithFormat:@"%d", y] forKey:@"Y"];
        [s setObject:[NSString stringWithFormat:@"%d", width] forKey:@"Width"];
        [s setObject:[NSString stringWithFormat:@"%d", height] forKey:@"Height"];
        NSString *name = [[ch chapterNames] objectAtIndex:i];
        if (!name) name = @"";
        [s setObject:[NSString stringWithFormat:@"GoToChapter(%@);\nStartPlaying;", name] forKey:@"Action"];
		
		if (cancelled) break;
		image = [self imageForSprite:s key:@"Image"];
		//image = [[QTImage alloc] initWithFile:[[[spriteMovieView movie] URL] path] transparency:nil time:tv];
		if (image)
			[s setObject:image forKey:@"image"];
		if (cancelled) break;
		image = [self imageForSprite:s key:@"Rollover"];
		if (image)
			[s setObject:image forKey:@"rimage"];
		
		
		
		// text version
		s = [[NSMutableDictionary alloc] init];
        [[track objectForKey:@"sprites"] addObject:s];
        [s setObject:[NSString stringWithFormat:@"Sprite %d", ++k] forKey:@"Name"];
        [s setObject:[NSNumber numberWithBool:YES] forKey:@"isLeaf"];
        [s setObject:track forKey:@"parent"];
		[s setObject:[NSString stringWithFormat:@"%d", x] forKey:@"X"];
        [s setObject:[NSString stringWithFormat:@"%d", y+height] forKey:@"Y"];
        [s setObject:[NSString stringWithFormat:@"%d", width] forKey:@"Width"];
        [s setObject:[NSString stringWithFormat:@"%d", height/4] forKey:@"Height"];
		[s setObject:[NSString stringWithFormat:@"\"%@\"", name] forKey:@"Image"];
        [s setObject:[NSString stringWithFormat:@"GoToChapter(%@);\nStartPlaying;", name] forKey:@"Action"];
		if (cancelled) break;
		image = [self imageForSprite:s key:@"Image"];
		if (image)
			[s setObject:image forKey:@"image"];
		
		[theImageView performSelectorOnMainThread:@selector(setNeedsDisplayObject:) withObject:[NSNumber numberWithBool:YES] waitUntilDone:NO];
    }
    tr.value.lo = originalTV;
    SetMovieTime(mov, &tr);
    return track;
}

- (void)addNavButton:(NSString *)nav destination:(TimeValue)dest
{
    NSMutableDictionary *menu = [[spriteMovieView movie] menu];
    if (!menu)
        menu = [self newMenu];
    
    NSMutableDictionary *aSprite = [NSMutableDictionary dictionary];
    NSMutableArray *tracks = [menu objectForKey:@"tracks"];
    if (![tracks count])
        [self newPage:nil];
    NSMutableDictionary *track = [tracks lastObject];
    [[track objectForKey:@"sprites"] addObject:aSprite];
    int x = (int)naturalSize.width - 24;
    if ([nav isEqualToString:@"prev"]) x -= 30;
    int y = (int)naturalSize.height - 20;
    [aSprite setObject:[NSString stringWithFormat:@"%d",x] forKey:@"X"];
    [aSprite setObject:[NSString stringWithFormat:@"%d",y] forKey:@"Y"];
    [aSprite setObject:@"22" forKey:@"Width"];
    [aSprite setObject:@"17" forKey:@"Height"];
    [aSprite setObject:[NSNumber numberWithBool:YES] forKey:@"isLeaf"];
    [aSprite setObject:track forKey:@"parent"];
    NSString *path = [[NSBundle mainBundle] pathForResource:nav ofType:@"png"];
    [aSprite setObject:path forKey:@"Image"];
	[aSprite setObject:[NSArchiver archivedDataWithRootObject:[NSColor colorWithCalibratedRed:1.0 green:1.0 blue:1.0 alpha:1.0]] forKey:@"keyColor"];
    NSImage *image = [self imageForSprite:aSprite key:@"Image"];
	if (image)
		[aSprite setObject:image forKey:@"image"];
    [aSprite setObject:nav forKey:@"Name"];
    [aSprite setObject:[NSString stringWithFormat:@"GoToPage(%ld);", dest] forKey:@"Action"];
}

- (NSMutableArray *)getTracksOfType:(OSType)type
{
	MHMovie *m = [spriteMovieView movie];
    Movie mov = [m QTMovie];
    Track track = nil;
    long index = 1;
    int size;
    Handle h = NewHandleClear(0);
    char *buffer;
    NSString *aTrackName, *typeName = @"";
    switch (type) {
        case SoundMediaType: typeName = @"Sound Track "; break;
        case TextMediaType: typeName = @"Text Track "; break;
        case VideoMediaType: typeName = @"Video Track "; break;
        case SpriteMediaType: typeName = @"Sprite Track "; break;
        case MPEGMediaType: typeName = @"MPEG Track "; break;
        default: typeName = @"";
    }
	
	// find the chapter track
	Track chap = nil;
	MHTrack *t = nil;
	for (index = 0; index < [m trackCount]; index++) {
		t = [m trackAtIndex:index];
		t = [t chapterTrack];
		if (t) {
			chap = [t QTTrack];
			break;
		}
	}
	index = 1;
	
    NSMutableArray *array = [[NSMutableArray alloc] init];
    if (!type) track = GetMovieIndTrack(mov, index++);
    else track = GetMovieIndTrackType(mov, index++, type, movieTrackMediaType);
    while (track) {
		if (track != chap) { // don't count the chapter track
			GetUserData(GetTrackUserData(track), h, kUserDataName, 1);
			size = GetHandleSize(h);
			if (size) {
				buffer = (char *)*h;
				aTrackName = [NSString stringWithCString:buffer length:size];
				if (aTrackName)
					[array addObject:aTrackName];
			} else {
				aTrackName = [NSString stringWithFormat:@"%@%d", typeName, [array count] + 1];
				buffer = malloc(sizeof(char) * ([aTrackName length] + 1));
				[aTrackName getCString:buffer];
				SetUserDataItem(GetTrackUserData(track), buffer, [aTrackName length], kUserDataName, 1);
				[array addObject:aTrackName];
				free(buffer);
			}
		}
		if (!type) track = GetMovieIndTrack(mov, index++);
		else track = GetMovieIndTrackType(mov, index++, type, movieTrackMediaType);
    }
	if (type == TextMediaType) { // check for perian subtitles
		index = 1;
		NSArray *subs = [NSArray arrayWithObjects:@"SRT ", @"SSA ", @"ASS ", @"SPU ", nil];
		int i, n = [m trackCount];
		for (i=0;i<n;i++) {
			t = [m trackAtIndex:i];
			NSString *fourCC = [t fourCC];
			if ([subs containsObject:fourCC]) {
				aTrackName = [t name];
				if (!aTrackName) {
					aTrackName = [NSString stringWithFormat:@"%@ Track %d", fourCC, index++];
					[t setName:aTrackName];
				}
				[array addObject:aTrackName];
			}
		}
	}
    DisposeHandle(h);
    return array;
}

- (void)awakeFromNib
{
    //NSLog(@"MHSprite");
    updateTask = nil;
    loadLock = [[NSLock alloc] init];
	imageLock = [[NSLock alloc] init];
    mhSpriteNibInstance = self;
    isAwake = YES;
    [actionValue setTarget:self];
    [(NSPopUpButton *)actionValue setAction:@selector(selectActionValue:)];
    [self restorePrefs];
    //[theOutline restoreFromAutosave];
	[theOutline setAllowsEmptySelection:NO];
	[theOutline setAllowsMultipleSelection:YES];    
    NSConnection *theConnection = [NSConnection defaultConnection];
    [theConnection setRootObject:self];
    if ([theConnection registerName:@"mhSprite"] == NO) {
      NSLog(@"Failed to register name mhSprite\n");
    }
    [theConnection retain];
    [theConnection setDelegate:self];
    
    NSToolbar *toolbar = [[NSToolbar alloc] initWithIdentifier:@"spriteToolbar"];
    [toolbar setDelegate:self];
    [toolbar setAllowsUserCustomization:NO];
    [toolbar setAutosavesConfiguration:YES];
    [spriteWindow setToolbar:toolbar];
    
    [qScriptBox setToolTip:@"Supported Actions (CaSe InSeNsItIvE):\n\nGoToChapter(chapter name);\nGoToTime(movie time);\nGoToPage(page number);\nStartPlaying;\nSetRate(rate);\nStepForward;\nStepBackward;\nGoToBeginning;\nGoToEnd;\nSaveTime;\nGoToSavedTime;\n\nTrackNamed(track name).SetEnabled(yes, no, true, false); TrackAtIndex(index number).SetEnabled(yes, no, true, false);\n\nSetMovieLanguage(language);\nSetVolume(-256/256);\n\nPreceed with OnFrameLoad. OnMouseEnter. OnMouseExit. or OnIdle. if desired"];
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(flagsChanged:) name:@"flagsChangedNotification" object:nil];
}

- (NSToolbarItem *)toolbar:(NSToolbar *)tb itemForItemIdentifier:(NSString *)tbit willBeInsertedIntoToolbar:(BOOL)flag
{
    NSToolbarItem *tbi;
    NSArray *idents = [NSArray arrayWithObjects:@"New Sprite", @"New Page", @"Delete", @"Duplicate", @"Inspector", @"Options", @"Wizard", @"Update Movie", nil];
    void *actions[] = { @selector(newSprite:) , @selector(newPage:) , @selector(deletePage:) , @selector(duplicate:), @selector(makeKeyAndOrderFront:), @selector(more:), @selector(wizard:), @selector(updateMovie:) };
    
    int i, n = [idents count];
    for (i=0;i<n;i++) {
        if ([tbit isEqualToString:[idents objectAtIndex:i]]) {
            tbi = [[NSToolbarItem alloc] initWithItemIdentifier:tbit];
            [tbi setAction:actions[i]];
            [tbi setImage:[NSImage imageNamed:tbit]];
            if (i == 4) { // Inspector
                [tbi setTarget:inspectorWindow];
            } else {
                [tbi setTarget:self];
            }
            [tbi setLabel:tbit];
            [tbi setPaletteLabel:tbit];
            [tbi autorelease];
            return tbi;
        }
    }
    return nil;
}

- (NSArray *)toolbarDefaultItemIdentifiers:(NSToolbar *)toolbar
{
    return [NSArray arrayWithObjects:@"New Sprite", @"New Page", @"Duplicate", @"Delete", @"Inspector", @"Wizard", @"Update Movie", @"Options", nil];
}

- (NSArray *)toolbarAllowedItemIdentifiers:(NSToolbar *)toolbar
{
    return [NSArray arrayWithObjects:@"New Sprite", @"New Page", @"Delete", @"Duplicate", @"Inspector", @"Options", @"Wizard", @"Update Movie", NSToolbarSeparatorItemIdentifier, NSToolbarSpaceItemIdentifier, NSToolbarFlexibleSpaceItemIdentifier, NSToolbarCustomizeToolbarItemIdentifier, nil];
}

- (IBAction)customizeToolbar:(id)sender
{
    if (![[spriteWindow toolbar] customizationPaletteIsRunning])
        [[spriteWindow toolbar] runCustomizationPalette:nil];
}


- (void)open
{
    [spriteMovieView stop:nil];
    [spriteWindow setFrame:[self rightSize] display:YES];
    NSColor *color;
    NSData *aData = [[NSUserDefaults standardUserDefaults] objectForKey:@"spriteBGColor"];
    if (!aData) color = [NSColor whiteColor];
    else color = [NSUnarchiver unarchiveObjectWithData:aData];
    [bgColor setColor:color];
    [self setColor:color];
    MHMovie *aMovie = (MHMovie *)[spriteMovieView movie];
    NSMutableDictionary *menu = [aMovie menu];
    if (!menu)
        menu = [self newMenu];
    NSMutableArray *tracks = [menu objectForKey:@"tracks"];
    if (![tracks count])
        [self newPage:nil];
    tracks = [menu objectForKey:@"tracks"];
    [NSThread detachNewThreadSelector:@selector(reloadImages:) toTarget:self withObject:tracks];
    [spriteWindow makeKeyAndOrderFront:nil];
}

- (void)reloadImages:(NSArray *)tracks
{
	//NSLog(@"reloadImages");
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
    NSArray *values = [NSArray arrayWithObjects:@"Image", @"Rollover", @"Click", nil];
    NSArray *keys = [NSArray arrayWithObjects:@"image", @"rimage", @"cimage", nil];
    NSImage *anImage;
    int height = 0;
    BOOL hasImage = NO;
	cancelled = NO;
    int i, j, k;
    for (i=0;i<[tracks count];i++) {
		if (cancelled) break;
        [loadLock lock];
        NSArray *sprites = [[tracks objectAtIndex:i] objectForKey:@"sprites"];
        NSString *val = [[tracks objectAtIndex:i] objectForKey:@"Image"];
        anImage = nil;
        hasImage = ([[tracks objectAtIndex:i] objectForKey:@"bgImage"] != nil);
        [loadLock unlock];
        if (val && !hasImage) {
            if ([val hasPrefix:@"/"])
                anImage = [[NSImage alloc] initWithContentsOfFile:val];
            else
                anImage = [self imageForSprite:[tracks objectAtIndex:i] key:@"Image"];
				//anImage = [self imageFromFrame:[val intValue] height:[spriteMovieView sizeForMagnification:1.0].height - 16];
            [loadLock lock];
            if (anImage && [tracks count] > i)
                [(NSMutableDictionary *)[tracks objectAtIndex:i] setObject:anImage forKey:@"bgImage"];
            [loadLock unlock];
        }
        for (j=0;j<[sprites count];j++) {
			if (cancelled) break;
            [loadLock lock];
            NSMutableDictionary *sprite = [sprites objectAtIndex:j];
            height = [[sprite objectForKey:@"Height"] intValue];
            [loadLock unlock];
            for (k=0;k<3;k++) {
                anImage = nil;
                [loadLock lock];
                val = [sprite objectForKey:[values objectAtIndex:k]];
                BOOL hasImage = ([sprite objectForKey:[keys objectAtIndex:k]] != nil);
                [loadLock unlock];
                if (val && !hasImage) {
                    anImage = [self imageForSprite:sprite key:[values objectAtIndex:k]];
                    [loadLock lock];
                    if (anImage)
                        [sprite setObject:anImage forKey:[keys objectAtIndex:k]];
                    [loadLock unlock];
                }
				[theImageView performSelectorOnMainThread:@selector(setNeedsDisplayObject:) withObject:[NSNumber numberWithBool:YES] waitUntilDone:NO];
            }
        }
        //[theImageView setNeedsDisplay:YES];
    }
    [pool release];
}

- (void)setMovie:(NSMovie *)aMovie
{
    [spriteMovieView setMovie:aMovie];
    [theOutline reloadData];
    [theImageView setNeedsDisplay:YES];
}

- (NSRect)rightSize
{
    MHMovie *aMovie = (MHMovie *)[spriteMovieView movie];
    NSRect frame = [spriteWindow frame];
    if (!aMovie) return frame;
    Movie myMovie = [aMovie QTMovie];
    Track videoTrack = GetMovieIndTrackType(myMovie, 1, VideoMediaType, movieTrackMediaType);
    Fixed w, h;
    GetTrackDimensions(videoTrack, &w, &h);
    Rect box;
    box.top = box.left = 0;
    box.right = Fix2Long(w);
    box.bottom = Fix2Long(h);
    MatrixRecord m;
    GetTrackMatrix(videoTrack, &m);
    TransformRect(&m, &box, NULL);
    naturalSize.height = box.bottom - box.top;
    naturalSize.width = box.right - box.left;
    frame.size.height = naturalSize.height + 213 + (frame.size.height - [[spriteWindow contentView] frame].size.height);
    frame.size.width = naturalSize.width + 40;
    NSRect sFrame = [[spriteWindow screen] frame];
    if (frame.size.height > sFrame.size.height - 20) frame.size.height = sFrame.size.height - 20;
    if (frame.size.width > sFrame.size.width - 20) frame.size.width = sFrame.size.width - 20;
    //[spriteWindow setFrame:frame display:YES];
    return frame;
}

// OutlineView Use
- (id)outlineView:(NSOutlineView *)outlineView child:(int)index ofItem:(id)item
{
	[loadLock lock];
	id i = nil;
    NSMutableDictionary *menu = [[spriteMovieView movie] menu];
    if (!menu)
        menu = [self newMenu];
    if (!item) {
        if ([(NSArray *)[menu objectForKey:@"tracks"] count] > index)
            i = [[menu objectForKey:@"tracks"] objectAtIndex:index];
		else
			NSLog(@"found a nil");
    } else if ([(NSArray *)[item objectForKey:@"sprites"] count] > index)
        i = [[item objectForKey:@"sprites"] objectAtIndex:index];
	else NSLog(@"found a nil");
    [loadLock unlock];
	return i;
}

- (BOOL)outlineView:(NSOutlineView *)outlineView isItemExpandable:(id)item 	{	return ![[item objectForKey:@"isLeaf"] boolValue];	}

- (int)outlineView:(NSOutlineView *)outlineView numberOfChildrenOfItem:(id)item
{
    NSMutableDictionary *menu = [[spriteMovieView movie] menu];
    if (!menu)
        menu = [self newMenu];
    if (!item)
        return [(NSArray *)[menu objectForKey:@"tracks"] count];
    int i = [(NSArray *)[item objectForKey:@"sprites"] count];
    return i;
}

- (id)outlineView:(NSOutlineView *)outlineView objectValueForTableColumn:(NSTableColumn *)tableColumn byItem:(id)item
{
    NSString *idString = [tableColumn identifier];
    NSString *s = [item objectForKey:idString];
    return s;
}

- (void)outlineViewItemWillCollapse:(NSNotification *)notification
{
	NSDictionary *item = [theOutline itemAtRow:[theOutline selectedRow]];
	if ([[item objectForKey:@"isLeaf"] boolValue]) item = [item objectForKey:@"parent"];
	[theOutline performSelector:@selector(selectItem:) withObject:item afterDelay:0];
}
/*
- (BOOL)outlineView:(NSOutlineView *)outlineView shouldEditTableColumn:(NSTableColumn *)tableColumn item:(id)item
{
    //NSString *idString = [tableColumn identifier];
    //if ([idString isEqualToString:@"Name"]) return NO;
    //if ([item isKindOfClass:[oSprite class]] && [item added]) return NO;
    //if ([item shouldEditKey:idString]) return YES;
    //return NO;
    return YES;
}
*/
- (void)outlineViewSelectionDidChange:(NSNotification *)notification
{
    [theImageView setNeedsDisplay:YES];
    [self updateWidgets:nil];
    [self syncWidgets:(NSMutableDictionary *)[theOutline selectedItem]];
}

//- (BOOL)outlineView:(NSOutlineView *)outlineView shouldSelectItem:(id)item;
//- (void)selectItem:(id)item;
//- (BOOL)outlineView:(NSOutlineView *)outlineView shouldExpandItem:(id)item;
//- (void)expandItem:(id)item;
- (void)outlineView:(NSOutlineView *)outlineView setObjectValue:(id)object forTableColumn:(NSTableColumn *)tableColumn byItem:(id)item
{
    //NSArray *array = [NSArray arrayWithObjects:@"X", @"Y", @"Width", @"Height", nil];
    NSString *idString = [tableColumn identifier];
    NSArray *keys = [NSArray arrayWithObjects:@"Image", @"Rollover", @"Click", nil];
    NSArray *keys2 = [NSArray arrayWithObjects:@"image", @"rimage", @"cimage", nil];
    int index = [keys indexOfObject:idString];
    //int i, n = [array count];
    //for (i=0;i<n;i++) {
    //    if ([idString isEqualToString:[array objectAtIndex:i]])
    //        object = [NSNumber numberWithInt:[object intValue]];
    //}
    if ([object isKindOfClass:[NSString class]] && [object isEqualToString:@""]) {
        [item removeObjectForKey:idString];
        if (index != NSNotFound) [item removeObjectForKey:[keys2 objectAtIndex:index]];
		if (![[item objectForKey:@"isLeaf"] boolValue] && index == 0) {
			[item removeObjectForKey:@"bgImageSource"];
			[item removeObjectForKey:@"bgImage"];
		}
    } else {
        [(NSMutableDictionary *)item setObject:object forKey:idString];
        if (index != NSNotFound) {
            NSImage *image = [self imageForSprite:item key:idString];
			NSString *ikey = [keys2 objectAtIndex:index];
			if (![[item objectForKey:@"isLeaf"] boolValue]) {
				ikey = @"bgImage";
				[item setObject:object forKey:@"bgImageSource"];
			}
            if (image)
                [(NSMutableDictionary *)item setObject:image forKey:ikey];
            else
                [(NSMutableDictionary *)item removeObjectForKey:ikey];
        }
    }
    [theImageView setNeedsDisplay:YES];
}

- (void)outlineView:(NSOutlineView *)oview rightClickWithEvent:(NSEvent *)event
{
    NSPoint point = [event locationInWindow];
    point = [oview convertPoint:point fromView:nil];
    int row = [oview rowAtPoint:point];
    NSMutableDictionary *item = [oview itemAtRow:row];
    if (!item) return;
    [self selectItem:item byExtendingSelection:YES];
    NSMenu *menu = [self rightClickMenu];
    if (menu)
        [NSMenu popUpContextMenu:menu withEvent:event forView:oview];
}
/*
- (void)outlineView:(NSOutlineView *)outlineView willDisplayCell:(id)cell forTableColumn:(NSTableColumn *)tableColumn item:(id)item
{
    if ([item isKindOfClass:[oSprite class]] && [item added])
        [cell setTextColor:[NSColor grayColor]];
    else
        [cell setTextColor:[NSColor blackColor]];
}
*/
@end
