/* Copyright (c) 2004, Noah Sorscher
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 	• 	Redistributions of source code must retain the above copyright
 * 	notice, this list of conditions and the following disclaimer.
 * 	• 	Redistributions in binary form must reproduce the above copyright
 * 	notice, this list of conditions and the following disclaimer in the
 * 	documentation and/or other materials provided with the distribution.
 * 	• 	Neither the name of the Noah Sorscher nor the names of its
 * 	contributors may be used to endorse or promote products derived from
 * 	this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
//
//  MHWeb.h
//  Hootenanny
//
//  Created by Noah Sorscher on Thu Apr 03 2003.
//  Copyright (c) 2003 __MyCompanyName__. All rights reserved.
//

#import <Foundation/Foundation.h>
#import <Cocoa/Cocoa.h>
//#import "MHBB.h"
//#import "MHIMDB.h"
//#import "MHEPGTVT.h"


@interface MHWeb : NSObject {
    IBOutlet id bbText,
                epgText, epgtvt, optionsButton, okayButton,
                picField,
                resultTable;
    IBOutlet id resultText;
    IBOutlet id searchField, sourceText, sourceLink;
    IBOutlet id theOutline,
                webTab,
                webWindow;
                
    NSLock *expandingLock;
    id bb;
    id imdb;
    id epg;
    NSString *myTitle;

}

// client code need not worry about connecting UI elements.  all you need to do is this:
// [MHWeb addWebSelectionListener:self];
// [MHWeb go];
// then it pops up the window, and when/if the user makes a selection, MHWeb sends you back the method -webSelectionMade:(id)metadata, which is explained below.

// Public:

// Public Use
+ (void) go;
+ (void) setTitle:(NSString *)aTitle;
+ (void) addWebSelectionListener:(id)listener;  // listeners must implement the method -webSelectionMade:(id)metadata; in the case of an IMDB or EPG choice, "metadata" will be an NSDictionary.  In the case of a BlockBuster choice, "metadata" will be an instance of MHWeb, on which you can call the following functions:
//- (MHChapters *)chapters;
//- (MHChapters *)chaptersScaledToMovie:(NSMovie *)aMovie;

// Optional
//+ (void)setDownloadsImages:(BOOL)flag;  // default is YES
+ (void)addOptionsButtonListener:(id)listener;  // options button listeners must implement the method -webOptionsButtonPressed;
+ (void)setSearchText:(NSString *)input;  // this must be called after calling +go

// Internal:

// Interface Use
- (void) awakeFromNib;
- (IBAction)search:(id)sender;
- (IBAction)bbSearchAgain:(id)sender;
- (IBAction)flush:(id)sender;
- (IBAction)okay:(id)sender;
- (IBAction)options:(id)sender;
//- (IBAction)source:(id)sender;

// OutlineView Use
- (id)outlineView:(NSOutlineView *)outlineView child:(int)index ofItem:(id)item;
- (BOOL)outlineView:(NSOutlineView *)outlineView isItemExpandable:(id)item;
- (int)outlineView:(NSOutlineView *)outlineView numberOfChildrenOfItem:(id)item;
- (id)outlineView:(NSOutlineView *)outlineView objectValueForTableColumn:(NSTableColumn *)tableColumn byItem:(id)item;
- (BOOL)outlineView:(NSOutlineView *)outlineView shouldEditTableColumn:(NSTableColumn *)tableColumn item:(id)item;
- (BOOL)outlineView:(NSOutlineView *)outlineView shouldSelectItem:(id)item;
- (void)selectItem:(id)item;
- (BOOL)outlineView:(NSOutlineView *)outlineView shouldExpandItem:(id)item;
- (void)expandItem:(id)item;

// Internal Use
- (IBAction)followLink:(id)sender;
- (void)pageLoaded:(NSNotification *)note;
- (void)setString:(NSString *)input;
- (void)setOptionsButton;
- (void)open;

@end
