//
//  QTImage.m
//  Hootenanny
//
//  Created by Noah Sorscher on 1/21/05.
//  Copyright 2005 __MyCompanyName__. All rights reserved.
//

#import "QTImage.h"
#import <QuickTime/QuickTime.h>
#import <QTKit/QTKit.h>


@implementation QTImage

- (QTImage *)init
{
	myMovies = [[NSMutableArray alloc] init];
	return [super init];
}

- (void)dealloc
{
	[myMovies release];
	[super dealloc];
}

- (QTMovie *)movieForPath:(NSString *)path
{
	int i, n = [myMovies count];
	NSDictionary *d = nil;
	for (i=0;i<n;i++) {
		d = [myMovies objectAtIndex:i];
		if ([[d objectForKey:@"path"] isEqualToString:path])
			return [d objectForKey:@"movie"];
	}
	QTMovie *q = [QTMovie movieWithFile:path error:nil];
	if (q) {
		// QTMovie only images one track (the forwardmost layer?), so we'll take out all but one.  We are guessing that the biggest track is the most interesting.
		n = GetMovieTrackCount([q quickTimeMovie]);
		wide size, biggestSize;
		size.lo = size.hi = biggestSize.lo = biggestSize.hi = 0;
		int biggestTrack = 0;
		for (i=1;i<=n;i++) {
			Track t = GetMovieIndTrack([q quickTimeMovie], i);
			GetTrackDataSize64(t, 0, GetTrackDuration(t), &size);
			if (WideCompare(&size, &biggestSize) > 0) {
				biggestTrack = i;
				biggestSize = size;
			}
		}
		if (biggestTrack) {
			for (i=n;i>0;i--)
				if (i != biggestTrack)
					SetTrackEnabled(GetMovieIndTrack([q quickTimeMovie], i), NO);
		}
		
		//d = [NSDictionary dictionaryWithObjectsAndKeys:q, @"movie", path, @"path", nil];
		//[myMovies addObject:d];
	}
	return q;
}

- (NSImage *)initWithFile:(NSString *)path transparency:(NSColor *)keyColor time:(TimeValue)time
{
	
	return [self imageForPath:path transparency:keyColor time:time];
}	
	
- (NSImage *)imageForPath:(NSString *)path transparency:(NSColor *)keyColor time:(TimeValue)time
{
	return [self imageForPath:path transparency:keyColor time:time size:NSMakeSize(0,0)];
}

- (NSImage *)imageForPath:(NSString *)path transparency:(NSColor *)keyColor time:(TimeValue)time size:(NSSize)size
{
	
    if (!path) return nil;
    //NSMovie *movie = [[NSMovie alloc] initWithURL:[NSURL fileURLWithPath:path] byReference:NO];
    //if (!movie) return nil;
	
	QTMovie *m = [self movieForPath:path];
	if (!m) return nil;
	
	QTTime t = [m currentTime];
	if (time < 0) {
		t = [m duration];
		t.timeValue /= 2;
	} else
		t.timeValue = time;
	
	if (size.width > 1 && size.height > 1) {
		[m setMovieAttributes:[NSDictionary dictionaryWithObject:[NSValue valueWithSize:size] forKey:QTMovieCurrentSizeAttribute]];
	}
	
	NSImage *i = [m frameImageAtTime:t];
	
	if (!keyColor) return i;
	return [self transparentImage:i keyColor:keyColor];
	
}

- (NSImage *)transparentImage:(NSImage *)i keyColor:(NSColor *)keyColor
{
		keyColor = [keyColor colorUsingColorSpaceName:NSCalibratedRGBColorSpace];
		unsigned char red = 255 * [keyColor redComponent];
		unsigned char green = 255 * [keyColor greenComponent];
		unsigned char blue = 255 * [keyColor blueComponent];
		
		int row, column, widthInPixels = [i size].width, heightInPixels = [i size].height;
		NSBitmapImageRep *rep = [[NSBitmapImageRep alloc] initWithBitmapDataPlanes:NULL pixelsWide:widthInPixels pixelsHigh:heightInPixels bitsPerSample:8 samplesPerPixel:4 hasAlpha:YES isPlanar:NO colorSpaceName:NSDeviceRGBColorSpace bytesPerRow:4*[i size].width bitsPerPixel:0];
		
		//unsigned char *bitmapData = [rep bitmapData];
		
		
		if (!rep) return i;
		NSBitmapImageRep *irep = nil;
		NSArray *reps = [i representations];
		int arep, n = [reps count];
		for (arep=0;arep<n;arep++)
			if ([[reps objectAtIndex:arep] isKindOfClass:[NSBitmapImageRep class]])
				irep = [reps objectAtIndex:arep];
		if (!irep) return i;
		
		unsigned char r, g, b, a;		
		NSColor *pixColor = nil;
		unsigned char *theData = [rep bitmapData];
		unsigned char *dst = theData;
		//unsigned char *src = [irep bitmapData];
		for (row = 0; row < heightInPixels; row++) {
			for (column = 0; column < widthInPixels; column++) {
				pixColor = [irep colorAtX:column y:row];
				r = 255 * [pixColor redComponent];
				g = 255 * [pixColor greenComponent];
				b = 255 * [pixColor blueComponent];
				a = 255;
				//NSLog(@"%g, %g, %g", pred, pgreen, pblue);
				if (r > red-6 && r < red+6 && g > green-6 && g < green+6 && b > blue-6 && b < blue+6)
					a = 0;
				
				*dst++ = r;		// set dest. Alpha component
				*dst++ = g;		// set dest. Red component
				*dst++ = b;		// set dest. Green component
				*dst++ = a;		// set dest. Blue component 
				
			}
		}
		
		i = [[[NSImage alloc] initWithSize:NSMakeSize(widthInPixels, heightInPixels)] autorelease];
		if (i)
		{
			[i addRepresentation:rep];
			[rep release];
		}
		
		//i = [[NSImage alloc] initWithData:[rep representationUsingType:NSPNGFileType properties:nil]];
	
	return i;
    
	////////////////////////////////////////////////////////////////////////////////////////////////
/*	
    GWorldPtr gworld;
    Rect srcRect, portRect;
    GetMovieBox([movie QTMovie], &srcRect);
    NSRect r = NSMakeRect(0, 0, srcRect.right-srcRect.left, srcRect.bottom-srcRect.top);
    
    NSWindow *win = [[NSWindow alloc] initWithContentRect:r styleMask:NSBorderlessWindowMask backing:NSBackingStoreBuffered defer:NO];
    r = [[win contentView] frame];
    NSMovieView *myView = [[NSMovieView alloc] initWithFrame:r];
    [myView showController:NO adjustingSize:NO];
    [win setContentView:myView];
    r = [myView frame];
    [myView setMovie:movie];
    
    TimeRecord tr;
    GetMovieTime([movie QTMovie], &tr);
    tr.value.lo = time;
    SetMovieTime([movie QTMovie], &tr);
    
    // http://developer.apple.com/qa/qtmtb/qtmtb57.html
    OSErr err = NewGWorld(&gworld,
                k32ARGBPixelFormat,
                &srcRect,
                NULL,
                NULL,
                0);
        /* make sure to lock the PixMap * /
    LockPixels(GetGWorldPixMap(gworld));
        /* erase * /
    SetGWorld(gworld,NULL);
    GetPortBounds(gworld, &portRect);
    EraseRect(&portRect);
        /* tell the movie to draw into the GWorld * /
    SetMovieGWorld([movie QTMovie], gworld,  GetGWorldDevice(gworld));
    
    MoviesTask([movie QTMovie], 0);
    NSImage *image = [self gworldToNSImage:gworld transparency:keyColor];
    [win release];
    [myView release];
    [movie release];
    return image;
 */
}

/*
-(NSImage *)gworldToNSImage:(GWorldPtr) gWorldPtr transparency:(NSColor *)keyColor
{
    //NSLog(@"gworldToNSImage");
    PixMapHandle 		pixMapHandle = NULL;
    Ptr 				pixBaseAddr = nil;
    NSBitmapImageRep 	*imageRep = nil;
    NSImage 			*image = nil;
    
    float keyR = 0, keyB = 0, keyG = 0, keyA = 0;
    unsigned char KeyR = 0, KeyG = 0, KeyB = 0, KeyA = 0;
    if (keyColor) {
        keyColor = [keyColor colorUsingColorSpaceName:@"NSCalibratedRGBColorSpace"];
        [keyColor getRed:&keyR green:&keyG blue:&keyB alpha:&keyA];
        KeyR = 255.0 * keyR;
        KeyG = 255.0 * keyG;
        KeyB = 255.0 * keyB;
        KeyA = 255.0 * keyA;
    }
    
    //NSAssert(gWorldPtr != nil, @"nil gWorldPtr");
    if (nil == gWorldPtr) return nil;

    // Lock the pixels
    pixMapHandle = GetGWorldPixMap(gWorldPtr);
    if (pixMapHandle)
    {
        Rect 		portRect;
        unsigned 	portWidth, portHeight;
        int 		bitsPerSample, samplesPerPixel;
        BOOL 		hasAlpha, isPlanar;
        int 		destRowBytes;

        NSAssert(LockPixels(pixMapHandle) != false, @"LockPixels returns false");
    
        GetPortBounds(gWorldPtr, &portRect);
        portWidth = (portRect.right - portRect.left);
        portHeight = (portRect.bottom - portRect.top);
    
        bitsPerSample 	= 8;
        samplesPerPixel = 4;
        hasAlpha		= YES;
        isPlanar 		= NO;
        destRowBytes 	= portWidth * samplesPerPixel;
        imageRep 		= [[NSBitmapImageRep alloc] initWithBitmapDataPlanes:NULL pixelsWide:portWidth pixelsHigh:portHeight bitsPerSample:bitsPerSample samplesPerPixel:samplesPerPixel hasAlpha:hasAlpha isPlanar:NO colorSpaceName:NSDeviceRGBColorSpace bytesPerRow:destRowBytes bitsPerPixel:0];
        if (imageRep)
        {
            char 	*theData;
            int 	pixmapRowBytes;
            int 	rowByte,rowIndex;

            theData = [imageRep bitmapData];
        
            pixBaseAddr = GetPixBaseAddr(pixMapHandle);
            if (pixBaseAddr)
            {
                pixmapRowBytes = GetPixRowBytes(pixMapHandle);
            
                for (rowIndex=0; rowIndex< portHeight; rowIndex++)
                {
                    unsigned char *dst = theData + rowIndex * destRowBytes;
                    unsigned char *src = pixBaseAddr + rowIndex * pixmapRowBytes;
                    unsigned char a,r,g,b;
                    
                    for (rowByte = 0; rowByte < portWidth; rowByte++)
                    {
                        a = *src++;		// get source Alpha component
                        r = *src++;		// get source Red component
                        g = *src++;		// get source Green component
                        b = *src++;		// get source Blue component  
                        
                        if (keyColor) {
                            if (r == KeyR && g == KeyG && b == KeyB)
                                a = 0;
                        }
            
                        *dst++ = r;		// set dest. Alpha component
                        *dst++ = g;		// set dest. Red component
                        *dst++ = b;		// set dest. Green component
                        *dst++ = a;		// set dest. Blue component  
                    }
                }
            
                image = [[[NSImage alloc]initWithSize:NSMakeSize(portWidth, portHeight)] autorelease];
                if (image)
                {
                    [image addRepresentation:imageRep];
                    [imageRep release];
                }
            }
        }
    }

    NSAssert(pixMapHandle != NULL, @"null pixMapHandle");
    NSAssert(imageRep != nil, @"nil imageRep");
    NSAssert(pixBaseAddr != nil, @"nil pixBaseAddr");
    NSAssert(image != nil, @"nil image");

    if (pixMapHandle)
    {
        UnlockPixels(pixMapHandle);
    }

    return image;
}
*/

@end
