/* Copyright (c) 2004, Noah Sorscher
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 	• 	Redistributions of source code must retain the above copyright
 * 	notice, this list of conditions and the following disclaimer.
 * 	• 	Redistributions in binary form must reproduce the above copyright
 * 	notice, this list of conditions and the following disclaimer in the
 * 	documentation and/or other materials provided with the distribution.
 * 	• 	Neither the name of the Noah Sorscher nor the names of its
 * 	contributors may be used to endorse or promote products derived from
 * 	this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
//
//  tFileSave.m
//  Hootenanny
//
//  Created by Noah Sorscher on Sun Nov 23 2003.
//  Copyright (c) 2003 __MyCompanyName__. All rights reserved.
//

#import "tFileSave.h"
#import <QTKit/QTKit.h>

@interface QTMovie (saving)
- (BOOL)writeToFile:(NSString *)path;
@end

@implementation QTMovie (saving)
- (BOOL)writeToFile:(NSString *)path {
	NSError *e = nil;
	if (![self writeToFile:path withAttributes:nil error:&e] || e) {
		NSLog(@"error: [QTMovie writeToFile:withAttributes:]: %@", e);
		return NO;
	}
	return YES;
}
@end
	

@interface tFileSave (private)

- (void) saveOneMovie:(NSNotification *)aNotification;

@end

@implementation tFileSave


- (id) init
{
    self = [super init];
    [self clear];
    filesToSave = [[NSMutableArray alloc] init];
	progTimer = nil;
    return self;
}

- (void) clear
{
    cancelled = sameDir = NO;
    flattening = -1;
    myTask = nil;
    dirPath = tempPath = usedPath = nil;
	currentFileSize = 0;
	batchSize = 0;
    resourcePath = [[[NSBundle mainBundle] pathForResource:@"qtFlatten" ofType:nil] retain];
}

- (void) dealloc
{
    [filesToSave release];
	[super dealloc];
}

- (void) setProgressBar:(id)bar
{   progBar = bar;          }

- (void) setTextField:(id)field
{   textField = field;      }

- (void) setWindow:(id)aWindow
{   window = aWindow;       }

- (void) setSaveBeep:(id)beep
{   saveBeep = beep;        }

- (void) setStartButton:(id)button
{   startButton = button;   }

- (void) setSameDir:(BOOL)same
{   sameDir = same;         }

- (void) setDirPath:(NSString *)aPath
{   dirPath = aPath;        }

- (void) addMovie:(MHMovie *)aMovie
{
    if (aMovie)
        [filesToSave addObject:aMovie];
}

- (void) removeMovie:(MHMovie *)aMovie
{
    if (aMovie)
        [filesToSave removeObject:aMovie];
}

- (void) removeMovieAtIndex:(int)index
{
    if (index < [filesToSave count])
        [filesToSave removeObjectAtIndex:index];
}

- (void) removeAllMovies
{
    [filesToSave removeAllObjects];
}

- (void) cancel
{
    cancelled = YES;
    [myTask terminate];
    [window performClose:self];
	[progTimer invalidate];
	[progBar setHidden:YES];
	progTimer = nil;
}

// 0 = flattening, 1 = reference, 2 = save-in-place
- (void) saveFilesFlattening:(int)toFlatten
{
    cancelled = NO;
    [startButton setEnabled:NO];
    flattening = toFlatten;
	[progBar setHidden:NO];
	[progBar setIndeterminate:YES];
    [progBar startAnimation:self];
    NSArray *comps = [dirPath componentsSeparatedByString:@"/"];
    int i, j;
    NSString *aDir;
    NSString *aTempPath, *newPath = @"/";
    BOOL isDir;
    NSFileManager *fm = [NSFileManager defaultManager];
    for (i=0;i<[comps count];i++) {
        aDir = [comps objectAtIndex:i];
        if ([aDir length]) {
            newPath = [newPath stringByAppendingPathComponent:aDir];
            if (![fm fileExistsAtPath:newPath isDirectory:&isDir]) {
                [fm createDirectoryAtPath:newPath attributes:nil];
            } else if (!isDir) {
                aTempPath = newPath;
                j = 1;
                while ([fm fileExistsAtPath:aTempPath])
                    aTempPath = [NSString stringWithFormat:@"%@ %d", newPath, j++];
                newPath = aTempPath;
                [fm createDirectoryAtPath:newPath attributes:nil];
            }
        }
    }
    dirPath = [newPath copy];
	batchSize = [filesToSave count];
	/*
    if (toFlatten == 0)
        [self saveOneMovie:nil];
    else
	 */
        [NSThread detachNewThreadSelector:@selector(saveAllMovies) toTarget:self withObject:nil];
	progTimer = [NSTimer scheduledTimerWithTimeInterval:0.25 target:self selector:@selector(updateProgress:) userInfo:nil repeats:YES];
}

- (void)updateProgress:(id)sender
{
	float p1 = (float)(batchSize - 1 - [filesToSave count]) / batchSize;
	float p2 = (float)[[[[NSFileManager defaultManager] fileAttributesAtPath:usedPath traverseLink:YES]
							objectForKey:NSFileSize]
							unsignedLongLongValue] / 1024 / 1024; // megabytes
	p2 /= currentFileSize;
	p1 += p2 / batchSize;
	if (p1) {
		[progBar setIndeterminate:NO];
		[progBar setDoubleValue:p1];
	}
}

- (void) saveAllMovies
{
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	
	if (![[NSUserDefaults standardUserDefaults] boolForKey:@"ignoreSpotlight"]) {
		NSArray *bsl = [MHMovie blindspotlight];
		int i, n = [filesToSave count];
		int j, m = [bsl count];
		//BOOL decided = NO;
		NSMutableString *list = [NSMutableString stringWithString:@""];
		for (i=0;i<n;i++) {
			//if (decided) break;
			for (j=0;j<m;j++) {
				//if (decided) break;
				if ([[filesToSave objectAtIndex:i] objectForKey:[bsl objectAtIndex:j]] && [list rangeOfString:[bsl objectAtIndex:j]].length < 3) {
					if ([list length]) [list appendString:@", "];
					[list appendString:[bsl objectAtIndex:j]];
				}
			}
		}
		if ([list length]) {
			NSString *infoText = [NSString stringWithFormat:@"The following metadata you are saving will be ignored by Spotlight indexing: %@.  What do you want to do about this?", list];
			NSAlert *alert = [NSAlert alertWithMessageText:@"Spotlight Bug" defaultButton:@"Nothing" alternateButton:@"CC to \"Keywords\""
								  otherButton:@"CC to file's Finder comment" informativeTextWithFormat:infoText];
			if ([alert respondsToSelector:@selector(setShowsSuppressionButton:)])
					[alert setShowsSuppressionButton:YES];
			int result = [alert runModal];
			if (result == NSAlertDefaultReturn) {
					[[NSUserDefaults standardUserDefaults] setBool:NO forKey:@"copyKeywords"];
					[[NSUserDefaults standardUserDefaults] setBool:NO forKey:@"copyComment"];
			} else if (result == NSAlertAlternateReturn) {
					[[NSUserDefaults standardUserDefaults] setBool:YES forKey:@"copyKeywords"];
					[[NSUserDefaults standardUserDefaults] setBool:NO forKey:@"copyComment"];
			} else if (result == NSAlertOtherReturn) {
					[[NSUserDefaults standardUserDefaults] setBool:NO forKey:@"copyKeywords"];
					[[NSUserDefaults standardUserDefaults] setBool:YES forKey:@"copyComment"];
			}
			if ([alert respondsToSelector:@selector(suppressionButton)])
				if ([[alert suppressionButton] state])
					[[NSUserDefaults standardUserDefaults] setBool:YES forKey:@"ignoreSpotlight"];
			[[NSUserDefaults standardUserDefaults] synchronize];
				//decided = YES;
				//break;
		}
	}
	
    while ([filesToSave count] && !cancelled)
        [self saveOneMovie:nil];
    [self saveOneMovie:nil]; // for the temp path and myTask?
	[progBar setDoubleValue:0];
	[progBar setHidden:YES];
	[progTimer invalidate];
	progTimer = nil;
    [pool release];
}

- (void)saveOneMovie:(NSNotification *)aNotification
{
    //[[NSNotificationCenter defaultCenter] removeObserver:self];
    [myTask release];
    myTask = nil;
	
	BOOL success = NO;
    if ([[NSFileManager defaultManager] fileExistsAtPath:tempPath])
        [[NSFileManager defaultManager] removeFileAtPath:tempPath handler:nil];
    tempPath = nil;
    if ([filesToSave count] && !cancelled) {
        MHMovie *aMovie = [filesToSave objectAtIndex:0];
		
		NSString *list = @"";
		NSArray *bsl = [MHMovie blindspotlight];
		int i, n = [bsl count];
		for (i=0;i<n;i++) {
			if ([aMovie objectForKey:[bsl objectAtIndex:i]]) {
				if ([list length]) list = [list stringByAppendingString:@" "];
				list = [list stringByAppendingString:[aMovie objectForKey:[bsl objectAtIndex:i]]];
			}
		}
		if ([[NSUserDefaults standardUserDefaults] boolForKey:@"copyKeywords"])
			[aMovie setObject:list forKey:@"Keywords"];
		
        //NSLog(@"got movie to save: %d", [aMovie retainCount]);
        [filesToSave removeObjectAtIndex:0];
		currentFileSize = [[aMovie objectForKey:@"File Size"] floatValue];
        //NSLog(@"removed from array: %d", [aMovie retainCount]);
        NSString *name = [aMovie stringValue];
        [textField performSelectorOnMainThread:@selector(setStringValue:) withObject:[NSString stringWithFormat:@"Saving:  %@", name] waitUntilDone:NO];
        //[[window contentView] setNeedsDisplay:YES];
        usedPath = [aMovie objectForKey:@"File Name"];
        if (!usedPath) usedPath = name;
        if ([usedPath hasPrefix:@"."])
            usedPath = [NSString stringWithFormat:@" %@", usedPath];
		usedPath = [usedPath stringByDeletingPathExtension];
        usedPath = [NSMutableString stringWithString:usedPath];
        if (![usedPath hasSuffix:@".mov"]) [(NSMutableString *)usedPath appendString:@".mov"];
        [(NSMutableString *)usedPath replaceOccurrencesOfString:@":" withString:@"-" options:NSLiteralSearch range:NSMakeRange(0, [usedPath length])];
        tempPath = [MHUtil uniquePath:[[MHUtil supportDir] stringByAppendingPathComponent:usedPath]];
        if (sameDir)
            dirPath = [[aMovie objectForKey:@"File Path"] stringByDeletingLastPathComponent];
        usedPath = [MHUtil uniquePath:[dirPath stringByAppendingPathComponent:usedPath]];
        if (flattening == 0) {
            //[aMovie saveToPath:tempPath flatten:NO];
			[aMovie updateQTMovie];
			[[aMovie qtKitMovie] performSelectorOnMainThread:@selector(writeToFile:) withObject:tempPath waitUntilDone:YES];
            myTask = [[NSTask alloc] init];
            [myTask setLaunchPath:resourcePath];
            [myTask setArguments:[NSArray arrayWithObjects:tempPath, usedPath, nil]];
            //[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(saveOneMovie:) name:NSTaskDidTerminateNotification object:myTask];
            [myTask launch];
            [myTask waitUntilExit];
            // Custom Icon
            NSImage *image = [aMovie image];
            if (image) {
				if (![[NSWorkspace sharedWorkspace] setIcon:image forFile:usedPath options:NSExclude10_4ElementsIconCreationOption])
				NSLog(@"setIcon returned NO");
                //IconFamily *icon = [IconFamily iconFamilyWithThumbnailsOfImage:image];
                //[icon setAsCustomIconForFile:usedPath];
            }
			
            [self saveOneMovie:nil];
        } else if (flattening == 1)
            success = [aMovie saveToPath:usedPath flatten:NO];
        else if (flattening == 2) {
            success = [aMovie saveInPlace];
			usedPath = [[aMovie URL] path];
		}
		
		if ([[NSUserDefaults standardUserDefaults] boolForKey:@"copyComment"]) {
			if ([list length]) {
				[myTask release];
				myTask = [[NSTask alloc] init];
				[myTask setLaunchPath:@"/usr/bin/xattr"];
				[myTask setArguments:[NSArray arrayWithObjects:@"-w", @"com.apple.metadata:kMDItemFinderComment", list, usedPath, nil]];
				[myTask launch];
				[myTask waitUntilExit];
			}
		}
		
        //NSLog(@"saved: %d", [aMovie retainCount]);
    } else {
        [textField setStringValue:@""];
        [window performClose:self];
        //[progBar stopAnimation:self];
        if ([saveBeep state]) {
			if (success)
				[[NSSound soundNamed:@"Glass.aiff"] play];
			else
				[[NSSound soundNamed:@"Morse.aiff"] play];
		}
        [startButton setEnabled:YES];
    }
}

@end
