/* Copyright (c) 2004, Noah Sorscher
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 	 	Redistributions of source code must retain the above copyright
 * 	notice, this list of conditions and the following disclaimer.
 * 	 	Redistributions in binary form must reproduce the above copyright
 * 	notice, this list of conditions and the following disclaimer in the
 * 	documentation and/or other materials provided with the distribution.
 * 	 	Neither the name of the Noah Sorscher nor the names of its
 * 	contributors may be used to endorse or promote products derived from
 * 	this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
#import "tInit.h"
#include "defaults.h"
//#include "tModel.h"
#import "tPlaylist.h"
#include "tController.h"
//#import "tPathFormatter.h"
#import "MHMovie.h"
#import "tPlayback.h"
#import "keyDownWindow.h"
#import <sys/event.h>

@interface NSAttributedString (MHAttributedString)
- (NSString *)rtfValue;
@end
@implementation NSAttributedString (MHAttributedString)
- (NSString *)rtfValue
{
	NSData *d = [self RTFFromRange:NSMakeRange(0, [self length]) documentAttributes:nil];
	NSString *s = [NSString stringWithCString:[d bytes] length:[d length]];
	return s;
}
@end


@interface NSComboBox (MHColorComboBox)
- (void)addItemsWithRTFValues:(NSArray *)items;
- (NSString *)rtfValue;
@end
@implementation NSComboBox (MHColorComboBox)
- (void)addItemsWithRTFValues:(NSArray *)items
{
	int i, n = [items count];
	for (i=0;i<n;i++) {
		NSString *s = [items objectAtIndex:i];
		NSData *d = [NSData dataWithBytes:[s UTF8String] length:[s length]];
		NSAttributedString *astring = [[[NSAttributedString alloc] initWithRTF:d documentAttributes:nil] autorelease];
		if (astring)
			[self addItemWithObjectValue:astring];
	}
}
- (NSString *)rtfValue
{
	NSAttributedString *astring = [self attributedStringValue];
	return [astring rtfValue];
	/*
	NSMutableString *head = [NSMutableString stringWithString:@"{\\rtf1 {\\colortbl;"];
	NSMutableString *body = [NSMutableString stringWithString:@""];
	NSAttributedString *astring = [self attributedStringValue];
	NSString *str = [self stringValue];
	BOOL foundColor = NO;
	BOOL isBlack = YES;
	int i, n=[astring length];
	for (i=0;i<n;i++) {
		NSDictionary *dict = [astring attributesAtIndex:i effectiveRange:nil];
		NSColor *color = [dict objectForKey:NSForegroundColorAttributeName];
		color = [color colorUsingColorSpaceName:NSCalibratedRGBColorSpace];
		if (![color isEqual:[NSColor blackColor]]) {
			if (!foundColor) {
				[head appendFormat:@"\\red%d\\green%d\\blue%d;", (int)[color redComponent]*255, (int)[color greenComponent]*255, (int)[color blueComponent]*255];
				foundColor = YES;
			}
			if (isBlack) {
				[body appendString:@"\\cf1 "];
				isBlack = NO;
			}
		} else {
			if (!isBlack) {
				[body appendString:@"\\cf0 "];
				isBlack = YES;
			}
		}
		[body appendFormat:@"%C", [str characterAtIndex:i]];
	}
	[head appendFormat:@"} %@}", body];
	return head;	
	 */
}
@end

@implementation NSComboBoxCell (MHColorComboBoxCell)

- (void)tableView:(NSTableView *)tview willDisplayCell:(NSCell *)cell forTableColumn:(NSTableColumn *)tcol row:(int)row
{/*
	NSWindow *w = [tview window];
	NSRect frame = [w frame];
	if (frame.size.width < 114) {
		frame.size.width = 115;
		frame.origin.x -= 5;
		[w setFrame:frame display:YES];
	}*/
}
@end

@implementation MHApplication
/*
- (void)sendEvent:(NSEvent *)e
{
	if ([e type] != NSKeyDown) return [super sendEvent:e];
	if ([[self keyWindow] performKeyEquivalent:e]) return;
	NSResponder *r = [[self keyWindow] firstResponder];
	while (r) {
		if ([r performKeyEquivalent:e]) return;
		r = [r nextResponder];
	}
	if ([[self mainMenu] performKeyEquivalent:e]) return;

	NSString *s = nil;
	int flags = [e modifierFlags];
	NSString *c = [e characters];
	if ([c length]) {
		unichar u = [c characterAtIndex:0];
		//NSLog(@"%d", u);
		NSArray *arrows = [NSArray arrayWithObjects:@"Up Arrow", @"Down Arrow", @"Left Arrow", @"Right Arrow", nil];
		if (u > 63231 && u < 63236) { // arrows
			s = [arrows objectAtIndex:u-63232];
			if (flags & NSCommandKeyMask) s = [NSString stringWithFormat:@"%C %@", 8984, s];
			else if (flags & NSAlternateKeyMask) s = [NSString stringWithFormat:@"%C %@", 8997, s];
			else if (flags & NSShiftKeyMask) s = [NSString stringWithFormat:@"%C %@", 8679, s];
			else if (flags & NSControlKeyMask) s = [NSString stringWithFormat:@"^ %@", s];
		}
	}
	if (![[self delegate] doControl:s]) [super sendEvent:e];
}*/
@end

@implementation tInit
float hlR = 0.938; float hlG = 0.962; float hlB = 1.0; float hlA = 1.0;
id tInitNibInstance;
- (BOOL)doControl:(NSString *)action
{
	if (!action) return NO;
	//NSLog(action);
	MHMovie *m = [[tPlayback nibInstance] currentMovie];
	int choice = [[[[NSUserDefaults standardUserDefaults] objectForKey:@"controls"] objectForKey:action] intValue];
	//NSLog(@"%d", choice);
	float v = 0;
	switch (choice) {
			case 1: //if ([[m qtKitMovie] rate])
					[[tPlayback nibInstance] play:nil];
				//else [[tPlayback nibInstance] pause:nil];
				break;
			case 2: [tPlayback fullScreen]; break;
			case 3: v = [[m qtKitMovie] volume];
				if (v < 0) v = 0-v;
				if (v < 1) v += 0.1;
				else v *= 1.1;
				[(tPlayback *)[tPlayback nibInstance] setVolume:[NSNumber numberWithDouble:v*256]]; break;
			case 4: v = [[m qtKitMovie] volume];
				if (v > 0) {
					if (v <= 1) v -= 0.1;
					else v /= 1.1;
				}
				[(tPlayback *)[tPlayback nibInstance] setVolume:[NSNumber numberWithDouble:v*256]]; break;
			case 5: [[tPlayback nibInstance] maxVolume:nil]; break;
			case 6: [[tPlayback nibInstance] setVolume:1000]; break;
			case 7: [(tPlayback *)[tPlayback nibInstance] setVolume:[NSNumber numberWithDouble:(0-[[m qtKitMovie] volume])*256]]; break;
			case 8: [[m qtKitMovie] stepForward]; break;
			case 9: [[m qtKitMovie] stepBackward]; break;
			case 10: [[tPlayback nibInstance] jumpF:nil]; break;
			case 11: [[tPlayback nibInstance] jumpB:nil]; break;
			case 12: [m goToNextChapter];
					[[tPlayback nibInstance] tick:nil]; break;
			case 13: [m goToPrevChapter];
					[[tPlayback nibInstance] tick:nil];  break;
			case 14: [[tPlayback nibInstance] makeMovieGoNow:nextMovie]; break;
			case 15: [[tPlayback nibInstance] makeMovieGoNow:lastMovie]; break;
			case 16: [[tPlayback nibInstance] stepRate:@"Faster"]; break;
			case 17: [[tPlayback nibInstance] stepRate:@"Slower"]; break;
			case 18: //[[tPlayback nibInstance] stepRate:@"-1"];
					[[m qtKitMovie] setRate:-1]; break;
			default: return NO;
	}
	return YES;
}

- (int)numberOfRowsInTableView:(NSTableView *)t
{
	return 26;
}

- (NSObject *)tableView:(NSTableView *)t objectValueForTableColumn:(NSTableColumn *)c row:(int)row
{	
	NSMutableArray *ctrl = [NSMutableArray arrayWithObjects:@"Click", @"Double Click", @"Scroll Up", @"Scroll Down", @"Scroll Left", @"Scroll Right", @"Up Arrow", @"Down Arrow", @"Left Arrow", @"Right Arrow", nil];
	int i=6;
	for (;i<10;i++) {
		NSString *s = [NSString stringWithFormat:@"%C %@", 8984, [ctrl objectAtIndex:i]]; // cmd
		[ctrl addObject:s];
		s = [NSString stringWithFormat:@"%C %@", 8997, [ctrl objectAtIndex:i]]; // option
		[ctrl addObject:s];
		s = [NSString stringWithFormat:@"%C %@", 8679, [ctrl objectAtIndex:i]]; // shift
		[ctrl addObject:s];
		s = [NSString stringWithFormat:@"^ %@", [ctrl objectAtIndex:i]];
		[ctrl addObject:s];
	}
	if ([[c identifier] isEqualToString:@"Control"])
		return [ctrl objectAtIndex:row];
	NSCell *cell = [c dataCell];
	if (![cell isKindOfClass:[NSPopUpButtonCell class]]) {
		cell = [[[NSPopUpButtonCell alloc] init] autorelease];
		[c setDataCell:cell];
		[cell setControlSize:NSSmallControlSize];
		[(NSPopUpButtonCell *)cell addItemsWithTitles:[NSArray arrayWithObjects:@"Do Nothing", @"Play", @"Fullscreen", @"Volume Up", @"Volume Down", @"100% volume", @"400% Volume", @"Mute", @"Step Forward", @"Step Backward", @"Jump Forward", @"Jump Backward", @"Next Chapter", @"Prev Chapter", @"Next Movie", @"Prev Movie", @"Faster", @"Slower", @"Reverse", nil]];
	}
	NSUserDefaults *defs = [NSUserDefaults standardUserDefaults];
	NSMutableDictionary *d = [defs objectForKey:@"controls"];
	if (!d) {
		d = [NSMutableDictionary dictionaryWithCapacity:26];
		int a[26] = {1,2, 3,4,8,9, 3,4,9,8, 16,5,0,0, 17,7,0,0, 18,13,0,0, 2,12,0,0};
		//NSArray *a = [NSArray arrayWithObjects:@"Play/Pause", @"Fullscreen", @"Volume Up", @"Volume Down", @"Step Forward", @"Step Backward", @"Volume Up", @"Volume Down", @"Step Forward", @"Step Backward", nil];
		//NSRunAlertPanel(@"poop", @"or die", @"trying", nil, nil);
		for (i=0;i<26;i++) {
			int old = 0;
			NSString *oldKey;
			if (i<2 && [defs objectForKey:@"click1"]) {
				if (i==1) oldKey = @"click2";
				else oldKey = @"click1";
				old = [defs integerForKey:oldKey];
				if (old == 0) old = 2;
				else if (old == 2) old = 0;
				[d setObject:[NSNumber numberWithInt:old] forKey:[ctrl objectAtIndex:i]];
			} else if (i < 6 && [defs objectForKey:@"scrollUp"]) {
				oldKey = [[NSArray arrayWithObjects:@"scrollUp", @"scrollDown", @"scrollRight", @"scrollLeft", nil] objectAtIndex:i-2];
				old = [defs integerForKey:oldKey];
				if (old < 2) old += 8;
				else if (old < 4) old += 14;
				else if (old < 6) old += 6;
				else if (old < 8) old -= 3;
				else old = 0;
				[d setObject:[NSNumber numberWithInt:old] forKey:[ctrl objectAtIndex:i]];
			}
			else [d setObject:[NSNumber numberWithInt:a[i]] forKey:[ctrl objectAtIndex:i]];
		}	
		[defs setObject:d forKey:@"controls"];
		[defs synchronize];
	}
	NSNumber *num = [d objectForKey:[ctrl objectAtIndex:row]];
	if (![num intValue]) num = nil;
	return num;
}

- (void)tableView:(NSTableView *)t setObjectValue:(NSObject *)o forTableColumn:(NSTableColumn *)c row:(int)row
{
	NSString *key = (NSString *)[self tableView:t objectValueForTableColumn:[[t tableColumns] objectAtIndex:0] row:row];
	NSMutableDictionary *d = [NSMutableDictionary dictionaryWithDictionary:[[NSUserDefaults standardUserDefaults] objectForKey:@"controls"]];
	if (!d) d = [NSMutableDictionary dictionary];
	if (!o) [d removeObjectForKey:key];
	else [d setObject:o forKey:key];
	[[NSUserDefaults standardUserDefaults] setObject:d forKey:@"controls"];
	[[NSUserDefaults standardUserDefaults] synchronize];
	if (row < 6) return;
	row -= 6;
	unichar u = 63232;
	if (row < 4) u += row;
	else u += (row - 4) / 4;
	NSString *s = [NSString stringWithFormat:@"%C", u];
	int mask = 0;
	if (row > 3) {
		if (row % 4 == 0) mask = NSCommandKeyMask;
		else if (row % 4 == 1) mask = NSAlternateKeyMask;
		else if (row % 4 == 2) mask = NSShiftKeyMask;
		else if (row % 4 == 3) mask = NSControlKeyMask;
	}
	if (!mask) return; // non-modifier-key arrow key events should be context-specific, not overridden by menu items
	NSMenuItem *item = nil;
	switch ([(NSNumber *)o intValue]) {
		case 3:  item = [[[[[[NSApp mainMenu] itemAtIndex:4] submenu] itemAtIndex:13] submenu] itemAtIndex:0]; break; // vol up
		case 4:  item = [[[[[[NSApp mainMenu] itemAtIndex:4] submenu] itemAtIndex:13] submenu] itemAtIndex:1]; break; // vol down
		case 7:  item = [[[[[[NSApp mainMenu] itemAtIndex:4] submenu] itemAtIndex:13] submenu] itemAtIndex:2]; break; // mute
		case 10: item = [[[[NSApp mainMenu] itemAtIndex:4] submenu] itemAtIndex:16]; break; // jump forward
		case 11: item = [[[[NSApp mainMenu] itemAtIndex:4] submenu] itemAtIndex:17]; break; // jump backward
		case 14: item = [[[[NSApp mainMenu] itemAtIndex:4] submenu] itemAtIndex:2]; break; // next
		case 15: item = [[[[NSApp mainMenu] itemAtIndex:4] submenu] itemAtIndex:3]; break; // prev
		case 16: item = [[[[NSApp mainMenu] itemAtIndex:4] submenu] itemAtIndex:8]; break; // faster
		case 17: item = [[[[NSApp mainMenu] itemAtIndex:4] submenu] itemAtIndex:9]; break; // slower
		default: break;
	}
	[item setKeyEquivalent:s];
	[item setKeyEquivalentModifierMask:mask];
}



+ (void)initialize
{
    //NSLog(@"tInit initialize");
    NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
    NSMutableDictionary *appDefs = [NSMutableDictionary dictionary];
    NSString *saveDir = [NSString stringWithFormat:@"%@/Movies", NSHomeDirectory()];
    BOOL isDir;
    if (![[NSFileManager defaultManager] fileExistsAtPath:saveDir isDirectory:&isDir] || !isDir)
        saveDir = NSHomeDirectory();
        
    // Preferences
    NSDictionary *controlSpeeds = [NSDictionary dictionaryWithObjectsAndKeys:@"2", @"FF Speed", @"2", @"RW Speed", @"10", @"Reverse Jump", @"30", @"Forward Jump", @".3", @"SloMo Speed", @".5", @"Snap Back", nil];
    [appDefs setObject:controlSpeeds forKey:@"controlSpeeds"];
    
    // Bools
    NSArray *yeses = [NSArray arrayWithObjects:t_remember, t_toSize, t_album, t_author, t_copy, t_credate, t_fname, t_desc, t_dir, t_info, t_perf, t_prod, t_writ, t_saveBeep, @"merge", @"rowColoring", @"hides", @"warn", @"optIcon", @"exitMouse", @"chapWin", @"brushed", @"optReduce", @"doubleClick", @"addSeparateAudio", nil];
    int i, n = [yeses count];
    for (i=0;i<n;i++) {
        [appDefs setObject:[NSNumber numberWithBool:YES] forKey:[yeses objectAtIndex:i]];
    }
    NSArray *noes = [NSArray arrayWithObjects:t_constrainAspectRatio, nil];
    n = [noes count];
    for (i=0;i<n;i++) {
        [appDefs setObject:[NSNumber numberWithBool:NO] forKey:[noes objectAtIndex:i]];
    }
    
    // Numbers
    int intValues[] = {10, 1, 1, 6, 4, 1, 1};
    NSArray *intKeys = [NSArray arrayWithObjects:t_numberOfAutoColumns, t_whichSize, t_interval, t_numPerformers, t_pathLevels, @"titleNumber", nil];
    n = [intKeys count];
    for (i=0;i<n;i++) {
        [appDefs setObject:[NSNumber numberWithInt:intValues[i]] forKey:[intKeys objectAtIndex:i]]; 
    }
    float floatValues[] = { hlA, hlB, hlG, hlR, 12, 17};
    NSArray *floatKeys = [NSArray arrayWithObjects:@"hlA", @"hlB", @"hlG", @"hlR", @"fontSize", @"rowHeight", nil];
    n = [floatKeys count];
    for (i=0;i<n;i++) {
        [appDefs setObject:[NSNumber numberWithFloat:floatValues[i]] forKey:[floatKeys objectAtIndex:i]];
    }
    
    // Strings
    NSArray *stringValues = [NSArray arrayWithObjects:saveDir, @"Encoded by YourNameHere", @"native", @"", [[MHUtil supportDir] stringByAppendingPathComponent:@"Metahoot Download Cache.xml"], @"Credits", @"h:mm:ss.xxx", @"1000", nil];
    NSArray *stringKeys = [NSArray arrayWithObjects:@"saveDir", t_commentText, @"fsAspect", @"iconText", @"cacheLocation", @"skipText", @"format", @"denom", nil];
    n = [stringKeys count];
    for (i=0;i<n;i++) {
        [appDefs setObject:[stringValues objectAtIndex:i] forKey:[stringKeys objectAtIndex:i]];
    }
    
    [appDefs setObject:[[MHUtil supportDir] stringByAppendingString:@"textDump"] forKey:@"textDump"];
    
    [defaults registerDefaults:appDefs];
}

+ (BOOL)boolForKey:(NSString *)key
{	return [tInitNibInstance settingForKey:key];	}

+ (void)setBool:(BOOL)flag forKey:(NSString *)key
{	[tInitNibInstance setSetting:flag forKey:key];	}

+ (NSString *)stringForKey:(NSString *)key
{	return [tInitNibInstance stringSettingForKey:key];	}

+ (void)setString:(NSString *)string forKey:(NSString *)key
{	[tInitNibInstance setStringSetting:string forKey:key];	}

+ (double)doubleForKey:(NSString *)key
{	return [tInitNibInstance doubleSettingForKey:key];	}

+ (void)setDouble:(double)value forKey:(NSString *)key
{       [tInitNibInstance setDoubleSetting:value forKey:key];   }

- (void) flagsChanged:(NSEvent *)theEvent
{
	// this isn't called, is it?
	NSLog(@"flags");
    [[NSNotificationCenter defaultCenter] postNotificationName:@"flagsChangedNotification" object:theEvent];
}

- (void)applicationDidBecomeActive:(NSNotification *)aNotification
{
}

- (BOOL)applicationShouldHandleReopen:(NSApplication *)theApplication hasVisibleWindows:(BOOL)flag
{
    //NSLog(@"tInit applicationShouldHandleReopen");
    NSArray *windows = [NSApp windows];
    NSMutableArray *visWins = [NSMutableArray array];
    if ([windows count]) {
        int i;
        for (i=0;i<[windows count];i++) {
            if ([[windows objectAtIndex:i] isVisible])
                [visWins addObject:[windows objectAtIndex:i]];
        }
    }
    if (!flag || [visWins count] == 0 || (([visWins count] == 1) && [visWins objectAtIndex:0] == [player fullscreenWindow])) { 
        [controller setMainTabViewName:@"Main"];
        [[controller mainWindow] makeKeyAndOrderFront:nil];
    }
    return NO;
}

- (NSArray *)controls
{
	return [NSArray arrayWithObjects:floaterView, nil];
}

+ (NSArray *)controls
{	return [tInitNibInstance controls];	}

- (void)reloadDatabase
{
    NSUserDefaults *defs = [NSUserDefaults standardUserDefaults];
	//[openingWindow makeKeyAndOrderFront:nil];
	NSArray *lists = [defs objectForKey:@"playlists"];
	if (nil == lists) lists = [NSArray array];
	/*
	 * look for old-version settings
	 */
	NSArray *oldList = nil;
	oldList = [defs objectForKey:t_movieData];
	if (oldList != nil) {
		NSMutableArray *newLists = [[NSMutableArray alloc] init];
		[newLists addObjectsFromArray:lists];
		NSMutableDictionary *playlist = [[NSMutableDictionary alloc] init];
		[playlist setObject:@"Old Playlist" forKey:@"name"];
		[playlist setObject:oldList forKey:@"movieData"];
		[newLists insertObject:playlist atIndex:0];
		lists = newLists;
		[defs removeObjectForKey:t_movieData];
	}
	
	tPlaylist *aModel;
	NSString *aName;
	NSDictionary *aList;
	int i;
	/*
	 * get the playlist names, and add to controller
	 */
	for (i=1;i<[lists count];i++) {
		aList = [lists objectAtIndex:i];
		aName = [aList objectForKey:@"name"];
		aModel = [(tPlaylist *)[tPlaylist alloc] initWithController:controller];
		[aModel setName:aName];
		[controller addPlaylist:aModel];
	}
	
	/*
	 * add options for smart playlists
	 */
	NSArray *addedLists = [controller playlists];
	NSArray *names = [defs objectForKey:@"playlistNames"];
	NSArray *options = [defs objectForKey:@"playlistOptions"];
	for (i=0;i<[names count];i++) {
		aName = [names objectAtIndex:i];
		if ([addedLists count] <= i+1 || ![aName isEqualToString:[[addedLists objectAtIndex:i+1] name]]) {
			aModel = [(tPlaylist *)[tPlaylist alloc] initWithController:controller];
			[aModel setName:aName];
			[controller addPlaylist:aModel];
			NSArray *option = [options objectAtIndex:i];
			if ([option count])
				[aModel setOptions:option];
		}
	}
	addedLists = [controller playlists];
	
	/*
	 * add the movies to the library playlist
	 */
	NSMutableArray *allMovies = [NSMutableArray array];
	for (i=0;i<[lists count];i++)
		[allMovies addObjectsFromArray:[[lists objectAtIndex:i] objectForKey:@"movieData"]];
	[[addedLists objectAtIndex:0] addFromDefsEntry:allMovies];
	
	// add the movies to the non-library playlists
	tPlaylist *library = [addedLists objectAtIndex:0];
	for (i=0;i<[allMovies count];i++) {
		NSArray *listNumbers = [[allMovies objectAtIndex:i] objectForKey:@"listNumbers"];
		int j, m = [listNumbers count];
		for (j=0;j<m;j++) {
			int index = [[listNumbers objectAtIndex:j] intValue];
			if (index > 0 && index < [addedLists count])
				[(tPlaylist *)[addedLists objectAtIndex:index] addMovie:[library movieForIndex:i]];
		}
	}
	
	// re-sort the playlists
	NSArray *sorts = [defs objectForKey:@"playlistSorts"];
	for (i=0;i<[sorts count];i++) {
		NSString *sort = [sorts objectAtIndex:i];
		if ([sort length])
			[[addedLists objectAtIndex:i] sortByIdentifier:sort];
	}
}

- (void)applicationDidFinishLaunching:(NSNotification *)aNotification
{
	//NSRunAlertPanel(@"a", @"c", @"b", nil, nil);
	
    //NSLog(@"tInit applicationDidFinishLaunching");
    /*
     *	Movie Change Listeners
     */
    NSArray *movieChangeListeners = [NSArray arrayWithObjects:chapters, timeline, www, nil];
    int i, n = [movieChangeListeners count];
    for (i=0;i<n;i++) {
        [controller addMovieChangeListener:[movieChangeListeners objectAtIndex:i]];
    }

     /*
     *	(Re)load the Database
     */
    // weird bug workaround
    NSDictionary *fakeMovie = [NSDictionary dictionaryWithObject:[[NSBundle mainBundle] pathForResource:@"dummy" ofType:@"txt"] forKey:@"File Path"];
    tPlaylist *fakePlaylist = [(tPlaylist *)[tPlaylist alloc] initWithController:controller];
    [fakePlaylist setName:@"fake playlist"];
    [fakePlaylist addFromDefsEntry:[NSArray arrayWithObject:fakeMovie]];
    [controller addPlaylist:fakePlaylist];
    [controller removePlaylist:nil];
    
    NSUserDefaults *defs = [NSUserDefaults standardUserDefaults];
    if ([defs boolForKey:t_remember]) {
        [self reloadDatabase];
		[theTable reloadData];
    } else { // if not remember
        [controller addPlaylist:nil];
    }
	
    //[controller tableView:nil shouldSelectRow:-1];
    [controller tableView:nil shouldSelectRow:0];
    [plTable selectRowIndexes:[NSIndexSet indexSetWithIndex:0] byExtendingSelection:NO];
    
    [self setBrushedMetal];
    [[controller mainWindow] makeKeyAndOrderFront:nil];
    [controller updateTotals];
    [[theTable dataSource] tableView:theTable shouldSelectRow:0];
    [theTable selectRowIndexes:[NSIndexSet indexSetWithIndex:0] byExtendingSelection:NO];
    //[[controller mainWindow] makeFirstResponder:theTable];
	
	[click1 setEnabled:YES];
}


- (NSApplicationTerminateReply)applicationShouldTerminate:(NSApplication *)sender
{
    //NSLog(@"tInit applicationShouldTerminate");
    NSUserDefaults *defs = [NSUserDefaults standardUserDefaults];
    BOOL remember;
    
    [player windowShouldClose:nil];
    remember = [[defs objectForKey:t_remember] boolValue];
    if (remember) {
        //[model savePrefs];
        [controller savePrefs];
    }
    [defs synchronize];
    //[MHUtil killSupportDir];
    
    id vo = [NSConnection rootProxyForConnectionWithRegisteredName:@"mhImageHelper" host:nil];
    if (nil != vo)
        [vo quit];
    
    [MHUtil clearSupportDir];
	[player windowWillUseStandardFrame:nil defaultFrame:NSMakeRect(0,0,0,0)];
	
    return YES;
}


- (NSColor *) tableView: (NSTableView *)tableView backgroundColorForRow: (int) row
{
    if (![rowColoring state]) return [NSColor whiteColor];
    if (tableView != theTable) return [NSColor whiteColor];
    int i = row % 2;
    if (!i)
        return [highlightColor color];
            //[aCell setBackgroundColor:[NSColor colorWithCalibratedRed:1.0 green:0.8586 blue:0.373 alpha:1.0]];
    return [NSColor whiteColor];
            //[aCell setBackgroundColor:[NSColor colorWithCalibratedRed:1.0 green:0.981 blue:0.738 alpha:1.0]];
}

- (NSColor *) tableView: (NSTableView *)tableView highlightColorForRow: (int) row
{
    double a = 1.0, r = (double)61/(double)256, g = (double)128/(double)256, b = (double)223/(double)256;
    return [NSColor colorWithCalibratedRed:r green:g blue:b alpha:a];
}

- (void)savePrefs
{
    //NSLog(@"tInit");
    NSUserDefaults *defs = [NSUserDefaults standardUserDefaults];
    [player savePrefs:self];
    
    // Bools
    int i, j, m, n = [theBoolButtons count];
    for (i=0;i<n;i++) {
        [defs setBool:[[theBoolButtons objectAtIndex:i] state] forKey:[theBoolKeys objectAtIndex:i]];
    }
    
    [MHMovie setAsksForMissingFiles:[askMovies state]];
    [MHMovie setIgnoresMissingFiles:![placeHolder state]];
    
	NSArray *tables = [NSArray arrayWithObjects:[controller mainTable], [controller plTable], [controller summaryTable], [controller tracksTable], nil];
	n = [tables count];
	for (i=0;i<n;i++) {
		NSTableView *t = [tables objectAtIndex:i];
		m = [[t tableColumns] count];
		for (j=0;j<m;j++) {
			NSCell *c = [[[t tableColumns] objectAtIndex:j] dataCell];
			if ([c respondsToSelector:@selector(setFont:)])
				[c setFont:[NSFont systemFontOfSize:[fontSize doubleValue]]];
		}
	}
    
    // Strings
    n = [theStringFields count];
    for (i=0;i<n;i++) {
        [defs setObject:[[theStringFields objectAtIndex:i] stringValue] forKey:[theStringKeys objectAtIndex:i]];
    }
	NSArray *pops = [NSArray arrayWithObjects:click1, click2, scrollUp, scrollDown, scrollRight, scrollLeft, nil];
	NSArray *skeys = [NSArray arrayWithObjects:@"click1", @"click2", @"scrollUp", @"scrollDown", @"scrollRight", @"scrollLeft", nil];
	n = [pops count];
	for (i=0;i<n;i++) {
		[defs setInteger:[[pops objectAtIndex:i] indexOfSelectedItem] forKey:[skeys objectAtIndex:i]];
	}
	
	// time format box
	NSString *form = [timeFormat rtfValue];
	if ([form length]) {
		[defs setObject:form forKey:@"format"];
		NSMutableArray *forms = [NSMutableArray arrayWithArray:[defs objectForKey:@"timeFormats"]];
		int i, n = [forms count];
		for (i=0;i<n;i++) {
			NSString *aFormat = [forms objectAtIndex:i];
			if ([aFormat hasSuffix:form]) break;
		}
		if (i == n) { //no break
			[forms addObject:form];
			[defs setObject:forms forKey:@"timeFormats"];
			[timeFormat addItemsWithRTFValues:[NSArray arrayWithObject:form]];
		}
	}
    
    // Integers
    NSArray *intButtons = [NSArray arrayWithObjects:autoCols, numPerformers, pathLevels, fontSize, rowHeight, nil];
    NSArray *intKeys = [NSArray arrayWithObjects:t_numberOfAutoColumns, t_numPerformers, t_pathLevels, @"fontSize", @"rowHeight", nil];
    n = [intKeys count];
    for (i=0;i<n;i++) {
        [defs setInteger:[[intButtons objectAtIndex:i] intValue] forKey:[intKeys objectAtIndex:i]];
    }
    
    int sizeSelection = [saveSizeRadio selectedRow];
    [defs setInteger:sizeSelection forKey:t_saveSize];
    //[defs setInteger:[openWhichSize indexOfSelectedItem] forKey:t_whichSize];
    // color well nonsense
    float r, g, b, a;
    [[highlightColor color] getRed:&r green:&g blue:&b alpha:&a];
    [defs setFloat:r forKey:@"hlR"];
    [defs setFloat:g forKey:@"hlG"];
    [defs setFloat:b forKey:@"hlB"];
    [defs setFloat:a forKey:@"hlA"];
    //[defs setFloat:[interval floatValue] forKey:t_interval];
    
    [lang savePrefs];
    
    [self setBrushedMetal];
    [defs synchronize];
}

- (IBAction)editAutoComplete:(id)sender
{
    //[[NSUserDefaults standardUserDefaults] removeObjectForKey:@"learned"];
    //[[NSUserDefaults standardUserDefaults] synchronize];
}

- (IBAction)acAdd:(id)sender
{
    NSMutableDictionary *learned = [NSMutableDictionary dictionaryWithDictionary:[[NSUserDefaults standardUserDefaults] objectForKey:@"learned"]];
    id item = nil;
    int index = [acTable selectedRow];
    if (-1 == index)
        [learned setObject:[NSMutableArray array] forKey:@"New Category"];
    else {
        item = [acTable itemAtRow:index];
        if ([item parent])
            item = [item parent];
        NSString *cat = [item value];
        NSMutableArray *anArray = [NSMutableArray arrayWithArray:[learned objectForKey:cat]];
        [anArray addObject:@"New Value"];
        [learned setObject:anArray forKey:cat];
    }
    [[NSUserDefaults standardUserDefaults] setObject:learned forKey:@"learned"];
    [acTable reloadData];
    if (item)
        [acTable expandItem:item];
}

- (IBAction)acRemove:(id)sender
{
    NSUserDefaults *defs = [NSUserDefaults standardUserDefaults];
    NSMutableDictionary *learned = [NSMutableDictionary dictionaryWithDictionary:[defs objectForKey:@"learned"]];
    id item = [acTable itemAtRow:[acTable selectedRow]];
    if (!item) return;
    id parent = [item parent];
    if (parent) {
        NSMutableArray *anArray = [NSMutableArray arrayWithArray:[learned objectForKey:[parent value]]];
        [anArray removeObject:[item value]];
        [learned setObject:anArray forKey:[parent value]];
    } else {
        [learned removeObjectForKey:[item value]];
    }
    [defs setObject:learned forKey:@"learned"];
    [acTable reloadData];
    if (parent)
        [acTable expandItem:parent];
}

- (IBAction)setCache:(id)sender
{
    NSSavePanel *sp = [NSSavePanel savePanel];
    
    int runResult = [sp runModalForDirectory:nil file:@"Metahoot Download Cache.xml"];

    if (runResult == NSOKButton) {
        [cacheText setStringValue:[sp filename]];
    }
}

- (IBAction)defaultHLColor:(id)sender
{
    [highlightColor setColor:[NSColor colorWithCalibratedRed:hlR green:hlG blue:hlB alpha:hlA]];
}

- (void)setBrushedMetal
{
    //NSLog(@"tInit setBrushedMetal");
    NSRect aRect = [[[controller mainWindow] contentView] frame];
    NSRect winRect = [[controller mainWindow] frame];
    NSScreen *screen = [[controller mainWindow] screen];
    aRect.origin = winRect.origin;
    aRect.origin.y += winRect.size.height - aRect.size.height - 22;
    unsigned int styleMask, normalStyleMask = [prefsWindow styleMask];
    unsigned int brushedStyleMask = normalStyleMask | NSTexturedBackgroundWindowMask;
    BOOL shouldBrushed = [brushedMetal state];
    BOOL isBrushed = [[controller mainWindow] styleMask] == brushedStyleMask;
    if ((shouldBrushed && !isBrushed) || (!shouldBrushed && isBrushed)) {
        if (shouldBrushed)
            styleMask = brushedStyleMask;
        else styleMask = normalStyleMask;
        NSBackingStoreType backingType = NSBackingStoreBuffered;
        NSWindow *newWindow = [[keyDownWindow alloc] initWithContentRect:aRect styleMask:styleMask backing:backingType defer:NO screen:screen];
        NSToolbar *tb = [[controller mainWindow] toolbar];
        NSArray *tbItems = [tb items];
        [newWindow setContentView:[[controller mainWindow] contentView]];
        [newWindow setDelegate:player];
        [newWindow setFrameAutosaveName:@"editingWindow"];
        [newWindow setMinSize:NSMakeSize(400,107)];
        //[newWindow makeKeyAndOrderFront:nil];
        [newWindow registerForDraggedTypes:[NSArray arrayWithObjects:NSFilenamesPboardType, NSStringPboardType, nil]];
        [newWindow setTitle:@"MetaHoot"];
        [newWindow setReleasedWhenClosed:NO];
        [controller setMainWindow:newWindow];
        int i;
        for (i=0;i<[tbItems count];i++) {
            [tb removeItemAtIndex:i];
            [tb insertItemWithItemIdentifier:[[tbItems objectAtIndex:i] itemIdentifier] atIndex:i];
        }
        [[controller mainWindow] makeKeyAndOrderFront:nil];
    }
}

- (BOOL)application:(NSApplication *)theApplication openFile:(NSString *)filename
{
    //NSLog(@"tInit application: openFile:");
    [controller performSelector:@selector(openFilesFromArray:) withObject:[NSArray arrayWithObject:filename] afterDelay:0];
    //[controller openFilesFromArray:[NSArray arrayWithObject:filename] model:[controller getTrueModel]];
    return YES;
}

- (void) awakeFromNib
{
	kqfd = 0;
	runLoopSocket = nil;
	watchlist = nil;
	subUpdateDate = nil;
    //NSLog(@"tInit awakeFromNib");
    tInitNibInstance = self;
    NSUserDefaults *defs = [NSUserDefaults standardUserDefaults];
    //NSString *idString;
    int autoColumns;
    int i, n;
    //NSArray *cols = [defs objectForKey:t_cols];
    NSTableColumn *aColumn;
	
	n = [self numberOfRowsInTableView:controlsTable];
	for (i=0;i<n;i++) {
		NSNumber *num = [self tableView:controlsTable objectValueForTableColumn:[[controlsTable tableColumns] objectAtIndex:1] row:i];
		[self tableView:controlsTable setObjectValue:num forTableColumn:[[controlsTable tableColumns] objectAtIndex:1] row:i];
	}
    
    //NSLog(@"init");
    /*
     *	For Drag and Drop
     */
    [[controller mainWindow] registerForDraggedTypes:[NSArray arrayWithObjects:NSFilenamesPboardType, NSStringPboardType, nil]];
    /*
     *	Initialize the Tableview shit
     */
    [theTable setAllowsColumnReordering:YES];
    [theTable setAllowsColumnResizing:YES];
    [theTable setAllowsMultipleSelection:YES];
    [theTable setAllowsEmptySelection:YES];
    [theTable setAllowsColumnSelection:YES];
    //[theTable setSizesToFit:NO];
    //[plTable setSizesToFit:YES];
    [theTable registerForDraggedTypes:[NSArray arrayWithObject:@"moviesFromSelf"]];
    [plTable registerForDraggedTypes:[NSArray arrayWithObjects:@"moviesFromOther", NSFilenamesPboardType, nil]];
    NSSize aSize;
    aSize.height = 0.0;
    aSize.width = 1.0;
    [theTable setIntercellSpacing:aSize];
    
    [chapterTable setDelegate:chapters];
   
    //chapter table
    aColumn = [[NSTableColumn alloc] initWithIdentifier:@"Content"];
    [[aColumn headerCell] setStringValue:@"Content"];
    [chapterTable addTableColumn:aColumn];
    aColumn = [[NSTableColumn alloc] initWithIdentifier:@"Start"];
    [[aColumn headerCell] setStringValue:@"Start"];
    [chapterTable addTableColumn:aColumn];
    [chapterTable setDelegate:chapters];
    [chapterTable reloadData];
    [chapterTable setNeedsDisplay:YES];
    
    /*
     *	Restore the Preferences
     */
     
     [lang restorePrefs];
     if (![defs objectForKey:@"controlSpeeds"]) {
        NSDictionary *controlSpeeds = [NSDictionary dictionaryWithObjectsAndKeys:@"2", @"FF Speed", @"2", @"RW Speed", @"10", @"Reverse Jump", @"30", @"Forward Jump", @".3", @"SloMo Speed", @".5", @"Snap Back", nil];
        [defs setObject:controlSpeeds forKey:@"controlSpeeds"];
     }
    
    // BOOLs
    theBoolButtons = [NSArray arrayWithObjects:rememberSwitch, openSizeSwitch, playOn, optAlbum, optAuthor, optComment, optCopy, optCreDate, optFName, optDesc, optDir, optInfo, optPerf, optProd, optWrit, optGenre, optWebsite, saveBeep, askMovies, updateFilename, exitPlay, chapScale, optMerge, floatWindow, hides, brushedMetal, doubleClick, editWarning, customIcon, optIcon, placeHolder, skip, autoComplete, scaleImportedChapters, exitMouse, editChapTimes, scaleDVDChapters, optReduce, dontAnimate, addSeparateAudio, optRating, nil];
    [theBoolButtons retain];
    theBoolKeys = [NSArray arrayWithObjects:t_remember, t_toSize, t_playOn, t_album, t_author, t_comment, t_copy, t_credate, t_fname, t_desc, t_dir, t_info, t_perf, t_prod, t_writ, @"optGenre", @"optWebsite", t_saveBeep, t_ask, @"filename", @"exitPlay", @"chapScale", @"merge", @"float", @"hides", @"brushed", @"doubleClick", @"warn", @"customIcon", @"optIcon", @"placeHolder", @"skip", @"autoComplete", @"scaleImportedChapters", @"exitMouse", @"editChapTimes", @"scaleDVDChapters", @"optReduce", @"dontAnimate", @"addSeparateAudio", @"optRating", nil];
    [theBoolKeys retain];
    n = [theBoolKeys count];
    for (i=0;i<n;i++) {
        [[theBoolButtons objectAtIndex:i] setState:[defs boolForKey:[theBoolKeys objectAtIndex:i]]];
    }
    
    [MHMovie setAsksForMissingFiles:[askMovies state]];
    [MHMovie setIgnoresMissingFiles:![placeHolder state]];
    
    // Strings
    theStringFields = [NSArray arrayWithObjects:saveDirPath, commentText, fsAspect, iconText, skipText, timeFormat, timeDenom, interval, nil];
    [theStringFields retain];
    theStringKeys = [NSArray arrayWithObjects:@"saveDir", t_commentText, @"fsAspect", @"iconText", @"skipText", @"format", @"denom", t_interval, nil];
    [theStringKeys retain];
    NSString *value;
    n = [theStringKeys count];
    for (i=0;i<n;i++) {
        value = [defs objectForKey:[theStringKeys objectAtIndex:i]];
		if (![value isKindOfClass:[NSString class]])
			if ([value respondsToSelector:@selector(stringValue)]) value = [value stringValue];
			else if ([value respondsToSelector:@selector(description)]) value = [value description];
			else value = nil;
        if (value) {
			if ([value hasPrefix:@"{\\rtf"]) {
				NSAttributedString *astring = [[[NSAttributedString alloc] initWithRTF:[NSData dataWithBytes:[value UTF8String] length:[value length]] documentAttributes:nil] autorelease];
				[[theStringFields objectAtIndex:i] setAttributedStringValue:astring];
			} else {
				//NSLog(@"nil value for %@", [theStringKeys objectAtIndex:i]);
				[[theStringFields objectAtIndex:i] setStringValue:value];
			}
        }
    }
	
	// time format box
	//NSRunAlertPanel(@"", @"", @"", @"", @"");
	NSArray *forms = [defs objectForKey:@"timeFormats"];
	if (![forms count]) {
		NSDictionary *dict = [NSDictionary dictionaryWithObject:[NSColor grayColor] forKey:NSForegroundColorAttributeName];
		NSMutableAttributedString *mstring = [[NSMutableAttributedString alloc] initWithString:@"(SRT) hh:mm:ss,xxx"];
		[mstring setAttributes:dict range:NSMakeRange(0, 6)];
		NSMutableAttributedString *mstring2 = [[NSMutableAttributedString alloc] initWithString:@"(SSA) h:mm:ss.xx"];
		[mstring2 setAttributes:dict range:NSMakeRange(0, 6)];
		NSMutableAttributedString *mstring3 = [[NSMutableAttributedString alloc] initWithString:@"(SMI) x"];
		[mstring3 setAttributes:dict range:NSMakeRange(0, 6)];
		NSMutableAttributedString *mstring4 = [[NSMutableAttributedString alloc] initWithString:@"(QT) [hh:mm:ss.xx]"];
		[mstring4 setAttributes:dict range:NSMakeRange(0, 5)];
		forms = [NSArray arrayWithObjects: [mstring4 rtfValue], [mstring rtfValue]
				 , [mstring2 rtfValue], [mstring3 rtfValue], nil];
		[defs setObject:forms forKey:@"timeFormats"];
	}
	[timeFormat addItemsWithRTFValues:forms];
    
    // Doubles
    theDoubleFields = [[NSArray arrayWithObjects:fontSize, rowHeight, nil] retain];
    theDoubleKeys = [[NSArray arrayWithObjects:@"fontSize", @"rowHeight", nil] retain];
    n = [theDoubleKeys count];
    for (i=0;i<n;i++) {
        value = [defs objectForKey:[theDoubleKeys objectAtIndex:i]];
        if (nil != value)
            [[theDoubleFields objectAtIndex:i] setStringValue:value];
    }
	[theTable setRowHeight:[rowHeight floatValue]];
    
    //[openWhichSize selectItemAtIndex:[defs integerForKey:t_whichSize]];
    //[openWhichSize setEnabled:[openSizeSwitch state]];
    autoColumns = [defs integerForKey:t_numberOfAutoColumns];
    [autoCols setIntValue:autoColumns];
    [saveSizeRadio selectCellAtRow:[defs integerForKey:t_saveSize] column:0];
    [saveTypeRadio selectCellAtRow:[defs integerForKey:t_saveType] column:0];
    //[interval setFloatValue:[defs floatForKey:t_interval]];
	id interv = [defs objectForKey:t_interval];
	if (!interv) [interval setStringValue:@"1"];
	else if ([interv isKindOfClass:[NSNumber class]]) [interval setStringValue:[interv stringValue]];
	else if ([interv isKindOfClass:[NSString class]]) [interval setStringValue:interv];
	
    //[ffSpeed setFloatValue:[defs floatForKey:t_ff]];
    [numPerformers setIntValue:[defs integerForKey:t_numPerformers]];
    [pathLevels setIntValue:[defs integerForKey:t_pathLevels]];
    
    if ([floatWindow state])
        [viewWindow setLevel:NSFloatingWindowLevel];
	
	
	NSArray *pops = [NSArray arrayWithObjects:click1, click2, scrollUp, scrollDown, scrollRight, scrollLeft, nil];
	NSArray *skeys = [NSArray arrayWithObjects:@"click1", @"click2", @"scrollUp", @"scrollDown", @"scrollRight", @"scrollLeft", nil];
	int sdefs[6] = {1,0,0,1,5,4};
	n = [pops count];
	for (i=0;i<n;i++) {
		if (![defs objectForKey:[skeys objectAtIndex:i]])
			[[pops objectAtIndex:i] selectItemAtIndex:sdefs[i]];
		else
			[[pops objectAtIndex:i] selectItemAtIndex:[defs integerForKey:[skeys objectAtIndex:i]]];
	}
	
    /*
    // color well nonsense
    float r, g, b, a;
    r = [defs floatForKey:@"hlR"];
    g = [defs floatForKey:@"hlG"];
    b = [defs floatForKey:@"hlB"];
    a = [defs floatForKey:@"hlA"];
    if (0 == r && 0 == g && 0 == b && 0 == a) {
        r = hlR;
        g = hlG;
        b = hlB;
        a = hlA;
    }
    [highlightColor setColor:[NSColor colorWithCalibratedRed:r green:g blue:b alpha:a]];
    */
    /*
     *	Set Up the Timer
     */
    NSTimeInterval timeInterval = [interval floatValue];
    theTimer = [NSTimer scheduledTimerWithTimeInterval:timeInterval target:[controller getPlayer] selector:@selector(tick:) userInfo:nil repeats:YES];
    
    //[self savePrefs];
}

- (BOOL)settingForKey:(NSString *)key
{
    int i, n = [theBoolKeys count];
    for (i=0;i<n;i++) {
        if ([key isEqualToString:[theBoolKeys objectAtIndex:i]])
            return [[theBoolButtons objectAtIndex:i] state];
    }
    return NO;
}

- (void)setSetting:(BOOL)flag forKey:(NSString *)key
{
    int i, n = [theBoolKeys count];
    for (i=0;i<n;i++) {
        if ([key isEqualToString:[theBoolKeys objectAtIndex:i]])
            [[theBoolButtons objectAtIndex:i] setState:flag];
    }
}

- (NSString *)stringSettingForKey:(NSString *)key
{
    int i, n = [theStringKeys count];
    for (i=0;i<n;i++) {
        if ([key isEqualToString:[theStringKeys objectAtIndex:i]])
            return [[theStringFields objectAtIndex:i] stringValue];
    }
    return @"";
}

- (void)setStringSetting:(NSString *)string forKey:(NSString *)key
{
    int i, n = [theStringKeys count];
    for (i=0;i<n;i++) {
        if ([key isEqualToString:[theStringKeys objectAtIndex:i]])
            [[theStringFields objectAtIndex:i] setStringValue:string];
    }
}

- (double)doubleSettingForKey:(NSString *)key
{
    int i, n = [theDoubleKeys count];
    for (i=0;i<n;i++) {
        if ([key isEqualToString:[theDoubleKeys objectAtIndex:i]])
            return [[theDoubleFields objectAtIndex:i] doubleValue];
    }
	n = [theStringKeys count];
	for (i=0;i<n;i++)
		if ([key isEqualToString:[theStringKeys objectAtIndex:i]])
			return [[theStringFields objectAtIndex:i] doubleValue];
    return 0;
}

- (void)setDoubleSetting:(double)value forKey:(NSString *)key
{
    int i, n = [theDoubleKeys count];
    for (i=0;i<n;i++) {
        if ([key isEqualToString:[theDoubleKeys objectAtIndex:i]])
            [[theDoubleFields objectAtIndex:i] setDoubleValue:value];
    }
}

- (IBAction)newTimer:(id)sender
{
    NSTimeInterval timeInterval = [interval floatValue];
    [theTimer invalidate];
    theTimer = [NSTimer scheduledTimerWithTimeInterval:timeInterval target:[controller getPlayer] selector:@selector(tick:) userInfo:nil repeats:YES];
}

- (IBAction)rowHeight:(id)sender
{
	float f = [sender floatValue];
	[theTable setRowHeight:f];
	[theTable reloadData];
	[[NSUserDefaults standardUserDefaults] setFloat:f forKey:@"rowHeight"];
}

- (IBAction)picOptionsMenu:(id)sender
{
	NSEvent *e = [NSApp currentEvent];
	NSMenu *m = [[[NSMenu alloc] init] autorelease];
	
	//[m addItemWithTitle:@"Image Display:" action:nil keyEquivalent:@""];
	//[m addItem:[NSMenuItem separatorItem]];
	
	NSMenuItem *item = [[[NSMenuItem alloc] initWithTitle:@"Image Column" action:@selector(addColumn:) keyEquivalent:@""] autorelease];
	[m addItem:item];
	[item setTarget:controller];
	if ([theTable columnWithIdentifier:@"Image"] != -1)
		[item setState:NSOnState];
	
	item = [[[NSMenuItem alloc] initWithTitle:@"	Preview" action:@selector(usePreview:) keyEquivalent:@""] autorelease];
	[m addItem:item];
	[item setTarget:self];
	BOOL prev = [[NSUserDefaults standardUserDefaults] boolForKey:@"showPreviewPic"];
	BOOL poster = [[NSUserDefaults standardUserDefaults] boolForKey:@"showPosterPic"];
	if (prev && !poster)
		[item setState:NSOnState];
	item = [[[NSMenuItem alloc] initWithTitle:@"	Icon" action:@selector(usePreview:) keyEquivalent:@""] autorelease];
	[m addItem:item];
	[item setTarget:self];
	if (!prev)
		[item setState:NSOnState];
	item = [[[NSMenuItem alloc] initWithTitle:@"	Poster Frame" action:@selector(usePreview:) keyEquivalent:@""] autorelease];
	[m addItem:item];
	[item setTarget:self];
	if (prev && poster)
		[item setState:NSOnState];
	
	MHMovie *movie = [controller currentMovie];
	if (movie && [movie hasImage]) {
		NSString *name = [movie stringValue];
		item = [[[NSMenuItem alloc] initWithTitle:[NSString stringWithFormat:@"Remove Image From %@", name] action:@selector(removeImageForLanguage:) keyEquivalent:@""] autorelease];
		[m addItem:item];
		[item setTarget:movie];
	}
	
	[NSMenu popUpContextMenu:m withEvent:e forView:sender];
	[sender setState:NSOffState];
}

- (BOOL)movieShouldTask:(id)movie
{
	//NSLog(@"%@ %@", [movie className], [[controller playingMovie] className]);
	if ([[controller playingMovie] qtKitMovie] == movie) return NO;
	return YES;
}

- (IBAction)usePreview:(id)sender
{
	BOOL prev = NO, poster = NO;
	if ([[sender title] isEqualToString:@"	Preview"])
		prev = ![sender state];
	else if ([[sender title] isEqualToString:@"	Icon"])
		prev = [sender state];
	else if ([[sender title] isEqualToString:@"	Poster Frame"])
		prev = poster = ![sender state];
	[[NSUserDefaults standardUserDefaults] setBool:prev forKey:@"showPreviewPic"];
	[[NSUserDefaults standardUserDefaults] setBool:poster forKey:@"showPosterPic"];
	[[NSUserDefaults standardUserDefaults] synchronize];
	[theTable reloadData];
	[controller setPicViewPic:[[controller currentMovie] image]];
}

- (id)outlineView:(NSOutlineView *)outlineView child:(int)index ofItem:(id)item
{
    NSDictionary *learned = [[NSUserDefaults standardUserDefaults] objectForKey:@"learned"];
    tOutlineItem *toi = [[tOutlineItem alloc] init];
    [toi setParent:item];
    NSString *value;
    if (!item)
        value = [[learned allKeys] objectAtIndex:index];
    else
        value = [[learned objectForKey:[item value]] objectAtIndex:index];
    [toi setValue:value];
    return toi;
}

- (BOOL)outlineView:(NSOutlineView *)outlineView isItemExpandable:(id)item
{
    return [[[[NSUserDefaults standardUserDefaults] objectForKey:@"learned"] allKeys] containsObject:[item value]];
}

- (int)outlineView:(NSOutlineView *)outlineView numberOfChildrenOfItem:(id)item
{
    NSDictionary *learned = [[NSUserDefaults standardUserDefaults] objectForKey:@"learned"];
    if (!item) return [learned count];
    return [(NSArray *)[learned objectForKey:[item value]] count];
}

- (id)outlineView:(NSOutlineView *)outlineView objectValueForTableColumn:(NSTableColumn *)tableColumn byItem:(id)item
{
    //NSDicitonary *learned = [[NSUserDefaults standardUserDefaults] objectForKey:@"learned"];
    return [item value];
}

- (void)outlineView:(NSOutlineView *)outlineView setObjectValue:(id)object forTableColumn:(NSTableColumn *)tableColumn byItem:(id)item
{

    NSUserDefaults *defs = [NSUserDefaults standardUserDefaults];
    NSMutableDictionary *learned = [NSMutableDictionary dictionaryWithDictionary:[defs objectForKey:@"learned"]];
    NSMutableArray *anArray;
    if ([item parent]) {
        anArray = [NSMutableArray arrayWithArray:[learned objectForKey:[[item parent] value]]];
        [anArray removeObject:[item value]];
        [anArray addObject:object];
        [learned setObject:anArray forKey:[[item parent] value]];
    } else {
        anArray = [NSMutableArray arrayWithArray:[learned objectForKey:[item value]]];
        [learned removeObjectForKey:[item value]];
        [learned setObject:anArray forKey:object];
    }
    [defs setObject:learned forKey:@"learned"];
    [defs synchronize];
}

void socketCallBack (CFSocketRef socketref, CFSocketCallBackType type, CFDataRef address, const void *data, void *info)
{
	tInit *me = (tInit *)info;
	struct kevent event;
	if (kevent(me->kqfd, NULL, 0, &event, 1, NULL) == -1) {
		NSLog(@"could not pick up event.");
		return;
	}
	NSString *path = (NSString *)event.udata;
	if (event.fflags & NOTE_RENAME)
		NSLog(@"rename: %@", path);
	if (event.fflags & NOTE_WRITE || event.fflags & NOTE_EXTEND)
		NSLog(@"write: %@", path);
	if (event.fflags & NOTE_DELETE) {
		// when the file is deleted, if it's re-created then we're going to lose our notices
		NSLog(@"file deleted: %@", path);
		[me->watchlist removeObject:path];
		// if this ever happens, we might want to re-watch the file
	}
	[me fileModified:path];
}
	
	//NSLog(@"(%d) detect write: %@", me->kqfd, path);
- (void)fileModified:(NSString *)path
{
	NSLog(@"modified: %@", path);
	// path is presumably the first watched file in the dir.  let's just check all the files
	NSMutableArray *modFiles = [NSMutableArray array];
	int i, n = [watchlist count];
	for (i=0;i<n;i++) {
		NSDate *newD = [[[NSFileManager defaultManager] fileAttributesAtPath:[watchlist objectAtIndex:i] traverseLink:NO] objectForKey:NSFileModificationDate];
		NSTimeInterval gap = [newD timeIntervalSinceNow];
		if (gap > -2 && [subUpdateDate compare:newD] != NSOrderedDescending) // fresher than 2 seconds and our last update was not more recent than the mod date
			[modFiles addObject:[watchlist objectAtIndex:i]];
		//if (gap > -2) [modFiles addObject:[watchlist objectAtIndex:i]]; // fresher than 2 seconds
	}
	if (![modFiles count]) return;
	
	if (![[NSUserDefaults standardUserDefaults] objectForKey:@"always reimport subs"]) {
		NSString *infoText = [NSString stringWithFormat:@"The subtitles file has just been edited.  Do you want to re-import it?"];
		NSAlert *alert = [NSAlert alertWithMessageText:@"Re-import Subtitles" defaultButton:@"Re-import" alternateButton:@"No" otherButton:nil informativeTextWithFormat:infoText];
		if ([alert respondsToSelector:@selector(setShowsSuppressionButton:)]) {
			[alert setShowsSuppressionButton:YES];
			[[alert suppressionButton] setTitle:@"Stop Asking (Make this selection automatic)"];
		}
		int result = [alert runModal];
		if ([alert respondsToSelector:@selector(suppressionButton)]) {
			if ([[alert suppressionButton] state]) {
				if (result == NSAlertDefaultReturn) 
					[[NSUserDefaults standardUserDefaults] setBool:YES forKey:@"always reimport subs"];
				else if (result == NSAlertAlternateReturn)
					[[NSUserDefaults standardUserDefaults] setBool:NO forKey:@"always reimport subs"];
				[[NSUserDefaults standardUserDefaults] synchronize];
			}
		}
		if (result == NSAlertAlternateReturn) return;
	} else if (![[NSUserDefaults standardUserDefaults] boolForKey:@"always reimport subs"]) return;
	
	if (subUpdateDate) [subUpdateDate performSelector:@selector(release) withObject:nil afterDelay:1];
	subUpdateDate = [[NSDate dateWithTimeIntervalSinceNow:0] retain];
	//n = [modFiles count]; // even though presumably the user only had time to modify one file at a time...
	//for (i=0;i<n;i++) {
		//NSString *noteName = [NSString stringWithFormat:@"MHModified %@", path];
		[[NSNotificationCenter defaultCenter] postNotificationName:@"MHModified" object:modFiles];
	//}
	[[tPlayback nibInstance] movieDidChangeTo:[[tPlayback nibInstance] currentMovie]];
}

- (BOOL)watchFile:(NSString *)path
{
	BOOL isDir = NO;
	if (!path || ![[NSFileManager defaultManager] fileExistsAtPath:path isDirectory:&isDir]) return NO;
	if (!kqfd) {
		kqfd = kqueue ();
		NSLog(@"%d", kqfd);
	}
	if ([watchlist containsObject:path]) return NO;
	if (!watchlist) watchlist = [[NSMutableArray alloc] initWithObjects:path, nil];
	else [watchlist addObject:path];
	
	int filefd = open ([path UTF8String], O_RDONLY);
	struct kevent fevent;
	EV_SET (&fevent, filefd, EVFILT_WRITE, EV_ADD | EV_CLEAR | EV_ENABLE, NOTE_WRITE | NOTE_DELETE | NOTE_EXTEND, 0, [path copy]);
	if (kevent(kqfd, &fevent, 1, NULL, 0, NULL) == -1) NSLog(@"could not kevent watching %@", path);
	
	NSString *dir = path;
	if (!isDir) dir = [path stringByDeletingLastPathComponent];
	BOOL found = NO;
	int i, n = [watchlist count]-1;
	for (i=0;i<n;i++)
		if ([[watchlist objectAtIndex:i] hasPrefix:dir]) // already watching this dir
			found = YES;
	if (!found) {
		int dirfd = open ([dir UTF8String], O_RDONLY);
		//struct kevent fevent;
		EV_SET (&fevent, dirfd, EVFILT_VNODE, EV_ADD | EV_CLEAR | EV_ENABLE, NOTE_WRITE, 0, [dir copy]);
		if (kevent(kqfd, &fevent, 1, NULL, 0, NULL) == -1) NSLog(@"could not kevent watching %@.", dir);
	}
	if (!runLoopSocket) {
		CFSocketContext context = { 0, self, NULL, NULL, NULL };
		runLoopSocket = CFSocketCreateWithNative(NULL, kqfd, kCFSocketReadCallBack, socketCallBack, &context);
		if (!runLoopSocket) NSLog(@"could not CFSocketCreateWithNative");
		CFRunLoopSourceRef rls = CFSocketCreateRunLoopSource(NULL, runLoopSocket, 0);
		if (!rls) NSLog(@"could not create a run loop source");
		CFRunLoopAddSource(CFRunLoopGetCurrent(), rls, kCFRunLoopDefaultMode);
		CFRelease (rls);
	}
	return YES;
}

-(IBAction)forget:(id)sender
{
	while ([timeFormat numberOfItems] > 4)
		[timeFormat removeItemAtIndex:[timeFormat numberOfItems]-1];
	[[NSUserDefaults standardUserDefaults] removeObjectForKey:@"timeFormats"];
	[[NSUserDefaults standardUserDefaults] synchronize];
}

-(IBAction)suppression:(id)sender
{
	NSArray *strings = [NSArray arrayWithObjects:@"Save in place", @"Multiple selection", @"Unwritable", @"Spotlight bug", @"Reload subtitles", nil];
	NSArray *defaults = [NSArray arrayWithObjects:@"noSaveWarn", @"MultiSelNoAsk", @"warn", @"ignoreSpotlight", @"always reimport subs", nil];
	NSUserDefaults *defs = [NSUserDefaults standardUserDefaults];
	int i, n = [strings count];
	if ([sender isKindOfClass:[NSButton class]]) {
		NSMenu *menu = [[[NSMenu alloc] init] autorelease];
		for (i=0;i<n;i++) {
			NSMenuItem *item = [[[NSMenuItem alloc] initWithTitle:[strings objectAtIndex:i] action:@selector(suppression:) keyEquivalent:@""] autorelease];
			[item setTarget:self];
			[menu addItem:item];
			NSNumber *set = [defs objectForKey:[defaults objectAtIndex:i]];
			if ([set boolValue] || (i==4 && set)) [item setState:NSOnState];
		}
		[NSMenu popUpContextMenu:menu withEvent:[NSApp currentEvent] forView:sender];
	} else if ([sender isKindOfClass:[NSMenuItem class]]) {
		i = [[sender menu] indexOfItem:sender];
		if ([sender state]) [defs removeObjectForKey:[defaults objectAtIndex:i]];
		else [defs setBool:YES forKey:[defaults objectAtIndex:i]];
		[defs synchronize];
	}
}

@end
