/* Copyright (c) 2004, Noah Sorscher
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 	 	Redistributions of source code must retain the above copyright
 * 	notice, this list of conditions and the following disclaimer.
 * 	 	Redistributions in binary form must reproduce the above copyright
 * 	notice, this list of conditions and the following disclaimer in the
 * 	documentation and/or other materials provided with the distribution.
 * 	 	Neither the name of the Noah Sorscher nor the names of its
 * 	contributors may be used to endorse or promote products derived from
 * 	this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
#import "tMovieView.h"
#import "tWindow.h"
#import "MHRectView.h"

@implementation tMovieView


- (void)drawRectangle:(NSRect)r
{
	if (!r.size.width && !r.size.height) {
		[square orderOut:self];
		return;
	}
	NSRect frame = [[self window] frame];
	frame.size = [self frame].size;
	if ([self isControllerVisible]) {
		frame.origin.y += 16;
		frame.size.height -= 16;
	}
	if (!square) {
		MHRectView *v = [[MHRectView alloc] initWithFrame:frame];
		square = [[NSWindow alloc] initWithContentRect:frame styleMask:NSBorderlessWindowMask backing:NSBackingStoreBuffered defer:NO screen:[[self window] screen]];
		[square setContentView:v];
		[square setBackgroundColor:[NSColor colorWithCalibratedWhite:0.0 alpha:0.0]];
		[square setOpaque:NO];
		[square setIgnoresMouseEvents:NO];
		[square setLevel:[[self window] level]];
	}
	[square setFrame:frame display:YES];
	NSRect c, m = [(MHMovie *)[[self movie] delegate] displayFrame];
	c.size.width = frame.size.width / m.size.width;
	c.size.height = frame.size.height / m.size.height;
	c.origin.x = m.origin.x;
	c.origin.y = m.origin.y - m.size.height;
	[[square contentView] setCoord:c];
	[[square contentView] setRect:r];
	
	[square makeKeyAndOrderFront:self];
	[square orderWindow:NSWindowAbove relativeTo:[[self window] windowNumber]];
}

- (void)showNote:(NSString *)s size:(float)f
{
	if (!s || ![s length]) {
		[note orderOut:self];
		return;
	}
	if (![[self window] isVisible]) return;
	NSButton *butt = nil;
	if (note) {
		butt = [note contentView];
	} else {
		butt = [[NSButton alloc] init];
		[butt setBezelStyle:NSRecessedBezelStyle];
	}
	[butt setFont:[NSFont labelFontOfSize:f]];
	[butt setTitle:s];
	NSRect a = [self frame], b = [[self window] frame];
	a.origin.x += b.origin.x + a.size.width;
	a.origin.y += b.origin.y + a.size.height;
	a.size = [[butt cell] cellSize];
	//if (a.size.width < 100) a.size.width = 100;
	//if (a.size.height < 100) a.size.height = 100;
	a.origin.x -= a.size.width;
	a.origin.y -= a.size.height;
	if (!note) {
		note = [[NSWindow alloc] initWithContentRect:a styleMask:NSBorderlessWindowMask backing:NSBackingStoreBuffered defer:NO screen:[[self window] screen]];
		[note setContentView:butt];
		[note setBackgroundColor:[NSColor colorWithCalibratedWhite:0.0 alpha:0.0]];
		[note setOpaque:NO];
	} else {
		[note setFrame:a display:YES];
	}
	
	[note setLevel:[[self window] level]];
	[note makeKeyAndOrderFront:self];
	[note orderWindow:NSWindowAbove relativeTo:[[self window] windowNumber]];
	[NSObject cancelPreviousPerformRequestsWithTarget:self selector:@selector(orderOut:) object:nil];
	[note performSelector:@selector(orderOut:) withObject:self afterDelay:2];
}

- (void)showNote:(NSString *)s
{
	float f = 20;
	if ([s characterAtIndex:0] > 10553) f = 50;
	[self showNote:s size:f];
}

- (void)setRate:(float)rate
{
	char *symbol = "\x25\xB6";
#ifdef __LITTLE_ENDIAN__
	symbol = "\xB6\x25";
#endif
	NSString *s = [NSString stringWithCharacters:(unichar *)symbol length:1];
	if (rate != 1)
		s = [NSString stringWithFormat:@" %@%.2gx", s, rate];
	else
		s = [NSString stringWithFormat:@" %@", s];
	[self showNote:s];
	[[self movie] setRate:rate];
}

- (void)slomo:(id)sender
{
	[[tPlayback nibInstance] slomo:sender];
}

- (void)noteSound
{
	char *symbol = "\x26\x6A";
#ifdef __LITTLE_ENDIAN__
	symbol = "\x6A\x26";
#endif
	NSString *s = [NSString stringWithCharacters:(unichar *)symbol length:1];
	float v = [[self movie] volume];
	if (v < 99)
		s = [NSString stringWithFormat:@"%@ %.2g", s, v];
	else
		s = [NSString stringWithFormat:@"%@ %.3g", s, v];
	[self showNote:s];
}

- (void)scrollWheel:(NSEvent *)e
{
	//NSLog(@"scroll: %g %g %g", [e deltaX], [e deltaY], [e deltaZ]);
	NSString *action = nil;
	char *symbol = nil;
	NSDictionary *dict = [[NSUserDefaults standardUserDefaults] objectForKey:@"controlSpeeds"];
	//NSLog([e description]);
	float biggest = 0, candidate = 0;
	BOOL neg = NO;
	candidate = [e deltaY];
	if (candidate < 0) {
		candidate = 0 - candidate;
		neg = YES;
	}
	if (candidate > biggest) {
		biggest = candidate;
		if (neg) action = @"Scroll Down";
		else action = @"Scroll Up";
	}
	candidate = [e deltaX];
	neg = NO;
	if (candidate < 0) {
		candidate = 0 - candidate;
		neg = YES;
	}
	if (candidate > biggest) {
		biggest = candidate;
		if (neg) action = @"Scroll Right";
		else action = @"Scroll Left";
	}
	if (action) {
		[[NSApp delegate] doControl:action];
		/*
		double seconds = 0;
	
		switch ([[NSUserDefaults standardUserDefaults] integerForKey:action]) {
		case 0: [self stepForward:self]; break;
		case 1: [self stepBackward:self]; break;
		case 2: [self setRate: [[dict objectForKey:@"FF Speed"] floatValue]]; break;
		case 3: [self setRate: 0-[[dict objectForKey:@"RW Speed"] floatValue]]; break;
		case 4: seconds = [[[self movie] delegate] currentTimeIndexSeconds];
			seconds += [[dict objectForKey:@"Forward Jump"] floatValue];
			[[[self movie] delegate] setCurrentTimeIndexSeconds:seconds];
				symbol = "\x29\x3B";
#ifdef __LITTLE_ENDIAN__
				symbol = "\x3B\x29";
#endif
				[self showNote:[NSString stringWithCharacters:(unichar *)symbol length:1] size:50];
				break;
		case 5: seconds = [[[self movie] delegate] currentTimeIndexSeconds];
			seconds -= [[dict objectForKey:@"Reverse Jump"] floatValue];
			[[[self movie] delegate] setCurrentTimeIndexSeconds:seconds];
				symbol = "\x29\x3A";
#ifdef __LITTLE_ENDIAN__
				symbol = "\x3A\x29";
#endif
				[self showNote:[NSString stringWithCharacters:(unichar *)symbol length:1] size:50];
				break;
		case 6: [[self movie] setVolume:[[self movie] volume]+0.2];
				[self noteSound]; break;
		case 7: [[self movie] setVolume:[[self movie] volume]-0.2];
				[self noteSound]; break;
		case 8: // do nothing
		default: break;
		}
		 */
	}
	[[tPlayback nibInstance] tick:nil];
	[tPlayback performSelector:@selector(refreshVolume:) withObject:nil afterDelay:0];
	[[self window] mouseMoved:e];
}

- (void)setMovie:(id)movie
{
	if ([movie isKindOfClass:[QTMovie class]]) {
		[super setMovie:movie];
	} else if ([movie isKindOfClass:[NSMovie class]]) {
		if ([movie respondsToSelector:@selector(qtKitMovie)]) {
			[super setMovie:[movie qtKitMovie]];
		}
	}
	[[self movie] setAttribute:[NSNumber numberWithBool:YES] forKey:QTMovieEditableAttribute];
}

- (void)showController:(BOOL)show adjustingSize:(BOOL)adjustSize
{
	[super setControllerVisible:show];
}

- (float) rate
{
	return [[self movie] rate];
}

- (BOOL)isPlaying
{
	return [[self movie] rate] != 0;
}

- (void)start:(id)sender
{
	[self play:sender];
}

- (void)stop:(id)sender
{
	[self pause:sender];
}

- (void)pause:(id)sender
{
	char *symbol = "\x27\x59";
#ifdef __LITTLE_ENDIAN__
	symbol = "\x59\x27";
#endif
	NSString *s = [NSString stringWithCharacters:(unichar *)symbol length:1];
	s = [NSString stringWithFormat:@"%@%@", s, s];
	[self performSelector:@selector(showNote:) withObject:s afterDelay:0];
	[super pause:sender];
	if ([[[self movie] delegate] respondsToSelector:@selector(thumb)])
		[[[self movie] delegate] thumb];
}

- (void)play:(id)sender
{
	float rate = 1;
	if ([[NSUserDefaults standardUserDefaults] objectForKey:@"playRate"]) rate = [[NSUserDefaults standardUserDefaults] floatForKey:@"playRate"];
	//NSLog(@"rate: %g", rate);
	[self setRate:rate];
	return;
}

- (void) awakeFromNib
{
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(closeWindow:)  name:@"movieWindowCloseNotification" object:nil];
	note = square = nil;
}

- (void) closeWindow:(NSNotification *)note
{
    //NSLog(@"close");
    [[self window] close];
}
/*
- (void)keyDown:(NSEvent *)e
{
	[self performKeyEquivalent:e];
}
*/
- (BOOL)performKeyEquivalent:(NSEvent *)theEvent
{
    NSString *chars;
    unichar uch;
	float rate = 0;
	
	if ([theEvent type] != NSKeyDown) return [super performKeyEquivalent:theEvent]; // this is why we get doubles
	
    chars = [theEvent characters];
	//NSLog(@"tmovieview: \"%@\", %d", chars, [theEvent type]);
    //result = [super performKeyEquivalent:theEvent];
    uch = [chars characterAtIndex:0];
    //if ([theEvent modifierFlags] & NSCommandKeyMask ||
      //  [theEvent modifierFlags] & NSControlKeyMask) // ||
        //[theEvent modifierFlags] & NSAlternateKeyMask)
        //    return NO;
	QTMovie *q = [self movie];
	MHMovie *movie = [q delegate];
    BOOL forward = NO;
	BOOL outty = NO;
	NSString *arrow = nil;
	int flags = [theEvent modifierFlags];
    switch (uch) {
		case ' ':
			rate = [q rate];
			if (rate) [self pause:self];
			else [self play:self];
			return YES;
		case 'f': case 27: // esc
			[tPlayback fullScreen];
			return YES;
		case '\t':
			if ([theEvent modifierFlags] & NSAlternateKeyMask)
				[tPlayback lastMovie];
			else
				[tPlayback nextMovie];
			return YES;
		case 'o':
			outty = YES;
		case 'i':
			if (/*doubles &&*/ q) {
				QTTime a = [q selectionStart], b = [q selectionEnd], c = [q currentTime];
				if (outty) { // we're setting the out point
					if (a.timeValue > c.timeValue) { // if the new out point is before the old in point
						a = b = c;
					} else { // if the new out point is after the old in point
						b = c;
					}
				} else { // we're setting the in point
					if (b.timeValue < c.timeValue) { // if the new in point is after the old out point
						a = b = c;
					} else {
						a = c;
					}
				}
				b.timeValue -= a.timeValue; // convert an end time into a duration
				[q setSelection:QTMakeTimeRange(a, b)];
			}
			//doubles = !doubles;
			return YES;
			
		case NSRightArrowFunctionKey: arrow = @"Right Arrow";
		case NSLeftArrowFunctionKey: if (!arrow) arrow = @"Left Arrow";
		case NSUpArrowFunctionKey: if (!arrow) arrow = @"Up Arrow";
		case NSDownArrowFunctionKey: if (!arrow) arrow = @"Down Arrow";
			doubles = !doubles;
			//if (doubles) return YES;
			if (flags & NSCommandKeyMask) arrow = [NSString stringWithFormat:@"%C %@", 8984, arrow];
			else if (flags & NSAlternateKeyMask) arrow = [NSString stringWithFormat:@"%C %@", 8997, arrow];
			else if (flags & NSShiftKeyMask) arrow = [NSString stringWithFormat:@"%C %@", 8679, arrow];
			else if (flags & NSControlKeyMask) arrow = [NSString stringWithFormat:@"^ %@", arrow];
			return [[NSApp delegate] doControl:arrow];
        /*case NSRightArrowFunctionKey: forward = YES;
        case NSLeftArrowFunctionKey:
            if (doubles && [theEvent modifierFlags] & NSAlternateKeyMask) {
                if ([movie isKindOfClass:[MHMovie class]]) {
                    double seconds = [movie currentTimeIndexSeconds];
                    if (forward) {
                        NSString *lang = nil;
                        if ([[movie chapters] multilingual]) lang = [movie currentLanguage];
                        seconds = [[movie chapters] nextTimeForTime:seconds language:lang];
                        if (seconds == -1) seconds = [movie durationSeconds] - 0.5;
                    } else
                        seconds = [[movie chapters] prevTimeForTime:seconds];
                    [movie setCurrentTimeIndexSeconds:seconds];
                } else return NO;
            } else if (doubles) {
                if (forward) [self stepForward:nil];
                else [self stepBackward:nil];
            }
            doubles = !doubles;
            return YES;
		case NSUpArrowFunctionKey: forward = YES;
		case NSDownArrowFunctionKey:
			if ([theEvent modifierFlags] & NSCommandKeyMask) return NO;
			if (doubles) {
				short vol = GetMovieVolume([movie QTMovie]);
				if (forward) vol += 50;
				else vol -= 50;
				if (vol < 0) vol = 0;
				SetMovieVolume([movie QTMovie], vol);
				[self noteSound];
			}
			doubles = !doubles;
			[tPlayback performSelector:@selector(refreshVolume:) withObject:nil afterDelay:0];
			return YES;*/
    }
    return [super performKeyEquivalent:theEvent];
}

- (void)mouseDragged:(NSEvent *)e
{
	[[self window] mouseDragged:e];
	[super mouseDragged:e];
}

- (void)mouseDown:(NSEvent *)event
{
	//NSLog(@"tMovieView mouseDown:");
	NSPoint p = [NSEvent mouseLocation];
	p.y -= [[self window] frame].origin.y;
	if ([self isControllerVisible] && p.y < 17) {
		[super mouseDown:event];
		return;
	}
	if ([[self window] respondsToSelector:@selector(reservedFrame)] && NSPointInRect([NSEvent mouseLocation], [(tWindow *)[self window] reservedFrame])) {
		[[self window] makeFirstResponder:[self window]];
		return;
	}
	//NSLog(@"%g, %g", p.x, p.y);
	float rate = 0;
    if ([event modifierFlags] & NSControlKeyMask)
        [self rightMouseDown:event];
    else {//[super mouseDown:event];
		NSString *action = nil;
		if ([event clickCount] == 1) action = @"Click";
		else if ([event clickCount] > 1) action = @"Double Click";
		if (![[NSApp delegate] doControl:action]) [super mouseDown:event];
		/*
		//NSLog(action);
		switch ([[NSUserDefaults standardUserDefaults] integerForKey:action]) {
			case 0: 
				if ([action isEqualToString:@"click2"] && [[NSUserDefaults standardUserDefaults] integerForKey:@"click1"] == 1)
					[self play:self];
				[tPlayback fullScreen]; break;
				
			case 1: rate = [[self movie] rate];
				if (rate) [self pause:self];
				else [self play:self];
				break;
			case 2: // do nothing
				[super mouseDown:event];
			default: break;
		}*/
	}
	if ([[self window] isKindOfClass:[tWindow class]])
		[[self window] makeFirstResponder:[self window]];
}

- (NSMenu *)menuForEvent:(NSEvent *)e
{
	//NSLog(@"menu");
	return nil;
}

- (void)rightMouseDown:(NSEvent *)theEvent
{
	//NSLog(@"right");
    NSMenu *menu = [[[NSMenu alloc] init] autorelease];
    NSMenuItem *item = [[[NSMenuItem alloc] initWithTitle:@"Play" action:@selector(play:) keyEquivalent:@""] autorelease];
    [item setTarget:self];
    [menu addItem:item];
    item = [[[NSMenuItem alloc] initWithTitle:@"Stop" action:@selector(pause:) keyEquivalent:@""] autorelease];
    [item setTarget:self];
    [menu addItem:item];
	NSString *title = nil;
	if (![self frame].origin.x) {  // hacky way to say if we're in the viewWindow
		title = @"Show Controller";
		if ([self isControllerVisible])
			title = @"Hide Controller";
		item = [[[NSMenuItem alloc] initWithTitle:title action:@selector(toggleController:) keyEquivalent:@""] autorelease];
		[item setTarget:[tPlayback nibInstance]];
		[menu addItem:item];
	}
	NSArray *tracks = [[[self movie] delegate] tracksOfType:@"Sound"];
	int i, n = [tracks count];
	if (n > 1) {
		NSMenu *submenu = [[[NSMenu alloc] init] autorelease];
		item = [[[NSMenuItem alloc] initWithTitle:@"Language" action:nil keyEquivalent:@""] autorelease];
		[item setSubmenu:submenu];
		[menu addItem:item];
		for (i=0;i<n;i++) {
			MHTrack *track = [tracks objectAtIndex:i];
			NSString *name = [track name];
			if (!name) name = [NSString stringWithFormat:@"Sound Track %d", i+1];
			item = [[[NSMenuItem alloc] initWithTitle:name action:@selector(setSoundTrack:) keyEquivalent:@""] autorelease];
			if ([track enabled]) [item setState:YES];
			[item setTarget:[tPlayback nibInstance]];
			[submenu addItem:item];
		}
	}
	
    if ([[[self movie] delegate] respondsToSelector:@selector(chapters)] && [(MHChapters *)[(MHMovie *)[[self movie] delegate] chapters] count]) {
		NSMenu *submenu = [[[NSMenu alloc] init] autorelease];
		item = [[[NSMenuItem alloc] initWithTitle:@"Chapters" action:nil keyEquivalent:@""] autorelease];
		[item setSubmenu:submenu];
		[menu addItem:item];
		NSString *lang = nil;
		if ([[(MHMovie *)[[self movie] delegate] chapters] multilingual]) lang = [(MHMovie *)[[self movie] delegate] currentLanguage];
		NSArray *names = [[(MHMovie *)[[self movie] delegate] chapters] chapterNamesForLanguage:lang];
		n = [names count];
		for (i=0;i<n;i++) {
			item = [[[NSMenuItem alloc] initWithTitle:[names objectAtIndex:i] action:@selector(gotoChapter:) keyEquivalent:@""] autorelease];
			[item setTarget:[tPlayback nibInstance]];
			[submenu addItem:item];
		}
    }
	
	if ([[self window] isKindOfClass:[tWindow class]])
		title = @"Exit Fullscreen";
	else title = @"Enter Fullscreen";
	item = [[[NSMenuItem alloc] initWithTitle:title action:@selector(fullScreen:) keyEquivalent:@""] autorelease];
	[item setTarget:[tPlayback nibInstance]];
	[menu addItem:item];
    
    [NSMenu popUpContextMenu:menu withEvent:theEvent forView:self];
	if ([[self window] isKindOfClass:[tWindow class]])
		[[self window] makeFirstResponder:[self window]];
}

@end
