/* Copyright (c) 2004, Noah Sorscher
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 	 	Redistributions of source code must retain the above copyright
 * 	notice, this list of conditions and the following disclaimer.
 * 	 	Redistributions in binary form must reproduce the above copyright
 * 	notice, this list of conditions and the following disclaimer in the
 * 	documentation and/or other materials provided with the distribution.
 * 	 	Neither the name of the Noah Sorscher nor the names of its
 * 	contributors may be used to endorse or promote products derived from
 * 	this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
#import <QuickTime/QuickTime.h>
#import "tPlayback.h"
//#import "Movies.h"
//#include "tModel.h"
//#include "QTFullScreen.h"
#import "tWindow.h"
#import "tChapters.h"
#import "tController.h"
#import "tInit.h"
#import "MHHoldButton.h"
#import "MHMenuButton.h"
#import "OverlayView.h"

@interface tPlayback (private)
- (void)setPlaying:(BOOL)playing;
@end

@interface NSCell (extra)
- (void)setImageRow:(int)row;
@end

@interface NSView (delay)
- (void)setFrameV:(NSValue *)rectValue;
@end
@implementation NSView (delay)
- (void)setFrameV:(NSValue *)rectValue
{	[self setFrame:[rectValue rectValue]];	}
@end


@implementation tPlayback

/////////////////////
//
//	GLOBALS
//
////////////////

//int currentPlayingMovieID;

id tPlayNibInstance;

//////////////////////////////////
//
//	FUNCTION DECLARATIONS
//
////////////////////////////

void setTimeIndex(MHMovie *theMovie, double timeIndex);

//////////////////////////////////
//
//	Methods
//
//////////////////////////////

+ (tPlayback *)nibInstance
{
    return tPlayNibInstance;
}

- (NSSize)windowWillResize:(NSWindow *)sender toSize:(NSSize)size
{
	[self windowWillMove:nil];
    //printf("%gx%g --> ", size.width, size.height);
    if (sender != viewWindow) return size;
    unsigned int currentFlags = [[NSApp currentEvent] modifierFlags];
    if (currentFlags & NSShiftKeyMask || !currentMovie) return size;
    BOOL vis = [movieView isControllerVisible];
    int width = [[currentMovie objectForKey:@"Correct Width"] intValue];
    if (vis) width -= 10;
    int height = [[currentMovie objectForKey:@"Correct Height"] intValue];
    if (vis) height -= 16;
    size.height = size.width * (double)height / (double)width;
    if (vis) {
        //size.width += 10;
        size.height += 16;
    }
    size.height += 22;
    //printf("%gx%g\n", size.width, size.height);
    return size;
}

- (void)windowWillMove:(NSNotification *)n
{
	[movieView showNote:nil];
	[movieView drawRectangle:NSMakeRect(0, 0, 0, 0)];
	[windowedMovieView showNote:nil];
}

-(BOOL)windowShouldClose:(id)sender
{
    if (sender == viewWindow)
        [movieView pause:self];
    else if (sender == [controller mainWindow])
        [windowedMovieView pause:self];
	[self windowWillMove:nil];
    //[tracksDrawer close];
    [self setPlaying:NO];
    if (nil == currentMovie) return YES;
    [currentMovie saveCurrentHeight];
	TimeValue t = [self getSampleTimeValue];
	if (t) [currentMovie setTimeIndex:t];
    return YES;
}

- (NSView *)overlayViewWithParent:(NSWindow *)parent rect:(NSRect)rect
{
    NSWindow *newWindow = [[NSWindow alloc] initWithContentRect:rect styleMask:NSBorderlessWindowMask backing:NSBackingStoreBuffered defer:NO];
    [newWindow setBackgroundColor:[NSColor clearColor]];
    [newWindow setAlphaValue:1.0];
    [newWindow setOpaque:NO];
    [newWindow setHasShadow:NO];
    [newWindow setContentView:[[OverlayView alloc] initWithFrame:[[newWindow contentView] frame]]];
    [newWindow setIgnoresMouseEvents:YES];
    if (parent) [parent addChildWindow:newWindow ordered:NSWindowAbove];
    return [newWindow contentView];
}

- (void)keyDown:(NSEvent *)e
{
    if ([e isKindOfClass:[NSNotification class]]) {
        NSString *chars = [[NSApp currentEvent] characters];
		//NSLog(@"playback: \"%@\"", chars);
        if ([chars length] && ' ' == [chars characterAtIndex:0])
            [playButton performClick:self];
        else
            [[(NSNotification *)e object] keyDown:(NSEvent *)self];
    }
}

- (void)flagsChanged:(NSEvent *)event
{
    if ([event isKindOfClass:[NSNotification class]])
        event = [(NSNotification *)event object];
    if (!event) return;
    unsigned int flags = [event modifierFlags];
	int i, j, m, n;
		NSArray *apics = [NSArray arrayWithObjects:@"ff", @"play2", @"rw", @"jumpB", @"jumpF", nil];
		NSArray *bpics = [NSArray arrayWithObjects:@"fff", @"slow", @"rrw", @"repeat", @"advance", nil];
		NSArray *atips = [NSArray arrayWithObjects:@"Fast Forward", @"Play", @"Rewind", @"Replay", @"Advance", @"Pause", nil];
		NSArray *btips = [NSArray arrayWithObjects:@"Faster Forward", @"Slow", @"Faster Rewind", @"Last Movie", @"Next Movie", nil];
		NSArray *aactions = [NSArray arrayWithObjects:@"ff:", @"play:", @"rw:", @"jumpB:", @"jumpF:", nil];
		NSArray *bactions = [NSArray arrayWithObjects:@"fff:", @"slomo:", @"rrw:", @"lastMovie:", @"nextMovie:", nil];
		
	NSArray *views = [NSArray arrayWithObjects:movieControls, floaterView, nil];
	m = [views count];
	for (j=0;j<m;j++) {
		NSArray *s = [[views objectAtIndex:j] subviews];
		n = [s count];
		for (i=0;i<n;i++) {
			NSButton *b = [s objectAtIndex:i];
			if ([b isKindOfClass:[MHHoldButton class]] && [(MHHoldButton *)b isDepressed]) continue;
			if ([b isKindOfClass:[NSButton class]]) {
				NSString *t = [b toolTip];
				if (t) {
					int x = [atips indexOfObject:t];
					if (x != NSNotFound) {
						if (x == 5) x = 1; // "Pause" -> "Play"
						if (flags & NSAlternateKeyMask) {
							[b setImage:[NSImage imageNamed:[bpics objectAtIndex:x]]];
							[b setToolTip:[btips objectAtIndex:x]];
							[b setAction:NSSelectorFromString([bactions objectAtIndex:x])];
							if ([b isKindOfClass:[MHHoldButton class]])
								[b setNormal:YES];
						}
					} else {
						x = [btips indexOfObject:t];
						if (x != NSNotFound) {
							if (flags & NSAlternateKeyMask) {
								//NSLog(@"huh?");
							} else {
								[b setImage:[NSImage imageNamed:[apics objectAtIndex:x]]];
								[b setToolTip:[atips objectAtIndex:x]];
								[b setAction:NSSelectorFromString([aactions objectAtIndex:x])];
								if ([b isKindOfClass:[MHHoldButton class]])
									[b setNormal:NO];
							}
						}
					}
				}
			}
		}
	}
}

- (NSRect)windowWillUseStandardFrame:(NSWindow *)sender defaultFrame:(NSRect)defaultFrame
{
	if (![miniWin isVisible]) return defaultFrame;
    [miniWin orderOut:self];
    //[movieControls removeFromSuperview];
    //[movieController removeFromSuperview];
    //[controller refreshToolbarItems:toollist];
	
	NSToolbar *tool = [[controller mainWindow] toolbar];
	int i, n = [[[miniWin toolbar] items] count];
	for (i=0;i<n;i++)
		[[miniWin toolbar] removeItemAtIndex:0];
	while ([[tool items] count])
		[tool removeItemAtIndex:0];
	n = [toollist count];
	for (i=0;i<n;i++)
		[tool insertItemWithItemIdentifier:[toollist objectAtIndex:i] atIndex:0];
	
    [[controller mainWindow] makeKeyAndOrderFront:nil];
    return [miniWin frame];
}

- (void)windowDidBecomeMain:(NSNotification *)aNotification
{
    if ([aNotification object] == miniWin) [controller windowDidBecomeMain:aNotification];
}

- (void)windowDidResignMain:(NSNotification *)aNotification
{
    if ([aNotification object] == miniWin) [controller windowDidResignMain:aNotification];
}

- (void)hideMini
{
	[miniWin orderOut:self];
}

- (NSWindow *)showMini
{
	NSToolbar *mtool = [[controller mainWindow] toolbar];
	int i;
	[toollist removeAllObjects];
	for (i=[[mtool items] count]-1;i>=0;i--) {
		NSString *s = [[[mtool items] objectAtIndex:i] itemIdentifier];
		[toollist addObject:s];
		if ([s isEqualToString:@"Movie Controller"] || [s isEqualToString:@"Movie Controls"])
			[mtool removeItemAtIndex:i];
	}
	
    //if (![miniWin toolbar]) {
        NSToolbar *toolbar = [[[NSToolbar alloc] initWithIdentifier:@"miniWindowToolbar"] autorelease];
        [toolbar setDelegate:controller];
        [toolbar setAllowsUserCustomization:NO];
        [toolbar setAutosavesConfiguration:NO];
        [toolbar setDisplayMode:NSToolbarDisplayModeIconOnly];
        [miniWin setToolbar:toolbar];
		[toolbar insertItemWithItemIdentifier:@"Movie Controller" atIndex:0];
		[toolbar insertItemWithItemIdentifier:@"Movie Controls" atIndex:0];
    //}
    NSRect norm = [[controller mainWindow] frame];
    NSRect mini = [miniWin frame];
    mini.origin.x = norm.origin.x;
    mini.origin.y = norm.origin.y + norm.size.height - mini.size.height;
    [miniWin setFrame:mini display:YES];
	[miniWin setLevel:NSNormalWindowLevel];
    [miniWin makeKeyAndOrderFront:self];
	return miniWin;
}

BOOL ocontrolsShown = NO;

- (void)showOverlayControls
{
    return;
    if (ocontrolsShown) return;
    ocontrolsShown = YES;
    float w = 300, h = 100;
    NSRect frame = [bigWindow frame];
    //NSRect sframe = [[bigWindow screen] frame];
    frame.origin.x = frame.size.width / 2 - w / 2;
    frame.origin.y = frame.size.height / 3 - h / 2;
    //frame.origin.x = frame.origin.y = 0;
    frame.size.width = w;
    frame.size.height = h;
    NSView *oview = [bigWindow overlayViewWithRect:frame];
    [oview addSubview:movieController];
    [movieController setFrame:NSMakeRect(0,0,w,h)];
}

- (int)numberOfRowsInTableView:(NSTableView *)aTableView
{
    return 1;
}

- (id)tableView:(NSTableView *)aTableView objectValueForTableColumn:(NSTableColumn *)aTableColumn row:(int)rowIndex
{
	[aTableView setRowHeight:27];
	NSString *iden = [aTableColumn identifier];
	if ([iden hasSuffix:@"Jump"]) {
		NSCell *cell = [aTableColumn dataCell];
		if (![cell isKindOfClass:[NSComboBoxCell class]]) {
			[aTableColumn setDataCell:[[NSComboBoxCell alloc] init]];
			[[aTableColumn dataCell] setEditable:YES];
			[[aTableColumn dataCell] addItemsWithObjectValues:[NSArray arrayWithObjects:@"7", @"15", @"30", @"Chapter", nil]];
		}
	}
    NSDictionary *dict = [[NSUserDefaults standardUserDefaults] objectForKey:@"controlSpeeds"];
    return [dict objectForKey:iden];
}

- (void)tableView:(NSTableView *)aTableView setObjectValue:(id)anObject forTableColumn:(NSTableColumn *)aTableColumn row:(int)rowIndex
{
    NSMutableDictionary *dict = [NSMutableDictionary dictionaryWithDictionary:[[NSUserDefaults standardUserDefaults] objectForKey:@"controlSpeeds"]];
    if (anObject)
        [dict setObject:anObject forKey:[aTableColumn identifier]];
    else
        [dict removeObjectForKey:[aTableColumn identifier]];
    [[NSUserDefaults standardUserDefaults] setObject:dict forKey:@"controlSpeeds"];
    [[NSUserDefaults standardUserDefaults] synchronize];
}

- (NSWindow *)fullscreenWindow
{
    return bigWindow;
}

- (IBAction)scrubberAction:(id)sender
{
    if (!currentMovie) return;
    float value = [sender floatValue];
    TimeRecord tr;
	//NSMovie *movie = [playingMovieView movie];
	Movie q = [currentMovie QTMovie]; // [movie quickTimeMovie];
    GetMovieTime(q, &tr);
    TimeValue dur = GetMovieDuration(q);
    tr.value.lo = value * (double)dur;
    SetMovieTime(q, &tr);
	int i, n = [timeE count];
	if (timescale) {
		for (i=0;i<n;i++) {
			[[timeE objectAtIndex:i] setStringValue:[MHUtil nSecondsToShortString:(double)tr.value.lo / (double)timescale]];
			[[timeR objectAtIndex:i] setStringValue:[MHUtil nSecondsToShortString:(double)(dur - tr.value.lo) / (double)timescale]];
		}
	}
}

- (IBAction)langMenu:(id)sender
{
    if (![sender isKindOfClass:[NSEvent class]]) return;
    if (!currentMovie) return;
    NSMenu *menu = [[[NSMenu alloc] init] autorelease];
    //NSMutableArray *array = [NSMutableArray array];
    //Movie mov = [currentMovie QTMovie];
    int i, n = [currentMovie trackCount];
    NSMutableArray *langs = [NSMutableArray array], *sounds = [NSMutableArray array], *enabled = [NSMutableArray array];
    int tally = 0;
    for (i=0;i<n;i++) {
        MHTrack *track = [currentMovie trackAtIndex:i];
        NSString *language = [track language];
        if (NSNotFound == [langs indexOfObject:language])
            [langs addObject:language];
        if ([track isSoundType]) {
            tally++;
            NSString *name = [track name];
            if (!name)
                name = [NSString stringWithFormat:@"Sound Track %d", tally];
            [sounds addObject:name];
            if ([track enabled]) {
                [enabled addObject:name];
                [enabled addObject:language];
            }
        }
    }
    n = [langs count];
    if (n > 1) {
        for (i=0;i<n;i++) {
            NSMenuItem *item = [[[NSMenuItem alloc] init] autorelease];
            [item setTitle:[langs objectAtIndex:i]];
            [item setTarget:self];
            [item setAction:@selector(setLanguage:)];
            if (NSNotFound != [enabled indexOfObject:[langs objectAtIndex:i]])
                [item setState:YES];
            [menu addItem:item];
       } 
    } else if (tally > 1) {
        n = [sounds count];
        for (i=0;i<n;i++) {
            NSMenuItem *item = [[[NSMenuItem alloc] init] autorelease];
            [item setTitle:[sounds objectAtIndex:i]];
            [item setTarget:self];
            [item setAction:@selector(setSoundTrack:)];
            if (NSNotFound != [enabled indexOfObject:[sounds objectAtIndex:i]])
                [item setState:YES];
            [menu addItem:item];
        }
    }
    if (n > 1) {
		int level = [bigWindow level];
		BOOL levelChanged = NO;
		if ([bigWindow isVisible] && level == NSScreenSaverWindowLevel) {
			[bigWindow setLevel:NSPopUpMenuWindowLevel];
			levelChanged = YES;
		}
        [NSMenu popUpContextMenu:menu withEvent:sender forView:nil];
		if (levelChanged) [bigWindow setLevel:level];
	}
}

- (IBAction)chapMenu:(id)sender
{
    if (![sender isKindOfClass:[NSEvent class]]) return;
    if (!currentMovie) return;
    NSMenu *menu = [[[NSMenu alloc] init] autorelease];
    NSString *lang = nil;
    if ([[currentMovie chapters] multilingual]) lang = [currentMovie currentLanguage];
    NSArray *chaps = [[currentMovie chapters] chapterNamesForLanguage:lang];
    NSArray *times = [[currentMovie chapters] chapterTimesForLanguage:lang];
    NSMenuItem *curChap = nil;
	Movie q = [[playingMovieView movie] quickTimeMovie];
	TimeRecord tr;
	TimeValue tv = GetMovieTime(q, &tr);
    double currTime = (double)tv / (double)tr.scale;
    int i, n = [chaps count];
    for (i=0;i<n;i++) {
        NSMenuItem *item = [[[NSMenuItem alloc] init] autorelease];
        NSString *title = [chaps objectAtIndex:i];
        [item setTitle:title];
        [item setTarget:self];
        [item setAction:@selector(gotoChapter:)];
        if (!curChap && (i == n-1 || [[times objectAtIndex:i+1] doubleValue] > currTime)) curChap = item;
        [menu addItem:item];
    }
    [curChap setState:YES];
	
    int level = [bigWindow level];
	BOOL levelChanged = NO;
	if ([bigWindow isVisible] && level == NSScreenSaverWindowLevel) {
		[bigWindow setLevel:NSPopUpMenuWindowLevel];
		levelChanged = YES;
	}
	[NSMenu popUpContextMenu:menu withEvent:sender forView:nil];
	if (levelChanged) [bigWindow setLevel:level];
}

- (IBAction)cycleText:(id)sender
{
    timeDisplayMode++;
    if (timeDisplayMode > 2) timeDisplayMode = 0;
    [self tick:self];
}

- (IBAction)setLanguage:(id)sender
{
    NSString *title = [sender title];
    long aLang = [tLanguages stringToNumber:title];
    SetMovieLanguage([currentMovie QTMovie], aLang);
    SelectMovieAlternates([currentMovie QTMovie]);
    [self setLangState];
    if ([[currentMovie chapters] multilingual]) {
        if ([movieView isControllerVisible]) {
            [movieView showController:NO adjustingSize:NO];
            [movieView showController:YES adjustingSize:NO];
        }
        [self setChapterTicks];
    }
    [[NSNotificationCenter defaultCenter] postNotificationName:@"movieLanguageChangedNotification" object:self];
}

- (IBAction)setSoundTrack:(id)sender
{
    int i, n = [currentMovie trackCount];
    int tally = 0;
    for (i=0;i<n;i++) {
        MHTrack *track = [currentMovie trackAtIndex:i];
        if ([track isSoundType]) {
            tally++;
            NSString *name = [track name];
            if (!name)
                name = [NSString stringWithFormat:@"Sound Track %d", tally];
            if ([name isEqualToString:[sender title]])
                [track setEnabled:YES];
            else
                [track setEnabled:NO];
        }
    }
    [self setLangState];
    [[NSNotificationCenter defaultCenter] postNotificationName:@"movieLanguageChangedNotification" object:self];
}

- (IBAction)playMovies:(id)sender
{
	//NSLog(@"play");
    float curSpeed = [playingMovieView rate];
    if (![playingMovieView isPlaying] && ![[playingMovieView window] isVisible])
        [self movieDidChangeTo:[[controller getModel] currentPlayingMovie]];
    if (curSpeed > 1 && [playingMovieView isPlaying])
        [self jump:0 - curSpeed * [[[[NSUserDefaults standardUserDefaults] objectForKey:@"controlSpeeds"] objectForKey:@"Snap Back"] floatValue]];
    if ([ffMute state]) {
        short vol = GetMovieVolume([currentMovie QTMovie]);
        if (vol < 0) SetMovieVolume([currentMovie QTMovie], 0 - vol);
    }
    [playingMovieView setMovie:currentMovie];
    //[[playingMovieView window] makeKeyAndOrderFront:self];
    [self makeMovieGoNow:thisMovie];
    [self setPlaying:YES];
    /*
	NSRect rect = [viewWindow frame];
    rect.size.width /= 2;
    rect.size.height /= 4;
    rect.origin.x += rect.size.width / 2;
    rect.origin.y += 5;
    return;
    NSView *oview = [viewWindow overlayViewWithRect:rect];
    [oview addSubview:movieControls];
    rect.origin = NSMakePoint(0,0);
    [movieControls setFrame:rect];
	*/
}

- (void)play:(id)sender
{
	[self playMovies:sender];
}

- (IBAction)stopMovies:(id)sender
{
    [playingMovieView pause:self];
    if (playingMovieView == fullscreenMovieView)
	[bigWindow exitFullScreen];
    [self setPlaying:NO];
}

- (void)setPlaying:(BOOL)playing
{
    if (playing && [playButton action] != @selector(pause:)) {
		[playButton setImage:[NSImage imageNamed:@"pause"]];
		[(NSButton *)playButton setAction:@selector(pause:)];
		[playButton setToolTip:@"Pause"];
		[playButton2 setImage:[NSImage imageNamed:@"pause"]];
		[(NSButton *)playButton2 setAction:@selector(pause:)];
		[playButton2 setToolTip:@"Pause"];
    } else if (!playing && [playButton action] != @selector(playMovies:)) {
		[playButton setImage:[NSImage imageNamed:@"play2"]];
		[(NSButton *)playButton setAction:@selector(playMovies:)];
		[playButton setToolTip:@"Play"];
		[playButton2 setImage:[NSImage imageNamed:@"play2"]];
		[(NSButton *)playButton2 setAction:@selector(play:)];
		[playButton2 setToolTip:@"Play"];
    }
}

- (IBAction)pause:(id)sender
{
    [playingMovieView pause:nil];
    [self setPlaying:NO];
    [currentMovie setTimeIndex:[self getSampleTimeValue]];
}

+ (IBAction)toggleController:(id)sender
{
    [tPlayNibInstance toggleController:sender];
}

- (IBAction)toggleController:(id)sender
{
    if (![sender respondsToSelector:@selector(title)]) sender = controllerMenuItem;
    //[viewWindow makeKeyAndOrderFront:nil];
    NSString *title = [sender title];
    NSUserDefaults *defs = [NSUserDefaults standardUserDefaults];
    NSRect r = [viewWindow frame];
    NSRect s = [movieView frame];
    if ([title hasPrefix:@"Show"]) {
        [movieView showController:YES adjustingSize:NO];
        [defs setBool:NO forKey:@"hideController"];
        [sender setTitle:@"Hide Movie Controller"];
        //NSLog(@"Show");
        s.size.width -= 10;
        [movieView setFrame:s];
        r.size.width += 10;
        r.size.height += 16;
	} else if ([title hasPrefix:@"Hide"]) {
        [movieView showController:NO adjustingSize:NO];
        [defs setBool:YES forKey:@"hideController"];
        [sender setTitle:@"Show Movie Controller"];
        //NSLog(@"Hide");
        s.size.width += 10;
        [movieView setFrame:s];
        r.size.width -= 10;
        r.size.height -= 16;
    }
    [viewWindow setFrame:r display:YES];
    [defs synchronize];
}

- (void) setPMovieView:(id)newView
{
	//if (newView == movieView) NSLog(@"set to movieView");
	//if (newView == fullscreenMovieView) NSLog(@"set to fullscreenView");
    if (newView && newView != playingMovieView) {
		//[playingMovieView setMovie:nil];
        playingMovieView = newView;
	}
	int i, j, n = [sizeMenu numberOfItems];
	for (i=0;i<n;i++)
		if ([[sizeMenu itemAtIndex:i] state])
			break;
	for (j=0;j<n;j++) {
		if (newView == fullscreenMovieView) [[sizeMenu itemAtIndex:j] setState:(j == 3)];
		else if (newView == windowedMovieView) [[sizeMenu itemAtIndex:j] setState:(j == 4)];
		else if (newView == movieView) {
			if (i > 2) i = 1;
			[[sizeMenu itemAtIndex:j] setState:(j == i)];
			[movieView showController:NO adjustingSize:NO];
			if (![[NSUserDefaults standardUserDefaults] boolForKey:@"hideController"])
				[movieView showController:YES adjustingSize:NO];
		}
	}
}

- (IBAction)nextMovie:(id)sender
{
    [self makeMovieGoNow:nextMovie];
    [self setPlaying:YES];
}

+ (void)nextMovie
{
    [tPlayNibInstance nextMovie:nil];
}

- (IBAction)lastMovie:(id)sender
{
    [self makeMovieGoNow:lastMovie];
    [self setPlaying:YES];
}

+ (void)lastMovie
{
    [tPlayNibInstance lastMovie:nil];
}

- (IBAction)fullScreen:(id)sender
{
	if ([bigWindow isVisible]) {
		[bigWindow exitFullScreen];
		return;
	}
    if (nil == currentMovie) return;
    NSArray *screens = [NSScreen screens];
    NSScreen *aScreen;
    int index = [whichScreen indexOfSelectedItem] - 1;
    if (index == -1) {
        aScreen = [[playingMovieView window] screen];
    } else {
        if (index >= [screens count])
            index = 0;
        aScreen = [screens objectAtIndex:index];
    }
    
	[viewWindow orderOut:self];
	
    NSRect screenRect = [aScreen frame];
	screenRect.size.width++;
    [bigWindow setFrame:screenRect display:NO];
	//NSLog(@"screen rect: %g x %g, (%g, %g)", screenRect.size.width, screenRect.size.height, screenRect.origin.x, screenRect.origin.y);
    screenRect.origin = NSMakePoint(0,0);
    [self setPMovieView:fullscreenMovieView];
	screenRect = [fullscreenMovieView frame];
	//NSLog(@"view rect: %g x %g, (%g, %g)", screenRect.size.width, screenRect.size.height, screenRect.origin.x, screenRect.origin.y);
    [bigWindow setHidesOnDeactivate:[hides state]];
    
    [self setMovieFrameInFrame:screenRect];
	
    int level = NSScreenSaverWindowLevel;
    int levels[] = { NSNormalWindowLevel, NSFloatingWindowLevel, NSSubmenuWindowLevel, NSTornOffMenuWindowLevel, NSModalPanelWindowLevel, NSMainMenuWindowLevel, NSStatusWindowLevel, NSPopUpMenuWindowLevel, NSScreenSaverWindowLevel };
    int chosen = [fullscreenLevelPopup indexOfSelectedItem];
    level = levels[chosen];
    [bigWindow setLevel:level];
    [bigWindow setAcceptsMouseMovedEvents:YES];
    [bigWindow makeKeyAndOrderFront:nil];
	[bigWindow makeFirstResponder:bigWindow];
    [NSCursor setHiddenUntilMouseMoves:YES];
    wasAutoSleep = !IsAutoSlpControlDisabled();
    if (wasAutoSleep)
	AutoSleepControl(NO);
    [playingMovieView start:self];
	[self setPlaying:YES];
}

+ (void) fullScreen
{
    [tPlayNibInstance fullScreen:nil];
}

- (IBAction)inMainWindow:(id)sender
{
	[viewWindow orderOut:self];
    [windowedMovieView showController:NO adjustingSize:NO];
    NSRect frame = [mainTabView frame];
    frame.size.height -= 20;
    frame.origin.y = 20;
    [self setMovieFrameInFrame:frame];
    [controller setMainTabViewName:@"Movie"];
    [[controller mainWindow] makeKeyAndOrderFront:nil];
}

- (void) setMovieFrameInFrame:(NSRect)outerFrame
{
	//NSLog(@"inframe: %@", NSStringFromRect(outerFrame));
    double multiplier, newRatio, outerRatio = outerFrame.size.width / outerFrame.size.height;
    int w = [[currentMovie objectForKey:@"Correct Width"] intValue];
    int h = [[currentMovie objectForKey:@"Correct Height"] intValue];
    if ([playingMovieView isControllerVisible])
        h += 16;
    NSString *aspect = [fsAspect stringValue];
    NSRect movieRect;
    if ([aspect isEqualToString:@"fill"]) {
        movieRect.origin.x = movieRect.origin.y = 0;
        movieRect.size = outerFrame.size;
    } else {
        if ([aspect isEqualToString:@"native"] || [aspect doubleValue] == 0)
            newRatio = (double)w / (double)h;
        else
            newRatio = (double)w * [aspect doubleValue] / (double)h;
        w = h * newRatio;
        BOOL isWide = newRatio > outerRatio;
        if ([overscan state]) isWide = !isWide;
        if (isWide) {
            multiplier = outerFrame.size.width / w;
            movieRect.origin.x = outerFrame.origin.x;
            movieRect.origin.y = outerFrame.origin.y + (outerFrame.size.height - h * multiplier) / 2;
        } else {
            multiplier = outerFrame.size.height / h;
            movieRect.origin.x = outerFrame.origin.x + (outerFrame.size.width - w * multiplier) / 2;
            movieRect.origin.y = outerFrame.origin.y;
        }
        movieRect.size.width = w * multiplier;
        movieRect.size.height = h * multiplier;
    }
    
    [playingMovieView performSelector: @selector(setFrameV:) withObject:[NSValue valueWithRect: movieRect] afterDelay:0];
    
    [[playingMovieView superview] setNeedsDisplay:YES];
    
    if (playingMovieView != movieView)
        [movieView setMovie:nil];
    [playingMovieView setMovie:nil];
    
    [playingMovieView setMovie:currentMovie];
	//movieRect = [playingMovieView frame];
	//NSLog(@"outFrame: %@", NSStringFromRect(movieRect));
	
}

- (IBAction)refreshMonitors:(id)sender
{
    NSRect screenFrame;
    NSArray *screens = [NSScreen screens];
    [whichScreen removeAllItems];
    int i, n = [screens count];
    [whichScreen addItemWithTitle:@"Current"];
    for (i=0;i<n;i++) {
        screenFrame = [[screens objectAtIndex:i] frame];
        [whichScreen addItemWithTitle:[NSString stringWithFormat:@"%g x %g", screenFrame.size.width, screenFrame.size.height]];
    }
    int monitorChoice = [[NSUserDefaults standardUserDefaults] integerForKey:@"whichScreen"];
    if (monitorChoice < 0 || monitorChoice >= [whichScreen numberOfItems])
	monitorChoice = 0;
    [whichScreen selectItemAtIndex:monitorChoice];
}

- (IBAction)pickMonitor:(id)sender
{
    int selection = [whichScreen indexOfSelectedItem];
    [[NSUserDefaults standardUserDefaults] setInteger:selection forKey:@"whichScreen"];
}

- (BOOL)playOnExitFullScreen
{
    return [exitPlay state];
}

- (IBAction)setRateStickiness:(id)sender
{
	int rstick = [sender indexOfSelectedItem];
	[[NSUserDefaults standardUserDefaults] setInteger:rstick forKey:@"rateStickiness"];
	[[NSUserDefaults standardUserDefaults] synchronize];
}

- (IBAction)stepRate:(id)sender
{
	float r = 0;
	NSString *title = nil;
	if ([sender respondsToSelector:@selector(title)])
		title = [sender title];
	else if ([sender isKindOfClass:[NSString class]])
		title = sender;
	
	if ([[NSUserDefaults standardUserDefaults] objectForKey:@"playRate"])
		r = [[NSUserDefaults standardUserDefaults] floatForKey:@"playRate"];
	else
		r = [playingMovieView rate];
	if (!r) r = 1;
	if ([title isEqualToString:@"Faster"]) r += 0.1;
	else if ([title isEqualToString:@"Slower"]) r -= 0.1;
	else if ([title isEqualToString:@"Original Speed"]) r = 1.0;
	else r = [sender floatValue];
	
	BOOL p = ([playingMovieView rate] != 0);
	if (p)
		[playingMovieView setRate:r];
	//if (!p) [[currentMovie qtKitMovie] stop];
	int i, n = [[rateView subviews] count];
	for (i=0;i<n;i++) {
		[[[rateView subviews] objectAtIndex:i] setFloatValue:r];
	}
	[[NSUserDefaults standardUserDefaults] setFloat:r forKey:@"playRate"];
	[[NSUserDefaults standardUserDefaults] synchronize];
}


- (IBAction)setRate:(id)sender
{
    NSMenu *menu = [sender menu];
    int i, n = [menu numberOfItems];
    for (i=0;i<n;i++) {
        NSMenuItem *item = [menu itemAtIndex:i];
        if (item == sender) {
            [item setState:YES];
            NSString *title = [sender title];
            if ([title hasSuffix:@"x"])
                title = [title substringToIndex:[title length]-1];
            NSUserDefaults *defs = [NSUserDefaults standardUserDefaults];
            NSMutableDictionary *dict = [NSMutableDictionary dictionaryWithDictionary:[defs objectForKey:@"controlSpeeds"]];
            [dict setObject:title forKey:@"FF Speed"];
            [dict setObject:title forKey:@"RW Speed"];
            [defs setObject:dict forKey:@"controlSpeeds"];
            [defs synchronize];
        } else
            [item setState:NO];
    }
}

- (IBAction)halfSize:(id)sender
{
    [self setViewSize:0.5];
    [self setSizeMenuStates:sender];
    [self setPMovieView:movieView];
	if ([[controller mainTabViewName] isEqualToString:@"Movie"] || [bigWindow isVisible] || [windowedMovieView isPlaying])
		[self playMovies:self];
}

- (IBAction)origSize:(id)sender
{
    [self setViewSize:1.0];
    [self setSizeMenuStates:sender];
    [self setPMovieView:movieView];
	if ([[controller mainTabViewName] isEqualToString:@"Movie"] || [bigWindow isVisible] || [windowedMovieView isPlaying]) {
		[self shouldExitFullScreen];
		[self playMovies:self];
	}
}

- (IBAction)doubleSize:(id)sender
{
    [self setViewSize:2.0];
    [self setSizeMenuStates:sender];
    [self setPMovieView:movieView];
	if ([[controller mainTabViewName] isEqualToString:@"Movie"] || [bigWindow isVisible] || [windowedMovieView isPlaying])
		[self playMovies:self];
}

- (IBAction)fullscreenSize:(id)sender
{
    
    [self setSizeMenuStates:sender];
	if ([[controller mainTabViewName] isEqualToString:@"Movie"] || [viewWindow isVisible] || [windowedMovieView isPlaying])
		[self fullScreen:self];
	else
		[self setPMovieView:fullscreenMovieView];
}

- (IBAction)inMainWindowSize:(id)sender
{
    
    [self setSizeMenuStates:sender];
    [self setPMovieView:windowedMovieView];
	if ([viewWindow isVisible] || [bigWindow isVisible]) {
		[viewWindow orderOut:self];
		[movieView stop:self];
		[self shouldExitFullScreen];
		[self inMainWindow:self];
		[self playMovies:self];
	}
}

- (void) setSizeMenuStates:(id)sender
{
    if (!sender) return;
    int i;
    id item;
    for (i=0;i<[sizeMenu numberOfItems];i++) {
        item = [sizeMenu itemAtIndex:i];
        if (item == sender) {
            [item setState:YES];
            [[NSUserDefaults standardUserDefaults] setInteger:i forKey:t_whichSize];
            [[NSUserDefaults standardUserDefaults] synchronize];
        } else
            [item setState:NO];
    }
}

- (IBAction)refreshVolume:(id)sender
{
	if (!currentMovie || ![playingMovieView rate]) return;
    short vol = GetMovieVolume([currentMovie QTMovie]);
    [volScroll setDoubleValue:vol];
	[volScroll2 setDoubleValue:vol];
	if (vol > loudVol) loudVol = vol;
}
	
- (IBAction)maxVolume:(id)sender
{
    NSMovie *movie = [playingMovieView movie];
	if (![movie isKindOfClass:[NSMovie class]])
		movie = [movie delegate];
    if (movie) {
        SetMovieVolume([movie QTMovie], 256);
		[playingMovieView noteSound];
	}
    [self refreshVolume:self];
}

- (IBAction)minVolume:(id)sender
{
    NSMovie *movie = [playingMovieView movie];
    if (movie) {
        SetMovieVolume([movie QTMovie], 0);
		[playingMovieView noteSound];
	}
    [self refreshVolume:self];
}

+ (IBAction)refreshVolume:(id)sender
{   [tPlayNibInstance refreshVolume:sender];    }

- (IBAction)setVolume:(id)sender
{
	if (!currentMovie) return;
	double input = 256;
    if ([sender respondsToSelector:@selector(doubleValue)])
        input = [sender doubleValue];
	else if ([sender respondsToSelector:@selector(title)]) {
		input = GetMovieVolume([currentMovie QTMovie]);
		if ([[sender title] isEqualToString:@"Up"]) {
			if (input < 0) input = 0 - input;
			else if (input < 256) input += 25;
			else input *= 1.1;
		} else if ([[sender title] isEqualToString:@"Down"]) {
			if (input < 0) {}
			else if (input < 256) input -= 25;
			else input /= 1.1;
		}
	} else return;
	
            SetMovieVolume([currentMovie QTMovie], (short)input);
			loudVol = input;
			[playingMovieView noteSound];
    
}

- (IBAction)bigger:(id)sender
{
    if (nil == currentMovie) return;
    NSRect aRect = [movieView bounds];
    int width = aRect.size.width;
    int height = aRect.size.height;
    if ([movieView isControllerVisible]) height -= 16;
    int origWidth = [[currentMovie objectForKey:@"Original Width"] intValue];
    int origHeight = [[currentMovie objectForKey:@"Original Height"] intValue];
    [self setViewWidth:width+origWidth/2 height:height+origHeight/2];
}

- (IBAction)smaller:(id)sender
{
    if (nil == currentMovie) return;
    NSRect aRect = [movieView bounds];
    int width = aRect.size.width;
    int height = aRect.size.height;
    if ([movieView isControllerVisible]) height -= 16;
    int origWidth = [[currentMovie objectForKey:@"Original Width"] intValue];
    int origHeight = [[currentMovie objectForKey:@"Original Height"] intValue];
    [self setViewWidth:width-origWidth/2 height:height-origHeight/2];
}

// sets the view window to "size" times the movie's preferred size.
- (void)setViewSize:(double)size
{
    //NSSize aSize;
    int width, height;
    
    width = (double)[[currentMovie objectForKey:@"Correct Width"] intValue] * size;
    height = (double)[[currentMovie objectForKey:@"Correct Height"] intValue] * size;
    [self setViewWidth:width height:height];
    
}

- (void)setViewWidth:(int)w height:(int)h
{
    NSSize aSize;
	NSPoint p = [viewWindow frame].origin;
    aSize.width = w;
    aSize.height = h;
    if (aSize.width < 16) aSize.width = 174;
    if (aSize.height < 16) aSize.height = 16;
    if ([movieView isControllerVisible]) {
        aSize.width += 10;
        aSize.height += 16;
    }
	NSRect sFrame = [[viewWindow screen] frame];
	if ([viewWindow screen] == [[NSScreen screens] objectAtIndex:0]) sFrame.size.height -= 22;
	NSSize sSize = sFrame.size;
	if (aSize.width > sSize.width) aSize.width = sSize.width;
	if (aSize.height > sSize.height) aSize.height = sSize.height;
	if (p.x + aSize.width > sSize.width + sFrame.origin.x) p.x = sSize.width - aSize.width + sFrame.origin.x;
	if (p.y + aSize.height > sSize.height + sFrame.origin.y) p.y = sSize.height - aSize.height + sFrame.origin.y;
	[self windowWillResize:viewWindow toSize:aSize];
	[viewWindow setFrameOrigin:p];
    [viewWindow setContentSize:aSize];
}

- (double)getTimeValue
{
    MHMovie *movie = (MHMovie *)[playingMovieView movie];
	if (![movie isKindOfClass:[NSMovie class]])
		movie = [movie delegate];
    if (nil == movie) return 0;
    return [movie currentTimeIndexSeconds];
}

- (long)getSampleTimeValue
{
    MHMovie *movie = (MHMovie *)[playingMovieView movie];
	if (![movie isKindOfClass:[NSMovie class]])
		movie = [movie delegate];
    if (nil == movie) return 0;
    return [movie currentTimeIndex];
}

- (IBAction)savePrefs:(id)sender
{
    //NSLog(@"playback");
    NSUserDefaults *defs = [NSUserDefaults standardUserDefaults];
    [defs setBool:[overscan state] forKey:@"overscan"];
    [defs setBool:[ffMute state] forKey:@"ffMute"];
    [defs setInteger:[fullscreenLevelPopup indexOfSelectedItem] forKey:@"fullscreenLevel"];
	[defs setBool:[frames state] forKey:@"frames"];
    [defs synchronize];
}

- (IBAction)chapScroll:(id)sender
{
    int selected = [chapScroll indexOfSelectedItem];
    [[NSUserDefaults standardUserDefaults] setInteger:selected forKey:@"chapScroll"];
    //if (0 == selected) [menuChap startScrolling];
    //else [menuChap stopScrolling];
    if (1 == selected) {
        [menuChap setTitle:@"Chapters"];
		[menuChap2 setTitle:@"Chapters"];
	}
    else [self tick:nil];
}

- (void)awakeFromNib
{
	
	toollist = [[NSMutableArray alloc] init];
    //[viewWindow setAlphaValue:0.5];
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(flagsChanged:) name:@"flagsChangedNotification" object:nil];
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(keyDown:) name:@"keyDownNotification" object:nil];
    //[[NSNotificationCenter defaultCenter] addObserver:controller selector:@selector(windowDidBecomeMain:) name:@"NSWindowDidBecomeMainNotification" object:nil];
    //[[NSNotificationCenter defaultCenter] addObserver:controller selector:@selector(windowDidResignMain:) name:@"NSWindowDidResignMainNotification" object:nil];
	timeE = [[NSMutableArray alloc] init];
	timeR = [[NSMutableArray alloc] init];
	int i, j, m, n;
	NSArray *views = [NSArray arrayWithObjects:movieController, floaterView, nil];
	m = [views count];
	for (j=0;j<m;j++) {
		NSView *v = [views objectAtIndex:j];
		n = [[v subviews] count];
		for (i=0;i<n;i++) {
			NSView *iview = [[v subviews] objectAtIndex:i];
			if ([iview isKindOfClass:[NSImageView class]]) {
				NSTextField *t = [[NSTextField alloc] initWithFrame:NSMakeRect(0, 14, 38, 12)];
				[t setEditable:NO];
				[t setDrawsBackground:NO];
				[t setBezeled:NO];
				[t setStringValue:@""];
				[[t cell] setFont:[NSFont labelFontOfSize:9]];
				[[t cell] setWraps:NO];
				if ([[[(NSImageView *)iview image] name] isEqualToString:@"jLeft"]) {
					[t setAlignment:NSRightTextAlignment];
					[iview addSubview:t];
					[timeE addObject:t];
				} else if ([[[(NSImageView *)iview image] name] isEqualToString:@"jRight"]) {
					[iview addSubview:t];
					[timeR addObject:t];
				}
			}
		}
	}
    
    tPlayNibInstance = self;
    NSUserDefaults *defs = [NSUserDefaults standardUserDefaults];
	[capLoud setState:[defs boolForKey:@"capLoud"]];
	capTarget = [defs floatForKey:@"capLoudTarget"];
	if (!capTarget) capTarget = 20;
	[capLoudTarget setFloatValue:capTarget];
	[capLoudTarget setHidden:![capLoud state]];
	[capLoudTargetLabel setHidden:![capLoud state]];
    NSNumber *num = [defs objectForKey:t_whichSize];
    if (!num)
        num = [NSNumber numberWithInt:1];
    [ffMute setState:[defs boolForKey:@"ffMute"]];
	[frames setState:[defs boolForKey:@"frames"]];
    [overscan setState:[defs boolForKey:@"overscan"]];
    NSNumber *level = [defs objectForKey:@"fullscreenLevel"];
    if (!level) level = [NSNumber numberWithInt:8];
    [fullscreenLevelPopup selectItemAtIndex:[level intValue]];
    int preference = [defs integerForKey:@"chapScroll"];
    if (preference >= [chapScroll numberOfItems]) preference = [chapScroll numberOfItems]-1;
    [chapScroll selectItemAtIndex:preference];
    //if (!preference) [menuChap startScrolling];
    //else
    if (preference == 1) {
		[menuChap setTitle:@"Chapters"];
		[menuChap2 setTitle:@"Chapters"];
	}
	
    for (i=0;i<[sizeMenu numberOfItems];i++)
        [[sizeMenu itemAtIndex:i] setState:i==[num intValue]];
    if ([num intValue] == 3)
        playingMovieView = fullscreenMovieView;
    else if ([num intValue] == 4)
        playingMovieView = windowedMovieView;
    else playingMovieView = movieView;
    [exit selectItemAtIndex:[defs integerForKey:@"whichExit"]];
    if ([defs boolForKey:@"hideController"]) {
        //[viewWindow makeKeyAndOrderFront:nil];
        //[viewWindow orderOut:nil];
        [self performSelector:@selector(toggleController:) withObject:controllerMenuItem afterDelay:0];
    }
	int rstick = [defs integerForKey:@"rateStickiness"];
	if (rstick)
		if ([[[rateView subviews] objectAtIndex:2] respondsToSelector:@selector(selectItemAtIndex:)])
			[[[rateView subviews] objectAtIndex:2] selectItemAtIndex:rstick];
	if (rstick == 2) { // 2 means manual
		float rate = [defs floatForKey:@"playRate"];
		if (!rate) rate = 1.0;
		n = [[rateView subviews] count];
		for (i=0;i<n;i++)
			[[[rateView subviews] objectAtIndex:i] setFloatValue:rate];
	} else
		[[NSUserDefaults standardUserDefaults] setFloat:1.0 forKey:@"playRate"];
    bigWindow = [[tWindow alloc] initWithContentRect:[[NSScreen mainScreen] frame] styleMask:NSBorderlessWindowMask backing:NSBackingStoreBuffered defer:NO screen:[NSScreen mainScreen]];
	
    //[bigWindow setAcceptsMouseMovedEvents:YES];
    //[bigWindow setLevel:CGShieldingWindowLevel()];
    [bigWindow setBackgroundColor:[NSColor blackColor]];
    [bigWindow setContentView:fullscreenContentView];
	[fullscreenContentView retain];
    [bigWindow setHidesOnDeactivate:[hides state]];
    currentMovie = nil;
    wasAutoSleep = NO;
    currentTitle = @"";
    timeDisplayMode = 0;
    timescale = 0;
    oldSpeed = 0;
    [fullscreenMovieView showController:NO adjustingSize:NO];
    [self performSelector:@selector(maxVolume:) withObject:nil afterDelay:0];
    [self performSelector:@selector(movieDidChangeTo:) withObject:controller afterDelay:0];
    
    //[movieView setPlayer:self];
    [movieView setEditable:YES];
    
    [self refreshMonitors:nil];
	muter = nil;
	
	aclayout = malloc(sizeof(AudioChannelLayout));
	AudioChannelLayout *aclptr = aclayout;
	aclptr->mChannelLayoutTag = 6553601; // mono
	aclptr->mChannelBitmap = 0;
	aclptr->mNumberChannelDescriptions = 0;
}

- (IBAction)setWhichExit:(id)sender
{
    [[NSUserDefaults standardUserDefaults] setInteger:[exit indexOfSelectedItem] forKey:@"whichExit"];
    [[NSUserDefaults standardUserDefaults] synchronize];
}

- (void)shouldExitFullScreen
{
    //if (CGDisplayRelease( kCGDirectMainDisplay ) != kCGErrorSuccess)
    //    NSLog( @"Couldn't release the display!");
    //[[bigWindow contentView] retain];
	[fullscreenMovieView showNote:nil];
    [bigWindow orderOut:nil];
	float rate = [playingMovieView rate];
    [playingMovieView stop:self];
    [bigWindow setAcceptsMouseMovedEvents:NO];
	BOOL play = YES;
	NSEvent *e = [bigWindow currentEvent];
	NSString *chars = @"";
	if ([e type] == NSKeyDown) chars = [e characters];
	if ([chars length] && [chars characterAtIndex:0] == 'w' && [e modifierFlags] & NSCommandKeyMask)
		play = NO;
	if ([exitPlay state] && play) {
		if (playingMovieView == fullscreenMovieView) [self setPMovieView:movieView];
		[self playMovies:self];
		[playingMovieView setRate:rate];
	} else {
		[viewWindow orderOut:self];
		[self setPlaying:NO];
	}
    if (wasAutoSleep)
	AutoSleepControl(YES);
    //playingMovieView = movieView;
}

+ (void)shouldExitFullScreen
{
    [tPlayNibInstance shouldExitFullScreen];
}

- (MHMovie *)currentMovie
{
    return currentMovie;
}

void setTimeIndex(MHMovie *theMovie, double timeIndex) {
    [theMovie setCurrentTimeIndex:timeIndex];
}
/*
static OSStatus insertProcessDataCallback (
										   void						*inUserData,
										   AudioUnitRenderActionFlags  *ioRenderFlags,
										   const AudioTimeStamp     *inTimeStamp,
										   UInt32                   inNumberFrames,
										   AudioBufferList          *inInputData,
										   AudioBufferList		     *outOutputData
)
{
	
	*outOutputData = *inInputData;
	
	int buf, sam, nSamples;
	float *s;
	
	for (buf=0;buf<outOutputData->mNumberBuffers;buf++) {
		nSamples = outOutputData->mBuffers[buf].mDataByteSize / 4;
		for (sam=0;sam<nSamples;sam++) {
			s = outOutputData->mBuffers[buf].mData;
			if (s[sam] > 0.01) {
				s[sam] = 0.01;
			}
		}
	}
	
	
	return noErr;
}

static OSStatus insertResetCallback(
									void			*inUserData,
									Float64			inSampleRate,
									UInt32			inMaxFrames,
									Float64			*outLatency,
									Float64			*outTailTime
									)
{
	NSLog(@"reset");
	return noErr;
}

static OSStatus insertFinalizeCallback(void	*inUserData)
{
	NSLog(@"finalize");
	return noErr;
}

- (void)regInfoPopulate:(QTAudioContextInsertRegistryInfo *)registryInfo
{
	registryInfo->userData = (void*)self;
	registryInfo->inputChannelLayoutSize = sizeof(AudioChannelLayout); 
	registryInfo->inputChannelLayout = aclayout; 
	registryInfo->outputChannelLayoutSize = sizeof(AudioChannelLayout); 
	registryInfo->outputChannelLayout = aclayout; 
	registryInfo->processDataCallback = insertProcessDataCallback;
	registryInfo->resetCallback = insertResetCallback;
	registryInfo->finalizeCallback = insertFinalizeCallback;
}
*/
- (IBAction)capLoud:(id)sender
{
	[[NSUserDefaults standardUserDefaults] setBool:[capLoud state]	forKey:@"capLoud"];
	[[NSUserDefaults standardUserDefaults] setFloat:[capLoudTarget floatValue] forKey:@"capLoudTarget"];
	[[NSUserDefaults standardUserDefaults] synchronize];
	[capLoudTarget setHidden:![capLoud state]];
	[capLoudTargetLabel setHidden:![capLoud state]];
	capTarget = [capLoudTarget floatValue];
	[self movieDidChangeTo:currentMovie];
}

- (void)movieDidChangeTo:(MHMovie *)aMovie
{
	if (![[NSUserDefaults standardUserDefaults] integerForKey:@"rateStickiness"]) { // 0 means each movie
		[self stepRate:@"1.0"];
	}
	
    //NSLog(@"tPlayback movieDidChangeTo %@", aMovie);
    if (aMovie && ![aMovie isKindOfClass:[MHMovie class]]) aMovie = [(tController *)aMovie currentMovie];
	
	/*
	QTAudioContextRef aContext = nil;
	OSStatus err = GetMovieAudioContext([currentMovie QTMovie], &aContext);
	QTAudioContextInsertRegistryInfo regInfo = {0};
	if ([[NSUserDefaults standardUserDefaults] boolForKey:@"capLoud"])
		[self regInfoPopulate:&regInfo];
	err = QTAudioContextRegisterInsert(aContext, sizeof(regInfo), &regInfo);
	 */
	
	
	if (muter)
		[muter invalidate];
	muter = nil;
	loudVol = GetMovieVolume([currentMovie QTMovie]);
	if (!loudVol) loudVol = 512;
	softCount = 0;
	static UInt32 numberOfBandLevels = 1;
	if ([[NSUserDefaults standardUserDefaults] boolForKey:@"capLoud"]) {
		SetMovieAudioFrequencyMeteringNumBands([aMovie QTMovie], kQTAudioMeter_MonoMix, &numberOfBandLevels);
		SetMovieAudioVolumeMeteringEnabled([aMovie QTMovie], kQTAudioMeter_MonoMix, YES);
		freqResults = malloc(offsetof(QTAudioFrequencyLevels, level[1]));
		QTAudioFrequencyLevels *freq = (QTAudioFrequencyLevels *)freqResults;
		freq->numChannels = 1;
		freq->numFrequencyBands = 1;
		muter = [NSTimer scheduledTimerWithTimeInterval:0.1 target:self selector:@selector(meters) userInfo:nil repeats:YES];
	}
	
	//if (aMovie == currentMovie) return;
    currentMovie = aMovie;
	 
    [scrubName setTitle:@""];
	[scrubName2 setTitle:@""];
	NSString *stringE = @"", *stringR = @"";
	
    [scrubber setMarkers:nil];
	[scrubber2 setMarkers:nil];
    [self setLangState];
    if (nil != aMovie) {
		double seconds = GetMovieTime([aMovie QTMovie], nil) / GetMovieTimeScale([aMovie QTMovie]);
		double duration = GetMovieDuration([aMovie QTMovie]) / GetMovieTimeScale([aMovie QTMovie]);
        [currentTitle release];
        currentTitle = [[aMovie stringValue] retain];
		[scrubName setStringValue:currentTitle];
        [scrubName2 setStringValue:currentTitle];
        timescale = GetMovieTimeScale([aMovie QTMovie]);
		stringE = [MHUtil nSecondsToShortString:seconds];
		stringR = [MHUtil nSecondsToShortString:duration - seconds];
		[viewWindow setTitle:[NSString stringWithFormat:@"%@ - %@", currentTitle, stringE]];
		if ([chapScroll indexOfSelectedItem] < 1) {
			NSString *language = nil;
			if ([[aMovie chapters] multilingual]) language = [aMovie currentLanguage];
			NSString *chapName = [[aMovie chapters] nameForTime:seconds language:language];
			if (!chapName) chapName = @"Chapters";
			//chapName = [chapName whitespaceTrimmed];
			[menuChap setTitle:chapName];
			[menuChap2 setTitle:chapName];
		}		
    } else {
        currentTitle = @"";
        timescale = 0;
		[playingMovieView stop:self];
		[viewWindow orderOut:self];
        [movieView setMovie:nil];
        [fullscreenMovieView setMovie:nil];
        [windowedMovieView setMovie:nil];
    }
	
	int i, n = [timeE count];
	for (i=0;i<n;i++) {
		[[timeE objectAtIndex:i] setStringValue:stringE];
		[[timeR objectAtIndex:i] setStringValue:stringR];
	}
    [self setChapterTicks];
    //NSLog(@"superview: %@", [[playingMovieView superview] className]);
    if ([[playingMovieView window] isVisible] || (playingMovieView == windowedMovieView && ![[[mainTabView selectedTabViewItem] identifier] isEqualToString:@"Movie"]))
        [playingMovieView setMovie:aMovie];
    else
        [playingMovieView setMovie:nil];
    [self setPlaying:NO];
    if (aMovie) {
        //short vol = GetMovieVolume([aMovie QTMovie]);
        //[volScroll setDoubleValue:vol];
        SetMovieVolume([aMovie QTMovie], (short)[volScroll doubleValue]);
    }
    TimeRecord tr;
    [scrubber setFloatValue:(float)GetMovieTime([currentMovie QTMovie], &tr) / (float)GetMovieDuration([currentMovie QTMovie])];
    [scrubber2 setFloatValue:(float)GetMovieTime([currentMovie QTMovie], &tr) / (float)GetMovieDuration([currentMovie QTMovie])];
    
    int row = [[controller getModel] indexForMovie:currentMovie];
    [[controller getModel] setCurrentPlayingIndex:row];
    //[[[theTable tableColumnWithIdentifier:@"Order"] dataCell] setImageRow:row];
    [theTable setNeedsDisplay:YES];
    [self performSelector:@selector(tick:) withObject:self afterDelay:0];
}

- (void)setChapterTicks
{
    NSMutableArray *markers = [NSMutableArray array];
    NSString *lang = nil;
    if ([[currentMovie chapters] multilingual]) lang = [currentMovie currentLanguage];
    NSArray *times = [[currentMovie chapters] chapterTimesForLanguage:lang];
    [menuChap setTransparent:![times count]];
	[menuChap2 setTransparent:![times count]];
    double total = [[currentMovie objectForKey:@"Seconds"] doubleValue];
    int i, n = [times count];
    for (i=0;i<n;i++) {
        double time = [[times objectAtIndex:i] doubleValue] / total;
        [markers addObject:[NSNumber numberWithDouble:time]];
    }
    [scrubber setMarkers:markers];
    [scrubber2 setMarkers:markers];
}

- (void)setLangState
{
    if ([self numLangs:currentMovie] < 2) {
        [menuLang setImage:nil];
		[scrubName2 setImage:nil];
        //[menuLang setTransparent:YES];
        //[menuLang setEnabled:NO];
        NSString *title = [currentMovie stringValue];
        if (!title) title = @"";
        [menuLang setTitle:title];
		[scrubName2 setTitle:title];
    } else {
        [menuLang setImage:[NSImage imageNamed:@"menuArrow"]];
        [scrubName2 setImage:[NSImage imageNamed:@"menuArrow"]];
        //[menuLang setTransparent:NO];
        //[menuLang setEnabled:YES];
        int i, n = [currentMovie trackCount];
        for (i=0;i<n;i++) {
            MHTrack *track = [currentMovie trackAtIndex:i];
            if ([track isSoundType] && [track enabled]) {
                NSString *lang = [track language];
                NSString *title = [currentMovie objectForKey:@"Title" language:lang];
                if (!title)
                    title = [currentMovie stringValue];
                if (!title) title = @"";
                lang = [tLanguages abbrev:lang];
                if (lang)
                    title = [title stringByAppendingFormat:@" (%@)", lang];
                    /*
                    NSDictionary *attributes = [NSDictionary dictionaryWithObjectsAndKeys:
                                [NSFont systemFontOfSize:10], NSFontAttributeName, nil];
                                //[NSColor colorWithCalibratedRed:0.418 green:0.426 blue:0.371 alpha:1],
                                    //NSBackgroundColorAttributeName,
                                //[NSColor colorWithCalibratedRed:0.836 green:0.855 blue:0.746 alpha:1],
                                    //NSForegroundColorAttributeName, nil];
                    NSAttributedString *title = [[NSAttributedString alloc] initWithString:lang attributes:attributes];
                    [menuLang setAttributedTitle:title];
					 */
                [menuLang setTitle:title];
                [scrubName2 setTitle:title];
                break;
            }
        }
    }
}

- (int)numLangs:(MHMovie *)aMovie
{
    int tally = 0;
    int i, n = [aMovie trackCount];
    for (i=0;i<n;i++)
        if ([[aMovie trackAtIndex:i] isSoundType])
            tally++;
    return tally;
}

- (void) makeMovieGoNow:(movieOrderT)whichMovie
{
    int index, nRows;
    double timeValue; // = [whichSize indexOfSelectedItem];
    MHMovie *theMovie = nil;
    tPlaylist *model = [theTable dataSource];
    
    timeValue = [self getSampleTimeValue];
    nRows = [theTable numberOfRows];
    if (nRows == 0) return;
    
    theMovie = (MHMovie *)[playingMovieView movie];
	if (![theMovie isKindOfClass:[NSMovie class]])
		theMovie = [theMovie delegate];
    if (theMovie) index = [model indexForMovie:theMovie];
    else index = [theTable selectedRow];
    if (theMovie != nil) {
        [theMovie saveCurrentHeight];
        if (timeValue < GetMovieDuration([theMovie QTMovie]) - GetMovieTimeScale([theMovie QTMovie]))
            [theMovie setTimeIndex:timeValue];
        else
            [theMovie setTimeIndex:0];
    }
    switch (whichMovie) {
        case thisMovie: if (index < 0) index = 0;
                        break;
        case nextMovie:	index++;
                        break;
        case lastMovie: index--;
                        break;
        default: break;
    }
    if (index < 0) index = nRows - 1;
    if (index >= nRows) index = 0;
    theMovie = [model movieForIndex:index];
    
    // Set the place in the movie
    timeValue = [[theMovie objectForKey:@"Seconds Elapsed"] doubleValue] * GetMovieTimeScale([theMovie QTMovie]);
    if ((int)timeValue != 0) setTimeIndex(theMovie, timeValue);
    
    // Keep the selection right
    //[model tableView:theTable shouldSelectRow:index];
    //[theTable selectRow:index byExtendingSelection:NO];
    //[theTable scrollRowToVisible:index];
    
    // Chapter Skipping
    if ([tInit boolForKey:@"skip"]) {
        MHChapters *ch = [theMovie chapters];
        int i;
        if (nil != ch) {
            NSString *string = [tInit stringForKey:@"skipText"];
            if (nil != string && ![string isEqualToString:@""]) {
                for (i=0;i<2;i++)
                    skips[i] = [ch rangeOfChapterContainingString:string index:i];
            }
        }
    }
    
    
    [self movieDidChangeTo:theMovie];
    
    [playingMovieView setMovie:theMovie];
    
    // Play the movie
        int i;
        for (i=0;i<[sizeMenu numberOfItems];i++)
            if ([[sizeMenu itemAtIndex:i] state])
                break;
		
    if ([toSize state]) {
        switch (i) {
            case 0: [self setViewSize:0.5]; break;
            case 1: case 2: [self setViewSize:i]; break;
            default: break;
        }
	}
        switch (i) {
            case 0: case 1: case 2: [viewWindow setTitle:currentTitle];
                if ([[controller mainTabViewName] isEqualToString:@"Movie"])
					[controller setMainTabViewName:@"Main"];
				if (![viewWindow isVisible])
                    [viewWindow makeKeyAndOrderFront:nil];
                break;
            case 3:  [self fullScreen:nil]; break;
            case 4:  [self inMainWindow:nil]; break;
            default: break;
        }
    
    //SetMovieVolume([theMovie QTMovie], (short)[volScroll doubleValue]);
    
	//float r = [[[rateView subviews] objectAtIndex:0] floatValue];
	//if (!r) r = 1.0;
    [playingMovieView start:self]; //setRate:r];
    
    [iconDrawer close];
}

- (IBAction) slomo:(id)sender
{
    float speed = [[[[NSUserDefaults standardUserDefaults] objectForKey:@"controlSpeeds"] objectForKey:@"SloMo Speed"] floatValue];
    if (!speed) speed = .3;
    if (speed < 0) speed = 0 - speed;
    [self play:nil];
	//[playingMovieView play:nil];
    [playingMovieView setRate:speed];
}

- (IBAction) ff:(id)sender
{
    //float curSpeed = [playingMovieView rate];
    //[playingMovieView start:nil];
    float speed = [[[[NSUserDefaults standardUserDefaults] objectForKey:@"controlSpeeds"] objectForKey:@"FF Speed"] floatValue];
    if (!speed) speed = 2;
    if (speed < 0) speed = 0 - speed;
    if ([sender respondsToSelector:@selector(isDepressed)] && ![sender isDepressed]) {
            speed = oldSpeed;
            [self jump:0 - [playingMovieView rate] * [[[[NSUserDefaults standardUserDefaults] objectForKey:@"controlSpeeds"] objectForKey:@"Snap Back"] floatValue]];
            if ([ffMute state]) {
                short vol = GetMovieVolume([currentMovie QTMovie]);
                if (vol < 0) SetMovieVolume([currentMovie QTMovie], 0 - vol);
            }
    } else {
        oldSpeed = [playingMovieView rate];
        if (![playingMovieView isPlaying])
            oldSpeed = 0;
        if ([ffMute state]) {
            short vol = GetMovieVolume([currentMovie QTMovie]);
            if (vol > 0) SetMovieVolume([currentMovie QTMovie], 0 - vol);
        }
    }
    if (0 == speed) {
        [playingMovieView pause:nil];
        [self setPlaying:NO];
    } else {
        if (![[playingMovieView window] isVisible])
			[self play:nil];
		//[playingMovieView play:nil];
        [playingMovieView setRate:speed];
        [self setPlaying:YES];
    }
}

- (IBAction) fff:(id)sender
{
    float speed = [[[[NSUserDefaults standardUserDefaults] objectForKey:@"controlSpeeds"] objectForKey:@"FF Speed"] floatValue];
    if (!speed) speed = 2;
    if (speed < 0) speed = 0 - speed;
    if ([playingMovieView rate] >= speed) speed = [playingMovieView rate] * 2;
    if ([ffMute state]) {
        short vol = GetMovieVolume([currentMovie QTMovie]);
        if (vol > 0) SetMovieVolume([currentMovie QTMovie], 0 - vol);
    }
    if (![[playingMovieView window] isVisible])
		[self play:nil];
	//[playingMovieView play:nil];
    [playingMovieView setRate:speed];
    [self setPlaying:NO];
}

- (IBAction) rw:(id)sender
{
    float speed = [[[[NSUserDefaults standardUserDefaults] objectForKey:@"controlSpeeds"] objectForKey:@"RW Speed"] floatValue];
    if (!speed) speed = 2;
    if (speed > 0) speed = 0-speed;
    if ([sender respondsToSelector:@selector(isDepressed)] && ![sender isDepressed]) {
            speed = oldSpeed;
            if ([ffMute state]) {
                short vol = GetMovieVolume([currentMovie QTMovie]);
                if (vol < 0) SetMovieVolume([currentMovie QTMovie], 0 - vol);
            }
    } else {
        oldSpeed = [playingMovieView rate];
        if (![playingMovieView isPlaying])
            oldSpeed = 0;
        if ([ffMute state]) {
            short vol = GetMovieVolume([currentMovie QTMovie]);
            if (vol > 0) SetMovieVolume([currentMovie QTMovie], 0 - vol);
        }
    }
    if (0 == speed) {
        [playingMovieView pause:nil];
        [self setPlaying:NO];
    } else {
        if (![[playingMovieView window] isVisible])
			[self play:nil];
		//[playingMovieView play:nil];
        [playingMovieView setRate:speed];
        [self setPlaying:YES];
    }
}

- (IBAction) rrw:(id)sender
{
    float speed = [[[[NSUserDefaults standardUserDefaults] objectForKey:@"controlSpeeds"] objectForKey:@"RW Speed"] floatValue];
    if (!speed) speed = 2;
    if (speed > 0) speed = 0 - speed;
    if ([playingMovieView rate] <= speed) speed = [playingMovieView rate] * 2;
    if ([ffMute state]) {
        short vol = GetMovieVolume([currentMovie QTMovie]);
        if (vol > 0) SetMovieVolume([currentMovie QTMovie], 0 - vol);
    }
    if (![[playingMovieView window] isVisible])
		[self play:nil];
	//[playingMovieView play:nil];
    [playingMovieView setRate:speed];
    [self setPlaying:NO];
}

- (IBAction) jumpF:(id)sender
{
    NSObject *j = [[[NSUserDefaults standardUserDefaults] objectForKey:@"controlSpeeds"] objectForKey:@"Forward Jump"];
	double jumpBy = 0;
	if ([j isKindOfClass:[NSString class]] && [j isEqualToString:@"Chapter"]) {
		NSString *lang = nil;
		jumpBy = [currentMovie currentTimeIndexSeconds];
        if ([[currentMovie chapters] multilingual]) lang = [currentMovie currentLanguage];
		int i = [[currentMovie chapters] indexForTime:jumpBy];
		i++;
        jumpBy = [[currentMovie chapters] timeForChapterAtIndex:i language:lang] - jumpBy;
	} else
		jumpBy = [(NSString *)j floatValue];
    if (jumpBy < 0) jumpBy = 0 - jumpBy;
    [self jump:jumpBy];
	[playingMovieView showNote:[NSString stringWithFormat:@"%C", 10555]];
}

- (IBAction) jumpB:(id)sender
{
    NSObject *j = [[[NSUserDefaults standardUserDefaults] objectForKey:@"controlSpeeds"] objectForKey:@"Forward Jump"];
	double jumpBy = 0;
	if ([j isKindOfClass:[NSString class]] && [j isEqualToString:@"Chapter"]) {
		NSString *lang = nil;
		double time = [currentMovie currentTimeIndexSeconds];
        if ([[currentMovie chapters] multilingual]) lang = [currentMovie currentLanguage];
		int i = [[currentMovie chapters] indexForTime:time];
        jumpBy = [[currentMovie chapters] timeForChapterAtIndex:i language:lang];
		if (time - jumpBy < 1 && i > 0)
			jumpBy = [[currentMovie chapters] timeForChapterAtIndex:i-1 language:lang];
		jumpBy = time - jumpBy;
	} else
		jumpBy = [(NSString *)j floatValue];
    if (jumpBy > 0) jumpBy = 0 - jumpBy;
    [self jump:jumpBy];
	[playingMovieView showNote:[NSString stringWithFormat:@"%C", 10554]];
}

- (IBAction)jumpToSheet: (id)sender
{
	if ( [viewWindow isKeyWindow] == NO)
	{
		[viewWindow makeKeyAndOrderFront:sender];
	}

	//else [[controller getPlayer] playMovies:self]
	[jumpToText setStringValue: [MHUtil nSecondsToString:[self getTimeValue]]];
	[NSApp beginSheet: jumpSheet
	modalForWindow: viewWindow
		modalDelegate: nil
	didEndSelector: nil
			contextInfo: nil];
}

- (IBAction)endJumpToSheet: (id)sender
{
    [jumpSheet orderOut: NULL];
    [NSApp endSheet: jumpSheet];
}

- (IBAction)jumpTo: (id)sender
{
    NSString *input = [jumpToText stringValue];
    double seconds = [(tController *)controller stringToNSeconds:input];
    //if (seconds == 0 ) return;
    
    MHMovie *theMovie = (MHMovie *)[movieView movie];
	if (![theMovie isKindOfClass:[NSMovie class]])
		theMovie = [theMovie delegate];
    if (nil == theMovie || ![theMovie isKindOfClass:[MHMovie class]]) return;
    [theMovie setCurrentTimeIndexSeconds:seconds];
	/*
	selectText is only useful if you keep the JumpSheet active
	[jumpToText selectText:sender];
	 */
    [jumpToText setStringValue: [MHUtil nSecondsToString: seconds]];
    [self endJumpToSheet: sender];
    [self tick:self];
}

- (void) jump:(double)seconds
{
    MHMovie *theMovie = (MHMovie *)[playingMovieView movie];
	//if (![theMovie isKindOfClass:[NSMovie class]])
	//	theMovie = [theMovie delegate];
    Movie qtMovie = nil;
	if ([theMovie respondsToSelector:@selector(quickTimeMovie)])
		qtMovie = [theMovie quickTimeMovie];
	if (!qtMovie) return;
    TimeRecord tr;
    TimeValue tv;
    
    tv = GetMovieTime(qtMovie, &tr);
    tr.value.lo += seconds * (double)tr.scale;
    if (tr.value.lo < 0) tr.value.lo = 0;
    SetMovieTime(qtMovie, &tr);
    [self tick:self];
}

- (void)copyTime:(id)sender
{
	float seconds = [self getTimeValue];
	long s, m, h, d, w, y;
	s = m = h = d = w = y = -1;
	TimeValue64 f = 0;
	
	float x = -1;
	
	NSString *form = [[[NSUserDefaults standardUserDefaults] objectForKey:@"format"] lowercaseString];
	if (!form) form = @"h:mm:ss.x";
	if ([form hasPrefix:@"{\\rtf"]) {
		NSAttributedString *astring = [[NSAttributedString alloc] initWithRTF:[NSData dataWithBytes:[form UTF8String] length:[form length]] documentAttributes:nil];
		int i, n = [astring length];
		NSMutableString *formula = [NSMutableString stringWithString:@""];
		for (i=0;i<n;i++) {
			NSDictionary *dict = [astring attributesAtIndex:i effectiveRange:nil];
			NSColor *color = [dict objectForKey:NSForegroundColorAttributeName];
			if (!color)
				[formula appendFormat:@"%C", [[astring string] characterAtIndex:i]];
		}
		form = formula;
		[astring release];
	}
	NSString *nom = [[NSUserDefaults standardUserDefaults] objectForKey:@"denom"];
	if (!nom) nom = @"1000";
	int n = [nom intValue];
	
	NSRange r = NSMakeRange(0, [form length]);
	NSMutableString *form1 = [NSMutableString stringWithString:form];
	if ([form1 replaceOccurrencesOfString:@"x" withString:@"1" options:nil range:r]) {
		x = seconds;
		x *= n;
	} if ([form1 replaceOccurrencesOfString:@"s" withString:@"2" options:nil range:r]) {
		s = seconds;
		x -= s * n;
	}
	if ([form1 replaceOccurrencesOfString:@"m" withString:@"3" options:nil range:r]) {
		m = seconds / 60;
		s -= m * 60;
	}
	if ([form1 replaceOccurrencesOfString:@"h" withString:@"4" options:nil range:r]) {
		h = seconds / 60 / 60;
		m -= h * 60;
		if (m < 0) s -= h * 60 * 60;
	}
	if ([form1 replaceOccurrencesOfString:@"d" withString:@"5" options:nil range:r]) {
		d = seconds / 60 / 60 / 24;
		h -= d * 24;
		if (h < 0) {
			m -= d * 24 * 60;
			if (m < 0)
				s -= d * 24 * 60 * 60;
		}
	}
	if ([form1 replaceOccurrencesOfString:@"w" withString:@"6" options:nil range:r]) {
		w = seconds / 60 / 60 / 24 / 7;
		d -= w * 7;
		if (d < 0) {
			h -= w * 7 * 24;
			if (h < 0) {
				m -= w * 7 * 24 * 60;
				if (m < 0)
					s -= w * 7 * 24 * 60 * 60;
			}
		}
	}
	if ([form1 replaceOccurrencesOfString:@"y" withString:@"7" options:nil range:r]) {
		y = seconds / 60 / 60 / 24 / 365;
		if (w > 0) w -= y * 52; // not quite but it will have to do
		else d -= y * 365;
		if (d < 0) {
			h -= y * 365 * 24;
			if (h < 0 ) {
				m -= y * 365 * 24 * 60;
				if (m < 0)
					s -= y * 365 * 24 * 60 * 60;
			}
		}
	}
	if ([form1 replaceOccurrencesOfString:@"f" withString:@"8" options:nil range:r]) {
		Movie q = [currentMovie QTMovie];
		Track t = GetMovieIndTrackType(q, 1, VideoMediaType, movieTrackMediaType);
		if (!t) t = GetMovieIndTrackType(q, 1, MPEGMediaType, movieTrackMediaType);
		NSLog(@"counting frames: %@", form);
		TimeRecord tr;
		TimeValue now = GetMovieTime(q, &tr);
		TimeValue start = (int)seconds * tr.scale;
		TimeValue end = GetMovieDuration(q);
		if (s < 0 && m < 0 && h < 00 && d < 0 && w < 0 && y < 0 && x < 0) // only frames: whole movie, like .sub
			start = 0;
		OSType myType = VideoMediaType;
		while (start >= 0 && start < now && start < end) {
			f++;
			GetTrackNextInterestingTime(t, nextTimeMediaSample, start, fixed1, &start, nil);
			//if (f % 10000 == 0) NSLog(@"%qi", (long long)f);
		}
		NSLog(@"count: %qi", (long long)f);
		
		Media m = GetTrackMedia(t);
		//long max = GetMediaSampleCount(m);
		TimeValue64 dtime = TrackTimeToMediaTime(now, t);
		SInt64 scount = 0;
		MediaDecodeTimeToSampleNum(m, dtime, &scount, nil, nil);
		NSLog(@"DecodeTimeToSampleNum: %qi", scount);
		scount = 0;
		MediaDisplayTimeToSampleNum(m, dtime, &scount, nil, nil);
		NSLog(@"DisplayTimeToSampleNum: %qi", scount);
		scount = 0;
		long samcount = 0;
		MediaTimeToSampleNum(m, TrackTimeToMediaTime(now, t), &samcount, nil, nil);
		NSLog(@"TimeToSampleNum: %ld", samcount);
		
	}
	NSScanner *scan = [NSScanner scannerWithString:form1];
	NSCharacterSet *set = [NSCharacterSet characterSetWithCharactersInString:@"12345678"];
	NSMutableString *form2 = [NSMutableString stringWithString:@""];
	
	int amount, digits;
	while (![scan isAtEnd]) {
		NSString *buffer = nil;
		amount = -1;
		[scan scanUpToCharactersFromSet:set intoString:&buffer];
		if ([buffer length]) [form2 appendString:buffer];
		[scan scanInt:&amount];
		if (amount > 0) {
			digits = log10(amount) + 1;
			long v = -1;
			switch (amount % 10) {
				case 1: v = x;
					if (digits < log10(n) + 0.999)// if the denom overflows the format, you want the denom
						digits = log10(n) + 0.999;
					break;
				case 2: v = s; break;
				case 3: v = m; break;
				case 4: v = h; break;
				case 5: v = d; break;
				case 6: v = w; break;
				case 7: v = y; break;
				case 8: v = f; break;
				default:
					break;
			}
			if (v > -1) {
				if (digits < log10(v)+0.999) // if the value overflows the format, you probably still want the value
					digits = log10(v)+0.999;
				if (digits > 1) buffer = [NSString stringWithFormat:@"%%.%dd", digits];
				else buffer = [NSString stringWithString:@"%d"];
			
				[form2 appendFormat:buffer, v];
			}
		}
	}
	
	//NSString *time = [NSString stringWithFormat:@"%d:%.2d:%.2d%@%@", h, m, s, lim, nom];
	//NSLog(@"%@ (%d)", form2, n);
	NSArray *types = [NSArray arrayWithObject:NSStringPboardType];
	[[NSPasteboard generalPasteboard] declareTypes:types owner:self];
	[[NSPasteboard generalPasteboard] setString:form2 forType:NSStringPboardType];
}

- (IBAction)gotoChapter:(id)sender
{
    if ([sender respondsToSelector:@selector(menu)]) {
        NSMenu *menu = [sender menu];
        int index = [menu indexOfItem:sender];
        NSString *lang = nil;
        if ([[currentMovie chapters] multilingual]) lang = [currentMovie currentLanguage];
        double time = [[currentMovie chapters] timeForChapterAtIndex:index language:lang];
        if (time >= 0) {
            TimeRecord tr;
			Movie q = [[playingMovieView movie] quickTimeMovie];
			TimeValue tv = GetMovieTime (q, &tr);
			tr.value.lo = time * (double)tr.scale;
			SetMovieTime(q, &tr);
            [scrubber setFloatValue:(float)GetMovieTime(q, &tr) / (float)GetMovieDuration(q)];
            [scrubber2 setFloatValue:(float)GetMovieTime(q, &tr) / (float)GetMovieDuration(q)];
        }
        //NSLog(@"%@", title);
    } else NSLog(@"gotoChapter: sender is does not respond to -menu!");
}

- (void)tick:(id)info
{
	MHMovie *movie = [playingMovieView movie];
	if (!movie) return;
	if (![movie isKindOfClass:[NSMovie class]])
		movie = [movie delegate];
	Movie q = [movie QTMovie];
	TimeRecord tr;
	TimeValue now = GetMovieTime(q, &tr);
	TimeValue max = GetMovieDuration(q);
	if (timescale && ![movie isOffline])
		[movie setTimeIndex:now];
	double seconds = [[movie objectForKey:@"Seconds Elapsed"] floatValue];
	int dsec = (seconds - (int)seconds) * 100 + 0.5; // for rounding
	double duration = [[movie objectForKey:@"Seconds"] floatValue];
	int i, n;
	[scrubber setFloatValue:seconds / duration];
	[scrubber2 setFloatValue:seconds / duration];
	NSString *timeStringE = [MHUtil nSecondsToShortString:seconds];
	NSString *timeStringR = [MHUtil nSecondsToShortString:duration - seconds];
	NSString *msec = @"";
	float interval = [tInit doubleForKey:t_interval];
	if (interval < 0.1) msec = [NSString stringWithFormat:@".%.2d", dsec];
	else if (interval < 1) msec = [NSString stringWithFormat:@".%d", dsec/10];
	if ([frames state]) {
		TimeValue prev = (int)seconds * tr.scale;
		int fr = 0;
		Track t = GetMovieIndTrackType(q, 1, VideoMediaType, movieTrackMediaType);
		if (t) {
			while (YES) {
			//const OSType myType = VideoMediaType;
				GetTrackNextInterestingTime(t, nextTimeMediaSample, prev, fixed1, &prev, nil);
				if (prev > now || prev >= max) break;
				fr++;
				if (fr > 200) break;
			}
			msec = [msec stringByAppendingFormat:@"(%.2d)", fr];
		}
	}
	[viewWindow setTitle:[NSString stringWithFormat:@"%@ - %@%@", currentTitle, timeStringE, msec]];
	n = [timeE count];
	for (i=0;i<n;i++) {
			[[timeE objectAtIndex:i] setStringValue:timeStringE];
			[[timeR objectAtIndex:i] setStringValue:timeStringR];
	}
	
    if ([playOn state] && IsMovieDone(q))
            [self makeMovieGoNow:nextMovie];
    if (nil == currentMovie && !info) return;
    unsigned int flags = [[NSApp currentEvent] modifierFlags];
    if (!(flags & NSAlternateKeyMask)) {
        if ([playingMovieView rate] != 0 && [playingMovieView isPlaying]) [self setPlaying:YES];
        else [self setPlaying:NO];
    }
    if (![playingMovieView isPlaying] && !info) return;
    if (!currentMovie) return;
    if ([tInit boolForKey:@"skip"]) {
        long time = [self getSampleTimeValue];
        int i;
        NSRange aRange;
        for (i=0;i<2;i++) {
            aRange = skips[i];
            if (aRange.length == 'end')
                aRange.length = GetMovieDuration(q) - aRange.location;
            if (time > aRange.location && time < aRange.location + aRange.length) {
                [currentMovie setCurrentTimeIndex:aRange.location + aRange.length];
                break;
            }
        }
    }
    
    if ([chapScroll indexOfSelectedItem] < 1) {
        NSString *language = nil;
        if ([[currentMovie chapters] multilingual]) language = [currentMovie currentLanguage];
        NSString *chapName = [[currentMovie chapters] nameForTime:seconds language:language];
        if (!chapName) chapName = @"Chapters";
		//chapName = [chapName whitespaceTrimmed];
        [menuChap setTitle:chapName];
		[menuChap2 setTitle:chapName];
    }
    [self refreshVolume:nil];
}

-(void)meters
{
	NSMovie *m = [playingMovieView movie];
	if (![m isKindOfClass:[NSMovie class]]) m = [m delegate];
	OSStatus err = GetMovieAudioFrequencyLevels([m QTMovie], kQTAudioMeter_MonoMix, (QTAudioFrequencyLevels *)freqResults);
	QTAudioFrequencyLevels *freq = (QTAudioFrequencyLevels *)freqResults;
	Float32 value = freq->level[0]*100;
	float ratio = 0;
	if (value)
		ratio = capTarget / value;
	else
		ratio = 2;
	if (ratio < 0.6 || ratio > 1.4) {
		if (ratio > 1.2) ratio = 1.2;
		short v = GetMovieVolume([m QTMovie]);
	//NSLog(@"r: %g value: %g", ratio, value);
		if (!v) v = 100;
		v = (float)v * ratio;
		if (v > loudVol) v = loudVol;
	//NSLog(@"v: %d", v);
		SetMovieVolume([m QTMovie], v);
		[volScroll setDoubleValue:v];
		[volScroll2 setDoubleValue:v];
	}
	//} else if (value > 25) {
	//	softCount = 0;
	//} else if (value < 10) {
	//	softCount++;
	//	if (softCount > 10)
	//		SetMovieVolume([m QTMovie], loudVol);
	//}
	
}

- (int)whichExit
{
    return [exit indexOfSelectedItem];
}

+ (int)whichExit
{
    return [tPlayNibInstance whichExit];
}

- (void)shouldFreeMovie:(MHMovie *)aMovie
{
    if (aMovie == currentMovie || aMovie == nil)
        currentMovie = nil;
}

@end
