/* Copyright (c) 2004, Noah Sorscher
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 	 	Redistributions of source code must retain the above copyright
 * 	notice, this list of conditions and the following disclaimer.
 * 	 	Redistributions in binary form must reproduce the above copyright
 * 	notice, this list of conditions and the following disclaimer in the
 * 	documentation and/or other materials provided with the distribution.
 * 	 	Neither the name of the Noah Sorscher nor the names of its
 * 	contributors may be used to endorse or promote products derived from
 * 	this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
//
//  tScanner.m
//  Hootenanny
//
//  Created by Noah Sorscher on Wed Mar 26 2003.
//  Copyright (c) 2003 __MyCompanyName__. All rights reserved.
//

#import "tScanner.h"

@implementation NSScanner(CSWAdditions)
/*" Additions to make parsing easier. "*/

/*"
 * Scans up to and past string, except if string immediately follows the current position, in which
 * case it just scans past it.
"*/ 
- (BOOL)scanPastString:(NSString *)string {
    if (![string length]) return NO;
    if ([self scanString:string intoString:NULL]) {
        return YES;
    } else {
        if ([self scanUpToString:string intoString:NULL]) {
            NSCharacterSet *skipped = [self charactersToBeSkipped];
            while ([string length] && [skipped characterIsMember:[string characterAtIndex:0]])
                string = [string substringFromIndex:1];
            if (![string length])
                return YES;
            if ([self scanString:string intoString:NULL])
                return YES;
        }
    }
    return NO;
}

- (BOOL)superScanString:(NSString *)string
{
    return [self superScanString:string intoString:nil];
}

- (BOOL)superScanString:(NSString *)string intoString:(NSString **)buffer
{
    NSString *next = [[self string] substringFromIndex:[self scanLocation]];
    if ([next hasPrefix:string]) {
        [self setScanLocation:[self scanLocation]+[string length]];
        if (buffer)
            *buffer = [string copy];
        return YES;
    }
    return [self scanString:string intoString:buffer];
}
@end