/* Copyright (c) 2004, Noah Sorscher
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 	 	Redistributions of source code must retain the above copyright
 * 	notice, this list of conditions and the following disclaimer.
 * 	 	Redistributions in binary form must reproduce the above copyright
 * 	notice, this list of conditions and the following disclaimer in the
 * 	documentation and/or other materials provided with the distribution.
 * 	 	Neither the name of the Noah Sorscher nor the names of its
 * 	contributors may be used to endorse or promote products derived from
 * 	this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
#import "tTracks.h"
#import <QTKit/QTKit.h>

@interface MHSliderCell : NSSliderCell
{
	BOOL myHidden;
}
- (void)setTransparent:(BOOL)flag;
@end

@implementation MHSliderCell
- (id)init
{	myHidden = NO;
	return [super init];	}
- (void)setTransparent:(BOOL)flag
{	myHidden = flag;	}
- (void)drawWithFrame:(NSRect)cellFrame inView:(NSView *)controlView
{	if (myHidden) return;
	[super drawWithFrame:cellFrame inView:controlView];	}
@end


@implementation MHComboBoxCell

-(id)init
{
	myHidden = NO;
	return [super init];
}
-(void)setTransparent:(BOOL)flag
{
	myHidden = flag;
}
- (void)drawWithFrame:(NSRect)cellFrame inView:(NSView *)controlView
{
	if (myHidden)return;
	[super drawWithFrame:cellFrame inView:controlView];
}

@end



@implementation tTracks


- (IBAction)open:(id)sender
{
    //[tracksWindow makeKeyAndOrderFront:nil];
    [controller setMainTabViewName:@"Tracks"];
    [self movieDidChangeTo:[controller currentMovie]];
    [self tableViewSelectionDidChange:nil];
}

- (void)textDidChange:(NSNotification *)aNotification
{
    //NSLog(@"textDidChange");
    if (recursion) return;
    NSTextView *box = [aNotification object];
    
    int col = [tracksTable editedColumn];
    if (col < 0) return;
    NSString *ident = [[[tracksTable tableColumns] objectAtIndex:col] identifier];
    if (![ident isEqualToString:@"Language"]) return;
    
    NSRange selection = [box selectedRange];
    NSString *string = [NSString stringWithString:[box string]];
    int initialLength = [string length];
    if (selection.location != initialLength) return;
    NSMutableArray *list = [NSMutableArray arrayWithArray:[tLanguages stringLangs]];
    [list addObject:@"Unspecified"];
    int i, n = [list count];
    recursion = YES;
    for (i=0;i<n;i++) {
        NSString *lang = [list objectAtIndex:i];
        if ([lang hasPrefix:string]) {
            lang = [lang substringFromIndex:initialLength];
            [box insertText:lang];
            selection.length = [lang length];
            [box setSelectedRange:selection];
            break;
        }
    }
    recursion = NO;
}

- (IBAction)delete:(id)sender
{
    if ([movie isOffline]) return;
	NSArray *list = [self multis];
	int j, m = [list count];
	for (j=0;j<m;j++) {
		MHMovie *m = [list objectAtIndex:j];
		BOOL changed = NO;
		int i, n = [m trackCount];
		TimeValue offset, lowOffset = [[m trackAtIndex:0] offset];
		for (i=n-1;i>=0;i--) {
			if ([tracksTable isRowSelected:i]) {
				[m removeTrackAtIndex:i];
				//[[m trackAtIndex:i] remove];
				changed = YES;
			} else {
				offset = [[m trackAtIndex:i] offset];
				if (offset < lowOffset) lowOffset = offset;
			}
		}
		if (changed) {
			[m setSaveInPlaceUnSafe];
			if (lowOffset) {
				n = [m trackCount];
				for (i=0;i<n;i++)
					[[m trackAtIndex:i] setOffset:[[m trackAtIndex:i] offset] - lowOffset];
			}
			[m gatherData];
		}
	}
	NSMovie *aMovie = movie;
	movie = nil;
	[self movieDidChangeTo:aMovie];
	[tracksTable reloadData];
}

- (IBAction)alternateGroup:(id)sender
{
    if ([movie isOffline]) return;
	NSArray *list = [self multis];
	int j, m = [list count];
	for (j=0;j<m;j++) {
		MHMovie *m = [list objectAtIndex:j];
		int i, n = [m trackCount];
		NSMutableArray *selected = [NSMutableArray array];
		for (i=0;i<n;i++)
			if ([tracksTable isRowSelected:i])
				[selected addObject:[m trackAtIndex:i]];
		if (![selected count])
			for (i=0;i<n;i++)
				if ([[m trackAtIndex:i] isSoundType])
					[selected addObject:[m trackAtIndex:i]];
		n = [selected count];
		if (n > 1) {
			MHTrack *lastTrack = nil, *thisTrack;
			for (i=0;i<n;i++) {
				thisTrack = [selected objectAtIndex:i];
				[lastTrack setAlternate:thisTrack];
				lastTrack = thisTrack;
			}
			[lastTrack setAlternate:[selected objectAtIndex:0]];
		}
	}
	[tracksTable reloadData];
}

- (IBAction)autoSetting:(id)sender
{
	NSArray *list = [self multis];
	int i, n = [list count];
	for (i=0;i<n;i++) {
		UserData ud = GetMovieUserData([[list objectAtIndex:i] QTMovie]);
		OSType t = 'ptv ';
		int n = 100;
		OSErr err = 0;
		Handle h = nil;
		char ch = 1;
		while (n = CountUserDataType(ud, t))
			err = RemoveUserData(ud, t, 1);
		if ([aFull state]) {
			h = NewHandleClear(8);
			unichar letters[4];
			letters[0] = 3;
			letters[1] = letters[2] = 0;
			letters[3] = 1;
#ifdef __LITTLE_ENDIAN__
			letters[0] = NSSwapShort(letters[0]);
			letters[3] = NSSwapShort(letters[3]);
#endif
			BlockMoveData(letters, *h, 4*sizeof(unichar));
			err = AddUserData(ud, h, t);
			DisposeHandle(h);
		}
		
		OSType types[3] = {'play', 'aclo', 'aqui' };
		NSArray *buttons = [NSArray arrayWithObjects:aPlay, aClose, aQuit, nil];
		int i;
		for (i=0;i<3;i++) {
			t = types[i];
			while (n = CountUserDataType(ud, t))
				err = RemoveUserData(ud, t, 1);
			if ([[buttons objectAtIndex:i] state]) {
				h = NewHandleClear(1);
				BlockMoveData(&ch, *h, 1);
				err = AddUserData(ud, h, t);
				DisposeHandle(h);
			}
		}
	}
}

- (void)movieDidChangeTo:(NSMovie *)aMovie
{
    //if (![tracksWindow isVisible]) return;
    movie = (MHMovie *)aMovie;
    if (!movie) return;
    [tracksTable reloadData];
	
	/*
	QTMovie *m = [movie qtKitMovie];
	NSString *attr = [m attributeForKey:QTMovieApertureModeAttribute];
	if (attr) NSLog(attr);
	else NSLog(@"no");
	*/
	[aFull setState:NO];
	[aPlay setState:NO];
	[aClose setState:NO];
	[aQuit setState:NO];
	Handle h = NewHandleClear(0);
	OSType t = 'ptv ';
	UserData ud = GetMovieUserData([movie QTMovie]);
	OSErr err = GetUserData(ud, h, t, 1);
	char ch = 0;
	if (!err) {
		unichar *samp = (unichar *)*h;
		long size = GetHandleSize(h);
		if (size >= 4*sizeof(unichar)) {
			short s = samp[3];
#ifdef __LITTLE_ENDIAN__
			s = NSSwapShort(s);
#endif
			if (1 == s)
				[aFull setState:YES];
		}
	}
	t = 'play';
	err = GetUserData(ud, h, t, 1);
	if (!err) {
		ch = **h;
		if (ch == 1)
			[aPlay setState:YES];
	}
	t = 'aclo';
	err = GetUserData(ud, h, t, 1);
	if (!err) {
		ch = **h;
		if (ch == 1)
			[aClose setState:YES];
	}
	t = 'aqui';
	err = GetUserData(ud, h, t, 1);
	if (!err) {
		ch = **h;
		if (ch == 1)
			[aQuit setState:YES];
	}
}

id tracksNibInstance;

- (void)awakeFromNib
{
	tracksNibInstance = self;
    //NSLog(@"tTracks");
    didDelete = recursion = NO;
	[tracksTable setColumnAutoresizingStyle:NSTableViewNoColumnAutoresizing];
	
	NSTableColumn *col = [[NSTableColumn alloc] initWithIdentifier:@"Transparency"];
	MHSliderCell *slide = [[MHSliderCell alloc] init];
	[slide setControlSize:NSSmallControlSize];
	[col setDataCell:slide];
	[[col headerCell] setStringValue:@"Transparency"];
	[tracksTable addTableColumn:col];
	
	col = [[NSTableColumn alloc] initWithIdentifier:@"Graphics Mode"];
	NSPopUpButtonCell *pop = [[NSPopUpButtonCell alloc] init];
	[pop setControlSize:NSSmallControlSize];
	[col setDataCell:pop];
	[pop addItemsWithTitles:[NSArray arrayWithObjects:@"srcCopy", @"srcOr", @"srcXor", @"srcBic", @"notSrcCopy", @"notSrcOr", @"notSrcXor", @"notSrcBic",
							 @"patCopy", @"patOr", @"patXor", @"patBic", @"notPatCopy", @"notPatOr", @"notPatXor", @"notPatBic", @"grayishTextOr",
							 @"hilite", @"blend", @"addPin", @"addOver", @"subPin", @"transparent", @"addMax", @"subOver", @"adMin", @"ditherCopy",
							 @"straightAlpha", @"prewhiteAlpha", @"preblackAlpha", @"composition", @"straightAlphaBlend", @"preMulColorAlpha",
							 @"perComponentAlpha", nil]];
	[[col headerCell] setStringValue:@"Graphics Mode"];
	[tracksTable addTableColumn:col];
	col = [[NSTableColumn alloc] initWithIdentifier:@"DataRef"];
	[[col headerCell] setStringValue:@"DataRef"];
	[col setWidth:300];
	[tracksTable addTableColumn:col];
	
	[tracksTable setIntercellSpacing:NSMakeSize(0,5)];
	[tracksTable setUsesAlternatingRowBackgroundColors:YES];
    [tracksTable setTarget:self];
    [tracksTable setDoubleAction:@selector(doubleClick)];
    [controller performSelector:@selector(addMovieChangeListener:) withObject:self afterDelay:0];
    [tracksTable registerForDraggedTypes:[NSArray arrayWithObject:NSFilenamesPboardType]];
	col = [tracksTable tableColumnWithIdentifier:@"Rotate"];
	MHComboBoxCell *cell = [[MHComboBoxCell alloc] init];
	[cell setEditable:YES];
	[cell addItemsWithObjectValues:[NSArray arrayWithObjects:@"90", @"180", @"270", @"Flip Horizontal", @"Flip Vertical", nil]];
	[col setDataCell:cell];	
	NSArray *bools = [NSArray arrayWithObjects:@"Enabled", @"Preload", @"High Quality", nil];
	int i, n = [bools count];
	for (i=0;i<n;i++) {
		col = [tracksTable tableColumnWithIdentifier:[bools objectAtIndex:i]];
		if (col) {
			NSButtonCell *box = [[NSButtonCell alloc] initTextCell:@""];
			[box setButtonType:NSSwitchButton];
			[box setControlSize:NSSmallControlSize];
			[col setDataCell:box];
		}
	}
	
	
	[[NSNotificationCenter defaultCenter] addObserver:tracksTable selector:@selector(reloadData) name:@"movieLanguageChangedNotification" object:nil];
}

- (void)tableView:(NSTableView *)tview willDisplayCell:(id)cell forTableColumn:(NSTableColumn *)tcol row:(int)row
{
	//NSLog([cell className]);
	MHTrack *t = [movie trackAtIndex:row];
	NSArray *visuals = [NSArray arrayWithObjects:@"Rotate", @"Transparency", @"Graphics Mode", nil];
	if ([visuals containsObject:[tcol identifier]])
	{
		NSString *type = [t type];
		if ([type isEqualToString:@"Sound"]) {
			if ([cell respondsToSelector:@selector(setTransparent:)])
				[cell setTransparent:YES];
			[cell setEnabled:NO];
		} else {
			if ([cell respondsToSelector:@selector(setTransparent:)])
				[cell setTransparent:NO];
			[cell setEnabled:YES];
		}
		return;
	}
	NSArray *bools = [NSArray arrayWithObjects:@"Enabled", @"Preload", @"High Quality", nil];
	int i, n = [bools count];
	for (i=0;i<n;i++) {
		if ([[tcol identifier] isEqualToString:[bools objectAtIndex:i]] && [cell respondsToSelector:@selector(setState:)]) {
			NSInteger state = NSOffState;
			switch (i) {
				case 0: if ([t enabled]) state = NSOnState; break;
				case 1: if ([t preloads]) state = NSOnState; break;
				case 2: if ([t highQuality]) state = NSOnState; break;
				default: break;
			}
			[cell setState:state];
			return;
		}
	}
}

- (int)numberOfRowsInTableView:(NSTableView *)aTableView
{	return [movie trackCount];	}

- (id)tableView:(NSTableView *)aTableView objectValueForTableColumn:(NSTableColumn *)aTableColumn row:(int)rowIndex
{
    if (rowIndex >= [movie trackCount] || rowIndex < 0) return nil;
    MHTrack *track = [movie trackAtIndex:rowIndex];
    NSString *s = [aTableColumn identifier];
    if ([s isEqualToString:@"Name"]) {
        NSString *name = [track name];
        if (!name) name = [NSString stringWithFormat:@"%d %@ Track", rowIndex+1, [track type]];
        return name;
    } else if ([s isEqualToString:@"Enabled"]) return ([track enabled] ? @"YES" : @"NO");
    else if ([s isEqualToString:@"Offset"]) return [NSString stringWithFormat:@"%ld", [track offset]];
    else if ([s isEqualToString:@"Duration"]) return [NSString stringWithFormat:@"%ld", [track duration]];
    else if ([s isEqualToString:@"Language"]) return [track language];
    else if ([s isEqualToString:@"Type"]) return [track type];
    else if ([s isEqualToString:@"FourCC"]) return [track fourCC];
    else if ([s isEqualToString:@"Vendor"]) return [track vendor];
    else if ([s isEqualToString:@"Size"]) return [NSString stringWithFormat:@"%g", [track mbSize]];
    else if ([s isEqualToString:@"TimeScale"]) return [NSString stringWithFormat:@"%ld", [track timeScale]];
    else if ([s isEqualToString:@"Format"]) return [track format];
    else if ([s isEqualToString:@"Preload"]) return ([track preloads] ? @"YES" : @"NO");
    else if ([s isEqualToString:@"High Quality"]) return ([track highQuality] ? @"YES" : @"NO");
    else if ([s isEqualToString:@"Layer"]) return [NSString stringWithFormat:@"%d", [track layer]];
    else if ([s isEqualToString:@"Frame"]) return NSStringFromRect([track displayFrame]);
	else if ([s isEqualToString:@"Volume"]) return [NSNumber numberWithFloat:[track volume]];
	else if ([s isEqualToString:@"Transparency"]) return [NSNumber numberWithFloat:[track transparency]];
	else if ([s isEqualToString:@"Graphics Mode"]) {
		long mode = [track graphicsMode];
		if (mode == 64) mode = 26;
		else if (mode == 272) mode = 33;
		else if (mode > 255) mode -= 229;
		else if (mode > 48) mode -= 33;
		else if (mode > 31) mode -= 14;
		//NSPopUpButtonCell *pop = [aTableColumn dataCellForRow:rowIndex];
		//[pop selectItemAtIndex:mode];
		return [NSNumber numberWithInt:mode];
	} else if ([s isEqualToString:@"DataRef"]) return [track dataRef];
    return nil;
}

- (NSArray *)multis
{
	int count = 0;
	NSTableView *mtable = [controller mainTable];
	MHMovie *m = nil;
	int i, n = [mtable numberOfRows];
	for (i=0;i<n;i++) {
		if ([mtable isRowSelected:i]) count++;
		//if (count > 1) break;
	}
	NSMutableArray *list = [NSMutableArray arrayWithObjects:movie, nil];
	int result = NSAlertDefaultReturn;
	if (count > 1 && ![[NSUserDefaults standardUserDefaults] boolForKey:@"multiSelNoAsk"]) {
		result = NSRunAlertPanel(@"Multiple Selection", [NSString stringWithFormat:@"Do you want to edit all %d selected movies or just one?", count], @"All", @"One", @"All, and don't ask again");
		if (result == NSAlertOtherReturn) {
			[[NSUserDefaults standardUserDefaults] setBool:YES forKey:@"multiSelNoAsk"];
			[[NSUserDefaults standardUserDefaults] synchronize];
			result = NSAlertDefaultReturn;
		}
	}
	if (result == NSAlertDefaultReturn) {
		for (i=0;i<n;i++) {
			if ([mtable isRowSelected:i]) {
				m = [[mtable dataSource] movieForIndex:i];
				if (m != movie) [list addObject:m];
			}
		}
	}
	return list;
}

- (void)tableView:(NSTableView *)aTableView setObjectValue:(id)anObject forTableColumn:(NSTableColumn *)aTableColumn row:(int)rowIndex
{
	[movieView drawRectangle:NSMakeRect(0, 0, 0, 0)];
	//NSLog([anObject className]);
    if (rowIndex >= [movie trackCount] || rowIndex < 0) return;
    if ([movie isOffline]) return;
	[movie setTimeIndex:[movie currentTimeIndex]];
	
	NSArray *list = [self multis];
    NSString *s = [aTableColumn identifier];
	int i, n = [list count];
	for (i=0;i<n;i++) {
		MHMovie *m = [list objectAtIndex:i];
		MHTrack *track = [m trackAtIndex:rowIndex];
		
		if ([s isEqualToString:@"Frame"]) {
			if ([track subtitleType]) {
				NSRect myRect = NSRectFromString(anObject);
				[movie removeSubs:nil];
				[movie addSeparateSubs:nil forceSize:myRect.size];
				int j;
				for (j=0;j<[movie trackCount];j++) {
					track = [m trackAtIndex:j];
					if ([track subtitleType]) {
						// set the origin, without changing size (again)
						myRect.size = [track displayFrame].size;
						[track setDisplayFrame:myRect];
					}
				}
			}
		} else if ([s isEqualToString:@"Graphics Mode"]) {
			long mode = [anObject intValue];
			if (mode == 26) mode = 64;
			else if (mode == 33) mode = 272;
			else if (mode > 25) mode += 229;
			else if (mode > 17) mode += 14;
			else if (mode > 15) mode += 33;
			//[track setGraphicsMode:mode];
			anObject = [NSString stringWithFormat:@"%d", mode];
		} else if ([s isEqualToString:@"Transparency"]) {
			QTMovieView *v = nil;
			if ([movieView movie] != [movie qtKitMovie]) {  // don't know why, but movies that are "virgin" to a movieview cause a crash when transparency is set.
				v = [[QTMovieView alloc] init];
				[v setMovie:[movie qtKitMovie]];
			}
			[track setTransparency:0]; // so it will redraw
			//[track setTransparency:[anObject floatValue]];
			if (v) {
				[v setMovie: nil];
				[v release];
			}
		}
		
		NSString *a = [NSString stringWithFormat:@"%@|%@", s, anObject];
		[track makeAlterations:a];
		
		if ([s isEqualToString:@"Language"]) [controller setCurrentMovie:movie];
		/*
    if ([s isEqualToString:@"Name"]) [track setName:anObject];
	else if ([s isEqualToString:@"Enabled"]) [track setEnabled:[anObject boolValue]];
    else if ([s isEqualToString:@"Offset"]) [track setOffset:[anObject intValue]];
    else if ([s isEqualToString:@"Duration"]) [track setDuration:[anObject intValue]];
    else if ([s isEqualToString:@"FourCC"]) [track setFourCC:anObject];
    else if ([s isEqualToString:@"Vendor"]) [track setVendor:anObject];
    else if ([s isEqualToString:@"TimeScale"]) [track setTimeScale:[anObject intValue]];
    else if ([s isEqualToString:@"Format"]) [track setFormat:anObject];
	else if ([s isEqualToString:@"Preload"]) [track setPreloads:[anObject boolValue]];
	else if ([s isEqualToString:@"High Quality"]) [track setHighQuality:[anObject boolValue]];
    else if ([s isEqualToString:@"Layer"]) [track setLayer:(short)[anObject intValue]];
    else  else if ([s isEqualToString:@"Rotate"]) [track rotate:anObject];
    else if ([s isEqualToString:@"Transparency"]) {
		QTMovieView *v = nil;
		if ([movieView movie] != [movie qtKitMovie]) {  // don't know why, but movies that are "virgin" to a movieview cause a crash when transparency is set.
			v = [[QTMovieView alloc] init];
			[v setMovie:[movie qtKitMovie]];
		}
		[track setTransparency:0]; // so it will redraw
		[track setTransparency:[anObject floatValue]];
		if (v) {
			[v setMovie: nil];
			[v release];
		}
	} else if ([s isEqualToString:@"Graphics Mode"]) {
		long mode = [anObject intValue];
		if (mode == 26) mode = 64;
		else if (mode == 33) mode = 272;
		else if (mode > 25) mode += 229;
		else if (mode > 17) mode += 14;
		else if (mode > 15) mode += 33;
		[track setGraphicsMode:mode];
	} else continue;*/
		[m setSaveInPlaceUnSafe];
		[m gatherData];
	}
	[tracksTable reloadData];
}

- (NSString *)reduce:(NSString *)input
{
    int i, n = [input length];
    char *buffer = malloc((n+1)*sizeof(char));
    for (i=0;i<n;i++)
        buffer[i]=[input characterAtIndex:i]-1;
    buffer[i] = 0;
    NSString *output = [NSString stringWithCString:buffer];
    free(buffer);
    return output;
}

- (BOOL) tableView: (NSTableView *) tableView shouldEditTableColumn: (NSTableColumn *) tableColumn row: (int) rowIndex
{
    if ([movie isOffline]) return NO;
    NSArray *bad = [NSArray arrayWithObjects:@"Type", @"Enabled", @"Size", @"Preload", @"High Quality", @"DataRef", nil];
    NSString *idString = [tableColumn identifier];
    int i, n = [bad count];
    for (i=0;i<n;i++)
        if ([idString isEqualToString:[bad objectAtIndex:i]])
            return NO;
    if ([idString isEqualToString:@"Format"] && ![[NSUserDefaults standardUserDefaults] boolForKey:[self reduce:@"tvqfsFeju"]])
        return NO;
	if ([idString isEqualToString:@"Frame"]) {
		if ([[movieView window] isVisible])
			[movieView drawRectangle:[[movie trackAtIndex:rowIndex] displayFrame]];
	}
    return YES;
}

- (void)setTrackRect:(NSRect)r
{
	NSText *t = [[tracksTable window] fieldEditor:YES forObject:tracksTable];
	[t selectAll:self];
	[t setString:NSStringFromRect(r)];
}

+ (void)setTrackRect:(NSRect)r
{
	[tracksNibInstance setTrackRect:r];
}

- (void)doubleClick
{
    if ([movie isOffline]) return;
    int clickedColumn = [tracksTable clickedColumn];
	NSString *idString = nil;
	int row = [tracksTable clickedRow];
	NSTableColumn *tcol = [[tracksTable tableColumns] objectAtIndex:clickedColumn];
	if (-1 != clickedColumn) idString = [tcol identifier];
	if (-1 != row) {
        if ([idString isEqualToString:@"Enabled"]) {
            BOOL enabled = [[(MHMovie *)movie trackAtIndex:row] enabled];
            [[(MHMovie *)movie trackAtIndex:row] setEnabled:!enabled];
        } else if ([idString isEqualToString:@"Preload"]) {
            MHTrack *track = [(MHMovie *)movie trackAtIndex:row];
            [track setPreloads:![track preloads]];
        } else if ([idString isEqualToString:@"High Quality"]) {
            MHTrack *track = [(MHMovie *)movie trackAtIndex:row];
            [track setHighQuality:![track highQuality]];
        } else {
			if ([self tableView:tracksTable shouldEditTableColumn:tcol row:row]) {
				[tracksTable editColumn:clickedColumn row:row withEvent:[NSApp currentEvent] select:YES];
			}
			return;
		}
        [tracksTable reloadData];
    }
}

- (void)tableViewSelectionDidChange:(NSNotification *)aNotification
{
    switch ([tracksTable numberOfSelectedRows]) {
        case 0: [deleteButton setEnabled:NO]; break;
        case 1: [deleteButton setEnabled:YES];
                [deleteButton setTitle:@"Delete Track"]; break;
        default: [deleteButton setEnabled:YES];
                [deleteButton setTitle:@"Delete Tracks"]; break;
    }
	[movieView drawRectangle:NSMakeRect(0, 0, 0, 0)];
}

- (NSDragOperation)tableView:(NSTableView*)tableView validateDrop:(id <NSDraggingInfo>)info proposedRow:(int)row proposedDropOperation:(NSTableViewDropOperation)operation
{
    //if (!row || row == [playlistTable selectedRow]) return NSDragOperationNone;
    //NSLog(@"copy");
    if ([movie isOffline]) return NSDragOperationNone;
    return NSDragOperationCopy;
}

- (BOOL)tableView:(NSTableView*)tableView acceptDrop:(id <NSDraggingInfo>)info row:(int)row dropOperation:(NSTableViewDropOperation)operation
{
    if ([movie isOffline]) return NO;
    NSPasteboard *pboard = [info draggingPasteboard];
    if (![[pboard types] containsObject:NSFilenamesPboardType] ) return NO;
    NSArray *files = [pboard propertyListForType:NSFilenamesPboardType];
    int i, n = [files count];
    BOOL changed = NO;
    for (i=0;i<n;i++) {
		NSString *file = [files objectAtIndex:i];
		if ([movie addMovie:file]) {
        //MHMovie *newMovie = [[MHMovie alloc] initWithPath:[files objectAtIndex:i]];
		/*NSError *e = nil;
		QTMovie *newMovie = [QTMovie movieWithFile:file error:&e];
		if (e) NSLog(@"error opening file: %@", e);
        if (newMovie) {
			Movie q = [newMovie quickTimeMovie];
            int oldTrackCount = GetMovieTrackCount([movie QTMovie]);
            SetMovieSelection(q, 0, GetMovieDuration(q));
			AddMovieSelection([movie QTMovie], q);
            int j, newTrackCount = [movie trackCount];
            for (j=oldTrackCount; j<newTrackCount; j++) {
                [[movie trackAtIndex:j] setOffset:0];
				[[movie trackAtIndex:j] setSource:file];
				if (newTrackCount - oldTrackCount < 2) [[movie trackAtIndex:j] makeAlterations:[NSString stringWithFormat:@"Name|%@", [[file lastPathComponent] extensionRemoved]]];
            }*/
            [tracksTable reloadData];
            changed = YES;

			
        }
    }
    //.if (changed) [movie gatherData];
    return YES;
}

- (void)setMatte:(id)sender
{
	int row = [tracksTable selectedRow];
	MHTrack *track = [movie trackAtIndex:row];
	if (track) {
		MediaHandler mh = GetMediaHandler(GetTrackMedia([track QTTrack]));
		long mode = 0;
		RGBColor color;
		OSErr err = MediaGetGraphicsMode(mh, &mode, &color);
		
		
		NSOpenPanel *p = [NSOpenPanel openPanel];
		[p runModalForTypes:nil];
		NSArray *f = [p filenames];
		if ([f count]) {
			NSString *n = [f objectAtIndex:0];
			QTMovie *q = [QTMovie movieWithFile:n error:nil];
			if (q) {				
				GWorldPtr gworld = nil;
				Rect r;
				GetMovieBox([q quickTimeMovie], &r);
				err = NewGWorld(&gworld, 32, &r, nil, nil, 0);
				SetMovieGWorld([q quickTimeMovie], gworld, nil);
				MoviesTask([q quickTimeMovie], 0);
				PixMapHandle theMatte = GetGWorldPixMap(gworld);
				SetTrackMatte([track QTTrack], theMatte);
				 
				//RgnHandle r = GetMovieClipRgn([q quickTimeMovie]);
				//SetTrackClipRgn([track QTTrack], r);
			}
		}
	}
}

- (void)reconstitute:(id)sender
{
	int row = [tracksTable selectedRow];
	MHTrack *track = [movie trackAtIndex:row];
	if (!track) return;
	Media med = GetTrackMedia([track QTTrack]);
	Fixed w, h;
	GetTrackDimensions([track QTTrack], &w, &h);
	//QTMovie *mov = [QTMovie movieWithFile:@"/1.mov" error:nil];
	//Handle myDataRefHandl = NewHandleClear(4);
	//SetMovieDefaultDataRef([mov quickTimeMovie], myDataRefHandl, HandleDataHandlerSubType);
	Track t = NewMovieTrack([movie quickTimeMovie], w, h, kFullVolume);
	OSType type;
	Str255 str;
	SampleDescriptionHandle sdh = NewHandleClear(sizeof(SampleDescriptionHandle));
	GetMediaHandlerDescription(med, &type, str, nil);
	Media m = NewTrackMedia(t, type, GetMediaTimeScale(med), nil, 0);
	//[mov writeToFile:@"/1.mov" withAttributes:[NSDictionary dictionaryWithObject:[NSNumber numberWithBool:YES] forKey:QTMovieFlatten]];
	OSErr err = BeginMediaEdits(m);
	UInt8 *dataOut = nil;
	long  size, numberOfSamples, i, sampleCount = GetMediaSampleCount(med);
	TimeValue64 time = 0, duration = 0;
	OSType flags = 0;
	for (i=0;i<sampleCount;i++) {
		//SampleReferencePtr P = malloc(sampleCount * sizeof(SampleReferenceRecord));
		//OSErr err = GetMediaSampleReference(med, 0, nil, sdh, nil, sampleCount, &sampleCount, P);
		err = GetMediaSample2(med, nil, 0, &size, time, nil, nil, nil, nil, nil, 1, nil, &flags);
		dataOut = malloc(size);
		err = GetMediaSample2(med, dataOut, size, &size, time, &time, &duration, nil, sdh, nil, 1, &numberOfSamples, &flags);
		err = AddMediaSample2(m, dataOut, size, duration, 0, sdh, numberOfSamples, flags, nil);
		free(dataOut);
		dataOut = nil;
		time += duration;
	}
	err = EndMediaEdits(m);
	TimeValue dur = GetMediaDuration(m);
	err = InsertMediaIntoTrack(t, 0, 0, dur, Long2Fix(1));
	//[mov writeToFile:@"/2.mov" withAttributes:[NSDictionary dictionaryWithObject:[NSNumber numberWithBool:YES] forKey:QTMovieFlatten]];
	[movie gatherData];
	[tracksTable reloadData];
}

- (void)tableView:(NSTableView *)t rightClickWithEvent:(NSEvent *)e
{
	NSPoint p = [e locationInWindow];
	p = [t convertPoint:p toView:0];
	int row = [t rowAtPoint:p];
	//NSLog(@"row %d", row);
	MHTrack *track = [movie trackAtIndex:row];
	if (track) {
		NSMenu *m = [[[NSMenu alloc] init] autorelease];
		NSMenuItem *i = [[[NSMenuItem alloc] initWithTitle:@"Reconstitute" action:@selector(reconstitute:) keyEquivalent:@""] autorelease];
		[i setTarget:self];
		[m addItem:i];
		[NSMenu popUpContextMenu:m withEvent:e forView:t];
	}
}

@end
