/* Copyright (c) 2004, Noah Sorscher
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 	 	Redistributions of source code must retain the above copyright
 * 	notice, this list of conditions and the following disclaimer.
 * 	 	Redistributions in binary form must reproduce the above copyright
 * 	notice, this list of conditions and the following disclaimer in the
 * 	documentation and/or other materials provided with the distribution.
 * 	 	Neither the name of the Noah Sorscher nor the names of its
 * 	contributors may be used to endorse or promote products derived from
 * 	this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
#import "tWindow.h"
#import "tPlayback.h"

@implementation tWindow

- (id)init
{
	floater = nil;
	NSLog(@"init");
	return [super init];
}

- (BOOL)acceptsFirstResponder
{
	return YES;
}

- (BOOL)canBecomeKeyWindow
{
    return YES;
}

- (BOOL)acceptsFirstMouse:(NSEvent *)e
{	return YES;		}

- (void)mouseDragged:(NSEvent *)e
{
    [NSObject cancelPreviousPerformRequestsWithTarget:self];
    [self performSelector:@selector(hideCursor) withObject:nil afterDelay:2];
	NSPoint p = [NSEvent mouseLocation];
	NSRect frame = [floater frame];
	if (NSPointInRect(p, frame)) {
		p.x -= 115;
		p.y -= 35;
		[floater setFrameOrigin:p];
		[floater makeKeyAndOrderFront:self];
		[[NSUserDefaults standardUserDefaults] setFloat:p.x	forKey:@"floaterx"];
		[[NSUserDefaults standardUserDefaults] setFloat:p.y	forKey:@"floatery"];
		[[NSUserDefaults standardUserDefaults] synchronize];
	}
	[self makeFirstResponder:self];
}

- (void)hideCursor
{
	if (![self isVisible]) return;
	NSPoint p = [NSEvent mouseLocation];
	if (p.y > [[self screen] frame].size.height - [NSMenuView menuBarHeight]) return;
	NSRect frame = [floater frame];
	if (NSPointInRect(p, frame)) return;
    [NSCursor setHiddenUntilMouseMoves:YES];
	int levels[] = { NSNormalWindowLevel, NSFloatingWindowLevel, NSSubmenuWindowLevel, NSTornOffMenuWindowLevel, NSModalPanelWindowLevel, NSMainMenuWindowLevel, NSStatusWindowLevel, NSPopUpMenuWindowLevel, NSScreenSaverWindowLevel };
	[self setLevel:levels[[[NSUserDefaults standardUserDefaults] integerForKey:@"fullscreenLevel"]]];
	[self makeKeyAndOrderFront:self];
	[floater orderOut:self];
	//[self makeKeyAndOrderFront:self];
}

- (void)mouseMoved:(NSEvent *)theEvent
{
	if ([NSEvent mouseLocation].y > [[self screen] frame].size.height - [NSMenuView menuBarHeight])
		[self setLevel:NSFloatingWindowLevel];
    [NSObject cancelPreviousPerformRequestsWithTarget:self];
    [self performSelector:@selector(hideCursor) withObject:nil afterDelay:2];
	
	if (!floater) {
		NSArray *controls = [[NSApp delegate] controls];
		NSRect frame = [self frame];
		frame.origin.x = 25; //frame.size.width / 2;
		frame.origin.y = 25; //frame.size.height / 5;
		if ([[NSUserDefaults standardUserDefaults] objectForKey:@"floaterx"]) {
			NSPoint p;
			p.x = [[NSUserDefaults standardUserDefaults] floatForKey:@"floaterx"];
			p.y = [[NSUserDefaults standardUserDefaults] floatForKey:@"floatery"];
			if (NSPointInRect(p, frame))
				frame.origin = p;
		}
		frame.size.width /= 2;
		frame.size.height = 64;
		//frame.origin.x -= frame.size.width / 2;
		floater = [[NSWindow alloc] initWithContentRect:frame styleMask: NSBorderlessWindowMask backing:NSBackingStoreBuffered defer:NO screen:[self screen]];
		[floater setLevel:[self level]];
		[floater setBackgroundColor:[NSColor colorWithCalibratedWhite:0.0 alpha:0.0]];
		[floater setOpaque:NO];
		[floater setMovableByWindowBackground:YES];
		[floater setHidesOnDeactivate:[self hidesOnDeactivate]];
		
		[floater setContentView:[controls objectAtIndex:0]];
	}
	[floater makeKeyAndOrderFront:self];
}

- (NSRect)reservedFrame
{
	if (!floater) return NSMakeRect(0, 0, 0, 0);
	return [floater frame];
}

- (void)setLevel:(int)l
{
	[super setLevel:l];
	[floater setLevel:l];
}

- (void)scrollWheel:(NSEvent *)e
{
	NSArray *subs = [[self contentView] subviews];
	int i, n = [subs count];
	for (i=0;i<n;i++) {
		if ([[[subs objectAtIndex:i] className] isEqualToString:@"tMovieView"])
			[[subs objectAtIndex:i] scrollWheel:e];
	}
}
/*
- (void)rightMouseDown:(NSEvent *)e
{
	NSLog(@"right");
}
*/
- (void)mouseDown:(NSEvent *)theEvent
{
	NSPoint p = [NSEvent mouseLocation];
	NSRect frame = [floater frame];
	if (NSPointInRect(p, frame)) return;
	NSArray *subs = [[self contentView] subviews];
	int i, n = [subs count];
	for (i=0;i<n;i++) {
		if ([[[subs objectAtIndex:i] className] isEqualToString:@"tMovieView"])
			[[subs objectAtIndex:i] mouseDown:theEvent];
	}
	/*
	//NSLog(@"tWindow mouseDown:");
    if ([tInit boolForKey:@"exitMouse"])
        [self exitFullScreen];
    else
        [super mouseDown:theEvent];
	 */
}

- (BOOL)performKeyEquivalent:(NSEvent *)e
{
	if ([[e characters] characterAtIndex:0] == 'w' && [e modifierFlags] & NSCommandKeyMask) {
		[self exitFullScreen];
		return YES;
	}
	return [super performKeyEquivalent:e];
}

- (void)keyDown:(NSEvent *)theEvent
{
    
    int exitKey = [tPlayback whichExit];
    
	//NSLog(@"tWindow: %@", [theEvent characters]);
	if (exitKey == 1 || [[theEvent characters] characterAtIndex:0] == 27)  // Esc seems to be ASCII 27, at least on my machine
        [self exitFullScreen];
    else
        [super keyDown:theEvent];
}

- (void) exitFullScreen
{
    //if (CGDisplayRelease( kCGDirectMainDisplay ) != kCGErrorSuccess)
    //    NSLog( @"Couldn't release the display!");
    [tPlayback shouldExitFullScreen];
	
    //[[self contentView] retain];
    //[self orderOut:nil];
}

- (void)performClose:(id)sender
{
	[self exitFullScreen];
}

- (void) orderOut:(id)sender
{
	[floater orderOut:self];
	[floater release];
	floater = nil;
	[super orderOut:sender];
}

@end
