/* Copyright (c) 2004, Noah Sorscher
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 	• 	Redistributions of source code must retain the above copyright
 * 	notice, this list of conditions and the following disclaimer.
 * 	• 	Redistributions in binary form must reproduce the above copyright
 * 	notice, this list of conditions and the following disclaimer in the
 * 	documentation and/or other materials provided with the distribution.
 * 	• 	Neither the name of the Noah Sorscher nor the names of its
 * 	contributors may be used to endorse or promote products derived from
 * 	this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
#import "spriteHelper.h"

@interface spriteHelper (Internals)

- (void)postUpdate:(NSString *)string done:(double)percent;
- (Track)spriteFromMovie:(Movie)mov time:(TimeValue)time beginning:(BOOL)beg;
//- (void)setMovieToSize:(NSSize)size;
- (void)emptyMovie:(QTMovie *)aMovie;
- (void)addSprites:(NSArray *)sprites toMovieSegment:(QTMovie *)movie;
- (void)addSprite:(NSDictionary *)sprite toMovie:(QTMovie *)movie;
- (NSString *)nameTrack:(Track)track withAction:(NSString *)actionNumber;
- (void)setAction:(NSString *)action forAtom:(QTAtomContainer)button;
- (void)newSpriteTrack:(NSDictionary *)track forMovie:(QTMovie *)newMovie;
- (NSString *)trackName:(Track)track;
- (void)name:(NSString *)name track:(Track)track;
- (void) addSprite:(NSDictionary *)sprite toKeyFrame:(QTAtomContainer)theKeyFrameSample invisibility:(BOOL)invis spriteIndex:(QTAtomID)spriteIndex;
- (BOOL) setTrackEnabledAction:(QTAtomContainer)theSprite trackType:(OSType)theType ident:(id)theIdent;
- (RGBColor)rgbColor:(NSColor *)aColor;

@end

@implementation spriteHelper

//static BOOL WriteMovieToPath(Movie movie, NSString *path, BOOL flatten);
FSSpec fsSpecFromPath2(NSString *path);
OSErr QTWired_MakeSpriteDraggable (QTAtomContainer theContainer, QTAtomID theID);

- (void)awakeFromNib
{
    //NSLog(@"awake");
    //[window setFrameOrigin:NSMakePoint(-16000, -16000)];
    //[window makeKeyAndOrderFront:nil];
    //[window orderOut:nil];
    //[NSApp hide:self];
    
    
    NSArray *args = [[NSProcessInfo processInfo] arguments];
	//NSLog(@"%@", args);
	debug = NO;
    if ([args count] == 1) {
        args = [NSArray arrayWithObjects:@"", @"/1.mov", @"/plist.xml", nil];
		debug = YES;
	}
	if (debug) [[NSAlert alertWithMessageText:@"poop" defaultButton:@"OK" alternateButton:@"nevermind" otherButton:@"too late" informativeTextWithFormat:@"or die"] runModal]; 
    if ([args count] != 3) {
        NSLog(@"incorrect arguments for mhspritehelper.  Movie location, input sprites location");
        [NSApp terminate:nil];
    }
    //NSLog([args objectAtIndex:0]);
    //NSLog(@"added");
    [self addSprites:[args objectAtIndex:2] toMovie:[args objectAtIndex:1]];
    [NSApp terminate:nil];
}
/*
- (NSDictionary *)addAutoStart:(NSDictionary *)track loopTo:(TimeValue)loo
{
	// start
	NSMutableArray *sprites = [NSMutableArray arrayWithArray:[track objectForKey:@"sprites"]];
	if (![sprites count]) {
		[sprites addObject:[NSMutableDictionary dictionary]];
	}
	NSMutableDictionary *sprite = [NSMutableDictionary dictionaryWithDictionary:[sprites objectAtIndex:0]];
	NSString *action = [sprite objectForKey:@"Action"];
	if (!action) action = @"";
	action = [action stringByAppendingString:@"\nOnFrameLoad.StartPlaying;"];
	[sprite setObject:action forKey:@"Action"];
	[sprites insertObject:sprite atIndex:0];
	[sprites removeObjectAtIndex:1];
	
	NSMutableDictionary *mtrack = [NSMutableDictionary dictionaryWithDictionary:track];
	[mtrack setObject:sprites forKey:@"sprites"];
	
	// end
	if (loo >= 0) [sprite setObject:[NSString stringWithFormat:@"%qi", (long long)loo] forKey:@"loop"]; // sprites shouldn't otherwise have loop objects, only tracks
	
	return mtrack;
}
*/
- (void)resizeMovie:(QTMovie *)movie
{
	Movie q = [movie quickTimeMovie];
	int i, n = GetMovieTrackCount(q);
	Rect target, box;
	GetMovieNaturalBoundsRect([srcMovie quickTimeMovie], &target);
	GetMovieNaturalBoundsRect(q, &box);
	short boxw = box.right - box.left,
		boxh = box.bottom - box.top,
		targetw = target.right - target.left,
		targeth = target.bottom - target.top;
	if (boxw == targetw && boxh == targeth) return;
	Fixed scaleX = FixRatio(targetw, boxw);
	Fixed scaleY = FixRatio(targeth, boxh);
	MatrixRecord m;	
	
	for (i=1;i<=n;i++) {
		Track t = GetMovieIndTrack(q, i);
		GetTrackMatrix(t, &m);
		ScaleMatrix(&m, scaleX, scaleY, Long2Fix(0), Long2Fix(0));
		SetTrackMatrix(t, &m);
	}
}

- (OSErr)addSprites:(NSString *)dataLocation toMovie:(NSString *)moviePath
{
    //NSLog(@"addSprites");
    NSDictionary *menu = [NSDictionary dictionaryWithContentsOfFile:dataLocation];
    if (!menu) {
        NSLog(@"menu data file mis-formed: %@", dataLocation);
        return 1;
    }
    srcMovie = [QTMovie movieWithFile:moviePath error:nil];
    QTMovie *newMovie = [QTMovie movieWithFile:moviePath error:nil];
    if (!srcMovie) {
        NSLog(@"couldn't create movie!");
        return 1;
    }
    [self emptyMovie:newMovie];
	Movie nmov = [newMovie quickTimeMovie];
	
	// hide subtitles for preview images
	int i, n = GetMovieTrackCount([srcMovie quickTimeMovie]);
	wide size, biggestSize;
	size.lo = size.hi = biggestSize.lo = biggestSize.hi = 0;
	int biggestTrack = 0;
	for (i=1;i<=n;i++) {
		Track t = GetMovieIndTrack([srcMovie quickTimeMovie], i);
		GetTrackDataSize64(t, 0, GetTrackDuration(t), &size);
		if (WideCompare(&size, &biggestSize) > 0) {
			biggestTrack = i;
			biggestSize = size;
		}
	}
	if (biggestTrack) {
		for (i=n;i>0;i--)
			if (i != biggestTrack)
				SetTrackEnabled(GetMovieIndTrack([srcMovie quickTimeMovie], i), NO);
	}
    
    //TimeValue origDur = GetMovieDuration([srcMovie QTMovie]);
    Rect box;
    GetMovieBox([srcMovie quickTimeMovie], &box);
    NSArray *tracks = [menu objectForKey:@"tracks"];
    // set the track names
    /*
    NSArray *trackNames = [menu objectForKey:@"trackNames"];
    long i;
    for (i=0;i<[trackNames count];i++) {
        Track aTrack = GetMovieIndTrack([srcMovie QTMovie], i+1);
        NSString *aTrackName = [trackNames objectAtIndex:i];
        char *buffer = malloc(sizeof(char) * ([aTrackName length] + 1));
        [aTrackName getCString:buffer];
        SetUserDataItem(GetTrackUserData(aTrack), buffer, [aTrackName length], kUserDataName, 1);
        free(buffer);
    }
    */
    // find the track start times
    nTracks = [tracks count];
    TimeValue totalTime = 0;
    TimeValue trackDur;
    QTMovie *movSeg;
    NSString *bg;
	
	nSprites = iSprite = 0;
	for (i=0;i<nTracks;i++)
		nSprites += [[[tracks objectAtIndex:i] objectForKey:@"sprites"] count];
	
    /*
    startTimes = malloc(sizeof(int) * nTracks);
    for (i=0;i<nTracks;i++) {
    startTimes[i] = totalTime;
	bg = [[tracks objectAtIndex:i] objectForKey:@"Image"];
	if ([bg hasSuffix:@".mov"]) {
	    movSeg = [[NSMovie alloc] initWithURL:[NSURL fileURLWithPath:bg] byReference:NO];
	    totalTime += GetMovieDuration([movSeg QTMovie]);
	    if ([[tracks objectAtIndex:i] objectForKey:@"loop"])
		totalTime += TRACK_DURATION * 2;
	} else
	    totalTime += TRACK_DURATION;
    }
    */
    
    totalTime = 0;
    NSMutableDictionary *aTrack, *sprite;
    for (iTrack=0;iTrack<nTracks;iTrack++) {
        //[self postUpdate:[NSString stringWithFormat:@"Adding Track %d of %d", iTrack, nTracks] done:(float)iTrack/(float)nTracks];
        aTrack = [NSMutableDictionary dictionaryWithDictionary:[tracks objectAtIndex:iTrack]];
		if (![[aTrack objectForKey:@"sprites"] count]) {
			[aTrack setObject:[NSMutableArray array] forKey:@"sprites"];
			sprite = [NSMutableDictionary dictionary];
			[[aTrack objectForKey:@"sprites"] addObject:sprite];
		} else
			sprite = [[aTrack objectForKey:@"sprites"] objectAtIndex:0];
		NSString *action = [sprite objectForKey:@"Action"];
		if (!action) action = @"";
        //sprite = [[NSMutableDictionary alloc] init];
        bg = [aTrack objectForKey:@"Image"];
        if (bg && [bg hasSuffix:@".mov"]) {
			[aTrack removeObjectForKey:@"Image"];
			action = [action stringByAppendingString:@"\nOnFrameLoad.StartPlaying;"];
			[sprite setObject:action forKey:@"Action"];
			movSeg = [QTMovie movieWithFile:bg error:nil];
			if (!movSeg) NSLog(@"No Background Movie for Track %d", iTrack);
			
			[self resizeMovie:movSeg];
			
			BOOL looping = [[aTrack objectForKey:@"loop"] boolValue];
			totalTime = GetMovieDuration(nmov);
			int index = GetMovieTrackCount(nmov) + 1;
			NSString *name = [aTrack objectForKey:@"Name"];
			if ([name isEqualToString:@"Page"]) name = [NSString stringWithFormat:@"Page %d", iTrack];
			name = [name stringByAppendingString:@" motion background"];
			// add background movie
			SetMovieSelection(nmov, GetMovieDuration(nmov), 0);  // zero selection at end
			SetMovieSelection([movSeg quickTimeMovie], 0, GetMovieDuration([movSeg quickTimeMovie]));
			AddMovieSelection(nmov, [movSeg quickTimeMovie]);
			
			unsigned short o = 0;
			RGBColor c;
			if ([aTrack objectForKey:@"opacity"])
				o = (float)[[aTrack objectForKey:@"opacity"] floatValue] * (float)65535;
			//NSLog(@"%g, %d", [[aTrack objectForKey:@"opacity"] floatValue], o);
			c.red = c.green = c.blue = o;
			
			short layer = 0;
			for (;index<=GetMovieTrackCount(nmov);index++) {
				Track t = GetMovieIndTrack(nmov, index);
				[self name:name track:t];
				short alayer = GetTrackLayer(t);
				if (alayer < layer) layer = alayer;
				if (o)
					MediaSetGraphicsMode(GetMediaHandler(GetTrackMedia(t)), blend, &c);
			}
			layer--;
			
			[aTrack setObject:[NSNumber numberWithLong:totalTime] forKey:@"offset"];
			[aTrack setObject:[NSNumber numberWithLong:GetMovieDuration(nmov)] forKey:@"duration"];
			
			// add the sprites
			//[self addSprites:[aTrack objectForKey:@"sprites"] toMovieSegment:movSeg];
			[self newSpriteTrack:aTrack forMovie:newMovie];
			Track track = GetMovieIndTrack(nmov, GetMovieTrackCount(nmov));
			//SetTrackOffset(track, totalTime);
			//trackDur = GetTrackDuration(track);
			//TimeValue movieDur = GetMovieDuration(nmov);
			//NSLog(@"totaltime: %ld, trackdur: %ld, moviedur: %ld", totalTime, trackDur, movieDur);
			//ScaleTrackSegment(track, totalTime, trackDur, movieDur);
			SetTrackLayer(track, layer);
			
			MediaHandler mh = GetMediaHandler(GetTrackMedia(track));
			RGBColor color;
			color.red = color.green = color.blue = 65535;
			MediaSetGraphicsMode(mh, 256, &color); // straight alpha 
			
			if (looping) {
				sprite = [NSMutableDictionary dictionaryWithObjectsAndKeys:[NSString stringWithFormat:@"OnIdle.GoToTime(%ld);", totalTime], @"Action", nil];
				aTrack = [NSMutableDictionary dictionaryWithObjectsAndKeys:[NSMutableArray arrayWithObjects:sprite, nil], @"sprites", nil];
				totalTime = GetMovieDuration(nmov);
				[aTrack setObject:[NSNumber numberWithLong:totalTime] forKey:@"duration"];
				[aTrack setObject:[NSNumber numberWithLong:totalTime - GetMovieTimeScale(nmov) / 4] forKey:@"offset"];
				[self newSpriteTrack:aTrack forMovie:newMovie];
				track = GetMovieIndTrack(nmov, GetMovieTrackCount(nmov));
				[self name:@"looper" track:track];
			}
			
		} else {
			action = [action stringByAppendingString:@"\nOnIdle.SetRate(0);"];
			[sprite setObject:action forKey:@"Action"];
			[self newSpriteTrack:aTrack forMovie:newMovie];
		}
    }
    
    //WriteMovieToPath([movie QTMovie], [outPath stringByAppendingString:@"~"], YES);
    SetMovieBox([srcMovie quickTimeMovie], &box);
    //TimeValue newDur = GetMovieDuration([srcMovie QTMovie]);
    //OSErr err = DeleteMovieSegment([movie QTMovie], origDur, newDur-origDur);
    //NSLog(@"err: %d", err);
    //WriteMovieToPath([movie QTMovie], @"/z.mov", NO);
	if (debug)
		[newMovie writeToFile:@"/2.mov" withAttributes:nil];
	NSMovie *nMovie = [[NSMovie alloc] initWithMovie:[newMovie quickTimeMovie]];
    id vo = [NSConnection rootProxyForConnectionWithRegisteredName:@"mhSprite" host:nil];
    NSData *data = [NSArchiver archivedDataWithRootObject:nMovie];
    if (vo && data)
        [vo submitMenus:data];
    return noErr;
}
/*
- (QTMovie *)loopingMovie:(NSString *)path targetTime:(TimeValue)time
{
    //NSLog(@"loopingMovie");
    QTMovie *movie = [QTMovie movieWithFile:path error:nil];
    if (!movie) return nil;
    Movie mov = [movie quickTimeMovie];
    NSString *aName = [self trackName:GetMovieIndTrackType(mov, 1, VideoMediaType, movieTrackMediaType)];
    if (!aName)
		aName = [path lastPathComponent];
    aName = [aName stringByAppendingString:@" Sprite"];
    Track aSpriteTrack = [self spriteFromMovie:mov time:time beginning:YES];
    [self name:aName track:aSpriteTrack];
    Track bSpriteTrack = [self spriteFromMovie:mov time:time beginning:NO];
    [self name:aName track:bSpriteTrack];
    return movie;
}
  */
/*
- (Track)spriteFromMovie:(Movie)mov time:(TimeValue)time beginning:(BOOL)beg
{
    //NSLog(@"spriteForLoop");
    Track vidTrack = GetMovieIndTrackType(mov, 1, VideoMediaType, movieTrackMediaType);
    Fixed myWidth, myHeight;
    GetTrackDimensions(vidTrack, &myWidth, &myHeight);
    MatrixRecord m;
    Rect box;
    box.top = box.left = 0;
    box.right = Fix2Long(myWidth);
    box.bottom = Fix2Long(myHeight);
    GetTrackMatrix(vidTrack, &m);
    TransformRect(&m, &box, NULL);
    myWidth = Long2Fix(2); //(box.right - box.left)/2);
    myHeight = Long2Fix(2); //(box.bottom - box.top)/2);
    
    OSErr err;
    AliasHandle alias = NULL;
    NSString *tempFile = [[MHUtil supportDir] stringByAppendingPathComponent:@"spriteDump"];
    if (![[NSFileManager defaultManager] fileExistsAtPath:tempFile])
        [tempFile writeToFile:tempFile atomically:NO];
    if (tempFile) {
        FSSpec spec;
        FSRef ref;
        err = FSPathMakeRef([tempFile fileSystemRepresentation], &ref, NULL);
        err = FSGetCatalogInfo(&ref, kFSCatInfoNone, NULL, NULL, &spec, NULL);
        err = NewAliasMinimal(&spec, &alias);
    }
    Handle dataRef = (Handle)alias;
    OSType dataRefType = 0;
    if (alias) dataRefType = rAliasType;
    
    Track spriteTrack = NewMovieTrack(mov, myWidth, myHeight, kNoVolume);
    Media spriteMedia = NewTrackMedia(spriteTrack, SpriteMediaType, GetMovieTimeScale(mov), dataRef, dataRefType);
    QTAtomContainer theKeyFrameSample;
    OSErr myErr = QTNewAtomContainer(&theKeyFrameSample);
    if (myErr) NSLog(@"failed to make key frame sample");
    SpriteUtils_AddPICTImageToKeyFrameSample(theKeyFrameSample, GetMoviePict(mov, 0), nil, nil, 1, NULL, NULL);
    BeginMediaEdits(spriteMedia);
    QTAtomContainer myButton, myActionButton;
    myErr = QTNewAtomContainer(&myButton);
    Point myLocation;
    myLocation.h = myLocation.v = 0;
    short myLayer = 1, myIndex = 1;
    short isVisible = 1;
    myErr = SpriteUtils_SetSpriteData(myButton, &myLocation, &isVisible, &myLayer, &myIndex, NULL, NULL, NULL);
    if (myErr) NSLog(@"failed to set sprite data");
    myErr = QTCopyAtom(myButton, kParentAtomIsContainer, &myActionButton);
    if (myErr) NSLog(@"failed to copy qt atom to action button");
    if (beg) {
		WiredUtils_AddMovieSetRateAction(theKeyFrameSample, kParentAtomIsContainer, kQTEventFrameLoaded, fixed1);
    } else {
		WiredUtils_AddMovieGoToTimeAction(theKeyFrameSample, kParentAtomIsContainer, kQTEventFrameLoaded, time);
        WiredUtils_AddMovieSetRateAction(theKeyFrameSample, kParentAtomIsContainer, kQTEventFrameLoaded, fixed1);
    }
    SpriteUtils_AddSpriteToSample(theKeyFrameSample, myActionButton, 1);
    QTDisposeAtomContainer(myActionButton);
    SpriteUtils_AddSpriteSampleToMedia(spriteMedia, theKeyFrameSample, TRACK_DURATION, true, NULL);
    myErr = EndMediaEdits(spriteMedia);
    if (myErr) NSLog(@"failed end media edits");
    TimeValue start = 0;
    if (!beg)
		start = GetMovieDuration(mov);
    InsertMediaIntoTrack(spriteTrack, start, 0, GetMediaDuration(spriteMedia), fixed1);
    RGBColor bgColor;
    bgColor.red = bgColor.green = bgColor.blue = 0; //black?  0xffff;  // white
    MHWired_SetTrackProperties(spriteMedia, 0, bgColor, YES);
    //long i, n = GetMovieTrackCount(mov);
    //if (beg)
	//for (i=1;i<=n;i++)
	    //if (spriteTrack != GetMovieIndTrack(mov, i))
		//SetTrackOffset(GetMovieIndTrack(mov, i), TRACK_DURATION);
    SetTrackLayer(spriteTrack, 10);
    return spriteTrack;
}
*/
- (void)emptyMovie:(QTMovie *)aMovie
{
    //NSLog(@"emptyMovie");
    Movie mov = [aMovie quickTimeMovie];
    long i = GetMovieTrackCount(mov);
    for(;i>0;i--)
        DisposeMovieTrack(GetMovieIndTrack(mov, i));
}
/*
- (void)addSprites:(NSArray *)sprites toMovieSegment:(QTMovie *)movie
{
    int i, n = [sprites count];
    for (i=0;i<n;i++) {
        [self postUpdate:@"Adding Sprite..." done:(double)i/(double)n];
        [self addSprite:[sprites objectAtIndex:i] toMovie:movie];
    }
}
*/
- (BOOL) isAction:(NSString *)action
{
	if (!action) return NO;
	NSArray *list = [action componentsSeparatedByString:@"\n"];
	int i, n = [list count];
	for (i=0;i<n;i++) {
		NSString *s = [[list objectAtIndex:i] lowercaseString];
		if (![s hasPrefix:@"//"] &&
			![s hasPrefix:@"nohittest"] &&
			![s hasPrefix:@"onframeload"] &&
			![s hasPrefix:@"onidle"])
			return YES;
	}
	return NO;
}
/*
- (void)addSprite:(NSDictionary *)sprite toMovie:(QTMovie *)movie
{
    //NSLog(@"addSprite");
    NSMutableDictionary *sprite2 = [NSMutableDictionary dictionaryWithDictionary:sprite];
    NSString *action = [sprite2 objectForKey:@"Action"];
    NSString *pic = [sprite2 objectForKey:@"Image"];
    NSSize picSize = NSMakeSize(100,75);
    if (pic && [pic hasPrefix:@"/"]) {
	NSImage *image = [[NSImage alloc] initWithContentsOfFile:pic];
	if (image)
	    picSize = [image size];
    }
    long w = [[sprite2 objectForKey:@"Width"] intValue];
    if (!w) w = picSize.width;
    long h = [[sprite2 objectForKey:@"Height"] intValue];
    if (!h) h = picSize.height;
    Fixed myWidth = Long2Fix(w);
    Fixed myHeight = Long2Fix(h);
    //NSLog(@"width: %d height: %d", (int)Fix2Long(myWidth), (int)Fix2Long(myHeight));
    
    OSErr err;
    AliasHandle alias = NULL;
    NSString *tempFile = [[MHUtil supportDir] stringByAppendingPathComponent:@"spriteDump"];
    if (![[NSFileManager defaultManager] fileExistsAtPath:tempFile])
        [tempFile writeToFile:tempFile atomically:NO];
    if (tempFile) {
        FSSpec spec;
        FSRef ref;
        err = FSPathMakeRef([tempFile fileSystemRepresentation], &ref, NULL);
        err = FSGetCatalogInfo(&ref, kFSCatInfoNone, NULL, NULL, &spec, NULL);
        err = NewAliasMinimal(&spec, &alias);
    }
    Handle dataRef = (Handle)alias;
    OSType dataRefType = 0;
    if (alias) dataRefType = rAliasType;
    
    Track theTrack = NewMovieTrack([movie quickTimeMovie], myWidth, myHeight, kNoVolume);
    Media theMedia = NewTrackMedia(theTrack, SpriteMediaType, GetMovieTimeScale([movie quickTimeMovie]), dataRef, dataRefType);
    QTAtomContainer theKeyFrameSample;
    OSErr myErr = QTNewAtomContainer(&theKeyFrameSample);
    BeginMediaEdits(theMedia);
    int x = [[sprite2 objectForKey:@"X"] intValue];
    int y = [[sprite2 objectForKey:@"Y"] intValue];
    [sprite2 removeObjectForKey:@"X"];
    [sprite2 removeObjectForKey:@"Y"];
	
	RGBColor c;
	c.red = c.green = c.blue = 65535;
	MediaHandler mh = GetMediaHandler(theMedia);
	long gmode = 260; // straight alpha blend
	NSNumber *opq = [sprite2 objectForKey:@"opacity"];
	if (opq) {
		float o = [opq floatValue];
		[sprite2 removeObjectForKey:@"opacity"];
		unsigned short i = o * 65535;
		c.red = c.green = c.blue = i;
		ComponentResult err = MediaSetGraphicsMode(mh, gmode, &c);
		if (err) NSLog(@"transparency error: %d", err);
	}
    
	NSString *name = [self nameTrack:theTrack withAction:[sprite2 objectForKey:@"Action"]];
	[self postUpdate:name done:-1];
	
	UInt32 freq = 0;
    picIndex = 1;
    [self addSprite:sprite2 toKeyFrame:theKeyFrameSample invisibility:NO spriteIndex:1];
   // if ([action length])
	if ([self isAction:action])
		QTWired_AddCursorChangeOnMouseOver(theKeyFrameSample, 1);
	if ([[action lowercaseString] rangeOfString:@"onidle"].length == 6)
		freq = 1;
    SpriteUtils_AddSpriteSampleToMedia(theMedia, theKeyFrameSample, GetMovieDuration([movie quickTimeMovie]), true, NULL);
	
    myErr = EndMediaEdits(theMedia);
    InsertMediaIntoTrack(theTrack, 0, 0, GetMediaDuration(theMedia), fixed1);
    BOOL visible = NO;
    if ([sprite2 objectForKey:@"Image"])
	visible = YES;
    MHWired_SetTrackProperties(theMedia, freq, [self rgbColor:[NSColor blackColor]], visible);
    MatrixRecord m;
    GetTrackMatrix(theTrack, &m);
    Fixed deltaH = Long2Fix(x);
    Fixed deltaV = Long2Fix(y);
    TranslateMatrix(&m, deltaH, deltaV);
    SetTrackMatrix(theTrack, &m);
	
	MediaSetGraphicsMode(mh, gmode, &c);
	
	NSString *loopTo = [sprite2 objectForKey:@"loop"]; // looks like there's no way but to make a whole extra track for this :(
	if (loopTo) {
		theTrack = NewMovieTrack([movie quickTimeMovie], myWidth, myHeight, kNoVolume);
		theMedia = NewTrackMedia(theTrack, SpriteMediaType, GetMovieTimeScale([movie quickTimeMovie]), dataRef, dataRefType);
		myErr = QTNewAtomContainer(&theKeyFrameSample);
		BeginMediaEdits(theMedia);
		NSString *actionString = [NSString stringWithFormat:@"OnFrameLoad.GoToTime(%@);", loopTo];
		[sprite2 setObject:actionString forKey:@"Action"];
		[sprite2 removeObjectForKey:@"Rollover"];
		[sprite2 removeObjectForKey:@"Click"];
		picIndex = 1;
		[self addSprite:sprite2 toKeyFrame:theKeyFrameSample invisibility:NO spriteIndex:1];
		SpriteUtils_AddSpriteSampleToMedia(theMedia, theKeyFrameSample, TRACK_DURATION, true, NULL);
		myErr = EndMediaEdits(theMedia);
		InsertMediaIntoTrack(theTrack, 0, 0, TRACK_DURATION, fixed1);
		MHWired_SetTrackProperties(theMedia, freq, [self rgbColor:[NSColor blackColor]], visible);
		TimeValue quarter = GetMovieTimeScale([movie quickTimeMovie]) / 4;
		ScaleTrackSegment(theTrack, 0, TRACK_DURATION, quarter);
		SetTrackOffset(theTrack, GetMovieDuration([movie quickTimeMovie]) - quarter);
		SetTrackLayer(theTrack, 10);
		[self name:@"loop" track:theTrack];
	}
	
	
}
*/

- (NSString *)nameTrack:(Track)track withAction:(NSString *)action
{
    //NSLog(@"nameTrack");
    if (!action) return nil;
    NSScanner *scan = [NSScanner scannerWithString:action];
    [scan setCaseSensitive:NO];
    NSString *buffer = nil;
    NSString *newName = nil;
    if ([scan scanString:@"GoToTime" intoString:nil]) newName = @"Navigation Sprite";
    else if ([scan scanString:@"GoToChapter(" intoString:nil]) {
        [scan scanUpToString:@")" intoString:&buffer];
        if (buffer) {
            NSRange r = [buffer rangeOfString:@" "];
            if (r.location > 10) r.location = 10;
			if ([buffer length] > r.location && r.location > 0)
				buffer = [buffer substringToIndex:r.location];
            newName = [NSString stringWithFormat:@"Jump To %@ Sprite", buffer];
        } else newName = @"Chapter Selection Sprite";
    } else if ([scan scanString:@"// set" intoString:nil]) {
        NSString *def = @"Track";
        if ([scan scanString:@"audio" intoString:nil]) def = @"Audio";
        else if ([scan scanString:@"text" intoString:nil]) def = @"Text";
        newName = [NSString stringWithFormat:@"%@ Selection Sprite", def];
    }
    
    if (newName) {
	char *buffer = malloc(([newName length]+1) * sizeof(char));
	[newName getCString:buffer];
	SetUserDataItem(GetTrackUserData(track), buffer, [newName length], kUserDataName, 1);
	free(buffer);
    }
    return newName;
}

- (OSErr)addAppAction:(int)action forAtom:(QTAtomContainer)button withEvent:(long)theEvent
{
	QTAtom myEventAtom = 0;
	QTAtom myActionAtom = 0;
	OSErr myErr;
	OSErr err = QTInsertChild(button, kParentAtomIsContainer, kQTEventType, theEvent, 1, 0, NULL, &myEventAtom);
	if (err) myErr = err;
	err = QTInsertChild(button, myEventAtom, kAction, 0, 0, 0, NULL, &myActionAtom);
	if (err) myErr = err;
	int myAction = EndianU32_NtoB(kActionSendAppMessage);
	err = QTInsertChild(button, myActionAtom, kWhichAction, 1, 1, sizeof(myAction), &myAction, NULL);
	if (err) myErr = err;
	int myMessage = EndianU32_NtoB(action);
	err = QTInsertChild(button, myActionAtom, kActionParameter, 0, 1, sizeof(myMessage), &myMessage, NULL);
	if (err) myErr = err;
	return myErr;
}

- (OSErr)keyStuff:(QTAtomContainer)button // doesn't work
{
	QTAtom myEventParamAtom = 0, myActionAtom = 0, myActionAtom2 = 0, myParameterAtom = 0, myExpressionAtom = 0, myOperandAtom = 0;
	short myIndex = EndianS16_NtoB(4);
	OSErr myErr = 0;
	OSErr err = WiredUtils_AddQTEventAndActionAtoms(button, kParentAtomIsContainer, kQTEventKey, kActionTextTrackPasteText, &myActionAtom);
	if (err) myErr = err;
	err = WiredUtils_AddActionParameterAtom(button, myActionAtom, 1, 0, NULL, &myParameterAtom);
	if (err) myErr = err;
	err = WiredUtils_AddExpressionContainerAtomType(button, myParameterAtom, &myExpressionAtom);
	if (err) myErr = err;
	err = QTInsertChild(button, myExpressionAtom, kOperandAtomType, 1, 1, 0, NULL, &myOperandAtom);
	if (err) myErr = err;
	err = QTInsertChild(button, myOperandAtom, kOperandEventParameter, 1, 1, 0, NULL, &myEventParamAtom);
	if (err) myErr = err;
	err = QTInsertChild(button, myEventParamAtom, kActionParameter, 1, 1, sizeof(myIndex), &myIndex, NULL);
	if (err) myErr = err;
	
	err = WiredUtils_AddQTEventAndActionAtoms(button, kParentAtomIsContainer, kQTEventIdle, kActionTextTrackSetEditable, &myActionAtom2);
	if (err) myErr = err;
	short myEditState = EndianS16_NtoB(2);
	myErr = WiredUtils_AddActionParameterAtom(button, myActionAtom2, 1, sizeof(myEditState), &myEditState, NULL);
	
	return myErr;
}

- (void)setAction:(NSDictionary *)sprite forAtom:(QTAtomContainer)button keyFrame:(QTAtomContainer)theKeyFrameSample
{
    //NSLog(@"setAction");
	NSString *action = [sprite objectForKey:@"Action"];
    if (!action) return;
	QTAtomContainer myTarget;
    OSErr err = noErr;
    NSScanner *scan = [NSScanner scannerWithString:action];
    [scan setCaseSensitive:NO];
    NSString *buffer = nil;
    OSType theEvent = kQTEventMouseClickEnd;
    void *theTrackTarget = nil;
    long theTrackTargetType = 0;
    while (![scan isAtEnd]) {
        theEvent = kQTEventMouseClickEnd;
		myTarget = button;
        buffer = nil;
        theTrackTarget = nil;
        theTrackTargetType = 0;
        BOOL foundSomething = NO;
		[scan scanString:@"NoHitTest" intoString:nil]; // ignore, it is handled elsewhere
        if ([scan scanString:@"//" intoString:nil]) {
            [scan scanPastString:@"\n"];
        }
        if ([scan scanString:@"OnFrameLoad" intoString:nil]) {
            foundSomething = YES;
            theEvent = kQTEventFrameLoaded;
			myTarget = theKeyFrameSample;
            [scan scanString:@"ed" intoString:nil];
            [scan scanString:@"()" intoString:nil];
            [scan scanString:@"." intoString:nil];
        } else if ([scan scanString:@"OnMouseEnter" intoString:nil]) {
            foundSomething = YES;
            theEvent = kQTEventMouseEnter;
            [scan scanString:@"ed" intoString:nil];
            [scan scanString:@"()" intoString:nil];
            [scan scanString:@"." intoString:nil];
        } else if ([scan scanString:@"OnMouseExit" intoString:nil]) {
            foundSomething = YES;
            theEvent = kQTEventMouseExit;
            [scan scanString:@"ed" intoString:nil];
            [scan scanString:@"()" intoString:nil];
			[scan scanString:@"." intoString:nil];
		} else if ([scan scanString:@"OnMouseMoved" intoString:nil]) {
            foundSomething = YES;
            theEvent = kQTEventMouseMoved;
            [scan scanString:@"ed" intoString:nil];
            [scan scanString:@"()" intoString:nil];
            [scan scanString:@"." intoString:nil];
		} else if ([scan scanString:@"OnKey" intoString:nil]) {
            foundSomething = YES;
            theEvent = kQTEventKey;
			if ([scan scanString:@"Up" intoString:nil])
				theEvent = kQTEventKeyUp;
            [scan scanString:@"Down" intoString:nil];
            if ([scan scanString:@"(" intoString:nil]) {
				[scan scanUpToString:@")" intoString:nil];
				err = [self keyStuff:button];
			}
            [scan scanString:@"." intoString:nil];
			QTAtom newButton;
			NSString *name = [sprite objectForKey:@"Name"];
			char *theSpriteTarget = malloc(sizeof(char) * ([name length] + 1));
			theSpriteTarget[0] = [name length];
			memcpy(theSpriteTarget+1, [name cString], [name length]);

			WiredUtils_AddQTEventAndActionAtoms(button, kParentAtomIsContainer, kQTEventIdle, kActionSetFocus, &newButton);
			WiredUtils_AddTrackAndSpriteTargetAtoms(button, newButton, 0, nil, 0, kTargetSpriteName, theSpriteTarget);
			
		} else if ([scan scanString:@"OnIdle" intoString:nil]) {
			foundSomething = YES;
			theEvent = kQTEventIdle;
			[scan scanString:@"()" intoString:nil];
			[scan scanString:@"." intoString:nil];
		}
        if ([scan scanString:@"GoToChapter(" intoString:nil]) {
            foundSomething = YES;
            [scan scanUpToString:@"\n" intoString:&buffer];
            if (buffer) {
                NSRange r = [buffer rangeOfString:@")" options:NSBackwardsSearch];
                if (r.length)
                    buffer = [buffer substringToIndex:r.location];
            }
            if ([buffer length]) {
                Str255 tmpStr;
                QTTextCopyCStringToPascal([buffer cString], tmpStr);
                WiredUtils_AddMovieGoToTimeByNameAction(myTarget, kParentAtomIsContainer, theEvent, tmpStr);
            }
        } else if ([scan scanString:@"GoToTime(" intoString:nil]) {
            foundSomething = YES;
            long long time = -1;
            [scan scanLongLong:&time];
            [scan scanString:@")" intoString:nil];
            [scan scanString:@";" intoString:nil];
            if (time >= 0) {
                WiredUtils_AddMovieGoToTimeAction(myTarget, kParentAtomIsContainer, theEvent, time);
            }
        } else if ([scan scanString:@"StartPlaying" intoString:nil]) {
            foundSomething = YES;
            [scan scanString:@";" intoString:nil];
            WiredUtils_AddMovieSetRateAction(myTarget, kParentAtomIsContainer, theEvent, fixed1);
        } else if ([scan scanString:@"SetRate(" intoString:nil]) {
            foundSomething = YES;
            int rate = NSNotFound;
            [scan scanInt:&rate];
            [scan scanString:@")" intoString:nil];
            [scan scanString:@";" intoString:nil];
            if (rate != NSNotFound)
                WiredUtils_AddMovieSetRateAction(myTarget, kParentAtomIsContainer, theEvent, Long2Fix((long)rate));
        } else if ([scan scanString:@"StepForward" intoString:nil]) {
            foundSomething = YES;
            [scan scanString:@";" intoString:nil];
            WiredUtils_AddMovieStepForwardAction(myTarget, kParentAtomIsContainer, theEvent);
        } else if ([scan scanString:@"StepBackward" intoString:nil]) {
            foundSomething = YES;
            [scan scanString:@";" intoString:nil];
            WiredUtils_AddMovieStepBackwardAction(myTarget, kParentAtomIsContainer, theEvent);
        } else if ([scan scanString:@"GoToBeginning" intoString:nil]) {
            foundSomething = YES;
            [scan scanString:@";" intoString:nil];
            WiredUtils_AddMovieGoToBeginningAction(myTarget, kParentAtomIsContainer, theEvent);
        } else if ([scan scanString:@"GoToEnd" intoString:nil]) {
            foundSomething = YES;
            [scan scanString:@";" intoString:nil];
            WiredUtils_AddMovieGoToEndAction(myTarget, kParentAtomIsContainer, theEvent);
        } else if ([scan scanString:@"SetMovieLanguage(" intoString:nil]) {
            foundSomething = YES;
            [scan scanUpToString:@")" intoString:&buffer];
            [scan scanString:@")" intoString:nil];
            [scan scanString:@";" intoString:nil];
            int lang = [tLanguages stringToNumber:buffer];
            if (lang >= 0)
                err = WiredUtils_AddMovieSetLanguage(myTarget, kParentAtomIsContainer, theEvent, lang);
	} else if ([scan scanString:@"SaveTime" intoString:nil]) {
            foundSomething = YES;
		[scan scanString:@"()" intoString:nil];
		[scan scanString:@";" intoString:nil];
		err = WiredUtils_AddPushCurrentTimeAction(myTarget, kParentAtomIsContainer, theEvent);
	} else if ([scan scanString:@"GoToSavedTime" intoString:nil]) {
            foundSomething = YES;
		[scan scanString:@"()" intoString:nil];
		[scan scanString:@";" intoString:nil];
		err = WiredUtils_AddPopAndGotoTopTimeAction(myTarget, kParentAtomIsContainer, theEvent);
	} else if ([scan scanString:@"SetVolume(" intoString:nil] || [scan scanString:@"SetMovieVolume(" intoString:nil]) {
            foundSomething = YES;
            int i = NSNotFound;
            [scan scanInt:&i];
            [scan scanString:@")" intoString:nil];
            [scan scanString:@";" intoString:nil];
            if (i <= 256 && i >= -256)
                err = WiredUtils_AddMovieSetVolumeAction(myTarget, kParentAtomIsContainer, theEvent, (short)i);
            else if (i != NSNotFound) NSLog(@"invalid value set for SetMovieVolume sprite: %d", i);
        } else if ([scan scanString:@"SetSelection(" intoString:nil] || [scan scanString:@"SetMovieSelection(" intoString:nil]) {
            foundSomething = YES;
            int start = NSNotFound, stop = NSNotFound;
            [scan scanInt:&start];
            [scan scanString:@"," intoString:nil];
            [scan scanInt:&stop];
            [scan scanString:@")" intoString:nil];
            [scan scanString:@";" intoString:nil];
            if (start != NSNotFound && stop != NSNotFound)
                err = WiredUtils_AddMovieSetSelectionAction(myTarget, kParentAtomIsContainer, theEvent, start, stop);
        } else if ([scan scanString:@"SetSelectionByChapter(" intoString:nil] || [scan scanString:@"SetMovieSelectionByChapter(" intoString:nil]) {
            foundSomething = YES;
            NSString *start = nil, *stop = nil;
            [scan scanUpToString:@", " intoString:&start];
            [scan scanString:@", " intoString:nil];
            [scan scanUpToString:@")" intoString:&stop];
            if (start && stop) {
                Str255 tmpStart, tmpStop;
                QTTextCopyCStringToPascal([start cString], tmpStart);
                QTTextCopyCStringToPascal([stop cString], tmpStop);
                err = WiredUtils_AddMovieSetSelectionByNameAction(myTarget, kParentAtomIsContainer, theEvent, tmpStart, tmpStop);
            }
        } else if ([scan scanString:@"PlaySelection" intoString:nil]) {
            foundSomething = YES;
            [scan scanString:@"()" intoString:nil];
            [scan scanString:@";" intoString:nil];
            err = WiredUtils_AddMoviePlaySelectionAction(myTarget, kParentAtomIsContainer, theEvent, YES);
        } else if ([scan scanString:@"ExitFullscreen" intoString:nil]) {
			foundSomething = YES;
			err = [self addAppAction:kQTAppMessageExitFullScreenRequested forAtom:myTarget withEvent:theEvent];
		} else if ([scan scanString:@"EnterFullscreen" intoString:nil]) {
			foundSomething = YES;
			err = [self addAppAction:kQTAppMessageEnterFullScreenRequested forAtom:myTarget withEvent:theEvent];
		} else if ([scan scanString:@"Close" intoString:nil]) {
			foundSomething = YES;
			err = [self addAppAction:kQTAppMessageWindowCloseRequested forAtom:myTarget withEvent:theEvent];
		} else if ([scan scanString:@"TrackNamed(" intoString:nil]) {
            foundSomething = YES;
            buffer = nil;
            [scan scanUpToString:@")." intoString:&buffer];
            [scan scanString:@")." intoString:nil];
            if (buffer) {
                theTrackTarget = (void *)QTUtils_ConvertCToPascalString((char *)[buffer cString]);
                theTrackTargetType = kTargetTrackName;
            }
        } else if ([scan scanString:@"TrackAtIndex(" intoString:nil]) {
            foundSomething = YES;
            int index = -1;
            [scan scanInt:&index];
            [scan scanString:@")." intoString:nil];
            if (index >= 0) {
                index++;  // stupid QT and its 1-based arrays
                theTrackTarget = (void *)index;
                theTrackTargetType = kTargetTrackID;
            }
        }
        BOOL enabled = YES, setEnabled = NO;
        if ([scan scanString:@"SetEnabled" intoString:nil]) {
            foundSomething = YES;
            setEnabled = YES;
            if ([scan scanString:@"(N" intoString:nil] || [scan scanString:@"(F" intoString:nil])
                enabled = NO;
			[scan scanPastString:@"\n"];
        } else if ([scan scanString:@"SetDisabled" intoString:nil]) {
            foundSomething = YES;
            setEnabled = YES;
            enabled = NO;
			[scan scanPastString:@"\n"];
        } else if ([scan scanString:@"SetTrackVariable(" intoString:nil]) {
			foundSomething = YES;
			int var = 0;
			float val = NSNotFound;
			[scan scanInt:&var];
			[scan scanString:@"," intoString:nil];
			[scan scanFloat:&val];
			if (var && val != NSNotFound)
				err = WiredUtils_AddSpriteTrackSetVariableAction(myTarget, kParentAtomIsContainer, theEvent, var, val, theTrackTargetType, theTrackTarget, 0);
		}
        if (setEnabled) {
            foundSomething = YES;
            [scan scanString:@"(YES)" intoString:nil];
            [scan scanString:@"(TRUE)" intoString:nil];
            [scan scanString:@"()" intoString:nil];
            [scan scanString:@";" intoString:nil];
            if (theTrackTarget && theTrackTargetType)
                err = WiredUtils_AddTrackSetEnabledAction(myTarget, kParentAtomIsContainer, theEvent, theTrackTargetType, theTrackTarget, 0, enabled);
        }
        long theSpriteTargetType = 0;
        char *theSpriteTarget = nil;
        if ([scan scanString:@"SpriteNamed(" intoString:nil]) {
            foundSomething = YES;
            NSString *name = nil;
            [scan scanUpToString:@")." intoString:&name];
            [scan scanString:@")." intoString:nil];
            if (name) {
                theSpriteTarget = malloc(sizeof(char) * ([name length] + 1));
                theSpriteTarget[0] = [name length];
                memcpy(theSpriteTarget+1, [name cString], [name length]);
                theSpriteTargetType = kTargetSpriteName;
            }
        }
        BOOL visible = YES, setVisible = NO;
        if ([scan scanString:@"SetVisible" intoString:nil]) {
            foundSomething = YES;
            setVisible = YES;
            if ([scan scanString:@"(NO)" intoString:nil] || [scan scanString:@"(FALSE)" intoString:nil])
                visible = NO;
        } else if ([scan scanString:@"SetInvisible" intoString:nil]) {
            foundSomething = YES;
            setVisible = YES;
            visible = NO;
        }
        [scan scanString:@"()" intoString:nil];
        [scan scanString:@";" intoString:nil];
        if (setVisible) {
            err = WiredUtils_AddSpriteSetVisibleAction(myTarget, kParentAtomIsContainer, theEvent, theTrackTargetType, theTrackTarget, 0, theSpriteTargetType, theSpriteTarget, visible, nil);
        }
        if ([scan scanString:@"Rotate(" intoString:nil]) {
            foundSomething = YES;
            int deg = 0;
            [scan scanInt:&deg];
            if (deg)
                err = WiredUtils_AddSpriteRotateAction(myTarget, kParentAtomIsContainer, theEvent, theTrackTargetType, theTrackTarget, 0, theSpriteTargetType, theSpriteTarget, Long2Fix(deg), nil);
			[scan scanPastString:@")"];
			[scan scanString:@";" intoString:nil];
        } else if ([scan scanString:@"Translate(" intoString:nil]) {
			foundSomething = YES;
			int x=0, y=0;
			[scan scanInt:&x];
			[scan scanString:@"," intoString:nil];
			[scan scanInt:&y];
			[scan scanString:@")" intoString:nil];
			[scan scanString:@";" intoString:nil];
			if (x || y) {
				err = WiredUtils_AddSpriteTranslateAction(myTarget, kParentAtomIsContainer, theEvent, theTrackTargetType, theTrackTarget, 0, theSpriteTargetType, theSpriteTarget, Long2Fix(x), Long2Fix(y), NO, nil);
			}
		} else if ([scan scanString:@"Scale(" intoString:nil]) {
			foundSomething = YES;
			float x=0, y=0;
			[scan scanFloat:&x];
			[scan scanString:@"," intoString:nil];
			[scan scanFloat:&y];
			[scan scanString:@")" intoString:nil];
			[scan scanString:@";" intoString:nil];
			if (x || y) {
				err = WiredUtils_AddSpriteScaleAction(myTarget, kParentAtomIsContainer, theEvent, theTrackTargetType, theTrackTarget, 0, theSpriteTargetType, theSpriteTarget, FloatToFixed(x), FloatToFixed(y), nil);
			}
		} else if ([scan scanString:@"SetOpacity(" intoString:nil]) {
			float o = -1;
			[scan scanFloat:&o];
			[scan scanString:@")" intoString:nil];
			[scan scanString:@";" intoString:nil];
			if (o >= 0) {
				ModifierTrackGraphicsModeRecord r;
				r.graphicsMode = blend;
				if (o == 0) r.graphicsMode = srcCopy;
				r.opColor.red = r.opColor.blue = r.opColor.green = o * 65535;
				err = WiredUtils_AddSpriteSetGraphicsModeAction(myTarget, kParentAtomIsContainer, theEvent, theTrackTargetType, theTrackTarget, 0, theSpriteTargetType, theSpriteTarget, &r, nil);
			}
		}
        if (theSpriteTarget) {
            free(theSpriteTarget);
            theSpriteTarget = nil;
        }
        if (!foundSomething) {
            NSLog(@"unknown sprite action found: %@", [action substringFromIndex:[scan scanLocation]]);
            [scan scanPastString:@"\n"];
        }
    }
}

- (void)newSpriteTrack:(NSDictionary *)track forMovie:(QTMovie *)newMovie
{
    //NSLog(@"newSpriteTrack");
    TimeValue oldDur = GetMovieDuration([newMovie quickTimeMovie]);
    Track vidTrack = GetMovieIndTrackType([srcMovie quickTimeMovie], 1, VideoMediaType, movieTrackMediaType);
    Fixed myWidth, myHeight;
    GetTrackDimensions(vidTrack, &myWidth, &myHeight);
    MatrixRecord m;
    Rect box;
    box.top = box.left = 0;
    box.right = Fix2Long(myWidth);
    box.bottom = Fix2Long(myHeight);
    GetTrackMatrix(vidTrack, &m);
    TransformRect(&m, &box, NULL);
    myWidth = Long2Fix(box.right - box.left);
    myHeight = Long2Fix(box.bottom - box.top);
    
    OSErr err;
    AliasHandle alias = NULL;
    NSString *tempFile = [[MHUtil supportDir] stringByAppendingPathComponent:@"spriteDump"];
    if (![[NSFileManager defaultManager] fileExistsAtPath:tempFile])
        [tempFile writeToFile:tempFile atomically:NO];
    if (tempFile) {
        FSSpec spec;
        FSRef ref;
        err = FSPathMakeRef([tempFile fileSystemRepresentation], &ref, NULL);
        err = FSGetCatalogInfo(&ref, kFSCatInfoNone, NULL, NULL, &spec, NULL);
        err = NewAliasMinimal(&spec, &alias);
    }
    Handle dataRef = (Handle)alias;
    OSType dataRefType = 0;
    if (alias) dataRefType = rAliasType;
    
    Track myTrack = NewMovieTrack([newMovie quickTimeMovie], myWidth, myHeight, kNoVolume);
    Media myMedia = NewTrackMedia(myTrack, SpriteMediaType, GetMovieTimeScale([newMovie quickTimeMovie]), dataRef, dataRefType);
    QTAtomContainer theKeyFrameSample;
    OSErr myErr = QTNewAtomContainer(&theKeyFrameSample);
    NSMutableArray *sprites = [NSMutableArray array];
    if ([track objectForKey:@"Image"]) {
		NSMutableDictionary *bgSprite = [NSMutableDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithInt:0], @"X",
		[NSNumber numberWithInt:0], @"Y", [NSNumber numberWithInt: Fix2Long(myWidth)], @"Width",
		[NSNumber numberWithInt: Fix2Long(myHeight)], @"Height", [track objectForKey:@"Image"], @"Image", nil];
		if ([track objectForKey:@"opacity"])
			[bgSprite setObject:[track objectForKey:@"opacity"] forKey:@"opacity"];
		[sprites addObject:bgSprite];
    }
    if ([track objectForKey:@"sprites"])
	[sprites addObjectsFromArray:[track objectForKey:@"sprites"]];
    
    BeginMediaEdits(myMedia);
    int i, n = [sprites count];
    picIndex = 1;
    NSString *goodAction = nil;
    NSString *name = [track objectForKey:@"Name"];
    if (name && ![name isEqualToString:@"Page"]) {
        char *buffer = malloc(sizeof(char) * ([name length] + 1));
        [name getCString:buffer];
        SetUserDataItem(GetTrackUserData(myTrack), buffer, [name length], kUserDataName, 1);
        free(buffer);
    } else {
        name = [self nameTrack:myTrack withAction:goodAction];
    }
	UInt32 freq = 100; // ms per tick
    for (i=0;i<n;i++) {
        //[self postUpdate:@"Adding Sprites..." done:(double)i/(double)n];
		NSString *spritename = [[sprites objectAtIndex:i] objectForKey:@"Name"];
		if (!spritename) spritename = @"no name sprite";
		float done = (float)iSprite / (float)nSprites;
		spritename = [NSString stringWithFormat:@"Adding Sprite: %@ : %@...", name, spritename];
		[self postUpdate:spritename done:done];
		iSprite++;
		[self addSprite:[sprites objectAtIndex:i] toKeyFrame:theKeyFrameSample invisibility:YES spriteIndex:i+1];
		NSString *action = [[sprites objectAtIndex:i] objectForKey:@"Action"];
		if ([action length]) {
			if ([self isAction:action])
				QTWired_AddCursorChangeOnMouseOver(theKeyFrameSample, i+1);
			if (!goodAction) {
				goodAction = action;
				//[self postUpdate:name done:-1];
			}
			if ([[action lowercaseString] rangeOfString:@"onidle"].length == 6)
				freq = 1;
		}
    }
    //WiredUtils_AddMovieSetRateAction(theKeyFrameSample, kParentAtomIsContainer, kQTEventFrameLoaded, Long2Fix(0));
	//NSDictionary *stop = [NSDictionary dictionaryWithObjectsAndKeys:@"OnIdle.SetRate(0);", @"Action", nil];
	//[self addSprite:stop toKeyFrame:theKeyFrameSample invisibility:YES spriteIndex:i+1];
	TimeValue offset = oldDur;
	if ([track objectForKey:@"offset"]) offset = [[track objectForKey:@"offset"] longValue];
	TimeValue duration = GetMovieTimeScale([newMovie quickTimeMovie]);
	if ([track objectForKey:@"duration"]) duration = [[track objectForKey:@"duration"] longValue] - offset;
    SpriteUtils_AddSpriteSampleToMedia(myMedia, theKeyFrameSample, duration, true, NULL);
    myErr = EndMediaEdits(myMedia);
    InsertMediaIntoTrack(myTrack, 0, 0, GetMediaDuration(myMedia), fixed1);
    RGBColor color;
    NSColor *bgColor = nil;
    NSData *bgColorD = [track objectForKey:@"color"];
    if (bgColorD)
	bgColor = [NSUnarchiver unarchiveObjectWithData:bgColorD];
    if (!bgColor)
	bgColor = [NSColor blackColor];
    color = [self rgbColor:bgColor];
    MHWired_SetTrackProperties(myMedia, freq, color, YES);
    SetTrackOffset(myTrack, offset);
}

- (NSString *)trackName:(Track)track
{
    Handle h = NewHandleClear(0);
    GetUserData(GetTrackUserData(track), h, kUserDataName, 1);
    char *buffer = (char *)*h;
    NSString *vidName = [NSString stringWithCString:buffer length:GetHandleSize(h)];
    if (!vidName || ![vidName length]) return nil;
    return vidName;
}

- (void)name:(NSString *)name track:(Track)track
{
    char *buffer = malloc(sizeof(char) * ([name length] + 1));
    [name getCString:buffer];
    SetUserDataItem(GetTrackUserData(track), buffer, [name length], kUserDataName, 1);
    free(buffer);
}
/*
- (void)copyImage:(NSImage *)image toGWorld:(GWorldPtr)gWorld
{
    NSArray *reps = [image representations];
    if (![reps count]) return;
    NSBitmapImageRep *rep = [reps objectAtIndex:0];
    if (![rep isKindOfClass:[NSBitmapImageRep class]]) return;
    void *myPixels = (void *)[rep bitmapData];
    int row, column, widthInPixels = [image size].width, heightInPixels = [image size].height;
    //int bitsPerPixel = [rep bitsPerPixel];
    int a, r, g, b;
    Ptr pixBaseAddr = GetPixBaseAddr(GetGWorldPixMap(gWorld));
    Rect portRect;
    GetPortBounds(gWorld, &portRect);
    for (row = 0; row < heightInPixels; row++) {
        unsigned char *dst = pixBaseAddr + row * (portRect.right - portRect.left);
        unsigned char *src = myPixels + row * widthInPixels;
        for (column = 0; column < widthInPixels; column++) {
                        r = *src++;		// get source Red component
                        g = *src++;		// get source Green component
                        b = *src++;		// get source Blue component  
                        a = *src++;		// get source Alpha component
            
                        *dst++ = a;		// set dest. Blue component 
                        *dst++ = r;		// set dest. Alpha component
                        *dst++ = g;		// set dest. Red component
                        *dst++ = b;		// set dest. Green component
        }
    }
}
*/
/*
- (NSRect)frameOfTrack:(Track)t
{
    Fixed w, h;
    GetTrackDimensions(t, &w, &h);
    Rect box;
    box.top = box.left = 0;
    box.right = Fix2Long(w);
    box.bottom = Fix2Long(h);
    MatrixRecord m;
    GetTrackMatrix(t, &m);
    TransformRect(&m, &box, NULL);
    NSRect rect;
    rect.origin.x = box.left;
    rect.origin.y = box.bottom;
    rect.size.width = box.right - box.left;
    rect.size.height = box.bottom - box.top;
    return rect;
}

- (void)setFrame:(NSRect)frame forMovie:(QTMovie *)movie
{
	
	NSArray *tracks = [movie tracksOfMediaType:QTMediaTypeVideo];
	if (![tracks count]) return;
	Track t = [[tracks objectAtIndex:0] quickTimeTrack];
	NSRect oldFrame = [self frameOfTrack:t];
	MatrixRecord m;
	GetTrackMatrix(t, &m);
	float dw = frame.size.width - oldFrame.size.width;
	float dh = frame.size.height - oldFrame.size.height;
	if (dw || dh) {
		Fixed scaleX = FixRatio(frame.size.width, oldFrame.size.width);
		Fixed scaleY = FixRatio(frame.size.height, oldFrame.size.height);
		ScaleMatrix(&m, scaleX, scaleY, Long2Fix(0), Long2Fix(0));
		SetTrackMatrix(t, &m);
	}
	/*
	oldFrame = [self frameOfTrack:t];
	int dx = frame.origin.x - oldFrame.origin.x;
	int dy = frame.origin.y - oldFrame.origin.y;
	if (dx || dy) {
		Fixed deltaH = Long2Fix(dx);
		Fixed deltaV = Long2Fix(dy);
		TranslateMatrix(&m, deltaH, deltaV);
	}
	SetTrackMatrix(t, &m);* /
}
*/
- (Handle)compressImage:(NSString *)imgSource withDescription:(ImageDescriptionHandle)myDesc size:(NSSize)size keyColor:(NSColor *)keyColor
{
    //NSImage *image = nil;
    //NSLog(@"source: %@ size:%gx%g", imgSource, size.width, size.height);
    QTMovie *movie = nil;
	NSString *path = nil;
    if ([imgSource hasPrefix:@"/"]) {
        //image = [[NSImage alloc] initWithContentsOfFile:imgSource];
        movie = [QTMovie movieWithFile:imgSource error:nil];
        if (!movie) return nil;
    } else if ([imgSource hasPrefix:@"\""]) {
		imgSource = [imgSource substringFromIndex:1];
		if ([imgSource hasSuffix:@"\""]) imgSource = [imgSource substringToIndex:[imgSource length]-1];
		NSString *text = [NSString stringWithFormat:@"{QTtext}{font:chicago}{outline}{size:15}{textColor: 65535, 65535, 65535}{backColor: 0, 0, 0}{justify:center}{keyedText:off}{dropShadow:off}{dropShadowOffset:1,1}{dropShadowTransparency:255}{timeScale:100}{width:%g}{height:%g}{timeStamps:absolute}{language:0}{textEncoding:0}{anti-alias:on}\n[00:00:00.00]\n%@\n[00:00:01.00]", size.width, size.height, imgSource];
		path = [MHUtil uniquePath:[[MHUtil supportDir] stringByAppendingPathComponent:@"text.txt"]];
		[text writeToFile:path atomically:NO];
		movie = [QTMovie movieWithFile:path error:nil];
		//NSImage *image = [movie currentFrameImage];
		//movie = [QTMovie movie];
		//QTTime dur = [movie currentTime];
		//dur.timeValue = dur.timeScale;
		//[movie addImage:image forDuration:dur withAttributes:nil];
	} else {
        movie = srcMovie;
		QTTime t = [movie currentTime];
		t.timeValue = [imgSource intValue];
		[movie setCurrentTime:t];
	}
    Rect                                        myRect, portRect;
    ImageSequence                               mySeqID = 0;			// compression sequence
    long                                        myDataSize = 0;
    Ptr                                         myData = NULL;
    UInt8                                       mySimilarity;
    OSErr                                       myErr = noErr;
    GWorldPtr                                   myGWorld = NULL;
    PixMapHandle                                myPixMap = NULL;
    Movie                                       mov = [movie quickTimeMovie];
    /*
    if (![imgSource hasPrefix:@"/"]) {
        TimeRecord tr;
        GetMovieTime(mov, &tr);
        tr.value.lo = [imgSource intValue];
        SetMovieTime(mov, &tr);
    }*/
    NSRect r = NSMakeRect(0,0, size.width, size.height);
	//NSLog(@"%g x %g", r.size.width, r.size.height);
    window = [[NSWindow alloc] initWithContentRect:r styleMask:NSBorderlessWindowMask backing:NSBackingStoreBuffered defer:NO];
	//[window setFrame:r display:YES];
    r = [[window contentView] frame];
	//NSLog(@"%g x %g", r.size.width, r.size.height);
    movieView = [[NSMovieView alloc] initWithFrame:r]; // grrr. QTMovieView doesn't let me scale the movie so....
    [movieView showController:NO adjustingSize:NO];
    [(NSWindow *)window setContentView:movieView];
	//if ([movieView movie] != movie)
	NSMovie *ns = [[NSMovie alloc] initWithMovie:mov];
		[movieView setMovie:ns];
    
    GetMovieBox(mov, &myRect);
	if (myRect.bottom - myRect.top == 0) myRect.bottom = size.height;
	if (myRect.right - myRect.left == 0) myRect.right = size.width;
	
	//[self setFrame:r forMovie:movie];
	//NSLog(@"%d x %d", myRect.right-myRect.left, myRect.bottom - myRect.top);
    //myRect.right = myRect.left + size.width;
    //myRect.bottom = myRect.top + size.height;
    //[window makeKeyAndOrderFront:nil];
    MoviesTask(mov, 0);
    //SetMovieBox(mov, &myRect);
    myErr = NewGWorld(&myGWorld, k32ARGBPixelFormat, &myRect, NULL, NULL, 0);
    myPixMap = GetGWorldPixMap(myGWorld);
    LockPixels(myPixMap);
    SetGWorld(myGWorld, NULL);
    GetPortBounds(myGWorld, &portRect);
    EraseRect(&portRect);
    int pixHeight = portRect.bottom - portRect.top;
    int rowBytes = GetPixRowBytes(myPixMap);
    
    //memset((*myPixMap)->baseAddr, 0, rowBytes * pixHeight);
    unsigned char *buffer = (*myPixMap)->baseAddr;
    int i, n = rowBytes * pixHeight;
	//NSLog(@"%d x %d", rowBytes/4, pixHeight);
    unsigned char red = 0, green = 0, blue = 0;
    keyColor = [keyColor colorUsingColorSpaceName:@"NSCalibratedRGBColorSpace"];
    if ([keyColor respondsToSelector:@selector(redComponent)] && [keyColor respondsToSelector:@selector(greenComponent)] && [keyColor respondsToSelector:@selector(blueComponent)]) {
        red = [keyColor redComponent] * 255.0;
        green = [keyColor greenComponent] * 255.0;
        blue = [keyColor blueComponent] * 255.0;
		for (i=0;i<n;i+=4) {
			buffer[i] = 255;
			buffer[i+1] = red;
			buffer[i+2] = green;
			buffer[i+3] = blue;
		}
	}
    
    
    myErr = CompressSequenceBegin(&mySeqID, myPixMap, NULL, &myRect, NULL, kCompressDepth, kAnimationCodecType, 0, codecLosslessQuality, codecLosslessQuality, 2, NULL, 0, myDesc);
    myErr = GetCSequenceMaxCompressionSize(mySeqID, myPixMap, &myDataSize);
    myData = NewPtr(myDataSize);
    myErr = MemError();
	if (myErr) NSLog(@"MemErr: %d", myErr);
    
    if (keyColor)
        myErr = CompressSequenceFrame(mySeqID, myPixMap, NULL, codecFlagUpdatePrevious, myData, &myDataSize, &mySimilarity, NULL);
    
    
    SetMovieGWorld(mov, myGWorld, GetGWorldDevice(myGWorld));
    //[movieView setMovie:nil];
    MoviesTask(mov, 0);
    
    if (keyColor)
        myErr = CompressSequenceFrame(mySeqID, myPixMap, NULL, codecFlagUpdatePrevious, myData, &myDataSize, &mySimilarity, NULL);
    else
        myErr = CompressSequenceFrame(mySeqID, myPixMap, NULL, 0, myData, &myDataSize, &mySimilarity, NULL);
    // end the compression sequence and clean up
    CDSequenceEnd(mySeqID);
    mySeqID = 0;
    
    /*
    NSBitmapImageRep *myRep = [[NSBitmapImageRep alloc] initWithData:[image TIFFRepresentationUsingCompression:NSTIFFCompressionNone factor:1.0]];
    
    char *theData = [myRep bitmapData];
    
    
    / *
    // get the current graphics port and device
    //GetGWorld(&mySavedPort, &mySavedDevice);
    myRect.left = myRect.top = 0;
    myRect.right = [image size].width;
    myRect.bottom = [image size].height;
    
    NSBitmapImageRep *rep = nil;
    NSArray *reps = [image representations];
    int i, n = [reps count];
    for (i=0;i<n;i++)
        if ([[reps objectAtIndex:i] isKindOfClass:[NSBitmapImageRep class]])
            rep = [reps objectAtIndex:i];
    if (!rep) return nil;
    
    //OSErr myErr = QTNewGWorldFromPtr(&myGWImage, k32RGBAPixelFormat, &myRect, NULL, NULL, 0, bytes, [image size].width * [[[image representations] objectAtIndex:0] bitsPerPixel]);
    
    NSBitmapImageRep *bmRep = [[NSBitmapImageRep alloc]	initWithBitmapDataPlanes:NULL 
                                                                pixelsWide:[image size].width 
                                                                pixelsHigh:[image size].height 
                                                            bitsPerSample:8 
                                                        samplesPerPixel:4
                                                                hasAlpha:YES
                                                                isPlanar:NO
                                                          colorSpaceName:NSDeviceRGBColorSpace 
                                                             bytesPerRow:[image size].width * 32
                                                            bitsPerPixel:0];
    
    
    
    
    PixMapHandle myPixMap = GetGWorldPixMap(myGWorld);
    
    
    // erase the image GWorld
    //SetGWorld(myGWImage, NULL);
    //RGBBackColor(theKeyColor);
    //ClipRect(&myRect);
    //EraseRect(&myRect);
    * /
    
    //char *buffer = malloc([image size].width * [image size].height * 4);
    //memset(buffer, 255, [image size].width * [image size].height * 4);
    //char *theData = [rep bitmapData];
    unsigned char *src;
    unsigned char *dst;
    int row, column;
    int i, a, r, g, b;
    for (row=0;row<myRect.bottom; row++) {
        //src = (*myPixMap)->baseAddr + row * GetPixRowBytes(myPixMap);
        src = theData + row * [myRep bytesPerRow];
        dst = GetPixBaseAddr(myPixMap) + row * [myRep bytesPerRow];
        for (column = 0;column<myRect.right;column++) {
            r = *src++;
            g = *src++;
            b = *src++;
            a = *src++;
        
            *dst++ = a;
            *dst++ = r;
            *dst++ = g;
            *dst++ = b;
        
            if ((a && a < 255) || (r && r < 255) || (g && g < 255) || (b && b < 255))
                i++;
        //((UInt32 *)buffer)[i] = 0xFF0000FF;
        }
    }
    / *
    (*myPixMap)->pixelFormat = k32ARGBPixelFormat;
    (*myPixMap)->baseAddr = buffer;
    (*myPixMap)->rowBytes = [image size].width * 4;
    (*myPixMap)->bounds = myRect;
    
    
    // lock our GWorld pixmap
    LockPixels(myPixMap);
    */
    // if transparency here
    //SetGWorld(myGWImage, NULL);
    // draw the image into the GWorld
    //[self copyImage:image toGWorld:myGWImage];
    // free the GWorld and drop references so we have more memory for PtrToHand
    //if (myGWImage != NULL)
    //    DisposeGWorld(myGWImage);
    //myGWImage = NULL;
    
    Handle theImageData;
    myErr = PtrToHand(myData, &theImageData, myDataSize);
    //SetGWorld(mySavedPort, mySavedDevice);
    if (myData) DisposePtr(myData);
    
	//[window performClose:self];
    //[window release];
    window = nil;
	//[movieView release];
    movieView = nil;
    
	if (path)
		[[NSFileManager defaultManager] removeFileAtPath:path handler:nil];
	
    return theImageData;
}

- (void) addSprite:(NSDictionary *)sprite toKeyFrame:(QTAtomContainer)theKeyFrameSample invisibility:(BOOL)invis spriteIndex:(QTAtomID)spriteIndex
{
    //NSLog(@"addSprite toKeyFrameSample");
    QTAtomID qtIDs[3];
    long w = [[sprite objectForKey:@"Width"] intValue];
    if (!w) w = 100;
    long h = [[sprite objectForKey:@"Height"] intValue];
    if (!h) h = 75;
    Rect box;
    box.top = box.left = 0;
    box.right = w;
    box.bottom = h;
    NSData *colorD = [sprite objectForKey:@"keyColor"];
    //RGBColor myKeyColor;
    NSColor *theKeyColor = NULL;
    if (colorD) {
        theKeyColor = [NSUnarchiver unarchiveObjectWithData:colorD];
    }
    NSArray *picKeys = [NSArray arrayWithObjects:@"Image", @"Rollover", @"Click", nil];
    int i;
	OSErr myErr = 0;
    BOOL visible = YES;
	int rx = 0, ry = 0;
    for (i=0;i<3;i++) {
        NSString *aPic = [sprite objectForKey:[picKeys objectAtIndex:i]];
        qtIDs[i] = 0;
        if (aPic || i==0) {
            if (!aPic)
                visible = NO;
            ImageDescriptionHandle imageDesc = (ImageDescriptionHandle)NewHandle(sizeof(ImageDescription));
            Handle hand = [self compressImage:aPic withDescription:imageDesc size:NSMakeSize(w, h) keyColor:theKeyColor];
            if (hand) {
				if ([[[sprite objectForKey:@"Action"] lowercaseString] hasPrefix:@"nohittest"]) {
					RgnHandle rh = (RgnHandle)NewHandleClear(0);
					//myErr = MediaGetInvalidRegion(GetMediaHandler(GetTrackMedia(GetMovieIndTrackType([srcMovie quickTimeMovie], 1, VideoMediaType, 0))), rh);
					//if (myErr) NSLog(@"get region: %d", myErr);
					myErr = ICUtils_RecompressCompressedImageWithTransparency(imageDesc, hand, nil, rh, &imageDesc, &hand);
					if (myErr) NSLog(@"recompress error: %d", myErr);
				}
                qtIDs[i] = picIndex++;
				NSString *regX = [sprite objectForKey:@"regX"], *regY = [sprite objectForKey:@"regY"];
				rx = w/2;
				ry = h/2;
				if (regX) rx = [regX intValue];
				if (regY) ry = [regY intValue];
				FixedPoint fp;
				fp.x = Long2Fix(rx);
				fp.y = Long2Fix(ry);
                myErr = SpriteExtras_AddCompressedImageToKeyFrameSample(theKeyFrameSample, hand, qtIDs[i], &fp, NULL, imageDesc);
            }
        }
    }
    QTAtomContainer myButton, myActionButton, myActions = NULL;
    short myIndex = qtIDs[0];
    myErr = QTNewAtomContainer(&myButton);
    Point myLocation;
    myLocation.h = [[sprite objectForKey:@"X"] intValue] + rx;
    myLocation.v = [[sprite objectForKey:@"Y"] intValue] + ry;
    short layer = 1;
    short isVisible = 1;
    if (invis && !visible)
	isVisible = 0;
    if ([[sprite objectForKey:@"invisible"] boolValue])
        isVisible = 0;
    StringPtr cName = nil;
    NSString *name = [sprite objectForKey:@"Name"];
    if (name) {
        cName = malloc(sizeof(char) * ([name length] + 1));
        memcpy(cName+1, [name cString], [name length]);
        cName[0] = [name length];
    }
	ModifierTrackGraphicsModeRecord record, *recordptr = nil;
	NSNumber *opq = [sprite objectForKey:@"opacity"];
	//NSLog(@"%@: %@", name, opq);
	if (opq) {
		float trans = [opq floatValue] * 65535;
		record.opColor.red = record.opColor.green = record.opColor.blue = trans;
		record.graphicsMode = blend;
		recordptr = &record;
	}
    myErr = SpriteUtils_SetSpriteData(myButton, &myLocation, &isVisible, &layer, &myIndex, recordptr, cName, myActions);
    if (cName) free(cName);
    myErr = QTCopyAtom(myButton, kParentAtomIsContainer, &myActionButton);
    if (qtIDs[1]) {
        // add rollover atoms
        WiredUtils_AddSpriteSetImageIndexAction(myActionButton, kParentAtomIsContainer, kQTEventMouseEnter, 0, NULL, 0, 0, NULL, qtIDs[1], NULL);
        if (qtIDs[0])
            WiredUtils_AddSpriteSetImageIndexAction(myActionButton, kParentAtomIsContainer, kQTEventMouseExit, 0, NULL, 0, 0, NULL, qtIDs[0], NULL);
	if (!visible) {
	    if (invis) { // the track is visible but the sprite isn't
		WiredUtils_AddSpriteSetVisibleAction(myActionButton, kParentAtomIsContainer, kQTEventMouseEnter, 0, NULL, 0, 0, nil, 1, NULL);
		WiredUtils_AddSpriteSetVisibleAction(myActionButton, kParentAtomIsContainer, kQTEventMouseExit, 0, NULL, 0, 0, nil, 0, NULL);
	    } else { // the sprite stays visible but the track doesn't
		// MediaSetGraphicsMode(handler, 260, &color);
	    }
	}
    }
    if (qtIDs[2]) {
        // add clicked image atoms
        WiredUtils_AddSpriteSetImageIndexAction(myActionButton, kParentAtomIsContainer, kQTEventMouseClick, 0, NULL, 0, 0, NULL, qtIDs[2], NULL);
        if (qtIDs[1]) // rollover
            WiredUtils_AddSpriteSetImageIndexAction(myActionButton, kParentAtomIsContainer, kQTEventMouseClickEnd, 0, NULL, 0, 0, NULL, qtIDs[1], NULL);
        else if (qtIDs[0]) // main
            WiredUtils_AddSpriteSetImageIndexAction(myActionButton, kParentAtomIsContainer, kQTEventMouseClickEnd, 0, NULL, 0, 0, NULL, qtIDs[0], NULL);
    }
    [self setAction:sprite forAtom:myActionButton keyFrame:theKeyFrameSample];
	if ([sprite objectForKey:@"Draggable"])
		QTWired_MakeSpriteDraggable(myActionButton, kParentAtomIsContainer); // note: this does not work
    myErr = SpriteUtils_AddSpriteToSample(theKeyFrameSample, myActionButton, spriteIndex);
    QTDisposeAtomContainer(myActionButton);
    

}

- (BOOL) setTrackEnabledAction:(QTAtomContainer)theSprite trackType:(OSType)theType ident:(id)theIdent
{
    //NSLog(@"setTrackEnabledAction");
    Track aTrack, targetTrack = nil;
    int i, n = GetMovieTrackCount([srcMovie quickTimeMovie]);
    BOOL enabled, byNumber;
    OSType aType;
    NSString *aTrackName;
    char *buffer;
    Handle h;
    OSErr myErr = noErr;
    void *theTrackTarget;
    long theTrackTargetType;
    if ([theIdent isKindOfClass:[NSNumber class]]) {
        byNumber = YES;
        targetTrack = GetMovieIndTrack([srcMovie quickTimeMovie], [theIdent intValue]);
    } else
        byNumber = NO;
    for (i=1; i<n; i++) {
        aTrack = GetMovieIndTrack([srcMovie quickTimeMovie], i);
        GetMediaHandlerDescription(GetTrackMedia(aTrack), &aType, 0, 0);
        if (aType == theType) {
            if (byNumber) { // by number
                if (aTrack == targetTrack)
                    enabled = YES;
                else
                    enabled = NO;
                theTrackTarget = (void *)i;
                theTrackTargetType = kTargetTrackID;
            } else { // by name
                h = NewHandleClear(0);
                if (![theIdent isKindOfClass:[NSString class]]) return NO;
                GetUserData(GetTrackUserData(aTrack), h, kUserDataName, 1);
                //buffer = malloc(sizeof(char) * GetHandleSize(h));
                //BlockMoveData(h, &buffer, GetHandleSize(h));
                buffer = (char *)*h;
                aTrackName = [NSString stringWithCString:buffer length:GetHandleSize(h)];
                if (aTrackName)
                    enabled = [theIdent isEqualToString:aTrackName];
                else
                    enabled = NO;
                theTrackTarget = (void *) QTUtils_ConvertCToPascalString(buffer);
                theTrackTargetType = kTargetTrackName;
                //DisposeHandle(h);
                //NSLog(@"set track %@ enabled: %d", aTrackName, enabled);
            }
            myErr = WiredUtils_AddTrackSetEnabledAction(theSprite, kParentAtomIsContainer, kQTEventMouseClickEnd, theTrackTargetType, theTrackTarget, 0, enabled);
            if (myErr) return NO;
        }
    }
    return YES;
}


- (RGBColor)rgbColor:(NSColor *)aColor
{
    //NSLog(@"rgbColor");
    RGBColor color;
    NSColor *newColor = [aColor colorUsingColorSpaceName:@"NSCalibratedRGBColorSpace"];
	//NSLog(@"colorspace converted");
    color.red = EndianU16_NtoB([newColor redComponent] * (double)0xffff);
    color.green =  EndianU16_NtoB([newColor greenComponent] * (double)0xffff);
    color.blue =  EndianU16_NtoB([newColor blueComponent] * (double)0xffff);
	//NSLog(@"end");
    return color;
}

- (void)postUpdate:(NSString *)string done:(double)done;
{
    double scale = 1.0 / (double)nTracks;
    double fullDone = done * scale + (double) iTrack / (double) nTracks;
    id vo = [NSConnection rootProxyForConnectionWithRegisteredName:@"mhSprite" host:nil];
    if (vo)
        [vo postUpdate:string done:fullDone];
}

- (void)quit
{	[NSApp terminate:nil];	}

/*
- (void)setMovieToSize:(NSSize)size
{
    //NSLog(@"Set Movie Width: %d Height: %d", (int)size.width, (int)size.height);
    Rect box;
    if (!size.height)
        size.height = 40;
    box.left = box.top = 0;
    box.right = size.width;
    box.bottom = size.height;
    Movie mov = [[movieView movie] QTMovie];
    SetMovieBox(mov, &box);
}
*/
/*
// from MyMovie.m
// http://developer.apple.com/samplecode/Sample_Code/Cocoa/bMoviePalette/MyMovie.m.htm
static BOOL WriteMovieToPath(Movie movie, NSString *path, BOOL flatten)
{
    OSErr    err;
    FSRef    fsRef;
    FSSpec   fsSpec;
    NSString *newPath, *newName, *tempPath, *newFolder;
    NSFileManager *fm = [NSFileManager defaultManager];
    Movie newmovie;
    short newResId;

    newPath = [NSString stringWithFormat:@"%@~",path];
    newName = [newPath lastPathComponent];
    if ([newName length] >= 25) {
        newName = [newName substringToIndex:25];
        newFolder = [newPath stringByDeletingLastPathComponent];
        newPath = [newFolder stringByAppendingPathComponent:newName];
        tempPath = newPath;
        int count = 1;
        while ([fm fileExistsAtPath:tempPath])
            tempPath = [NSString stringWithFormat:@"%@%d", newPath, count++];
        newPath = tempPath;
    }
    
    if (![fm fileExistsAtPath:newPath])
        [[NSString stringWithString:@"this file is empty"] writeToFile:newPath atomically:NO];
    err = FSPathMakeRef([newPath fileSystemRepresentation],&fsRef, NULL / * isDirectory * /);
    if (err == noErr){
        err = FSGetCatalogInfo(&fsRef, kFSCatInfoNone, NULL / *catalogInfo* /, NULL / *outName* /, &fsSpec, NULL / *parentRef* /);

        if (err == noErr){
            short resId;
            if (flatten) {
                FlattenMovie(movie,
                         flattenAddMovieToDataFork | flattenForceMovieResourceBeforeMovieData | flattenCompressMovieResource,
                         &fsSpec,'TVOD',
                         smSystemScript,
                         createMovieFileDeleteCurFile|createMovieFileDontCreateResFile,
                         &resId,nil);
                CloseMovieFile(resId);
            } else {
                //Movie newmovie;
                //short newResId;
                OSType creatorType = FOUR_CHAR_CODE('TVOD');
                err = CreateMovieFile(&fsSpec, creatorType, smSystemScript, nil, &resId, &newmovie); 
                / * 
                nil can be replaced by createMovieFileFlags
                CreateMovieFile has been superseded by CreateMovieStorage in QuickTime 6
                CreateMovieFile    http://developer.apple.com/techpubs/quicktime/qtdevdocs/APIREF/SOURCESI/createmoviefile.htm
                CreateMovieStorage http://developer.apple.com/techpubs/quicktime/qtdevdocs/APIREF/SOURCESI/createmoviestorage.htm
                * /
                err = AddMovieResource (movie, resId, &newResId, nil);
                CloseMovieFile(resId);
            }
        }
        //rename([newPath fileSystemRepresentation],[path fileSystemRepresentation]);
        [[NSFileManager defaultManager] movePath:newPath toPath:path handler:nil];
        return YES;
    }
        / * Clean up here * /
    //unlink([newPath fileSystemRepresentation]);
    return NO;
}
*/

FSSpec fsSpecFromPath2(NSString *path)
{
    FSRef fsRef;
    FSSpec fsSpec;
    OSErr myErr = 0;
    myErr = FSPathMakeRef((const UInt8 *)[path fileSystemRepresentation], &fsRef, NULL /* isDirectory */);
    myErr = FSGetCatalogInfo(&fsRef, kFSCatInfoNone, NULL /*catalogInfo*/, NULL /*outName*/, &fsSpec, NULL /*parentRef*/);
    return fsSpec;
}

//////////
//
// QTWired_MakeSpriteDraggable
// Add atoms to the specified sprite to make it draggable.
//
// To make a sprite draggable, we need to install handlers for three kinds of events: mouse click, mouse click end,
// and idle. On mouse click (within the sprite), we set a track variable to 1; on mouse click end, we set that variable
// to 0. Thus, the value of that variable can be used to tel us if the mouse is down within the sprite. On idle, we
// look to see if the value of the variable is 1; if it is, we set the position of the sprite to the current mouse position.
//
//////////

OSErr QTWired_MakeSpriteDraggable (QTAtomContainer theContainer, QTAtomID theID)
{
	QTAtom								mySpriteAtom = 0;
	QTAtom								myEventAtom = 0;
	QTAtom								myActionAtom = 0;
	QTAtom								myParamAtom = 0;
	QTAtom								myConditionalAtom, myExpressionAtom, myOperatorAtom, myOperandAtom, myOperandTypeAtom, myActionListAtom, myParameterAtom;
	short								myOperandIndex;
	long								myAction;
	QTAtomID							myVariableID;
	float								myVariableValue;
	float								myConstantValue;
	Boolean								myIsAbsolute;
	OSErr								myErr = noErr;
	
	// find the sprite atom with the specified ID in the specified container
	mySpriteAtom = QTFindChildByID(theContainer, kParentAtomIsContainer, kSpriteAtomType, theID, NULL);
	if (mySpriteAtom == 0) {
		// if there is none, insert a new sprite atom into the specified container
		myErr = QTInsertChild(theContainer, kParentAtomIsContainer, kSpriteAtomType, theID, 1, 0, NULL, &mySpriteAtom);
		if (myErr != noErr)
			goto bail;
	}
	
	//////////
	//
	// add a mouse click event handler
	//
	//////////
	
	// find the event atom of type kQTEventMouseClick in the sprite atom
	myEventAtom = QTFindChildByID(theContainer, mySpriteAtom, kQTEventType, kQTEventMouseClick, NULL);
	if (myEventAtom == 0) {
		// if there is none, insert a new event atom of type kQTEventMouseClick into the sprite atom
		myErr = QTInsertChild(theContainer, mySpriteAtom, kQTEventType, kQTEventMouseClick, 1, 0, NULL, &myEventAtom);
		if (myErr != noErr)
			goto bail;
	}
	
	// add an action atom to the mouse click event handler
	myErr = QTInsertChild(theContainer, myEventAtom, kAction, 0, 0, 0, NULL, &myActionAtom);
	if (myErr != noErr)
		goto bail;
	
	myAction = EndianU32_NtoB(kActionSpriteTrackSetVariable);
	myErr = QTInsertChild(theContainer, myActionAtom, kWhichAction, 1, 1, sizeof(myAction), &myAction, NULL);
	if (myErr != noErr)
		goto bail;
	
	// add parameters to the set variable action: variable ID (QTAtomID) and value (float)
	myVariableID = EndianU32_NtoB(kMouseStateVariableID);
	myErr = QTInsertChild(theContainer, myActionAtom, kActionParameter, 0, (short)kFirstParam, sizeof(myVariableID), &myVariableID, NULL);
	if (myErr != noErr)
		goto bail;
	
	myVariableValue = (float)1;
	EndianUtils_Float_NtoB(&myVariableValue);
	myErr = QTInsertChild(theContainer, myActionAtom, kActionParameter, 0, (short)kSecondParam, sizeof(myVariableValue), &myVariableValue, NULL);
	if (myErr != noErr)
		goto bail;
	
	//////////
	//
	// add a mouse click end event handler
	//
	//////////
	
	// find the event atom of type kQTEventMouseClick in the sprite atom
	myEventAtom = QTFindChildByID(theContainer, mySpriteAtom, kQTEventType, kQTEventMouseClickEnd, NULL);
	if (myEventAtom == 0) {
		// if there is none, insert a new event atom of type kQTEventMouseClick into the sprite atom
		myErr = QTInsertChild(theContainer, mySpriteAtom, kQTEventType, kQTEventMouseClickEnd, 1, 0, NULL, &myEventAtom);
		if (myErr != noErr)
			goto bail;
	}
	
	// add an action atom to the mouse click event handler
	myErr = QTInsertChild(theContainer, myEventAtom, kAction, 0, 0, 0, NULL, &myActionAtom);
	if (myErr != noErr)
		goto bail;
	
	myAction = EndianU32_NtoB(kActionSpriteTrackSetVariable);
	myErr = QTInsertChild(theContainer, myActionAtom, kWhichAction, 1, 1, sizeof(myAction), &myAction, NULL);
	if (myErr != noErr)
		goto bail;
	
	// add parameters to the set variable action: variable ID (QTAtomID) and value (float)
	myVariableID = EndianU32_NtoB(kMouseStateVariableID);
	myErr = QTInsertChild(theContainer, myActionAtom, kActionParameter, 0, (short)kFirstParam, sizeof(myVariableID), &myVariableID, NULL);
	if (myErr != noErr)
		goto bail;
	
	myVariableValue = (float)0;
	EndianUtils_Float_NtoB(&myVariableValue);
	myErr = QTInsertChild(theContainer, myActionAtom, kActionParameter, 0, (short)kSecondParam, sizeof(myVariableValue), &myVariableValue, NULL);
	if (myErr != noErr)
		goto bail;
	
	//////////
	//
	// add an idle event handler
	//
	//////////
	
	// find the event atom of type kQTEventIdle in the sprite atom
	myEventAtom = QTFindChildByID(theContainer, mySpriteAtom, kQTEventType, kQTEventIdle, NULL);
	if (myEventAtom == 0) {
		// if there is none, insert a new event atom of type kQTEventIdle into the sprite atom
		myErr = QTInsertChild(theContainer, mySpriteAtom, kQTEventType, kQTEventIdle, 1, 0, NULL, &myEventAtom);
		if (myErr != noErr)
			goto bail;
	}
	
	// add an action atom to the mouse click event handler
	myErr = QTInsertChild(theContainer, myEventAtom, kAction, 0, 0, 0, NULL, &myActionAtom);
	if (myErr != noErr)
		goto bail;
	
	myAction = EndianU32_NtoB(kActionCase);
	myErr = QTInsertChild(theContainer, myActionAtom, kWhichAction, 1, 1, sizeof(myAction), &myAction, NULL);
	if (myErr != noErr)
		goto bail;
	
	// add a parameter atom to the kActionCase action atom; this will serve as a parent to hold the expression and action atoms
	myErr = QTInsertChild(theContainer, myActionAtom, kActionParameter, 1, kFirstParam, 0, NULL, &myParamAtom);
	if (myErr != noErr)
		goto bail;
	
	// the condition atom
	myErr = QTInsertChild(theContainer, myParamAtom, kConditionalAtomType, 0, 1, 0, NULL, &myConditionalAtom);
	if (myErr != noErr)
		goto bail;
	
	myErr = QTInsertChild(theContainer, myConditionalAtom, kExpressionContainerAtomType, 1, 1, 0, NULL, &myExpressionAtom);
	if (myErr != noErr)
		goto bail;
	
	myErr = QTInsertChild(theContainer, myExpressionAtom, kOperatorAtomType, kOperatorEqualTo, 1, 0, NULL, &myOperatorAtom);
	if (myErr != noErr)
		goto bail;
	
	// add the operands to the operator atom
	myOperandIndex = 1;	
	myConstantValue = 1;
	myErr = QTInsertChild(theContainer, myOperatorAtom, kOperandAtomType, 0, myOperandIndex, 0, NULL, &myOperandAtom);
	if (myErr != noErr)
		goto bail;
	
	myErr = QTInsertChild(theContainer, myOperandAtom, kOperandConstant, 1, 1, 0, NULL, &myOperandTypeAtom);
	if (myErr != noErr)
		goto bail;
	
	EndianUtils_Float_NtoB(&myConstantValue);
	myErr = QTSetAtomData(theContainer, myOperandTypeAtom, sizeof(myConstantValue), &myConstantValue);
	
	myOperandIndex = 2;
	myVariableID = kMouseStateVariableID;
	myErr = QTInsertChild(theContainer, myOperatorAtom, kOperandAtomType, 0, myOperandIndex, 0, NULL, &myOperandAtom);
	if (myErr != noErr)
		goto bail;
	
	myErr = QTInsertChild(theContainer, myOperandAtom, kOperandSpriteTrackVariable, 1, 1, 0, NULL, &myOperandTypeAtom);
	if (myErr != noErr)
		goto bail;
	
	myVariableID = EndianU32_NtoB(myVariableID);
	myErr = QTInsertChild(theContainer, myOperandTypeAtom, kActionParameter, 1, 1, sizeof(myVariableID), &myVariableID, NULL);
	if (myErr != noErr)
		goto bail;
	
	// add an action list atom
	myErr = QTInsertChild(theContainer, myConditionalAtom, kActionListAtomType, 1, 1, 0, NULL, &myActionListAtom);
	if (myErr != noErr)
		goto bail;
	
	// add sprite translate action
	myErr = QTInsertChild(theContainer, myActionListAtom, kAction, 0, 0, 0, NULL, &myActionAtom);
	if (myErr != noErr)
		goto bail;
	
	myAction = EndianU32_NtoB(kActionSpriteTranslate);
	myErr = QTInsertChild(theContainer, myActionAtom, kWhichAction, 1, 1, sizeof(myAction), &myAction, NULL);
	if (myErr != noErr)
		goto bail;
	
	//////////
	//
	// add parameters to the translate action: Fixed x, Fixed y, Boolean isAbsolute
	//
	//////////
	
	// first parameter: get current mouse position x
	myErr = QTInsertChild(theContainer, myActionAtom, kActionParameter, 0, (short)kFirstParam, 0, NULL, &myParameterAtom);
	if (myErr != noErr)
		goto bail;
	
	myErr = QTInsertChild(theContainer, myParameterAtom, kExpressionContainerAtomType, 1, 1, 0, NULL, &myExpressionAtom);
	if (myErr != noErr)
		goto bail;
	
	myErr = QTInsertChild(theContainer, myExpressionAtom, kOperandAtomType, 0, 1, 0, NULL, &myOperandAtom);
	if (myErr != noErr)
		goto bail;
	
	myErr = QTInsertChild(theContainer, myOperandAtom, kOperandMouseLocalHLoc, 1, 1, 0, NULL, NULL);
	if (myErr != noErr)
		goto bail;
	
	// second parameter: get current mouse position y
	myErr = QTInsertChild(theContainer, myActionAtom, kActionParameter, 0, (short)kSecondParam, 0, NULL, &myParameterAtom);
	if (myErr != noErr)
		goto bail;
	
	myErr = QTInsertChild(theContainer, myParameterAtom, kExpressionContainerAtomType, 1, 1, 0, NULL, &myExpressionAtom);
	if (myErr != noErr)
		goto bail;
	
	myErr = QTInsertChild(theContainer, myExpressionAtom, kOperandAtomType, 0, 1, 0, NULL, &myOperandAtom);
	if (myErr != noErr)
		goto bail;
	
	myErr = QTInsertChild(theContainer, myOperandAtom, kOperandMouseLocalVLoc, 1, 1, 0, NULL, NULL);
	if (myErr != noErr)
		goto bail;
	
	myIsAbsolute = true;
	myErr = QTInsertChild(theContainer, myActionAtom, kActionParameter, 0, (short)kThirdParam, sizeof(myIsAbsolute), &myIsAbsolute, NULL);
	
bail:
	return(myErr);
}

@end
